# -*- coding: utf-8 -*-
"""
Created on Mon Sep 21 12:36:29 2020

@author: Nick
"""

import os
import hyperspy.api as hs
import datetime
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar


def nonzero_len(hsvid):
    return np.where(hsvid.data.sum(1).sum(1) != 0)[0].max()

def mag_consec(hsvid, pxwidth, lengthcut=5):
    from more_itertools import run_length
    maglist = np.array([float(x) for x in (hsvid.original_metadata.ser_header_parameters.CalibrationDeltaX)])
    count = 0
    outpairs = []
    for value, length in list(run_length.encode(maglist)):
        if length >= lengthcut and value == pxwidth:
            outpairs.append((count, count + length))
        count += length
    return outpairs
    

infolder = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20\20200914_OldPtliquidcell"
outfolder = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\1_preprocessing"

set_pxwidth = 3.450450459619629e-11

outfolder_video = outfolder + os.sep + 'video' + os.sep
outfolder_picture = outfolder + os.sep + 'picture' + os.sep
for folder in [outfolder, outfolder_video, outfolder_picture]:
    os.makedirs(folder, exist_ok=True)

video_dicts = []
picture_dicts = []
    
for root, dirs, files in os.walk(infolder):

            
    for file in files:
        if file.split('.')[-1] == 'emi':
            print(file)
            filein = hs.load(root + os.sep + file)
            filename = (root[len(infolder):] + os.sep + file).replace('\\', '/')
            
            if type(filein) != list:
                
                md = filein.original_metadata
                dt = datetime.datetime.strptime(md.ObjectInfo.AcquireDate, '%a %b %d %H:%M:%S %Y')
                boxes = [md.ObjectInfo.AcquireInfo, md.ObjectInfo.ExperimentalConditions.MicroscopeConditions, md.ObjectInfo.ExperimentalDescription]
                
                if len(filein.data.shape) == 3:
                    
                    correct_mag_indices = mag_consec(filein, set_pxwidth)
                    for i in range(len(correct_mag_indices)):
                        meta_dict = {k:v for box in boxes for k, v in box}
                        pairindex = correct_mag_indices[i]
                        meta_dict['orig_filename'] = filename
                        
                        if len(filename.split('/')) > 1:
                            meta_dict['folder'], meta_dict['filename'] = filename.rsplit('/', 1)
                        else:
                            meta_dict['folder'], meta_dict['filename'] = '', filename.replace('.emi', '.hspy')
                        
                        if len(correct_mag_indices) > 1:
                            alphabet = ['a', 'b', 'c', 'd', 'e', 'f']
                            meta_dict['filename'] = meta_dict['filename'][:-4] + '_%s' %(alphabet[i]) + '.hspy'
                        else:
                            meta_dict['filename'] = meta_dict['filename'].replace('.emi', '.hspy')
                            
                        meta_dict['orig_num_frames'] = nonzero_len(filein)
                        meta_dict['cropped_num_frames'] = pairindex[1] - pairindex[0]
                        meta_dict['correct_mag_start'] = pairindex[0]
                        meta_dict['correct_mag_end'] = pairindex[1]
                        meta_dict['date'] = dt.date()
                        meta_dict['time'] = dt.time()
                        meta_dict['pxwidth'] = filein.axes_manager[1].scale
                        meta_dict['pxnum'] = filein.axes_manager[1].size
                        meta_dict['pxunits'] = filein.axes_manager[1].units
                        meta_dict['framewidth'] = filein.axes_manager[1].scale * img.axes_manager[0].size
                        video_dicts.append(meta_dict)
                        cropped = filein.inav[pairindex[0]: pairindex[1]].deepcopy()
                        cropped.save(outfolder_video + meta_dict['filename'], overwrite=True)
                        plt.imsave(outfolder_video + meta_dict['filename'].replace('.hspy', '.png'), cropped.data[0], cmap='inferno')

                elif len(filein.data.shape) == 2:
                    meta_dict = {k:v for box in boxes for k, v in box}
                    meta_dict['original_filename'] = filename
                    meta_dict['out_filename'] = filename.split('/')[-1][:-4] + '.hspy'
                    meta_dict['date'] = dt.date()
                    meta_dict['time'] = dt.time()
                    meta_dict['pxwidth'] = filein.axes_manager[0].scale
                    meta_dict['pxnum'] = filein.axes_manager[0].size
                    meta_dict['pxunits'] = filein.axes_manager[0].units
                    meta_dict['framewidth'] = filein.axes_manager[0].scale * img.axes_manager[0].size
                    picture_dicts.append(meta_dict)
                    filein.save(outfolder_picture + meta_dict['out_filename'], overwrite=True)
                    plt.imsave(outfolder_picture + meta_dict['out_filename'].replace('.hspy', '.png'), filein.data, cmap='inferno')
                    
                    
            else:
                for i in range(len(filein)):
                    alphabet = ['a', 'b', 'c', 'd', 'e', 'f']
                    img = filein[i]
                    
                    md = img.original_metadata
                    dt = datetime.datetime.strptime(md.ObjectInfo.AcquireDate, '%a %b %d %H:%M:%S %Y')
                    boxes = [md.ObjectInfo.AcquireInfo, md.ObjectInfo.ExperimentalConditions.MicroscopeConditions, md.ObjectInfo.ExperimentalDescription]
                    
                    meta_dict = {k:v for box in boxes for k, v in box}
                    meta_dict['original_filename'] = filename
                    meta_dict['out_filename'] = filename.split('/')[-1][:-4] + '_%s.hspy' %alphabet[i]
                    meta_dict['date'] = dt.date()
                    meta_dict['time'] = dt.time()
                    meta_dict['pxwidth'] = img.axes_manager[0].scale
                    meta_dict['pxnum'] = img.axes_manager[0].size
                    meta_dict['pxunits'] = img.axes_manager[0].units
                    meta_dict['framewidth'] = img.axes_manager[0].scale * img.axes_manager[0].size
                    picture_dicts.append(meta_dict)
                    img.save(outfolder_picture + meta_dict['out_filename'], overwrite=True)
                    plt.imsave(outfolder_picture + meta_dict['out_filename'].replace('.hspy', '.png'), img.data, cmap='inferno')


                    
video_df_long = pd.DataFrame(video_dicts)
picture_df_long = pd.DataFrame(picture_dicts)





dates = picture_df_long.date.unique()

for date in dates:
    vids = video_df_long[video_df_long.date == date]
    pics = picture_df_long[picture_df_long.date == date]
    
    fig = plt.figure(figsize=(12, 12))
    ax = fig.add_axes([0, 0, 1, 1], aspect='equal')
    
    vidlocs = []
    for index, row in vids.iterrows():
        color = [np.random.rand(1)[0], np.random.rand(1)[0], np.random.rand(1)[0], 1]
        width = row.pxnum * set_pxwidth
        pos = [row.Stage_X_um * 1000.0, row.Stage_Y_um * 1000.0]
        XY = [pos[0] - width/2.0, pos[1] - width/2.0]
        rect = Rectangle(XY, width, width, fill=False, edgecolor=color)
        ax.add_artist(rect)
        ax.scatter([XY[0] + width], [XY[1] + width], s=0)
        ax.text(XY[0], XY[1], row.orig_filename, color=color)
        vidlocs.append(XY)
        
    for index, row in pics.iterrows():
        color = [np.random.rand(1)[0], np.random.rand(1)[0], np.random.rand(1)[0], 1]
        width = row.pxnum * row.pxwidth
        pos = [row.Stage_X_um * 1000.0, row.Stage_Y_um * 1000.0]
        XY = [pos[0] - width/2.0, pos[1] - width/2.0]
        rect = Rectangle(XY, width, width, fill=False, edgecolor=color)
        ax.add_artist(rect)
        ax.scatter([XY[0] + width], [XY[1] + width], s=0)
        ax.text(XY[0], XY[1], row.original_filename, color=color)

        
    sb = ScaleBar(1, 'nm')
    ax.add_artist(sb)
    
    ax.autoscale()
    fig.savefig(outfolder + os.sep + 'map_%s.svg' %date)
    
    video_locations = np.array(vidlocs)
    
    # fig, ax = plt.subplots(1)
    # ax.scatter(video_locations[:, 0], video_locations[:, 1])
        
    
        
        
                        
                        
                













video_changing = video_df_long.eq(video_df_long.iloc[0,:], axis=1).all(0)
video_df_nochanges = video_df_long.iloc[:,video_changing.values].iloc[0]
video_df_changes = video_df_long.iloc[:,video_changing.values == False]

video_df_nochanges.to_csv(outfolder_video + 'common_settings.csv')
video_df_changes.to_csv(outfolder_video + 'metadata_settings.csv')
video_df_long.to_csv(outfolder_video + 'all_settings.csv')


                

picture_changing = picture_df_long.eq(picture_df_long.iloc[0,:], axis=1).all(0)
picture_df_nochanges = picture_df_long.iloc[:,picture_changing.values].iloc[0]
picture_df_changes = picture_df_long.iloc[:,picture_changing.values == False]

picture_df_nochanges.to_csv(outfolder_picture + 'common_settings.csv')
picture_df_changes.to_csv(outfolder_picture + 'metadata_settings.csv')
picture_df_long.to_csv(outfolder_picture + 'all_settings.csv')




     
                
               
