# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 11:14:50 2020

@author: Nick
"""
import os
import gc
import math
import warnings

import matplotlib.pyplot as plt
import hyperspy.api as hs
import trackpy as tp
import numpy as np
import atomap.api as am
import pandas as pd
import seaborn as sns


from copy import deepcopy
from tqdm import tqdm
from scipy.signal import medfilt2d
from scipy.ndimage import gaussian_filter
from scipy.optimize import leastsq, curve_fit
from skimage.feature import peak_local_max, match_template
from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar
from lmfit.models import SkewedGaussianModel
from scipy.ndimage import gaussian_filter

def round_templates(segment, tempwidth, pos_int):
    # create template image for decomposition - masked using 2D gaussian
    templates = []
    pos_int_pass = []
    mask = gauss2Dmask(tempwidth, (tempwidth/2)**2)
    for pos in pos_int[:]:
        template = segment.data[pos[1] - tempwidth: pos[1] + tempwidth + 1, pos[0] - tempwidth: pos[0] + tempwidth + 1]
        if template.shape == (tempwidth * 2 + 1, tempwidth * 2 + 1):
            templates.append(template * mask)
            pos_int_pass.append(pos)
    template_arr = np.array(templates)
    template_im = hs.signals.Signal2D(template_arr)
    return template_im, pos_int_pass

def shift_2d_replace(data, dx, dy, constant=0.0):
    shifted_data = np.roll(data, dx, axis=1)
    if dx < 0:
        shifted_data[:, dx:] = constant
    elif dx > 0:
        shifted_data[:, 0:np.abs(dx)] = constant

    shifted_data = np.roll(shifted_data, dy, axis=0)
    if dy < 0:
        shifted_data[dy:, :] = constant
    elif dy > 0:
        shifted_data[0:np.abs(dy), :] = constant
    return shifted_data

def gauss2Dmask(tempwidth, sigma, normed=False):
    # generate gaussian 2D mask
    import numpy as np
    prefac = 1  # / np.sqrt(2 * np.pi * sigma)
    if normed is True:
        prefac = 1 / np.sqrt(2 * np.pi * sigma)
    X, Y = np.meshgrid(range(0, tempwidth*2 + 1), range(0, tempwidth*2 + 1))
    distX, distY = X - tempwidth, Y - tempwidth
    dist2 = (distX**2) + (distY**2)
    exp = np.exp2(-dist2/(2*sigma))
    return prefac * exp


def padvideo(video, minshiftx, maxshiftx, minshifty, maxshifty):
    padframe = np.pad(video.copy().data[:], ((0, 0), (np.abs(np.int(minshiftx)), np.abs(np.int(maxshiftx))), (np.abs(np.int(minshifty)), np.abs(np.int(maxshifty)))), mode='constant')
    outvideo = hs.signals.Signal2D(padframe)
    outvideo.axes_manager[1].size = video.axes_manager[1].size
    outvideo.axes_manager[2].size = video.axes_manager[2].size
    return outvideo


def ring_mask(size, radius, blur):
    [r_low, r_high] = radius
    mask = np.zeros((size, size), dtype=float)
    cent = np.int(size / 2)
    for i in range(0, size):
        for j in range(0, size):
            dist = np.sqrt(((i - cent)**2) + ((j - cent)**2))
            if r_low < dist < r_high:
                mask[i, j] = 1.0
    return gaussian_filter(mask, blur)


def cross_mask(size, cinner, cwidth, clength, blur):
    mask = np.ones((size, size), dtype=float)
    cent = np.int(size / 2)
    mask[cent - cwidth:cent + cwidth, (cent - cinner) - clength:cent - cinner] = 0.0
    mask[cent - cwidth:cent + cwidth, (cent + cinner):cent + cinner + clength] = 0.0
    mask[(cent - cinner) - clength:cent - cinner, cent - cwidth:cent + cwidth] = 0.0
    mask[(cent + cinner):cent + cinner + clength, cent - cwidth:cent + cwidth] = 0.0
    return gaussian_filter(mask, blur)


def template_rebuild(input_im, pos_int_pass, decomp_model, twidth, name, med_filt=False, gfilt=0):
    output_im = input_im.deepcopy()
    output_arr = np.zeros(input_im.data.shape, input_im.data.dtype)
    norm_arr = output_arr.copy()
    mask = gauss2Dmask(twidth, (twidth/2)**2)
    decomp_arr = decomp_model.data
    for i in range(0, decomp_model.data.shape[0]):
        pos = pos_int_pass[i]
        output_arr[pos[1] - twidth: pos[1] + twidth + 1, pos[0] - twidth: pos[0] + twidth + 1] += decomp_arr[i]
        norm_arr[pos[1] - twidth: pos[1] + twidth + 1, pos[0] - twidth: pos[0] + twidth + 1] += mask
    output_arr = np.nan_to_num(output_arr / norm_arr)
    if med_filt == True:
        output_arr = medfilt(output_arr, 3)
    output_arr = gaussian_filter(output_arr, gfilt)
    output_im.metadata.General.title = name
    output_im.data = output_arr
    return output_im

# prefs


warnings.filterwarnings("ignore")
hs.preferences.General.logging_level = 'CRITICAL'

plt.close('all')
gc.collect()

prevsettings = pd.read_csv(r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\2_input_settings - PCA.csv")
prevsettings = prevsettings.set_index('Unnamed: 0')
vidsfolder = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\1_preprocessing\video" + os.sep
savefolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\2_ALT_PCAdenoise" + os.sep


'''load video and cut------------------------------------------------------------------------------------------------------------------'''

for vidnum in tqdm(prevsettings.index):
    settings = prevsettings.loc[vidnum]
    

    vidinfn = vidsfolder + settings.filename
    vidin = hs.load(vidinfn)
    
    previd = vidin.inav[settings.crop2_start: settings.crop2_end].deepcopy()

    
    previd.change_dtype('float64')
    
    savefolder = savefolder_base + settings.filename + os.sep
    
    if not os.path.exists(savefolder):
        os.makedirs(savefolder)

        
        previd.save(savefolder + 'input_vid_cropped.hspy', overwrite=True)
        
        # get frame size and vid length
        
        length = previd.axes_manager[0].size
        rows, cols = previd.inav[0].data.shape
        size = 'Not Square Image'
        if rows == cols:
            size = rows
        crow, ccol = int(rows / 2), int(cols / 2)
        centre = (crow, ccol)
        
        partition_number = 512
        
        # plot histograms and align
        fig = plt.figure(figsize=(16, 10))
        ax1 = plt.subplot2grid((4, 8), (0, 0), colspan=4, rowspan=2)
        ax1.get_xaxis().set_visible(False)
        ax2 = plt.subplot2grid((4, 8), (2, 0), colspan=4, rowspan=2, sharex=ax1)
        ax3 = plt.subplot2grid((4, 8), (0, 4), colspan=4, rowspan=1)
        ax3.get_xaxis().set_visible(False)
        ax4 = plt.subplot2grid((4, 8), (1, 4), colspan=4, rowspan=1, sharex=ax3)
        ax4.get_xaxis().set_visible(False)
        ax5 = plt.subplot2grid((4, 8), (2, 4), colspan=4, rowspan=1, sharex=ax3)
        ax5.get_xaxis().set_visible(False)
        ax6 = plt.subplot2grid((4, 8), (3, 4), colspan=4, rowspan=1, sharex=ax3)
        int_datarr = previd.data.flatten()
        lobin = np.partition(int_datarr, partition_number)[partition_number]
        hibin = np.partition(int_datarr, -partition_number)[-partition_number]
        bins = np.linspace(lobin , hibin, num=128)
        hist, bin_edges = np.histogram(int_datarr, bins=bins, density=True)
        shiftcentre = bin_edges[hist.argmax()]
        centers = []
        amplitudes = []
        sigmas = []
        gammas = []
        arrowstart = [0,0]
        arrowend = [0,0]
        for i in tqdm(range(0, previd.data.shape[0]), desc='Measuring contrast offsets'):
            datarr = previd.inav[i].data.flatten()
            hist, bin_edges = np.histogram(datarr, bins=bins, density=True)
            bin_centres = (bin_edges[:-1] + bin_edges[1:]) / 2
            ax1.plot(bin_centres, hist, alpha=0.1)
            if i == 0:
                arrowstart[0] = bin_centres[hist.argmax()]
                arrowstart[1] = hist.max()
            elif i == previd.data.shape[0] - 1:
                arrowend[0] = bin_centres[hist.argmax()]
                arrowend[1] = hist.max()
            model = SkewedGaussianModel()
            params = model.make_params(amplitude=1, center=bin_centres[hist.argmax()], sigma=3000, gamma=0)
            result = model.fit(hist, params, x=bin_centres)
            fithist = result.eval(x=bin_centres)
            centers.append(result.best_values['center'])
            amplitudes.append(result.best_values['amplitude'])
            sigmas.append(result.best_values['sigma'])
            gammas.append(result.best_values['gamma'])
            shift = shiftcentre - result.best_values['center']
            shiftarr = previd.inav[i].data.copy() + shift
            hist, bin_edges = np.histogram(shiftarr.flatten(), bins=bins, density=True)
            bin_centres = (bin_edges[:-1] + bin_edges[1:]) / 2
            ax2.plot(bin_centres, hist, alpha=0.1)
            previd.inav[i].data[:] = shiftarr
        ax1.set_title('Individual Frames')
        ax1.set_xlim(lobin, hibin)
        ax1.annotate("", (arrowend[0], arrowend[1]), (arrowstart[0], arrowstart[1]), arrowprops=dict(arrowstyle="->"))     
        ax2.set_title('After Linear Shift')
        ax2.set_xlabel('Pixel Intensity')
        ax3.plot(centers)
        ax3.set_ylabel('Centre')
        ax4.plot(amplitudes)
        ax4.set_ylabel('Amplitude')
        ax5.plot(sigmas)
        ax5.set_ylabel('Sigma')
        ax6.plot(gammas)
        ax6.set_ylabel('Gamma')
        ax6.set_xlabel('Frame Number')
        plt.suptitle(settings.filename)
        plt.tight_layout()
        plt.savefig(savefolder + 'intensity_histograms.png')
        plt.savefig(savefolder + 'intensity_histograms.pdf')
        plt.show()
        
        previd.save(savefolder + 'vid_contrast_shift.hspy', overwrite=True)
        
        
        # plot int histogram and cut data
        
        
        datarr = previd.data.flatten()
        maxval = np.max(datarr)
        minval = np.min(datarr)
        
        
        normvid = (previd.copy() - minval) / (maxval - minval)
        normsavevid = normvid * 10000000
        normsavevid.change_dtype('int')
        normsavevid.save(savefolder + 'normvid.hspy', overwrite=True)
        del normsavevid
        
        plt.figure()
        normdatarr = normvid.data.flatten()
        normmaxval = np.max(normdatarr)
        normminval = np.min(normdatarr)
        hist, bin_edges = np.histogram(normdatarr, bins=np.linspace(normminval, normmaxval, num=512), density=True)
        hist = hist / hist.sum()
        cumhist = np.cumsum(hist)
        plt.plot(bin_edges[:-1], hist)
        plt.plot(bin_edges[1:-1], np.diff(hist))
        plt.xlabel('Intensity')
        plt.ylabel('freq. density')
        ax2 = plt.twinx()
        ax2.plot(bin_edges[:-1], cumhist, 'r')
        plt.ylabel('cum. freq. density')
        plt.savefig(savefolder + 'normalised_total_intensity_histogram.png')
        plt.savefig(savefolder + 'normalised_total_intensity_histogram.pdf')
        
        
        highcumthresh = 0.990
        lowcumthresh = 0.1
        highmult = 2.5
        
        highthreshold = bin_edges[: - 1][np.abs(cumhist - highcumthresh).argmin()]
        highvid = normvid.deepcopy() - highthreshold
        highvid.data[highvid.data < 0.0] = 0.0
        highvid.plot(navigator=None, vmin=0, vmax=0.1)
        
        lowthreshold = bin_edges[: - 1][np.abs(cumhist - lowcumthresh).argmin()]
        lowvid = normvid - (highvid * highmult) - lowthreshold
        lowvid.data[lowvid.data < 0.0] = 0.0
        lowvid = lowvid / np.max(lowvid.data)
        lowvid.plot(navigator=None, vmin=0, vmax=1)
        
          
        # self similarity matching loop
        
        smoothrad = settings.smoothrad
        atomsep = settings.atomsep
        tempwidth = settings.tempwidth
        tempwidth_2 = settings.tempwidth_2
        bss_num_1 = settings.bss_num_1
        bss_num_2 = settings.bss_num_2
        
        padwidth = max(tempwidth, tempwidth_2)
        padlowvid = padvideo(lowvid, padwidth, padwidth, padwidth, padwidth)
        
        reportlist = []
        outframelist = []
        
        for fnum in tqdm(range(0, lowvid.data.shape[0]), desc='PCA denoising and Rebuilding ' + settings.filename):
            framename = savefolder + 't_matched_frame_%d.gz' % (fnum)
            imname = savefolder + 't_matched_frame_%d.png' % (fnum)
            
            report = pd.Series()
            report['frame'] = fnum
            
            frame = padlowvid.inav[fnum]
            smoothframe = frame.deepcopy()
            smoothframe.data = gaussian_filter(frame.data, smoothrad)
            
            atom_pos = am.get_atom_positions(smoothframe, separation=atomsep)
            sublattice = am.Sublattice(atom_pos, image=frame.data)
            sublattice.find_nearest_neighbors()
            sublattice.refine_atom_positions_using_center_of_mass(show_progressbar=False)      
            pos_int = np.round(sublattice.atom_positions).astype(int)
            
            report['found_templates_1'] = len(pos_int)
            
            template_im, pos_int_pass = round_templates(frame, tempwidth, pos_int)
            
            report['passed_templates_1'] = len(pos_int_pass)
            
            template_im.decomposition(print_info=False)
            template_im.blind_source_separation(bss_num_1, print_info=False)

            bss_dmodel_1 = template_im.get_bss_model(bss_num_1)
            template_rebuilt_bss_1 = template_rebuild(frame, pos_int_pass, bss_dmodel_1, tempwidth, 'BSS_1')
            
            atom_pos_2 = am.get_atom_positions(template_rebuilt_bss_1, separation=atomsep)
            sublattice_2 = am.Sublattice(atom_pos_2, image=template_rebuilt_bss_1.data)
            sublattice_2.find_nearest_neighbors()
            sublattice_2.refine_atom_positions_using_center_of_mass(show_progressbar=False)      
            pos_int_2 = np.round(sublattice_2.atom_positions).astype(int)
            
            report['found_templates_2'] = len(pos_int_2)
            
            template_im_2, pos_int_pass_2 = round_templates(frame, tempwidth_2, pos_int_2)
            
            report['passed_templates_2'] = len(pos_int_pass_2)
            
            template_im_2.decomposition(print_info=False)
            template_im_2.blind_source_separation(bss_num_2, print_info=False)
            bss_dmodel_2 = template_im_2.get_bss_model(bss_num_2)
            template_rebuilt_bss_2 = template_rebuild(frame, pos_int_pass_2, bss_dmodel_2, tempwidth_2, 'BSS_2')
            
            plt.imsave(savefolder + 'frame_%d.png' %fnum, template_rebuilt_bss_2.data[padwidth: -padwidth, padwidth: -padwidth], cmap='inferno')

            reportlist.append(report)
            outframelist.append(template_rebuilt_bss_2)
            
        plt.close('all')
        gc.collect()
            
        padoutvideo = hs.stack(outframelist)
        outvideo = padoutvideo.isig[padwidth: -padwidth, padwidth: -padwidth].deepcopy()
        
        outvideo.save(savefolder + 'bss_denoised_low_video.hspy')
        
        report = pd.concat(reportlist, axis=1).T.set_index('frame')
        report.to_csv(savefolder + 'report.csv')
        
        






