import os
import gc
import math


import matplotlib.pyplot as plt
import hyperspy.api as hs
import trackpy as tp
import numpy as np
import atomap.api as am
import pandas as pd
import seaborn as sns


from copy import deepcopy
from tqdm import tqdm
from scipy.signal import medfilt2d
from scipy.ndimage import gaussian_filter
from scipy.optimize import leastsq, curve_fit
from skimage.feature import peak_local_max, match_template
from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar
from skimage.filters import threshold_niblack, threshold_sauvola
from skimage.registration import phase_cross_correlation
from skimage.morphology import erosion, dilation, disk
from scipy.ndimage import rank_filter


def padvideo(video, minshiftx, maxshiftx, minshifty, maxshifty):
    padframe = np.pad(video.copy().data[:], ((0, 0), (np.abs(np.int(minshiftx)), np.abs(np.int(maxshiftx))), (np.abs(np.int(minshifty)), np.abs(np.int(maxshifty)))), mode='constant')
    outvideo = hs.signals.Signal2D(padframe)
    outvideo.axes_manager[1].size = video.axes_manager[1].size
    outvideo.axes_manager[2].size = video.axes_manager[2].size
    return outvideo


def shift_2d_replace(data, dx, dy, constant=0.0):
    shifted_data = np.roll(data, dx, axis=1)
    if dx < 0:
        shifted_data[:, dx:] = constant
    elif dx > 0:
        shifted_data[:, 0:np.abs(dx)] = constant

    shifted_data_2 = np.roll(shifted_data, dy, axis=0)
    if dy < 0:
        shifted_data_2[dy:, :] = constant
    elif dy > 0:
        shifted_data_2[0:np.abs(dy), :] = constant
    return shifted_data_2

def drift_pad(video, drift):
    minyd = np.abs(drift.y.min())
    minxd = np.abs(drift.x.min())
    maxyd = np.abs(drift.y.max())
    maxxd = np.abs(drift.x.max())
    outvid = padvideo(video, maxyd, minyd, maxxd, minxd)
    return outvid

def drift_shifts(video, drift):
    padvideo = drift_pad(video, drift)
    outvideo = padvideo.deepcopy()
    for i in range(1, padvideo.data.shape[0]):
        try:
            absshiftx = drift[drift.index <= i].x.values[-1]
            absshifty = drift[drift.index <= i].y.values[-1]
        except:
            absshiftx = 0
            absshifty = 0
        shiftx = np.int(np.round(absshiftx))
        shifty = np.int(np.round(absshifty))
        outvideo.inav[i].data[:] = shift_2d_replace(padvideo.inav[i].data[:], -shiftx, -shifty)
    return outvideo



# prefs

import warnings
warnings.filterwarnings("ignore")

plt.close('all')
gc.collect()

prevsettings = pd.read_csv(r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\5_input_settings_lattice_drift.csv")
prevsettings = prevsettings.set_index('Unnamed: 0')
loadfolder_lattice = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\3_latticefilter" + os.sep
loadfolder_vid = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\4_atomfilter" + os.sep
savefolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\5_driftcorr" + os.sep
origfolder = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\2_ALT_PCAdenoise" + os.sep

vidnum = 80

settings = prevsettings.loc[vidnum]

savefolder = savefolder_base + settings.filename + os.sep
    
os.makedirs(savefolder, exist_ok=True)

vidinfn = loadfolder_vid + settings.filename + os.sep + 'compoundvid.hspy'
vid = hs.load(vidinfn)
vid = vid - vid.data.min()

lat_vid = hs.load(loadfolder_lattice + settings.filename + os.sep + 'filtvid.hspy')
contvid = hs.load(loadfolder_lattice + settings.filename + os.sep + 'inv_diff_filtvid.hspy')


length = vid.data.shape[0]


        


loadorig = origfolder + settings.filename + os.sep +'input_vid_cropped.hspy'
origvid=hs.load(loadorig)
origvid = origvid - origvid.data.min()


xcorr_dc = True
settings.xcorr_dc = xcorr_dc

if xcorr_dc == True:

    vid_dat = vid.data
    overlaps = []
    errs = []
    for i in tqdm(range(1, length)):
        max_overlap, corr_err, corr_phase = phase_cross_correlation(vid_dat[i-1], vid_dat[i], overlap_ratio=0.1)
        overlaps.append(max_overlap)
        errs.append(corr_err)
    rel_drift = np.array(overlaps)
    abs_drift = pd.DataFrame(np.cumsum(rel_drift, axis=0), columns=['y', 'x'])
    abs_drift['frame'] = abs_drift.index + 1
    abs_drift = abs_drift.set_index('frame', drop=True)
    
    fig, ax = plt.subplots(2)
    ax[0].plot(abs_drift.x, label='x-drift')
    ax[0].plot(abs_drift.y, label='y-drift')
    ax[0].legend()
    ax[1].plot(errs, label='correlation error')
    ax[1].set_ylim(0, ax[1].get_ylim()[1])
    ax[1].legend()
    fig.suptitle('Initial Xcorr drift detection')
    fig.savefig(savefolder + 'initial_guess.png')
    
    
            
    vid_drift1 = drift_shifts(vid, -abs_drift)
    lat_vid_drift1 = drift_shifts(lat_vid, -abs_drift)
    contvid_drift1 = drift_shifts(contvid, -abs_drift)
    orig_drift1 = drift_shifts(origvid, -abs_drift)
    
else:
    vid_drift1 = vid
    lat_vid_drift1 = lat_vid
    contvid_drift1 = contvid
    orig_drift1 = origvid
          
    
vid_drift1.plot()








# plot histograms
contvid_dat = np.sort(contvid.data.flatten(), kind='mergesort')
bins = np.linspace(contvid_dat.min(), contvid_dat.max(), num=501)
hist, bin_edges = np.histogram(contvid_dat, bins=bins, density=True)
bin_centres = (bin_edges[:-1] + bin_edges[1:]) / 2

fig, (ax1, ax2) = plt.subplots(1, 2)
ax1.plot(bin_centres, hist)

for i in range(0, length):
    framedat = contvid.inav[i].data[:].flatten()
    hist, bin_edges = np.histogram(framedat, bins=bins, density=True)
    ax2.plot(bin_centres, hist, alpha=0.1)
fig.savefig(savefolder + 'contvid_intensity_hist.png')





























contour_rank_disk = 3
contour_erosion_disk = 6
contour_dilation_disk = 9
contour_thresh = 1.2e7

settings['contour_rank_disk'] = contour_rank_disk
settings['contour_erosion_disk'] = contour_erosion_disk
settings['contour_dilation_disk'] = contour_dilation_disk
settings['contour_thresh'] = contour_thresh



# rank filter test

footp = disk(contour_rank_disk)
ranks = range(0, footp.sum(), 2)

savefolder_rank_test = savefolder + 'rank_test' + os.sep
os.makedirs(savefolder_rank_test, exist_ok=True)

chooseframe = 80
framedat = gaussian_filter(vid_drift1.inav[chooseframe].data[:], 0)
for rank in tqdm(ranks):
    outim = rank_filter(framedat, rank, footprint=disk(contour_rank_disk))
    savename = 'frame_%d_disksize_%d_rank_%d.png' % (chooseframe, contour_rank_disk, rank)
    plt.imsave(savefolder_rank_test + savename, outim)

contour_rank=14
settings['contour_rank'] = contour_rank









# assign labels
contourplots = []
threshplots = []
for i in tqdm(range(0, length), desc='Calculating Contours'):
    framedat = gaussian_filter(vid_drift1.inav[i].data[:], 0)
    thresh_image = rank_filter(framedat, contour_rank, footprint = footp)
    binary_image = erosion(dilation(thresh_image > contour_thresh, disk(contour_erosion_disk)), disk(contour_dilation_disk))
    contourplots.append(binary_image)
    threshplots.append(thresh_image)




contourvid = hs.signals.Signal2D(np.asarray(contourplots), dtype=int)
threshvid = hs.signals.Signal2D(np.asarray(threshplots), dtype=int)



contourvid.plot(navigator=None)
threshvid.plot(navigator=None)












































tpsize = 5
minmass = 0.1
cutmass = 0.22
cutsize = 1.19

settings['drift_tpsize'] = tpsize
settings['drift_minmass'] = minmass
settings['drift_cutmass'] = cutmass
settings['drift_cutsize'] = cutsize











 





test = tp.locate(lat_vid_drift1.data[-1], 5, invert=False, minmass=0.1)
tp.annotate(test, lat_vid_drift1.data[-1])





pos_list = tp.batch(lat_vid_drift1.data, tpsize, invert=False, minmass=minmass)





grid = sns.JointGrid('mass', 'size', data=pos_list, height=10)
cut_rect = plt.Rectangle((cutmass, cutsize), 6000000, 1.1, alpha=0.1)
grid.ax_joint.set_xlim(pos_list.mass.min(), pos_list.mass.max())
grid.ax_joint.set_ylim(pos_list['size'].min(), pos_list['size'].max())
grid.ax_joint.add_patch(cut_rect)
grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.05)
grid.plot_marginals(sns.histplot, kde=False, color="b", bins=100)
plt.savefig(savefolder + 'all_points.png')


cut_pos_list = pos_list[(pos_list['mass'] > cutmass) & (pos_list['size'] > cutsize)]















index_pass_list = []
dist = 1
for fnum in tqdm(range(0, vid.data.shape[0]), desc='Removing contour bounded'):
    contframe = contourvid.inav[fnum].data[:]
    frame_pos_list = pos_list[pos_list['frame'] == fnum]
    for index, particle in frame_pos_list.iterrows():
        x = np.rint(particle.x).astype(int)
        y = np.rint(particle.y).astype(int)
        meanval = contframe[y - dist: y + dist + 1, x - dist: x + dist + 1].mean()
        if meanval == 1:
            index_pass_list.append(index)


cut_pos_list_2 = cut_pos_list[cut_pos_list.index.isin(index_pass_list)]
excluded_points = cut_pos_list[~cut_pos_list.index.isin(index_pass_list)]




savefolder_contvid = savefolder + 'contour_video/'
if not os.path.exists(savefolder_contvid):
    os.mkdir(savefolder_contvid)

for fnum in tqdm(range(0, vid.data.shape[0]), desc='outputting contour video frames'):
    plt.figure(figsize=(6, 6), dpi=300)
    points_list = cut_pos_list_2[cut_pos_list_2['frame'] == fnum]
    ex_points_list = excluded_points[excluded_points['frame'] == fnum]
    plt.gca().contour(contourvid.inav[fnum].data, [0.5,], color='red')
    plt.gca().imshow(vid_drift1.inav[fnum].data)
    plt.gca().scatter(ex_points_list.x, ex_points_list.y, s=40, marker='o', facecolor='none', edgecolor='red')
    plt.gca().scatter(points_list.x, points_list.y, s=10, marker='o', facecolor='none', edgecolor='blue')
    plt.axis('off')
    plt.tight_layout()
    plt.savefig(savefolder_contvid + 'vid%d.png' % (fnum), bbox_inches='tight')
    plt.close('all')







grid = sns.JointGrid('mass', 'size', data=cut_pos_list, height=10)
grid.ax_joint.set_xlim(cut_pos_list.mass.min(), cut_pos_list.mass.max())
grid.ax_joint.set_ylim(cut_pos_list['size'].min(), cut_pos_list['size'].max())
grid = grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.05)
grid.plot_marginals(sns.histplot, kde=False, color="b", bins=100)
plt.savefig(savefolder + 'all_points_cut.png')















































trajsearch = 6
settings['drift_trajsearch'] = trajsearch
trajmem = 2
settings['drift_trajmem'] = trajmem
trajstubs = 3
settings['drift_trajstubs'] = trajstubs





trajectories = tp.link_df(cut_pos_list_2, trajsearch, memory=trajmem)
trajectories = tp.filter_stubs(trajectories, threshold=trajstubs)
trajectories = trajectories.reset_index(drop=True)



meanbinnedparticle_trajectories = trajectories.copy().groupby('particle').mean()
trajectories = trajectories.reset_index(drop=True)






cutsize2 = [1.235, 1.255]
cutmass2 = [0.28, 0.41]

settings['drift_cutsize2'] = cutsize2
settings['drift_cutmass2'] = cutmass2

plt.figure()
plt.scatter(meanbinnedparticle_trajectories['mass'], meanbinnedparticle_trajectories['size'], color='blue', alpha=0.5);


teajcut_rect = Rectangle((cutmass2[0], cutsize2[0]), cutmass2[1] - cutmass2[0], cutsize2[1] - cutsize2[0], color='blue', alpha=0.2)
plt.gca().add_patch(teajcut_rect)


checklomass = cutmass2[0] < meanbinnedparticle_trajectories['mass']
checkhimass = meanbinnedparticle_trajectories['mass'] < cutmass2[1]
checklosize = cutsize2[0] < meanbinnedparticle_trajectories['size']
checkhisize = meanbinnedparticle_trajectories['size'] < cutsize2[1]
meanbinnedparticle_trajectories_cut = meanbinnedparticle_trajectories[checklomass & checkhimass & checklosize & checkhisize]

plt.scatter(meanbinnedparticle_trajectories_cut['mass'], meanbinnedparticle_trajectories_cut['size'], color='red', alpha=0.8)

plt.gca().set_xlabel('size (px)')
plt.gca().set_ylabel('mass')
plt.suptitle('Thresholding lifetime averaged particles (dc)')
plt.savefig(savefolder + 'lifetime_averaged_thresholds(dc).png')



trajectories_cut = trajectories[trajectories.particle.isin(list(meanbinnedparticle_trajectories_cut.index))]






savefolder_traj1 = savefolder + 'traj1' + os.sep
os.makedirs(savefolder_traj1, exist_ok=True)


for i in tqdm(range(0, lat_vid.data.shape[0]), desc='outputting video frames'):
    scalebar = ScaleBar(origvid.axes_manager['x'].scale, 'nm', box_alpha=0, color='white')
    fig = plt.figure(figsize=((vid_drift1.data.shape[2]/100.0) * 4, (vid_drift1.data.shape[1]/100.0) * 4), dpi=100)
    ax = fig.add_axes([0, 0, 1, 1])
    frame_traj = trajectories_cut[trajectories_cut['frame'] <= i]
    ax.imshow(lat_vid_drift1.inav[i].data, cmap='inferno')
    tp.plot_traj(frame_traj, ax=ax)    
    ax.add_artist(scalebar)
    fig.savefig(savefolder_traj1 + 'latvid%d.png' % (i), bbox_inches='tight')
    plt.close('all')
















drift_2 = tp.compute_drift(trajectories_cut)
drift_2.plot()
plt.savefig(savefolder + 'lattice_drift(dc).png')


drift_2.to_csv(savefolder + 'lattice_extracted drift.csv')


        
vid_drift2 = drift_shifts(vid_drift1, drift_2)
lat_vid_drift2 = drift_shifts(lat_vid_drift1, drift_2)
contvid_drift2 = drift_shifts(contvid_drift1, drift_2)
orig_drift2 = drift_shifts(orig_drift1, drift_2)
   
























pos_list = tp.batch(lat_vid_drift2.data, tpsize, invert=False, minmass=minmass)
cut_pos_list = pos_list[(pos_list['mass'] > cutmass) & (pos_list['size'] > cutsize)]

index_pass_list = []
dist = 1
for fnum in tqdm(range(0, vid.data.shape[0]), desc='Removing contour bounded'):
    contframe = contourvid.inav[fnum].data[:]
    frame_pos_list = pos_list[pos_list['frame'] == fnum]
    for index, particle in frame_pos_list.iterrows():
        x = np.rint(particle.x).astype(int)
        y = np.rint(particle.y).astype(int)
        meanval = contframe[y - dist: y + dist + 1, x - dist: x + dist + 1].mean()
        if meanval == 1:
            index_pass_list.append(index)


cut_pos_list_2 = cut_pos_list[cut_pos_list.index.isin(index_pass_list)]
excluded_points = cut_pos_list[~cut_pos_list.index.isin(index_pass_list)]

trajectories = tp.link_df(cut_pos_list_2, trajsearch, memory=trajmem)
trajectories = tp.filter_stubs(trajectories, threshold=trajstubs)
trajectories = trajectories.reset_index(drop=True)

meanbinnedparticle_trajectories = trajectories.copy().groupby('particle').mean()
trajectories = trajectories.reset_index(drop=True)

checklomass = cutmass2[0] < meanbinnedparticle_trajectories['mass']
checkhimass = meanbinnedparticle_trajectories['mass'] < cutmass2[1]
checklosize = cutsize2[0] < meanbinnedparticle_trajectories['size']
checkhisize = meanbinnedparticle_trajectories['size'] < cutsize2[1]
meanbinnedparticle_trajectories_cut = meanbinnedparticle_trajectories[checklomass & checkhimass & checklosize & checkhisize]

trajectories_cut = trajectories[trajectories.particle.isin(list(meanbinnedparticle_trajectories_cut.index))]







savefolder_traj2 = savefolder + 'traj2' + os.sep
os.makedirs(savefolder_traj2, exist_ok=True)


for i in tqdm(range(0, lat_vid_drift2.data.shape[0]), desc='outputting video frames'):
    scalebar = ScaleBar(origvid.axes_manager['x'].scale, 'nm', box_alpha=0, color='white')
    fig = plt.figure(figsize=((vid_drift2.data.shape[2]/100.0) * 4, (vid_drift2.data.shape[1]/100.0) * 4), dpi=100)
    ax = fig.add_axes([0, 0, 1, 1])
    frame_traj = trajectories_cut[trajectories_cut['frame'] <= i]
    ax.imshow(lat_vid_drift2.inav[i].data, cmap='inferno')
    tp.plot_traj(frame_traj, ax=ax)    
    ax.add_artist(scalebar)
    fig.savefig(savefolder_traj2 + 'latvid%d.png' % (i), bbox_inches='tight')
    plt.close('all')














drift_3 = tp.compute_drift(trajectories_cut)
drift_3.plot()
plt.savefig(savefolder + 'lattice_drift(dc)_2ndpass.png')


drift_3.to_csv(savefolder + 'lattice_extracted drift_2ndpass.csv')


        
vid_drift3 = drift_shifts(vid_drift2, drift_3)
lat_vid_drift3 = drift_shifts(lat_vid_drift2, drift_3)
contvid_drift3 = drift_shifts(contvid_drift2, drift_3)
orig_drift3 = drift_shifts(orig_drift2, drift_3)
   

























vid_drift3.save(savefolder + 'filtered_video.hspy')
lat_vid_drift3.save(savefolder + 'lattice_video.hspy')
contvid_drift3.save(savefolder + 'contour_filtered_video.hspy')
orig_drift3.save(savefolder + 'original_video.hspy')

settings.to_csv(savefolder + 'settings_out.csv')





