import numpy as np
import os
import hyperspy.api as hs
import trackpy as tp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import gc

from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar
from tqdm import tqdm

import matplotlib
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42


import warnings
warnings.filterwarnings("ignore")

plt.close('all')
gc.collect()

vidnum=80
prevsettings = pd.read_csv(r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\5_input_settings_lattice_drift.csv")
prevsettings = prevsettings.set_index('Unnamed: 0')
settings = prevsettings.loc[vidnum]

loadfolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\5_driftcorr" + os.sep
savefolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\6_latticepoints" + os.sep

loadfolder = loadfolder_base + settings.filename + os.sep
savefolder = savefolder_base + settings.filename + os.sep

for folder in [loadfolder, savefolder]:
    os.makedirs(folder, exist_ok=True)
    


tm_vid = hs.load(loadfolder + 'lattice_video.hspy')









tpsize = 5
minmass = 0.01
cutmass = 0.22
cutsize = 1.2

settings['lattice_tpsize'] = tpsize
settings['lattice_minmass'] = minmass
settings['lattice_cutmass'] = cutmass
settings['lattice_cutsize'] = cutsize





test = tp.locate(tm_vid.data[-1], 5, invert=False, minmass=0.01)
tp.annotate(test, tm_vid.data[-1])






pos_list = tp.batch(tm_vid.data, tpsize, invert=False, minmass=minmass)

grid = sns.JointGrid('mass', 'size', data=pos_list, height=10)
cut_rect = plt.Rectangle((cutmass, cutsize), 6000000, 1.1, alpha=0.1)
grid.ax_joint.set_xlim(pos_list.mass.min(), pos_list.mass.max())
grid.ax_joint.set_ylim(pos_list['size'].min(), pos_list['size'].max())
grid.ax_joint.add_patch(cut_rect)
grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.05)
grid.plot_marginals(sns.histplot, color="b", bins=100)
plt.savefig(savefolder + 'all_points.png')


cut_pos_list = pos_list[(pos_list['mass'] > cutmass) & (pos_list['size'] > cutsize)]

cut_frame_pos_lists = []
for fnum in range(0, tm_vid.data.shape[0]):
    frame_pos_list = cut_pos_list[cut_pos_list.frame == fnum]
    maxx = frame_pos_list.x.max()
    maxy = frame_pos_list.y.max()
    xcheck = frame_pos_list.x < maxx - 2
    ycheck = frame_pos_list.y < maxy - 2
    cut_frame_pos_list = frame_pos_list[xcheck & ycheck]
    cut_frame_pos_lists.append(cut_frame_pos_list)

cut_pos_list = pd.concat(cut_frame_pos_lists)




grid = sns.JointGrid('mass', 'size', data=cut_pos_list, height=10)
grid.ax_joint.set_xlim(cut_pos_list.mass.min(), cut_pos_list.mass.max())
grid.ax_joint.set_ylim(cut_pos_list['size'].min(), cut_pos_list['size'].max())
grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.05)
grid.plot_marginals(sns.histplot, color="b", bins=100)
plt.savefig(savefolder + 'all_points_cut.png')








trajsearch = 6
settings['lattice_trajsearch'] = trajsearch
trajmem = 3
settings['lattice_trajmem'] = trajmem
trajstubs = 4
settings['lattice_trajstubs'] = trajstubs






trajectories = tp.link_df(cut_pos_list, trajsearch, memory=trajmem)
trajectories = tp.filter_stubs(trajectories, threshold=trajstubs)





cutsize2 = [1.23, 1.255]
cutmass2 = [0.26, 0.41]

settings['lattice_cutsize2'] = cutsize2
settings['lattice_cutmass2'] = cutmass2



meanbinnedparticle_trajectories = trajectories.groupby('particle').mean()

plt.figure()
plt.scatter(meanbinnedparticle_trajectories['mass'], meanbinnedparticle_trajectories['size'], color='blue', alpha=0.5);


teajcut_rect = Rectangle((cutmass2[0], cutsize2[0]), cutmass2[1] - cutmass2[0], cutsize2[1] - cutsize2[0], color='blue', alpha=0.2)
plt.gca().add_patch(teajcut_rect)


checklomass = cutmass2[0] < meanbinnedparticle_trajectories['mass']
checkhimass = meanbinnedparticle_trajectories['mass'] < cutmass2[1]
checklosize = cutsize2[0] < meanbinnedparticle_trajectories['size']
checkhisize = meanbinnedparticle_trajectories['size'] < cutsize2[1]
meanbinnedparticle_trajectories_cut = meanbinnedparticle_trajectories[checklomass & checkhimass & checklosize & checkhisize]

plt.scatter(meanbinnedparticle_trajectories_cut['mass'], meanbinnedparticle_trajectories_cut['size'], color='red', alpha=0.8)

plt.gca().set_xlabel('size (px)')
plt.gca().set_ylabel('mass')
plt.suptitle('Thresholding lifetime averaged particles (dc)')
plt.savefig(savefolder + 'lifetime_averaged_thresholds(dc).png')


trajectories_cut = trajectories[trajectories.particle.isin(list(meanbinnedparticle_trajectories_cut.index))]



drift = tp.compute_drift(trajectories_cut)
drift.plot()
plt.savefig(savefolder + 'lattice_drift(dc).png')




vid = hs.load(loadfolder + 'original_video.hspy')
vid_cut = [0.01, 99.9]
vid_dat = np.sort(vid.data.flatten(), kind='mergesort')
vid_cut_inds = np.asarray(vid_cut) * (len(vid_dat) / 100)
vid_cut_vals = [vid_dat[int(vid_cut_inds[0])], vid_dat[int(vid_cut_inds[1])]]
del vid_dat, vid_cut_inds, vid_cut

tmvid_cut = [0.01, 99.9]
tmvid_dat = np.sort(tm_vid.data.flatten(), kind='mergesort')
tmvid_cut_inds = np.asarray(tmvid_cut) * (len(tmvid_dat) / 100)
tmvid_cut_vals = [tmvid_dat[int(tmvid_cut_inds[0])], tmvid_dat[int(tmvid_cut_inds[1])]]
del tmvid_dat, tmvid_cut_inds, tmvid_cut

vidscale = vid.axes_manager[1].scale


for i in tqdm(range(0, tm_vid.data.shape[0]), desc='outputting video frames'):
    scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
    fig = plt.figure(figsize=((vid.data.shape[2]/100.0) * 4, (vid.data.shape[1]/100.0) * 4), dpi=100)
    frame_traj = trajectories_cut[trajectories_cut['frame'] == i]
    # tp.plot_traj(trajectories_cut[(trajectories_cut['frame'] <= i)])
    # for index, particle in trajectories_cut[(trajectories_cut['frame'] == i)].iterrows():
    #     plt.gca().scatter(particle.x, particle.y, s=50, marker='o', facecolor='none', edgecolor='blue')
    plt.gca().imshow(vid.inav[i].data, vmin=vid_cut_vals[0], vmax=vid_cut_vals[1], cmap='inferno')
    plt.gca().scatter(frame_traj.x, frame_traj.y, s=50, marker='o', facecolor='none', edgecolor='blue')
    plt.axis('off')
    plt.gca().add_artist(scalebar)
    plt.savefig(savefolder + 'vid%d.png' % (i), bbox_inches='tight')
    plt.close('all')
    
    scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
    fig = plt.figure(figsize=((vid.data.shape[2]/100.0) * 4, (vid.data.shape[1]/100.0) * 4), dpi=100)
    frame_traj = trajectories_cut[trajectories_cut['frame'] == i]
    # tp.plot_traj(trajectories_cut[(trajectories_cut['frame'] <= i)])
    # for index, particle in trajectories_cut[(trajectories_cut['frame'] == i)].iterrows():
    #     plt.gca().scatter(particle.x, particle.y, s=50, marker='o', facecolor='none', edgecolor='blue')
    plt.gca().imshow(tm_vid.inav[i].data, vmin=tmvid_cut_vals[0], vmax=tmvid_cut_vals[1], cmap='inferno')
    plt.gca().scatter(frame_traj.x, frame_traj.y, s=50, marker='o', facecolor='none', edgecolor='blue')
    plt.axis('off')
    plt.gca().add_artist(scalebar)
    plt.savefig(savefolder + 'latvid%d.png' % (i), bbox_inches='tight')
    plt.close('all')

    grid = sns.JointGrid('mass', 'size', data=trajectories_cut[(trajectories_cut['frame'] == i)], height=10)
    grid.ax_joint.set_xlim(trajectories_cut['mass'].min(), trajectories_cut['mass'].max())
    grid.ax_joint.set_ylim(trajectories_cut['size'].min(), trajectories_cut['size'].max())
    grid.plot_joint(plt.scatter, c='blue', linewidth=0.5, marker='o', alpha=0.2)
    grid.plot_marginals(sns.histplot, color="b", bins=16)
    plt.savefig(savefolder + 'mass_size%d.png' % (i))
    plt.close('all')






trajectories = trajectories.reset_index(drop=True)

trajectories.to_csv(savefolder + 'trajectories.csv')

drift = tp.compute_drift(trajectories)
drift.plot()
plt.savefig(savefolder + 'driftcheck.png')

settings.to_csv(savefolder + 'settings_out.csv')
