import numpy as np
import os
import gc
import hyperspy.api as hs
import trackpy as tp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from scipy.signal import wiener
from scipy.ndimage import median_filter
from sklearn.cluster import KMeans
from glob import glob


def linesvisible(lines, num):
    for i in range(0,4):
        lines[i]._visible = False
    lines[num]._visible = True


def get_specdots(mo_roi, specticks, xmax):
    mo_vec = np.array([mo_roi.x2 - mo_roi.x1 ,mo_roi.y2 - mo_roi.y1])
    mo_vec_new = np.array([mo_vec * i for i in specticks/xmax])
    startpoint = np.array([mo_roi.x1, mo_roi.y1])
    mo_vec_out = np.array([startpoint + point for point in mo_vec_new])
    return mo_vec_out

def framesplit(vid):
    framelist = []
    for fnum in range(vid.data.shape[0]):
        frame = vid.inav[fnum].deepcopy()
        framelist.append(frame)
    return framelist

def framecombine(framelist):
    signalout = hs.stack(framelist)#
    return signalout

def gaussfilter(framelist, size=1):
    outlist = []
    for frame in framelist:
        newframe = frame.deepcopy()
        newframe.data =  gaussian_filter(frame.data, size)
        outlist.append(newframe)
    return outlist

def wienerfilter(framelist, size=3):
    outlist = []
    for frame in framelist:
        newframe = frame.deepcopy()
        newframe.data =  wiener(frame.data, size)
        outlist.append(newframe)
    return outlist

def medfilter(framelist, size=3):
    outlist = []
    for frame in framelist:
        newframe = frame.deepcopy()
        newframe.data =  median_filter(frame.data, size)
        outlist.append(newframe)
    return outlist

def points_2vec(v0, v1):
    outpoints = []
    outpoints.append(v0)
    outpoints.append(v1)
    outpoints.append(v1 - v0)
    outpoints.append(- v0)
    outpoints.append(- v1)
    outpoints.append(v0 - v1)
    return np.array(outpoints)

def points_2vec_residuals(calcvectors, lv_pos0, lv_pos1, lv_pos2, lv_pos3, lv_pos4, lv_pos5):
    lv_pos = np.array([lv_pos0, lv_pos1, lv_pos2, lv_pos3, lv_pos4, lv_pos5])
    v0 = np.array([calcvectors[0], calcvectors[1]])
    v1 = np.array([calcvectors[2], calcvectors[3]])
    calcpoints = points_2vec(v0, v1)
    dists = []
    for i in range(0,6):
        dist = (((calcpoints[i] - lv_pos[i])**2).sum(axis=0))**0.5
        dists.append(dist)
    distarr = np.array(dists)
    RMS_dist = (((distarr**2).sum())/3)**0.5
    return RMS_dist

def site_intensity_score(trajectories, video):
    dat = video.data
    scores = []
    for index, row in tqdm(trajectories.iterrows(), total=trajectories.shape[0]):
        frame = int(row.frame)
        x = int(round(row.x))
        y = int(round(row.y))
        try:
            score = dat[frame, y, x]
            scores.append(score)
        except:
            pass
    avscore = np.mean(scores)
    return avscore

from scipy.optimize import least_squares
from matplotlib.patches import Rectangle
from matplotlib_scalebar.scalebar import ScaleBar
from tqdm import tqdm
from matplotlib.patches import Circle
from scipy.ndimage import gaussian_filter

plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42
#%matplotlib

pd.options.mode.chained_assignment = None 

metafolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20\20200914_OldPtliquidcell"
loadfolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual" + os.sep
savefolder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual\8_points_plot" + os.sep

settings_fname_list = glob(loadfolder_base + '7_adatompoints' + os.sep + '*' + os.sep + 'settings_out.csv')


for settings_fname in settings_fname_list:

    settings = pd.read_csv(settings_fname, index_col=0, squeeze=True)

    
    vidnum = int(settings.name)

    vidfolder = loadfolder_base + '5_driftcorr' + os.sep + settings.filename + os.sep
    latticefolder = loadfolder_base + '6_latticepoints' + os.sep + settings.filename + os.sep
    adatomfolder = loadfolder_base + '7_adatompoints' + os.sep + settings.filename + os.sep
    savefolder = savefolder_base + settings.filename + os.sep
    
    if not os.path.exists(savefolder):
        os.makedirs(savefolder, exist_ok=True)
        
        if settings.orig_filename != '/Scanning Preview13b.emi':
            metavid = hs.load(metafolder_base + settings.orig_filename)
        else:
            metavid = hs.load(metafolder_base + '/Scanning Preview13.emi')
            
        
        
        vid = hs.load(vidfolder + 'original_video.hspy')
        vid_cut = [0.2, 98]
        vid_dat = np.sort(vid.data.flatten(), kind='mergesort')
        vid_cut_inds = np.asarray(vid_cut) * (len(vid_dat) / 100)
        vid_cut_vals = [vid_dat[int(vid_cut_inds[0])], vid_dat[int(vid_cut_inds[1])]]
        del vid_dat, vid_cut_inds, vid_cut
        vidscale = vid.axes_manager[1].scale
        
        advid = hs.load(vidfolder + 'filtered_video.hspy')
        
        
        latvid = hs.load(vidfolder + 'lattice_video.hspy')
        latvid_cut = [0.1, 99.9]
        latvid_dat = np.sort(latvid.data.flatten(), kind='mergesort')
        latvid_cut_inds = np.asarray(latvid_cut) * (len(latvid_dat) / 100)
        latvid_cut_vals = [latvid_dat[int(latvid_cut_inds[0])], latvid_dat[int(latvid_cut_inds[1])]]
        del latvid_dat, latvid_cut_inds, latvid_cut
        latvidscale = latvid.axes_manager[1].scale
        
        origvid = hs.load(vidfolder + 'original_video.hspy')
        vidscale = origvid.axes_manager[1].scale
        
        trajectories_lattice = pd.read_csv(latticefolder + 'trajectories.csv')
        trajectories_adatom = pd.read_csv(adatomfolder + 'trajectories.csv')
        
        # savefolder_vid1 = savefolder + 'vid_initial/'
        # if not os.path.exists(savefolder_vid1):
        #     os.mkdir(savefolder_vid1)
        
        # for i in tqdm(range(0, vid.data.shape[0]), desc='outputting video frames'):
        #     scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
        #     fig = plt.figure(figsize=((vid.data.shape[2]/100.0) * 3, (vid.data.shape[1]/100.0) * 3), dpi=100)
        #     frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
        #     frame_ad = trajectories_adatom[trajectories_adatom['frame'] == i]
        #     ax = fig.add_axes([0, 0, 1, 1])
        #     tp.plot_traj(trajectories_adatom[(trajectories_adatom['frame'] <= i)], ax=ax)
        #     ax.imshow(vid.inav[i].data, vmin=vid_cut_vals[0], vmax=vid_cut_vals[1] * 1.5, cmap='inferno')
        #     ax.scatter(frame_ad.x, frame_ad.y, s=20, marker='o', facecolor='none', edgecolor='red')
        #     ax.scatter(frame_lat.x, frame_lat.y, s=8, marker='.', facecolor='blue', edgecolor='none')
        #     ax.add_artist(scalebar)
        #     plt.savefig(savefolder_vid1 + 'vid%d.png' % (i), dpi=100)
        #     plt.close('all')
            
            
        # savefolder_vidlat1 = savefolder + 'lattice_vid_initial/'
        # if not os.path.exists(savefolder_vidlat1):
        #     os.mkdir(savefolder_vidlat1)
        
        # for i in tqdm(range(0, vid.data.shape[0]), desc='outputting video frames'):
        #     scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
        #     fig = plt.figure(figsize=((vid.data.shape[2]/100.0) * 3, (vid.data.shape[1]/100.0) * 3), dpi=100)
        #     frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
        #     frame_ad = trajectories_adatom[trajectories_adatom['frame'] == i]
        #     ax = fig.add_axes([0, 0, 1, 1])
        #     tp.plot_traj(trajectories_lattice[(trajectories_lattice['frame'] <= i)], ax=ax)
        #     ax.imshow(latvid.inav[i].data, vmin=latvid_cut_vals[0], vmax=latvid_cut_vals[1] * 1.5, cmap='inferno')
        #     ax.scatter(frame_ad.x, frame_ad.y, s=20, marker='o', facecolor='none', edgecolor='red')
        #     ax.scatter(frame_lat.x, frame_lat.y, s=8, marker='.', facecolor='blue', edgecolor='none')
        #     ax.add_artist(scalebar)
        #     plt.savefig(savefolder_vidlat1 + 'vid%d.png' % (i), dpi=100)
        #     plt.close('all')   
            
            
        
        fakevid_str = [2, 0.5, 0]
        settings['fakevid_str'] = fakevid_str
        
        try:
            fakevid = (advid/advid.data.max()) * fakevid_str[0] + (latvid/latvid.data.max()) * fakevid_str[1] + (vid/vid.data.max()) * fakevid_str[2]
        except:
            vid = vid.deepcopy().inav[:advid.data.shape[0]]
            advid = advid.deepcopy().inav[:]
            latvid = latvid.deepcopy().inav[:]
            fakevid = (advid/advid.data.max()) * fakevid_str[0] + (latvid/latvid.data.max()) * fakevid_str[1] + (vid/vid.data.max()) * fakevid_str[2]
        
        fakevid.plot()
        
        fakevid_cut = [0.1, 99.9]
        fakevid_dat = np.sort(fakevid.data.flatten(), kind='mergesort')
        fakevid_cut_inds = np.asarray(fakevid_cut) * (len(fakevid_dat) / 100)
        fakevid_cut_vals = [fakevid_dat[int(fakevid_cut_inds[0])], fakevid_dat[int(fakevid_cut_inds[1])]]
        del fakevid_dat, fakevid_cut_inds, fakevid_cut
        fakevidscale = fakevid.axes_manager[1].scale
        
        
        
        savefolder_vidfake1 = savefolder + 'combined_vid_initial/'
        if not os.path.exists(savefolder_vidfake1):
            os.mkdir(savefolder_vidfake1)
        
        for i in tqdm(range(0, vid.data.shape[0]), desc='outputting video frames'):
            scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
            fig = plt.figure(figsize=((vid.data.shape[2]/100.0) * 3, (vid.data.shape[1]/100.0) * 3), dpi=100)
            frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
            frame_ad = trajectories_adatom[trajectories_adatom['frame'] == i]
            ax = fig.add_axes([0, 0, 1, 1])
            traj_frameplot = trajectories_adatom[(trajectories_adatom['frame'] <= i)]
            if len(traj_frameplot) > 0:
                tp.plot_traj(traj_frameplot, ax=ax)
            ax.imshow(fakevid.inav[i].data, vmin=fakevid_cut_vals[0], vmax=fakevid_cut_vals[1] * 1.5, cmap='inferno')
            ax.scatter(frame_ad.x, frame_ad.y, s=20, marker='o', facecolor='none', edgecolor='red')
            ax.scatter(frame_lat.x, frame_lat.y, s=8, marker='.', facecolor='blue', edgecolor='none')
            ax.add_artist(scalebar)
            plt.savefig(savefolder_vidfake1 + 'vid%d.png' % (i), dpi=100)
            plt.close('all') 
        
        
        
        
        
        
        
        
        
        rr_ref_frames = 25
        settings['rr_ref_frames'] = rr_ref_frames
        
        pad = lambda x: np.hstack([x, np.ones((x.shape[0], 1))])
        unpad = lambda x: x[:,:-1]
        
        savefolder_vid2 = savefolder + 'vid_RR/'
        if not os.path.exists(savefolder_vid2):
            os.mkdir(savefolder_vid2)
        
        res_list = []
        num_list = []
        Tmat_list = []
        for fnum in tqdm(range(vid.data.shape[0]), desc='rigid registration'):
            pos_cur_list = trajectories_lattice[trajectories_lattice.frame == fnum]
            pos_las_list = trajectories_lattice[trajectories_lattice.frame <= rr_ref_frames].groupby('particle').mean()
            pos_las_list['particle'] = pos_las_list.index
            p_list = np.array(list(set(pos_cur_list.particle.values).intersection(pos_las_list.particle.values)))
            if len(p_list) > 100:
                pointlist_previous = []
                pointlist_current = []
                for pnum in p_list:
                    pos_cur = pos_cur_list[pos_cur_list.particle == pnum]
                    pos_las = pos_las_list[pos_las_list.particle == pnum]
                    point_cur = pos_cur.x.values[0], pos_cur.y.values[0]
                    point_las = pos_las.x.values[0], pos_las.y.values[0]
                    pointlist_current.append(point_cur)
                    pointlist_previous.append(point_las)
                pointlist_previous = np.array(pointlist_previous)
                pointlist_current = np.array(pointlist_current)
                X = pad(pointlist_current)
                Y = pad(pointlist_previous)
                Tmat, res, rank, s = np.linalg.lstsq(X,Y)
                
                pointlist_transformed = unpad(np.dot(pad(pointlist_current), Tmat))
                pointlist_shift = pointlist_transformed - pointlist_previous
                res_list.append((((res[:-1]**2).sum())**0.5) / len(p_list))
        
                fig = plt.figure(figsize=(12,12))
                ax = fig.add_axes([0, 0, 1, 1], aspect='equal')
                ax.scatter(pointlist_previous[:,0], pointlist_previous[:,1], color='red', s=10)
                ax.scatter(pointlist_current[:,0], pointlist_current[:,1], color='blue', s=10)
                ax.scatter(pointlist_transformed[:,0], pointlist_transformed[:,1], color='green', s=5)
                ax.set_xlim(0, vid.data.shape[1])
                ax.set_ylim(0, vid.data.shape[2])
                fig.savefig(savefolder_vid2 + 'shifts_%d.png' % fnum)
                
            else:
                Tmat = np.array([[1, 0, 0],
                                 [0, 1, 0],
                                 [0, 0, 1]],
                                 dtype=float)
                res_list.append(np.nan)
                
            num_list.append(len(p_list))
            Tmat_list.append(Tmat)
            plt.close('all')
            gc.collect()
        
        np.save(savefolder + 'transformation_matrix_list.npy', Tmat_list)
        
        
        old_latpoints = pad(np.vstack([trajectories_lattice.x.values, trajectories_lattice.y.values]).transpose())
        framekey = trajectories_lattice.frame.values
        new_latpoints = np.zeros(old_latpoints.shape, dtype=float)
        for i in tqdm(range(len(trajectories_lattice))):
            old_latpoint = old_latpoints[i]
            Tmat = Tmat_list[framekey[i]]
            new_latpoints[i] = np.dot(old_latpoint, Tmat)
        
        
        trajectories_lattice['old_x'] = trajectories_lattice['x']
        trajectories_lattice['old_y'] = trajectories_lattice['y']
        trajectories_lattice['x'] = new_latpoints[:, 0]
        trajectories_lattice['y'] = new_latpoints[:, 1]
        
        
        
        
        old_adpoints = pad(np.vstack([trajectories_adatom.x.values, trajectories_adatom.y.values]).transpose())
        framekey = trajectories_adatom.frame.values
        new_adpoints = np.zeros(old_adpoints.shape, dtype=float)
        for i in tqdm(range(len(trajectories_adatom))):
            old_adpoint = old_adpoints[i]
            Tmat = Tmat_list[framekey[i]]
            new_adpoints[i] = np.dot(old_adpoint, Tmat)
        
        trajectories_adatom['old_x'] = trajectories_adatom['x']
        trajectories_adatom['old_y'] = trajectories_adatom['y']
        trajectories_adatom['x'] = new_adpoints[:, 0]
        trajectories_adatom['y'] = new_adpoints[:, 1]
        
        
        vid_T = vid.deepcopy()
        lvid_T = latvid.deepcopy()
        
        
        
        
        timescale = float(metavid.original_metadata.ObjectInfo.AcquireInfo.FrameTime)
        dosescale = 5183831.629
        
        
        savefolder_vid3 = savefolder + 'vid3/'
        if not os.path.exists(savefolder_vid3):
            os.mkdir(savefolder_vid3)
        
        plt.close('all')
        gc.collect()
        
        p_list = [1,]
        
        from skimage.transform import AffineTransform, warp
        
        
        for fnum in tqdm(range(0, vid.data.shape[0]), desc='multiframe output'):
            fig = plt.figure(figsize=(10, 6), dpi=150)
            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)
            ax1.axis('off')
            ax2.axis('off')
        
            tform = AffineTransform(matrix = Tmat_list[fnum].transpose())
            vid_transform = warp(vid.deepcopy().inav[fnum].data, tform, preserve_range=True)
            latvid_transform = warp(latvid.deepcopy().inav[fnum].data, tform, preserve_range=True)
        
        
            vid_T.data[fnum] = vid_transform
            lvid_T.data[fnum] = latvid_transform
        
        
        
            ax1.imshow(vid_transform, cmap='inferno', vmin=vid_cut_vals[0], vmax=vid_cut_vals[1])
            ax2.imshow(latvid_transform, cmap='inferno', vmin=latvid_cut_vals[0], vmax=latvid_cut_vals[1])
        
            scalebar1 = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
            scalebar2 = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
        
            ax1.add_artist(scalebar1)
            ax2.add_artist(scalebar2)
        
            ax1.annotate('Time: %d s' % (fnum * timescale), (8,580), color='white')
            # dosestr = 'Dose: %.1f $x 10^6e/nm^2$' % (fnum * dosescale)
            dosestr = 'Not Set'
            ax1.annotate(dosestr, (8, 500), color='white')
        
            fig.tight_layout()
            fig.savefig(savefolder_vid3 + 'vid%03d.png' % fnum)
            plt.close('all')
            gc.collect()
        
        
        
        vid_T.save(savefolder + 'transformed_vid_orig.hspy', overwrite=True)
        lvid_T.save(savefolder + 'transformed_latvid_orig.hspy', overwrite=True)
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        drift = tp.compute_drift(trajectories_adatom)
        firstpoint = pd.DataFrame({"x": 0.0, "y": 0.0}, index=[0])
        drift = drift.append(firstpoint, sort=False)
        drift = drift.sort_index()
        
        drift_lat = tp.compute_drift(trajectories_lattice)
        drift_lat = drift_lat.append(firstpoint, sort=False)
        drift_lat = drift_lat.sort_index()
        
        
        savefolder_drift = savefolder + 'sum_drift/'
        if not os.path.exists(savefolder_drift):
            os.mkdir(savefolder_drift)
        
        
        
        
        
        cmap = plt.cm.get_cmap('jet', 66)
        fig, (ax, ax_lat) = plt.subplots(1, 2, figsize=(8, 4))
        lim = np.ceil(np.max(np.abs(np.vstack([drift.values, drift_lat.values]))))
        ax.set_xlim(-lim, lim)
        ax_lat.set_xlim(-lim, lim)
        ax.set_ylim(-lim, lim)
        ax_lat.set_ylim(-lim, lim)
        ax.set_title('Adatoms (Mean)')
        ax_lat.set_title('Lattice (Mean)')
        for i in range(0, vid.data.shape[0]):
            ax.plot(drift.x[i: i + 2], -drift.y[i: i + 2], color=cmap(i))
            ax_lat.plot(drift_lat.x[i: i + 2], -drift_lat.y[i: i + 2], color=cmap(i))
            fig.savefig(savefolder_drift + 'frame_%d.png' % (i))
        
        
        
        
        
        
        
        
        
        fig, ax = plt.subplots(1)
        p_max = trajectories_adatom['particle'].max()
        for pnum in tqdm(range(0, p_max)):
            track = trajectories_adatom[trajectories_adatom.particle == pnum]
            if track.shape[0] > 5:
                start = track.iloc[0]
                track.x = track.x - start.x
                track.y = start.y - track.y
                ax.plot(track.x, track.y, alpha=0.15)
        fig.savefig(savefolder + 'all_tracks.pdf')
        fig.savefig(savefolder + 'all_tracks.png')
        
        
        
        
        
        
        tracks = []
        particles = trajectories_adatom.particle.unique()
        for pnum in tqdm(particles, desc = 'Populating Track Data'):
            track = trajectories_adatom[trajectories_adatom.particle == pnum]
            track['d_y'] = track['y'] - track['y'].shift(1)
            track['d_x'] = track['x'] - track['x'].shift(1)
            track['d_dist'] = (track['d_x']**2 + track['d_y']**2)**0.5
            track['d_theta'] = np.rad2deg(np.arctan2(track['d_y'], track['d_x']))
            track['gapframes'] = track['frame'] - track['frame'].shift(1)
            int_y = track.iloc[0].y
            int_x = track.iloc[0].x
            track['sum_dy'] = track['y'] - int_y
            track['sum_dx'] = track['x'] - int_x
            track['sum_dist'] = (track['sum_dx']**2 + track['sum_dy']**2)
            track['sum_theta'] = np.rad2deg(np.arctan2(track['sum_dy'], track['sum_dx']))
            tracks.append(track)
        tracks_df = pd.concat(tracks, axis=0, sort=False)
        tracks_df.to_csv(savefolder + 'tracks_particlegrouped.csv')
        
        
        
        
        lattice_tracks = []
        lattice_particles = trajectories_lattice.particle.unique()
        for pnum in tqdm(lattice_particles, desc = 'Populating Track Data'):
            track = trajectories_lattice[trajectories_lattice.particle == pnum]

            track['d_y'] = track['y'] - track['y'].shift(1)
            track['d_x'] = track['x'] - track['x'].shift(1)
            track['d_dist'] = (track['d_x']**2 + track['d_y']**2)**0.5
            track['gapframes'] = track['frame'] - track['frame'].shift(1)
            int_y = track.iloc[0].y
            int_x = track.iloc[0].x
            track['sum_dy'] = track['y'] - int_y
            track['sum_dx'] = track['x'] - int_x
            track['sum_dist'] = (track['sum_dx']**2 + track['sum_dy']**2)


            lattice_tracks.append(track)
        lattice_tracks_df = pd.concat(lattice_tracks, axis=0, sort=False)
        lattice_tracks_df.to_csv(savefolder + 'lattice_tracks.csv')
        
        
        
        
        
        
        
        
        
        maxjump = tracks_df.d_dist.max()
        minjump = tracks_df.d_dist.min()
        
        
        jump_bins = np.linspace(minjump, maxjump, num=30, endpoint=True)
        
        savefolder_jumps = savefolder + 'jumps/'
        if not os.path.exists(savefolder_jumps):
            os.mkdir(savefolder_jumps)
        
        for fnum in range(1, vid.data.shape[0]):
            jumps = tracks_df[tracks_df.frame == fnum].d_dist.values
            jumps_cut = jumps[jumps > 0]
            fig, ax = plt.subplots(1, figsize=(5, 5), dpi=300)
            sns.distplot(jumps_cut, bins=jump_bins, ax=ax)
            ax.set_xlim(0, 30)
            ax.set_ylim(0, 0.3)
            fig.savefig(savefolder_jumps + 'jumps_frame_%d.png' % (fnum))
            plt.close(fig=fig)
        
        all_jumps = tracks_df.d_dist.values
        all_jumps_cut = all_jumps[all_jumps > 0]
        fig, ax = plt.subplots(1, figsize=(5, 5), dpi=300)
        sns.distplot(all_jumps_cut, ax=ax)
        ax.set_xlim(0, 30)
        ax.set_ylim(0, 0.3)
        fig.savefig(savefolder + 'all_jumps.png')
        fig.savefig(savefolder + 'all_jumps.pdf')
        
        all_jump_angles = tracks_df.d_theta.values
        all_jump_angles = all_jump_angles[~np.isnan(all_jump_angles)]
        fig = plt.figure()
        ax = plt.subplot(111, projection='polar')
        bins = np.linspace(-180, 180, endpoint=True, num = 91)
        bin_centres = (bins[1:] + bins[:-1]) / 2
        hist, bin_edges = np.histogram(all_jump_angles, bins=bins, density=True)
        plt.plot(np.deg2rad(bin_centres), hist)
        
        cut_jump_angles = tracks_df[tracks_df.d_dist > 8].d_theta.values
        cut_jump_angles = cut_jump_angles[~np.isnan(cut_jump_angles)]
        fig = plt.figure()
        ax = plt.subplot(111, projection='polar')
        bins = np.linspace(-180, 180, endpoint=True, num = 46)
        bin_centres = (bins[1:] + bins[:-1]) / 2
        hist, bin_edges = np.histogram(cut_jump_angles, bins=bins, density=True)
        plt.plot(np.deg2rad(bin_centres), hist)
        plt.savefig(savefolder + 'cut_jump_angles.png')
        plt.savefig(savefolder + 'cut_jump_angles.pdf')
        
        
        jump_diffs = all_jump_angles[1:] - all_jump_angles[:-1]
        jump_diffs[jump_diffs < -180] += 360
        jump_diffs[jump_diffs > 180] -= 360
        
        fig = plt.figure()
        ax = plt.subplot(111, projection='polar')
        bins = np.linspace(-180, 180, endpoint=True, num = 91)
        bin_centres = (bins[1:] + bins[:-1]) / 2
        hist, bin_edges = np.histogram(jump_diffs, bins=bins, density=True)
        plt.plot(np.deg2rad(bin_centres), hist)
        plt.savefig(savefolder + 'directioncorrelation.png')
        plt.savefig(savefolder + 'directioncorrelation.pdf')
        
        
        
        
        
        
        
        
        
        if settings.filename == 'old_dry_vid.hspy':
            distmin, distmax = 2, 16
        else:
            distmin, distmax = 1, 12
        
        
        lattice_nn_list = []
        
        for fnum in tqdm(range(vid.data.shape[0]), desc = 'finding lattice points'):
        
            latframe = latvid.inav[fnum].data
            latframe_points = trajectories_lattice[trajectories_lattice.frame == fnum]
            latframe_x = latframe_points.x.values
            latframe_y = latframe_points.y.values
            latframe_points_list = np.asarray(list(zip(latframe_x, latframe_y)))
            lat_nn_x = []
            lat_nn_y = []
            lat_nn_theta = []
            lat_nn_mod = []
            lat_nn_sextant = []
            for i in range(0,latframe_x.shape[0]):
                point = latframe_points_list[i]
                dists = (np.sum((latframe_points_list - point)**2, axis=1))**0.5
                closepoints = latframe_points_list[(dists < distmax) & (dists > distmin)]
                for nn in closepoints:
                    nn_diff = nn - point
                    lat_nn_x.append(nn_diff[0])
                    lat_nn_y.append(nn_diff[1])
                    theta = np.rad2deg(np.arctan2(nn_diff[1], nn_diff[0]))
                    lat_nn_theta.append(theta)
                    lat_nn_mod.append(np.sqrt(nn_diff[0]**2 + nn_diff[1]**2))
            lattice_nn_df = pd.DataFrame({
                'x': lat_nn_x,
                'y': lat_nn_y,
                'theta': lat_nn_theta,
                'r': lat_nn_mod,
                'frame': fnum})
            lattice_nn_list.append(lattice_nn_df)
            
            
            
            
        lattice_nn_allframes = pd.concat(lattice_nn_list).reset_index(drop=True)
        
        kmeans = KMeans(n_clusters=6, random_state=0).fit(lattice_nn_allframes[['x', 'y']])
        lattice_nn_allframes['km_label'] = kmeans.labels_
        lattice_nn_allaveraged = lattice_nn_allframes.groupby('km_label').mean()
        lattice_nn_allaveraged['theta'] = np.rad2deg(np.arctan2(lattice_nn_allaveraged.y, lattice_nn_allaveraged.x))
        lattice_nn_allaveraged = lattice_nn_allaveraged.sort_values('theta')
        lattice_nn_allaveraged['sextant'] = range(6)
        
        
        
        
        
        
        savefolder_cluster = savefolder + 'clustering' + os.sep
        os.makedirs(savefolder_cluster, exist_ok=True)
        
        
        latticevector_df_list = []    
        for fnum in tqdm(range(vid.data.shape[0]), desc='clustering points'):    
            
            lattice_nn_df = lattice_nn_allframes[lattice_nn_allframes.frame == fnum]
            
            fig = plt.figure()
            ax = fig.add_axes([0, 0, 1, 1], aspect='equal')
        
            sextant_x = []
            sextant_y = []
            sextant_theta = []
            sextant_r = []
            for sextant in range(6):
                kmlabel = lattice_nn_allaveraged.index[sextant]
                sextant_df = lattice_nn_df[lattice_nn_df.km_label == kmlabel]
                x_val = sextant_df.x.mean()
                y_val = sextant_df.y.mean()
                theta_val = np.rad2deg(np.arctan2(y_val, x_val))
                r_val = np.sqrt(x_val**2 + y_val**2)
                sextant_x.append(x_val)
                sextant_y.append(y_val)
                sextant_theta.append(theta_val)
                sextant_r.append(r_val)
                ax.scatter(sextant_df.x, sextant_df.y, s=2, alpha=0.1)
        
        
            lattice_vectors = pd.DataFrame({
                'x': sextant_x,
                'y': sextant_y,
                'theta': sextant_theta,
                'r': sextant_r})
            lattice_vectors.index += 1
        
            lv_sorted = lattice_vectors
            sorted_x = lv_sorted.x.values
            sorted_y = lv_sorted.y.values
            lv_pos = np.column_stack([sorted_x, sorted_y])
        
            try:
                fitvec = least_squares(points_2vec_residuals, lv_pos[:2].flatten(), args=lv_pos, xtol=1e-15)
            
                fitv0 = np.array([fitvec.x[0], fitvec.x[1]])
                fitv1 = np.array([fitvec.x[2], fitvec.x[3]])
            
                calc_lattice_points = points_2vec(fitv0, fitv1)
                for sextant in range(6):
                    ax.plot([0, calc_lattice_points[sextant, 0]], [0, calc_lattice_points[sextant, 1]])
                    
                ax.set_xlim(-14, 14)
                ax.set_ylim(-14, 14)
                fig.savefig(savefolder_cluster + 'clustered_points_frame_%d.png' % fnum)
                
                plt.close('all')
                gc.collect()
                
                calc_lattice_vectors = pd.DataFrame({'x': calc_lattice_points[:,0], 'y': calc_lattice_points[:,1]})
                calc_lattice_vectors['theta'] = np.rad2deg(np.arctan2(calc_lattice_points[:,1], calc_lattice_points[:,0]))
                calc_lattice_vectors['dist'] = ((calc_lattice_points**2).sum(axis=1))**0.5
            
                lattice_vectors = lv_sorted.reset_index(drop=True)
                lattice_vectors['frame'] = fnum
                lattice_vectors['calcx'] = calc_lattice_vectors['x']
                lattice_vectors['calcy'] = calc_lattice_vectors['y']
                lattice_vectors['calctheta'] = calc_lattice_vectors['theta']
                lattice_vectors['calcdist'] = calc_lattice_vectors['dist']
                latticevector_df_list.append(lattice_vectors)
            except:
                latticevector_df_list.append(latticevector_df_list[-1])
        
        lattice_vector_frames_df = pd.concat(latticevector_df_list)
        lattice_vector_frames_df['num'] = lattice_vector_frames_df.index
        lattice_vector_frames_df.reset_index(drop=True, inplace=True)
        
        lattice_vector_frames_df.to_csv(savefolder + 'mo_lattice_vectors_by_frame.csv')
        
        
        
        
        
        
        
        
        
        lv_0_x ,lv_0_y = lattice_vector_frames_df[lattice_vector_frames_df.num==0].calcx.values, lattice_vector_frames_df[lattice_vector_frames_df.num==0].calcy.values
        lv_1_x ,lv_1_y  = lattice_vector_frames_df[lattice_vector_frames_df.num==1].calcx.values, lattice_vector_frames_df[lattice_vector_frames_df.num==1].calcy.values
        
        s1_vec_x, s1_vec_y = -(lv_0_x + lv_1_x) / 3, -(lv_0_y + lv_1_y) / 3
        s2_vec_x, s2_vec_y = -s1_vec_x, -s1_vec_y
        
        
        trajectories_mo = trajectories_lattice.copy()
        trajectories_s1 = trajectories_lattice.copy()
        trajectories_s2 = trajectories_lattice.copy()
        
        for index, row in tqdm(trajectories_s1.iterrows(), total = trajectories_s1.shape[0]):
            fnum = row.frame
            trajectories_s1.at[index, 'x'] += s1_vec_x[int(fnum)]
            trajectories_s1.at[index, 'y'] += s1_vec_y[int(fnum)]
            trajectories_s1.at[index, 'old_x'] += s1_vec_x[int(fnum)]
            trajectories_s1.at[index, 'old_y'] += s1_vec_y[int(fnum)]
        
        for index, row in tqdm(trajectories_s2.iterrows(), total = trajectories_s2.shape[0]):
            fnum = row.frame
            trajectories_s2.at[index, 'x'] += s2_vec_x[int(fnum)]
            trajectories_s2.at[index, 'y'] += s2_vec_y[int(fnum)]
            trajectories_s2.at[index, 'old_x'] += s2_vec_x[int(fnum)]
            trajectories_s2.at[index, 'old_y'] += s2_vec_y[int(fnum)]
            
        
        
        if site_intensity_score(trajectories_s1, vid) > site_intensity_score(trajectories_s2, vid):
            trajectories_s = trajectories_s1
            trajectories_vac = trajectories_s2
            s_vec_x, s_vec_y = s1_vec_x, s1_vec_y
            vac_vec_x, vac_vec_y = s2_vec_x, s2_vec_y
            s_vec_frameav = [s1_vec_x.mean(), s1_vec_y.mean()]
            vac_vec_frameav = [s2_vec_x.mean(), s2_vec_y.mean()]
        else:
            trajectories_s = trajectories_s2
            trajectories_vac = trajectories_s1
            s_vec_x, s_vec_y = s2_vec_x, s2_vec_y
            vac_vec_x, vac_vec_y = s1_vec_x, s1_vec_y
            s_vec_frameav = [s2_vec_x.mean(), s2_vec_y.mean()]
            vac_vec_frameav = [s1_vec_x.mean(), s1_vec_y.mean()]
        
        
        num = 3
        fig, ax = plt.subplots(1)
        for i in range(-num, num):
            for j in range(-num, num):
                mo_vec1_x = i * lv_0_x
                mo_vec1_y = i * lv_0_y
                mo_vec2_x = j * lv_1_x
                mo_vec2_y = j * lv_1_y
                mo_vec_x = mo_vec1_x + mo_vec2_x
                mo_vec_y = mo_vec1_y + mo_vec2_y
                ax.scatter(mo_vec_x, mo_vec_y, color='purple')
                ax.scatter(mo_vec_x + s_vec_x, mo_vec_y + s_vec_y, color='yellow')
                ax.scatter(mo_vec_x + vac_vec_x, mo_vec_y + vac_vec_y, color='red')
        lim = 10
        ax.set_xlim(-lim, lim)
        ax.set_ylim(-lim, lim)
        fig.savefig(savefolder_base + 'vecpoints_frame_' + settings.filename[:-5] + '.png')
                
                
            
        
        
        
        
        trajectories_mo.to_csv(savefolder + 'trajectories_mo.csv')
        trajectories_vac.to_csv(savefolder + 'trajectories_vac.csv')
        trajectories_s.to_csv(savefolder + 'trajectories_s.csv')
        
        
        
        # savefolder_latvid = savefolder + 'lat_vid/'
        # if not os.path.exists(savefolder_latvid):
        #     os.mkdir(savefolder_latvid)
        
        # from scipy.ndimage import gaussian_filter
        
        # for i in tqdm(range(0, vid.data.shape[0]), desc='outputting lattice_video frames'):
        
        #     fig = plt.figure(figsize=((vid.data.shape[2]/100.0) * 3, (vid.data.shape[1]/100.0) * 3), dpi=100)
        #     ax = fig.add_axes([0,0,1,1])
        #     ax.imshow(gaussian_filter(latvid.inav[i].data,0.8), vmin=latvid_cut_vals[0], vmax=latvid_cut_vals[1] * 1.5, interpolation='none')
        #     lims = ax.get_xlim()
        #     vac_df = trajectories_vac[trajectories_vac.frame == i]
        #     mo_df = trajectories_mo[trajectories_mo.frame == i]
        #     s_df = trajectories_s[trajectories_s.frame == i]
        #     ax.scatter(vac_df.old_x, vac_df.old_y, marker='o', edgecolor='red', facecolor='None', s=1)
        #     ax.scatter(mo_df.old_x, mo_df.old_y, marker='o', edgecolor='purple', facecolor='None', s=8)
        #     ax.scatter(s_df.old_x, s_df.old_y, marker='o', edgecolor='yellow', facecolor='None', s=6)
        #     ax.set_xlim(lims)
        #     fig.savefig(savefolder_latvid + 'vid%d.png' % (i), dpi=100)
        
        
        #     plt.close('all')
        #     gc.collect()
        
        
        
        
        
        
        
        
        
        
        
        allframes = []
        for fnum in tqdm(range(0, vid.data.shape[0]), desc='Finding Nearest Neighbours'):
            mo_points_df = trajectories_mo[trajectories_mo.frame == fnum]
            mo_indexes = mo_points_df.particle.values
            mo_points = np.asarray(list(zip(mo_points_df.x.values, mo_points_df.y.values)))
            
            s_points_df = trajectories_s[trajectories_s.frame == fnum]
            s_indexes = s_points_df.particle.values
            s_points = np.asarray(list(zip(s_points_df.x.values, s_points_df.y.values)))
            
            vac_points_df = trajectories_vac[trajectories_vac.frame == fnum]
            vac_indexes = vac_points_df.particle.values
            vac_points = np.asarray(list(zip(vac_points_df.x.values, vac_points_df.y.values)))
            
            adatoms_frame = tracks_df[tracks_df.frame == fnum]
            pt_indexes = adatoms_frame.particle.values
            pt_points = np.asarray(list(zip(adatoms_frame.x.values, adatoms_frame.y.values)))
        
            for index in adatoms_frame.index:
                point = (adatoms_frame.at[index, 'x'], adatoms_frame.at[index, 'y'])
                
                try:
                    mo_dists = (np.sum((mo_points - point)**2, axis=1))**0.5
                    s_dists = (np.sum((s_points - point)**2, axis=1))**0.5
                    vac_dists = (np.sum((vac_points - point)**2, axis=1))**0.5
                    pt_dists = (np.sum((pt_points - point)**2, axis=1))**0.5
            
                    pt_dists[pt_dists == 0] = 500
                    
                    minptdist = pt_dists.min()
                    mo_mindist = mo_dists.min()
                    s_mindist = s_dists.min()
                    vac_mindist = vac_dists.min()
            
                    nn_mo = mo_points[mo_dists.argmin()]
                    nn_s = s_points[s_dists.argmin()]
                    nn_vac = vac_points[vac_dists.argmin()]
                    nn_pt = pt_points[pt_dists.argmin()]
            
                    mo_index = mo_indexes[mo_dists.argmin()]
                    vac_index = vac_indexes[vac_dists.argmin()]
                    s_index = s_indexes[s_dists.argmin()]
                    pt_index = pt_indexes[pt_dists.argmin()]
            
                    nn_vec_mo = - (nn_mo - point)
                    nn_vec_s = -(nn_s - point)
                    nn_vec_vac = - (nn_vac - point)
                    nn_vec_pt = - (nn_pt - point)
            
                    adatoms_frame.at[index, 'lat_x'] = nn_vec_mo[0]
                    adatoms_frame.at[index, 'lat_y'] = nn_vec_mo[1]
            
                    adatoms_frame.at[index, 'pt_dist'] = minptdist
                    adatoms_frame.at[index, 'pt_theta'] = np.rad2deg(np.arctan2(nn_vec_pt[1], nn_vec_pt[0]))
                    adatoms_frame.at[index, 'pt_index'] = pt_index
            
                    if mo_mindist < s_mindist and mo_mindist < vac_mindist:
                        adatoms_frame.at[index, 'close_cat'] = 'Mo'
                        adatoms_frame.at[index, 'nn_x'] = nn_vec_mo[0]
                        adatoms_frame.at[index, 'nn_y'] = nn_vec_mo[1]
                        adatoms_frame.at[index, 'nn_dist'] = mo_mindist
                        adatoms_frame.at[index, 'nn_index'] = mo_index
                        adatoms_frame.at[index, 'fake_x'] = nn_mo[0]
                        adatoms_frame.at[index, 'fake_y'] = nn_mo[1]
            
            
                    elif s_mindist < mo_mindist and s_mindist < vac_mindist:
                        adatoms_frame.at[index, 'close_cat'] = 'S'
                        adatoms_frame.at[index, 'nn_x'] = nn_vec_s[0]
                        adatoms_frame.at[index, 'nn_y'] = nn_vec_s[1]
                        adatoms_frame.at[index, 'nn_dist'] = s_mindist
                        adatoms_frame.at[index, 'nn_index'] = s_index
                        adatoms_frame.at[index, 'fake_x'] = nn_s[0]
                        adatoms_frame.at[index, 'fake_y'] = nn_s[1]
            
                    elif vac_mindist < mo_mindist and vac_mindist < s_mindist:
                        adatoms_frame.at[index, 'close_cat'] = 'Vac'
                        adatoms_frame.at[index, 'nn_x'] = nn_vec_vac[0]
                        adatoms_frame.at[index, 'nn_y'] = nn_vec_vac[1]
                        adatoms_frame.at[index, 'nn_dist'] = vac_mindist
                        adatoms_frame.at[index, 'nn_index'] = vac_index
                        adatoms_frame.at[index, 'fake_x'] = nn_vac[0]
                        adatoms_frame.at[index, 'fake_y'] = nn_vac[1]
                
                except:
                    pass
                    
        
            allframes.append(adatoms_frame)
        tracks_nn = pd.concat(allframes, axis=0, sort=False)
        
        
        near_tracks = tracks_nn[tracks_nn.d_dist < 8]
        
        angle_difflist = (near_tracks['pt_theta'].values - near_tracks['d_theta'].values) % 360
        jumpsizelist = near_tracks['d_dist'].values
        
        jumpsizelist = jumpsizelist[~np.isnan(angle_difflist)]
        angle_difflist = angle_difflist[~np.isnan(angle_difflist)]
        
        bins = np.linspace(0, 360, 17)
        
        hist, bin_edges = np.histogram(angle_difflist, bins, weights=jumpsizelist**2)
        bin_centres = (bin_edges[:-1] + bin_edges[1:])/2
        
        ax = plt.subplot(111, polar=True)
        bars = ax.bar(np.deg2rad(bin_centres), hist, width=0.1)
        
        
        
        
        
        
        g = sns.jointplot(x='lat_x', y='lat_y', kind='kde', data=tracks_nn)
        g.savefig(savefolder + 'position_likelihood.png')
        g.savefig(savefolder + 'position_likelihood.pdf')
        
        
        
        
        
        calc_lattice_vectors = lattice_vector_frames_df.groupby('num').mean()
        calc_lattice_vectors.theta = np.rad2deg(np.arctan2(calc_lattice_vectors.y, calc_lattice_vectors.x))
        lattice_vector_list = list(zip(calc_lattice_vectors.calcx.values, calc_lattice_vectors.calcy.values))
        lattice_vector_list.append((0, 0))
        l_vec_arr = np.asarray(lattice_vector_list)
        points = l_vec_arr.copy()
        
        
        
        
        tracks_nn_fold = tracks_nn.copy()
        
        for index in tracks_nn_fold.index.values:
            while True:
                x = tracks_nn_fold.at[index, 'lat_x']
                y = tracks_nn_fold.at[index, 'lat_y']
                if np.isnan([x,y]).any():
                    break
                else:
                    point = x, y
                    dist_arr = np.sum((points - point)**2, axis=1)
                    closest_point = points[dist_arr.argmin()]
                    if np.array_equal(closest_point, [0, 0]):
                        break
                    else:
                        newx = x - closest_point[0]
                        newy = y - closest_point[1]
                        newdist = newx**2 + newy**2
                        newtheta = np.arctan2(newy, newx)
                        tracks_nn_fold.at[index, 'lat_x'] = newx
                        tracks_nn_fold.at[index, 'lat_y'] = newy
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        centreshift = np.asarray([0, 0])
        mo_points = points.copy() + centreshift
        
        real_mo_dist = 0.3122
        
        mo_dists = []
        for point in mo_points:
            dist = np.sqrt(np.sum(point**2-centreshift**2))
            if dist > 3:
                mo_dists.append(dist)
        mo_dist = np.asarray(mo_dists).mean()
        mo_sd = np.asarray(mo_dists).std()
        
        nm_per_px = real_mo_dist / mo_dist
        scale_err = np.abs((1 - (nm_per_px / metavid.axes_manager['x'].scale))) * 100
        scale_str = "calc nm per Pixel: %.6f, metadata nm per pixel: %.6f, percentage_error: %.1f" % (nm_per_px, metavid.axes_manager['x'].scale, scale_err)
        print(scale_str)
        with open(savefolder + "scale.txt", "w") as text_file:
            text_file.write(scale_str)
            
            
            
        
        
        
        point1 = (mo_points[0] + mo_points[1]) / 3
        point2 = (mo_points[1] + mo_points[2]) / 3
        point3 = (mo_points[2] + mo_points[3]) / 3
        point4 = (mo_points[3] + mo_points[4]) / 3
        point5 = (mo_points[4] + mo_points[5]) / 3
        point6 = (mo_points[5] + mo_points[0]) / 3
        
        set1_points = np.asarray([point1, point3, point5])
        set2_points = np.asarray([point2, point4, point6])
        
        
        set1_dist = np.min(((set1_points - s_vec_frameav)**2).sum(1))
        set2_dist = np.min(((set2_points - s_vec_frameav)**2).sum(1))
        
        if set1_dist < set2_dist:
            s_points = set1_points
            vac_points = set2_points
        else:
            s_points = set2_points
            vac_points = set1_points
        
        
        
                
        num = 3
        fig, ax = plt.subplots(1, figsize=(8,8))
        ax.set_aspect('equal')
        for i in range(-num, num):
            for j in range(-num, num):
                mo_vec1_x = i * lv_0_x
                mo_vec1_y = i * lv_0_y
                mo_vec2_x = j * lv_1_x
                mo_vec2_y = j * lv_1_y
                mo_vec_x = mo_vec1_x + mo_vec2_x
                mo_vec_y = mo_vec1_y + mo_vec2_y
                ax.scatter(mo_vec_x, mo_vec_y, color='purple', alpha=0.01, s=60)
                ax.scatter(mo_vec_x + s_vec_x, mo_vec_y + s_vec_y, color='yellow', alpha=0.01, s=60)
                ax.scatter(mo_vec_x + vac_vec_x, mo_vec_y + vac_vec_y, color='red', alpha=0.01, s=60)
        lim = 10
        ax.scatter(mo_points[:, 0], mo_points[:, 1], color='purple', marker='x', s=600, facecolor=None)
        ax.scatter(s_points[:, 0], s_points[:, 1], color='yellow', marker='x', s=600, facecolor=None)
        ax.scatter(vac_points[:, 0], vac_points[:, 1], color='red', marker='x', s=600, facecolor=None)
        ax.set_xlim(-lim, lim)
        ax.set_ylim(-lim, lim)
        ax.set_facecolor('grey')
        fig.savefig(savefolder_base + 'finpoints_frame_' + settings.filename[:-5] + '.png')
                
                
            
                    
        
        
        
        
        crit_angles = np.rad2deg(np.arctan2(mo_points[:, 1], mo_points[:, 0]))
        crit_angles = np.sort(crit_angles[crit_angles != 0])
        
        crit_points = []
        r = 16
        for theta in crit_angles:
            crit_points.append([r*np.cos(np.deg2rad(theta)), r*np.sin(np.deg2rad(theta))])
        
        
        fig, ax = plt.subplots(1)
        ax.scatter(mo_points[:, 0], mo_points[:, 1], color='purple')
        ax.scatter(s_points[:, 0], s_points[:, 1], color='yellow')
        ax.scatter(vac_points[:, 0], vac_points[:, 1], color='red')
        for point in crit_points:
            x, y = point
            ax.plot([centreshift[0], x], [centreshift[1],y], color='black')
        ax.set_aspect('equal')
        
        
        np.savetxt(savefolder + 'mo_points.txt', mo_points)
        np.savetxt(savefolder + 's_points.txt', s_points)
        np.savetxt(savefolder + 'vac_points.txt', vac_points)
        fig.savefig(savefolder + 'points_check.png')
        
        
        
        g = sns.jointplot(x='lat_x', y='lat_y', kind='kde', bw=0.6, data=tracks_nn_fold, fill=True)
        g.ax_joint.scatter(mo_points[:, 0], mo_points[:, 1], color='purple')
        g.ax_joint.scatter(s_points[:, 0], s_points[:, 1], color='yellow')
        g.ax_joint.scatter(vac_points[:, 0], vac_points[:, 1], color='red')
        g.savefig(savefolder + 'folded_position_likelihood.pdf')
        
        
        
        
        
        
        
        
        
        
        
        maxrad_dist = [1.2]
        maxrad_px = maxrad_dist / (nm_per_px * 10)
        
        np.savetxt(savefolder + 'maxrads_dist.txt', maxrad_dist)
        np.savetxt(savefolder + 'maxrads_px.txt', maxrad_px)
        
        
        
        
        
        for index, row in tracks_nn_fold.iterrows():
            if row.nn_dist < maxrad_px:
                tracks_nn_fold.at[index, 'cat'] = row['close_cat']
                tracks_nn_fold.at[index, 'cat_index'] = row['nn_index']
            else:
                tracks_nn_fold.at[index, 'cat'] = 'None'
                tracks_nn_fold.at[index, 'cat_index'] = 0
                tracks_nn_fold.at[index, 'fake_x'] = row.x
                tracks_nn_fold.at[index, 'fake_y'] = row.y
        
        
        
        np.unique(tracks_nn_fold.cat.values, return_counts=True)
        
        
        
        
        
        tracks_nn_fold.to_csv(savefolder + 'tracks_info.csv')
        
        
        
        
        
        
        
        
        
        savefolder_classvid = savefolder + 'class_vid/'
        if not os.path.exists(savefolder_classvid):
            os.mkdir(savefolder_classvid)
        
        
        for i in tqdm(range(0, vid.data.shape[0]), desc='outputting nn(position) frames'):
            scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
            fig = plt.figure(figsize=(vid.data.shape[2]/100.0, vid.data.shape[1]/100.0), dpi=100)
            ax = fig.add_axes([0, 0, 1, 1])
            ax.imshow(vid_T.inav[i].data, vmin=vid_cut_vals[0], vmax=vid_cut_vals[1] * 1.5, interpolation='hermite', cmap='Greys')
            frame_df = tracks_nn_fold[tracks_nn_fold['frame'] == i]
            vac_df = frame_df[frame_df['cat'] == 'Vac']
            mo_df = frame_df[frame_df['cat'] == 'Mo']
            s_df = frame_df[frame_df['cat'] == 'S']
            none_df = frame_df[frame_df['cat'] == 'None']
            frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
            plt.scatter(frame_lat.x, frame_lat.y, color='green', s=1)
            plt.scatter(vac_df.x, vac_df.y, marker='o', edgecolor='red', facecolor='None', s=8)
            for index, row in vac_df.iterrows():
                x1 = row.x
                y1 = row.y
                x2 = x1 + row.nn_x
                y2 = y1 + row.nn_y
                x_pair  = x1, x2
                y_pair = y1, y2
                plt.plot(x_pair, y_pair, linewidth = 0.5, color='red')
            plt.scatter(mo_df.x, mo_df.y, marker='o', edgecolor='purple', facecolor='None', s=8)
            for index, row in mo_df.iterrows():
                x1 = row.x
                y1 = row.y
                x2 = x1 + row.nn_x
                y2 = y1 + row.nn_y
                x_pair  = x1, x2
                y_pair = y1, y2
                plt.plot(x_pair, y_pair, linewidth = 0.5, color='purple')
            plt.scatter(s_df.x, s_df.y, marker='o', edgecolor='yellow', facecolor='None', s=8)
            for index, row in s_df.iterrows():
                x1 = row.x
                y1 = row.y
                x2 = x1 + row.nn_x
                y2 = y1 + row.nn_y
                x_pair  = x1, x2
                y_pair = y1, y2
                plt.plot(x_pair, y_pair, linewidth = 0.5, color='yellow')
            plt.scatter(none_df.x, none_df.y, marker='o', edgecolor='black', facecolor='None', s=8)
            plt.axis('off')
            plt.gca().add_artist(scalebar)
            plt.savefig(savefolder_classvid + 'vid%d.png' % (i), dpi=100)
            
        
            plt.close('all')
            gc.collect()
        
        
        
        
        
        
        
        
        nn_x_arr = tracks_nn_fold[tracks_nn_fold.frame > -1].lat_x.values
        nn_y_arr = tracks_nn_fold[tracks_nn_fold.frame > -1].lat_y.values
        
        repeat_nn_x_arr = []
        repeat_nn_y_arr = []
        
        for i in tqdm(range(0, len(nn_x_arr))):
            for point in points:
                repeat_nn_x_arr.append(nn_x_arr[i] + point[0])
                repeat_nn_y_arr.append(nn_y_arr[i] + point[1])
        
        g = sns.jointplot(x=repeat_nn_x_arr, y=repeat_nn_y_arr, kind='kde', bw=0.2, fill=True)
        for point in mo_points:
            circpatch = Circle(point, radius=maxrad_px, edgecolor='purple', facecolor='None', linewidth=1)
            g.ax_joint.add_artist(circpatch)
        for point in s_points:
            circpatch = Circle(point, radius=maxrad_px, edgecolor='yellow', facecolor='None', linewidth=1)
            g.ax_joint.add_artist(circpatch)
        for point in vac_points:
            circpatch = Circle(point, radius=maxrad_px, edgecolor='red', facecolor='None', linewidth=1)
            g.ax_joint.add_artist(circpatch)
        g.ax_joint.set_ylim(-9,8)
        g.ax_joint.set_xlim(-9,9)
        g.savefig(savefolder + 'nn_average_tiled.png')
        g.savefig(savefolder + 'nn_average_tiled.pdf')
        
        
        
        
        
        
        
        
        
        tracksnew = []
        
        tracklengths = []
        for pnum in tqdm(particles, desc='Populating Cat Refined Track Data'):
            track = tracks_nn_fold[tracks_nn_fold.particle == pnum]
            tracklength = len(track)
            tracklengths.append(tracklength)
        
            if tracklength > 2:
        
                track['lastcat'] = track.cat.shift(1)
                track['nextcat'] = track.cat.shift(-1)
                track['cons_cat'] = track.cat.groupby((track.cat != track.cat.shift()).cumsum()).cumcount()
                track['cons_len'] = track.cat.groupby((track.cat != track.cat.shift()).cumsum()).transform('size')
        
                track['same_pt'] = (track.pt_index == track.pt_index.shift(1))
                track['cons_same_pt'] = track.same_pt.groupby((track.same_pt != track.same_pt.shift()).cumsum()).cumcount() * track['same_pt']
                track['cons_same_pt_len'] = track.same_pt.groupby((track.same_pt != track.same_pt.shift()).cumsum()).transform('size') * track['same_pt']
        
        
                track['stuck'] = (track.nn_index == track.nn_index.shift(1)) & (track.cat == track.cat.shift(1))
                track['cons_stuck'] = track.stuck.groupby((track.stuck != track.stuck.shift()).cumsum()).cumcount() * track['stuck']
                track['cons_stuck_len'] = track.stuck.groupby((track.stuck != track.stuck.shift()).cumsum()).transform('size') * track['stuck']
        
        
                tracksnew.append(track)
        
        tracksnew = pd.concat(tracksnew, axis=0, sort=False).sort_index()
        tracksnew.to_csv(savefolder + 'tracks_cat.csv')
        
        
        
        cons_pt_df = tracksnew[(tracksnew.same_pt == True)]
        sns.distplot(cons_pt_df.pt_dist)
        plt.plot(cons_pt_df[(cons_pt_df.pt_dist < 28) & (cons_pt_df.pt_dist > 14)].groupby('cons_same_pt_len').mean().pt_dist)
        plt.plot(cons_pt_df[(cons_pt_df.pt_dist < 100) & (cons_pt_df.pt_dist > 28)].groupby('cons_same_pt_len').mean().pt_dist)
        plt.plot(cons_pt_df[(cons_pt_df.pt_dist < 14) & (cons_pt_df.pt_dist > 0)].groupby('cons_same_pt_len').mean().pt_dist)
        
        
        
        
        
        
        # maskcentre = [[518.0, 129.0],
        #               [31.0, 396.0]]
        # maskrad = [0, 0]
        
        # settings['mask_centre'] = maskcentre
        # settings['mask_radius'] = maskrad
        
        # maskfolder = savefolder + 'maskvid' + os.sep
        # if not os.path.exists(maskfolder):
        #     os.mkdir(maskfolder)
        
        
        # for fnum in tqdm(range(vid.data.shape[0])):
        #     fig = plt.figure(figsize=(8, 8))
        #     ax = fig.add_axes([0, 0, 1, 1])
        #     ax.imshow(gaussian_filter(vid.inav[fnum].data, 0.8), cmap='inferno')
        #     for i in range(len(maskrad)):
        #         maskcircle = Circle(maskcentre[i], radius = maskrad[i], alpha=0.2)
        #         ax.add_artist(maskcircle)
        #     fig.savefig(maskfolder + '%d.png' % fnum)
        #     plt.close('all')
        #     gc.collect()
        
        
        
        
        
        
        
        
        
        
        
        
        
        # tracks_cut_list = []
        # for i in range(len(maskrad)):
        #     for pnum in tqdm(tracksnew.particle.unique(), desc='removing lump'):
        #         track = tracksnew[tracksnew.particle == pnum]
        #         xdists = track.x.values - maskcentre[i][0]
        #         ydists = track.y.values - maskcentre[i][1]
        #         dists = (xdists**2 + ydists**2)**0.5
        #         if dists.mean() < maskrad[i]:
        #             tracks_cut_list.append(track)
        
        # if len(tracks_cut_list) > 0:
        #     tracksnewcut_remove = pd.concat(tracks_cut_list, axis=0, sort=False).sort_index()
        #     tracksnewcut = tracksnew[~np.isin(tracksnew.index, tracksnewcut_remove.index)]
        # else:
        #     tracksnewcut = tracksnew
        
        # tracksnewcut.to_csv(savefolder + 'tracks_cat_cut.csv')
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        driftframes = framesplit(fakevid)
        smoothdriftframes = gaussfilter(driftframes, 0.3)
        sdvid = framecombine(smoothdriftframes)
        sdvid.plot(cmap='inferno')
        
        sdvid.sum(0).plot()
        plt.savefig(savefolder + 'interstitial_mo.pdf')
        
        
        
        
        
        
        
        vac_df = tracks_nn_fold[tracks_nn_fold['cat'] == 'Vac']
        mo_df = tracks_nn_fold[tracks_nn_fold['cat'] == 'Mo']
        s_df = tracks_nn_fold[tracks_nn_fold['cat'] == 'S']
        none_df = tracks_nn_fold[tracks_nn_fold['cat'] == 'None']
        
        fig = plt.figure()
        
        sns.distplot(vac_df.nn_dist.values, kde_kws={"bw":0.1}, color='red', label='HC site')
        sns.distplot(mo_df.nn_dist.values, kde_kws={"bw":0.1}, color='purple', label='Mo site')
        sns.distplot(s_df.nn_dist.values, kde_kws={"bw":0.1}, color='yellow', label='S site')
        sns.distplot(none_df.nn_dist.values, kde_kws={"bw":0.2}, color='black', label='No site')
        
        ax = plt.gca()
        ax.legend()
        ax.set_xlim(0, 20)
        ax.set_xlabel('NN_distance (by NN site type)')
        
        fig.savefig(savefolder + 'NN_distance_hists_bysite.png')
        
        
        
        edgesize = 18
        
        savefolder_singleparts = savefolder + 'singleparts2/'
        if not os.path.exists(savefolder_singleparts):
            os.mkdir(savefolder_singleparts)
        
        savefolder_single_mo = savefolder_singleparts + 'mo/'
        if not os.path.exists(savefolder_single_mo):
            os.mkdir(savefolder_single_mo)
        
        savefolder_single_s = savefolder_singleparts + 's/'
        if not os.path.exists(savefolder_single_s):
            os.mkdir(savefolder_single_s)
        
        savefolder_single_vac = savefolder_singleparts + 'vac/'
        if not os.path.exists(savefolder_single_vac):
            os.mkdir(savefolder_single_vac)
        
        mo_templates = []
        mo_templates_extra = []
        mo_templates_pnumlist = []
        mo_templates_fnumlist = []
        mo_counts = 0
        for i in tqdm(range(0, vid.data.shape[0]), desc='test'):
            scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
            origframe = vid.inav[i].deepcopy().data
            fakeframe = sdvid.inav[i].deepcopy().data
            frame_df = tracks_nn_fold[tracks_nn_fold['frame'] == i]
            mo_df = frame_df[frame_df['cat'] == 'Mo']
            frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
            for index, row in mo_df.iterrows():
                if row.nn_dist < 1.1:
                    if row.pt_dist > 20:
                        xcent = int(np.rint(row.fake_x))
                        ycent = int(np.rint(row.fake_y))
                        if xcent > edgesize + 1 and xcent < origframe.shape[1] - edgesize - 1:
                            if ycent > edgesize + 1 and ycent < origframe.shape[0] - edgesize - 1:
                                origtemplate = origframe[ycent - edgesize:ycent + edgesize + 1, xcent - edgesize:xcent + edgesize + 1]
                                faketemplate = fakeframe[ycent - edgesize:ycent + edgesize + 1, xcent - edgesize:xcent + edgesize + 1]
                                if origtemplate.sum(0).min() != 0 and origtemplate.sum(1).min() != 0:
                                    mo_templates.append(origtemplate)
                                    mo_templates_extra.append(faketemplate)
                                    mo_templates_pnumlist.append(row.particle)
                                    mo_templates_fnumlist.append(i)
                                area_lat_df = frame_lat[(frame_lat.x > xcent - edgesize) & (frame_lat.x < xcent + edgesize + 1) & (frame_lat.y > ycent - edgesize) & (frame_lat.y <  ycent + edgesize + 1)]
                                # fig = plt.figure(figsize=(6,6))
                                # ax = fig.add_axes([0, 0, 1, 1])
                                # ax.imshow(origtemplate, cmap='inferno')
                                # scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
                                # fig.savefig(savefolder_single_mo + 'frame_%d_pnum_%d.png' % (i, row.particle), bbox_inches='tight')
                                # plt.close('all')
                                # gc.collect()
                                # fig = plt.figure(figsize=(6,6))
                                # ax = fig.add_axes([0, 0, 1, 1])
                                # ax.imshow(faketemplate, cmap='inferno')
                                # scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
                                # fig.savefig(savefolder_single_mo + 'frame_%d_pnum_%d_fake.png' % (i, row.particle), bbox_inches='tight')
                                mo_counts += 1
        print(mo_counts)
        
        s_templates = []
        s_templates_extra = []
        s_templates_pnumlist = []
        s_templates_fnumlist = []
        s_counts = 0
        for i in tqdm(range(0, vid.data.shape[0]), desc='test'):
            scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
            origframe = vid.inav[i].deepcopy().data
            fakeframe = sdvid.inav[i].deepcopy().data
            frame_df = tracks_nn_fold[tracks_nn_fold['frame'] == i]
            s_df = frame_df[frame_df['cat'] == 'S']
            frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
            for index, row in s_df.iterrows():
                if 0.7 < row.nn_dist < 1.3:
                    if row.pt_dist > 20:
                        xcent = int(np.rint(row.fake_x))
                        ycent = int(np.rint(row.fake_y))
                        if xcent > edgesize + 1 and xcent < origframe.shape[1] - edgesize - 1:
                            if ycent > edgesize + 1 and ycent < origframe.shape[0] - edgesize - 1:
                                origtemplate = origframe[ycent - edgesize:ycent + edgesize + 1, xcent - edgesize:xcent + edgesize + 1]
                                faketemplate = fakeframe[ycent - edgesize:ycent + edgesize + 1, xcent - edgesize:xcent + edgesize + 1]
                                if origtemplate.sum(0).min() != 0 and origtemplate.sum(1).min() != 0:
                                    s_templates.append(origtemplate)
                                    s_templates_extra.append(faketemplate)
                                    s_templates_pnumlist.append(row.particle)
                                    s_templates_fnumlist.append(i)
                                area_lat_df = frame_lat[(frame_lat.x > xcent - edgesize) & (frame_lat.x < xcent + edgesize + 1) & (frame_lat.y > ycent - edgesize) & (frame_lat.y <  ycent + edgesize + 1)]
                                # fig = plt.figure(figsize=(6,6))
                                # ax = fig.add_axes([0, 0, 1, 1])
                                # ax.imshow(origtemplate, cmap='inferno')
                                # scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
                                # fig.savefig(savefolder_single_s + 'frame_%d_pnum_%d.png' % (i, row.particle), bbox_inches='tight')
                                # plt.close('all')
                                # gc.collect()
                                # fig = plt.figure(figsize=(6,6))
                                # ax = fig.add_axes([0, 0, 1, 1])
                                # ax.imshow(faketemplate, cmap='inferno')
                                # scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
                                # fig.savefig(savefolder_single_s + 'frame_%d_pnum_%d_fake.png' % (i, row.particle), bbox_inches='tight')
                                # plt.close('all')
                                s_counts += 1
        print(s_counts)
        
        
        vac_templates = []
        vac_templates_extra = []
        vac_templates_pnumlist = []
        vac_templates_fnumlist = []
        vac_counts = 0
        for i in tqdm(range(0, vid.data.shape[0]), desc='test'):
            scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
            origframe = vid.inav[i].deepcopy().data
            fakeframe = sdvid.inav[i].deepcopy().data
            frame_df = tracks_nn_fold[tracks_nn_fold['frame'] == i]
            vac_df = frame_df[frame_df['cat'] == 'Vac']
            frame_lat = trajectories_lattice[trajectories_lattice['frame'] == i]
            for index, row in vac_df.iterrows():
                if 0.4 < row.nn_dist < 1.6:
                    if row.pt_dist > 20:
                        xcent = int(np.rint(row.fake_x))
                        ycent = int(np.rint(row.fake_y))
                        if xcent > edgesize + 1 and xcent < origframe.shape[1] - edgesize - 1:
                            if ycent > edgesize + 1 and ycent < origframe.shape[0] - edgesize - 1:
                                origtemplate = origframe[ycent - edgesize:ycent + edgesize + 1, xcent - edgesize:xcent + edgesize + 1]
                                faketemplate = fakeframe[ycent - edgesize:ycent + edgesize + 1, xcent - edgesize:xcent + edgesize + 1]
                                if origtemplate.sum(0).min() != 0 and origtemplate.sum(1).min() != 0:
                                    vac_templates.append(origtemplate)
                                    vac_templates_extra.append(faketemplate)
                                    vac_templates_pnumlist.append(row.particle)
                                    vac_templates_fnumlist.append(i)
                                area_lat_df = frame_lat[(frame_lat.x > xcent - edgesize) & (frame_lat.x < xcent + edgesize + 1) & (frame_lat.y > ycent - edgesize) & (frame_lat.y <  ycent + edgesize + 1)]
                                # fig = plt.figure(figsize=(6,6))
                                # ax = fig.add_axes([0, 0, 1, 1])
                                # ax.imshow(origtemplate, cmap='inferno')
                                # scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
                                # fig.savefig(savefolder_single_vac + 'frame_%d_pnum_%d.png' % (i, row.particle), bbox_inches='tight')
                                # plt.close('all')
                                # gc.collect()
                                # fig = plt.figure(figsize=(6,6))
                                # ax = fig.add_axes([0, 0, 1, 1])
                                # ax.imshow(faketemplate, cmap='inferno')
                                # scalebar = ScaleBar(vidscale, 'nm', box_alpha=0, color='white')
                                # fig.savefig(savefolder_single_vac + 'frame_%d_pnum_%d_fake.png' % (i, row.particle), bbox_inches='tight')
                                # plt.close('all')
                                vac_counts += 1
        print(vac_counts)
        
        
        mo_template_stackdata = np.stack(mo_templates)
        mo_template_stack = hs.signals.Signal2D(mo_template_stackdata)
        
        s_template_stackdata = np.stack(s_templates)
        s_template_stack = hs.signals.Signal2D(s_template_stackdata)
        
        vac_template_stackdata = np.stack(vac_templates)
        vac_template_stack = hs.signals.Signal2D(vac_template_stackdata)
        
        mo_template_extra_stackdata = np.stack(mo_templates_extra)
        mo_template_extra_stack = hs.signals.Signal2D(mo_template_extra_stackdata)
        
        s_template_extra_stackdata = np.stack(s_templates_extra)
        s_template_extra_stack = hs.signals.Signal2D(s_template_extra_stackdata)
        
        vac_template_extra_stackdata = np.stack(vac_templates_extra)
        vac_template_extra_stack = hs.signals.Signal2D(vac_template_extra_stackdata)
        
        mo_template_stack.save(savefolder + 'mo_template_stack.hspy', overwrite=True)
        s_template_stack.save(savefolder + 's_template_stack.hspy', overwrite=True)
        vac_template_stack.save(savefolder + 'vac_template_stack.hspy', overwrite=True)
        
        mo_template_extra_stack.save(savefolder + 'mo_template_stack_enh.hspy', overwrite=True)
        s_template_extra_stack.save(savefolder + 's_template_stack_enh.hspy', overwrite=True)
        vac_template_extra_stack.save(savefolder + 'vac_template_stack_enh.hspy', overwrite=True)
        
        
        
        
        
        
        
        mo_template_stack.change_dtype('float')
        mo_template_stack.decomposition()
        # mo_template_stack.plot_decomposition_results()
        
        mo_single = mo_template_stack.get_decomposition_model(1)
        mo_single_res = (((mo_single - mo_template_stack)**2).sum(1).sum(1)).data.argmin()
        mo_templates_fnumlist[mo_single_res]
        mo_templates_pnumlist[mo_single_res]
        
        mo_single.sum(0).plot()
        
        
        
        s_template_stack.change_dtype('float')
        s_template_stack.decomposition()
        # s_template_stack.plot_decomposition_results()
        
        s_single = s_template_stack.get_decomposition_model(1)
        s_single_res = (((s_single - s_template_stack)**2).sum(1).sum(1)).data.argmin()
        s_templates_fnumlist[s_single_res]
        s_templates_pnumlist[s_single_res]
        
        s_single.sum(0).plot()
        
        
        
        vac_template_stack.change_dtype('float')
        vac_template_stack.decomposition()
        # vac_template_stack.plot_decomposition_results()
        
        vac_single = vac_template_stack.get_decomposition_model(1)
        vac_single_res = (((vac_single - vac_template_stack)**2).sum(1).sum(1)).data.argmin()
        vac_templates_fnumlist[vac_single_res]
        vac_templates_pnumlist[vac_single_res]
        
        vac_single.sum(0).plot()
        
        
        
        
        mo_template_extra_stack.change_dtype('float')
        mo_template_extra_stack.decomposition()
        # mo_template_extra_stack.plot_decomposition_results()
        
        mo_single_e = mo_template_extra_stack.get_decomposition_model(1)
        mo_single_e_res = (((mo_single - mo_template_extra_stack)**2).sum(1).sum(1)).data.argmin()
        
        mo_single_e.sum(0).plot()
        
        
        
        
        s_template_extra_stack.change_dtype('float')
        s_template_extra_stack.decomposition()
        # s_template_extra_stack.plot_decomposition_results()
        
        s_single_e = s_template_extra_stack.get_decomposition_model(1)
        s_single_e_res = (((s_single - s_template_extra_stack)**2).sum(1).sum(1)).data.argmin()
        
        
        
        s_single_e.sum(0).plot()
        
        
        
        vac_template_extra_stack.change_dtype('float')
        vac_template_extra_stack.decomposition()
        # vac_template_extra_stack.plot_decomposition_results()
        
        vac_single_e = vac_template_extra_stack.get_decomposition_model(1)
        vac_single_e_res = (((vac_single - vac_template_extra_stack)**2).sum(1).sum(1)).data.argmin()
        
        
        vac_single_e.sum(0).plot()
        
        
        
        
        mo_template_e = mo_single_e.sum(0).deepcopy() / mo_single_e.data.shape[0]
        s_template_e = s_single_e.sum(0).deepcopy() / s_single_e.data.shape[0]
        vac_template_e = vac_single_e.sum(0).deepcopy() / vac_single_e.data.shape[0]
        
        mo_template = mo_single.sum(0).deepcopy() / mo_single.data.shape[0]
        s_template = s_single.sum(0).deepcopy() / s_single.data.shape[0]
        vac_template = vac_single.sum(0).deepcopy() / vac_single.data.shape[0]
        
        
        
        mo_template_e.save(savefolder + 'mo_template_enhanced.tiff', overwrite=True)
        s_template_e.save(savefolder + 's_template_enhanced.tiff', overwrite=True)
        vac_template_e.save(savefolder + 'vac_template_enhanced.tiff', overwrite=True)
        
        mo_template.save(savefolder + 'mo_template.tiff', overwrite=True)
        s_template.save(savefolder + 's_template.tiff', overwrite=True)
        vac_template.save(savefolder + 'vac_template.tiff', overwrite=True)
        
        
        from skimage.exposure import match_histograms
        
        
        mo_sum = mo_single.sum(0).data
        s_sum = s_single.sum(0).data
        vac_sum = vac_single.sum(0).data
        
        
        
        mo_diffs = np.zeros(len(mo_templates), dtype=float)
        for i in range(len(mo_templates)):
            template = mo_templates[i]
            template_matched = match_histograms(template, mo_sum)
            diff = (((template_matched - mo_sum)**2).sum())**0.5
            mo_diffs[i] = diff
        mo_minindexes = mo_diffs.argsort()[:10]
        
        s_diffs = np.zeros(len(s_templates), dtype=float)
        for i in range(len(s_templates)):
            template = s_templates[i]
            template_matched = match_histograms(template, mo_sum)
            diff = (((template_matched - mo_sum)**2).sum())**0.5
            s_diffs[i] = diff
        s_minindexes = s_diffs.argsort()[:10]
        
        vac_diffs = np.zeros(len(vac_templates), dtype=float)
        for i in range(len(vac_templates)):
            template = vac_templates[i]
            template_matched = match_histograms(template, mo_sum)
            diff = (((template_matched - mo_sum)**2).sum())**0.5
            vac_diffs[i] = diff
        vac_minindexes = vac_diffs.argsort()[:10]
        
        
        
        
        
        for i in mo_minindexes:
            print('Frame_%d, Particle_%d\n' %(mo_templates_fnumlist[i], mo_templates_pnumlist[i]))
        
        
        
        for i in s_minindexes:
            print('Frame_%d, Particle_%d\n' %(s_templates_fnumlist[i], s_templates_pnumlist[i]))
        
        
        
        for i in vac_minindexes:
            print('Frame_%d, Particle_%d\n' %(vac_templates_fnumlist[i], vac_templates_pnumlist[i]))
        
        
        
        
        
        def get_templatesum_from_indexes(templates, indexlist):
            template_pass_list = []
            for index in indexlist:
                template = templates[index]
                template_pass_list.append(template)
            return np.stack(template_pass_list).sum(0)
                
        
        savefolder_mo_averaged = savefolder + 'averaged' + os.sep + 'Mo' + os.sep
        os.makedirs(savefolder_mo_averaged, exist_ok=True)
        savefolder_s_averaged = savefolder + 'averaged' + os.sep + 'S' + os.sep
        os.makedirs(savefolder_s_averaged, exist_ok=True)
        savefolder_vac_averaged = savefolder + 'averaged' + os.sep + 'Vac' + os.sep
        os.makedirs(savefolder_vac_averaged, exist_ok=True)
        
        for cutlen in tqdm(range(1,100)):
            mo_minindexes = mo_diffs.argsort()[:cutlen]
            s_minindexes = s_diffs.argsort()[:cutlen]
            vac_minindexes = vac_diffs.argsort()[:cutlen]
            mo_averaged = get_templatesum_from_indexes(mo_templates, mo_minindexes)
            s_averaged = get_templatesum_from_indexes(s_templates, s_minindexes)
            vac_averaged = get_templatesum_from_indexes(vac_templates, vac_minindexes)
            plt.imsave(savefolder_mo_averaged + '%d.png' % cutlen, mo_averaged, cmap='inferno')
            plt.imsave(savefolder_s_averaged + '%d.png' % cutlen, s_averaged, cmap='inferno')
            plt.imsave(savefolder_vac_averaged + '%d.png' % cutlen, vac_averaged, cmap='inferno')
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        from matplotlib_scalebar.scalebar import ScaleBar
        
        fig = plt.figure(figsize=(4,4))
        ax = fig.add_axes([0,0,1,1])
        ax.imshow(mo_single.sum(0).data,interpolation='none', cmap='inferno')
        lims = ax.get_xlim()
        scalebar = ScaleBar(vidscale, 'nm', color='white', box_alpha=0)
        ax.add_artist(scalebar)
        ax.set_xlim(lims)
        ax.set_ylim(lims)
        fig.savefig(savefolder + 'template_mo.png')
        fig.savefig(savefolder + 'template_mo.pdf')
        
        fig = plt.figure(figsize=(4,4))
        ax = fig.add_axes([0,0,1,1])
        ax.imshow(s_single.sum(0).data,interpolation='none', cmap='inferno')
        lims = ax.get_xlim()
        scalebar = ScaleBar(vidscale, 'nm', color='white', box_alpha=0)
        ax.add_artist(scalebar)
        ax.set_xlim(lims)
        ax.set_ylim(lims)
        fig.savefig(savefolder + 'template_s.png')
        fig.savefig(savefolder + 'template_s.pdf')
        
        fig = plt.figure(figsize=(4,4))
        ax = fig.add_axes([0,0,1,1])
        ax.imshow(vac_single.sum(0).data,interpolation='none', cmap='inferno')
        lims = ax.get_xlim()
        scalebar = ScaleBar(vidscale, 'nm', color='white', box_alpha=0)
        ax.add_artist(scalebar)
        ax.set_xlim(lims)
        ax.set_ylim(lims)
        fig.savefig(savefolder + 'template_vac.png')
        fig.savefig(savefolder + 'template_vac.pdf')
            
        
        plt.close('all')
        gc.collect()
        
        
        
        
        settings.to_csv(savefolder + 'settings_out.csv')
