import numpy as np
import os
import gc
import warnings
import random

import hyperspy.api as hs
import trackpy as tp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import matplotlib
import pickle

from matplotlib.patches import Rectangle, Circle, FancyArrowPatch
from matplotlib_scalebar.scalebar import ScaleBar
from matplotlib.cm import get_cmap, ScalarMappable
from mpl_toolkits.mplot3d import Axes3D
from tqdm import tqdm
from gc import collect
from scipy.optimize import curve_fit
from scipy.ndimage import gaussian_filter
from lmfit.models import SkewedGaussianModel
from itertools import chain
from scipy.stats import gaussian_kde
from matplotlib import ticker
from scipy.optimize import least_squares, brute
from skimage.exposure import match_histograms
from matplotlib.offsetbox import AnchoredText
from glob import glob
from scipy.signal import find_peaks
import time
from scipy.spatial import KDTree, cKDTree
from matplotlib.colors import LogNorm
import time
from grid_strategy import strategies
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.signal import savgol_filter
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
from skimage.measure import profile_line
from scipy.spatial import distance
from itertools import combinations, compress

from compfunctions_v4 import *


#set Preferences

plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42

pd.options.mode.chained_assignment = None 

origvidfolder = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_output\1_preprocessing\video" + os.sep
folder_base = r"C:\Users\Nick\Dropbox (The University of Manchester)\Projects\Particle Tracking\new_data\14_09_20_individual" + os.sep
loadfolder_base = folder_base + "8_points_plot" + os.sep

# choose savefolder

savefolder = folder_base + '/compare_v7/' + os.sep
savefolder_figs = savefolder + 'output_figures' + os.sep
for s in [savefolder, savefolder_figs]:
    os.makedirs(s, exist_ok=True)
    
savefolder_msd = savefolder + 'msd' + os.sep
os.makedirs(savefolder_msd, exist_ok=True)

metascale = 0.034504504596196285

# load data loop

filename_in_list = os.listdir(loadfolder_base)

filename_list = [fname for fname in filename_in_list if os.path.isdir(loadfolder_base + fname)]

settings_list = []

vid_list = []
latvid_list = []
origvid_list = []

vid_cut_vals_list = []
latvid_cut_vals_list = []
origvid_cut_vals_list = []

scale_list = []
calc_scale_list = []
timescale_list = []

lattice_list = []
mo_traj_list = []
s_traj_list = []
vac_traj_list = []
atoms_list = []

beam_current_list = []
dosescale_list = []

lattice_norm_list = []
mo_traj_norm_list = []
s_traj_norm_list = []
vac_traj_norm_list = []
atoms_norm_list = []

mo_points_list = []
s_points_list = []
vac_points_list = []

doseparams_list = []

for filename in tqdm(filename_list, desc = 'loading data and normalising'):
    if os.path.isdir(loadfolder_base + filename):
        
        print('loading: ' + filename)
    
        # load videos, settings, scales
        vidfolder = folder_base + "5_driftcorr" + os.sep + filename + os.sep
        latticefolder = folder_base + "6_latticepoints" + os.sep + filename + os.sep
        adatomfolder = folder_base + "8_points_plot" + os.sep + filename + os.sep
        
        settings = pd.read_csv(adatomfolder + 'settings_out.csv')
        settings = settings.set_index('Unnamed: 0').iloc[:,0]
        settings_list.append(settings)
        
        vid = hs.load(adatomfolder + 'transformed_vid_orig.hspy')
        latvid = hs.load(adatomfolder + 'transformed_latvid_orig.hspy')
        origvid = hs.load(origvidfolder + settings.filename)
        
        if (origvid.data.shape[0] > vid.data.shape[0]):
            origvid = origvid.deepcopy().inav[:vid.data.shape[0]]
        vid_list.append(vid)
        latvid_list.append(latvid)
        origvid_list.append(origvid)
        
        vid_cut_vals = vid_intensity_cut(vid, [0.2, 98])
        latvid_cut_vals = vid_intensity_cut(latvid, [0.1, 99.9])
        origvid_cut_vals = vid_intensity_cut(origvid, [0.2, 98])
        vid_cut_vals_list.append(vid_cut_vals)
        latvid_cut_vals_list.append(latvid_cut_vals)
        origvid_cut_vals_list.append(origvid_cut_vals)
        
        if settings.filename == 'old_dry_vid.hspy':
            vidscale = origvid.axes_manager['x'].scale
        else:
            vidscale = metascale
            
        with open(adatomfolder + 'scale.txt','r') as f:
            output = f.read()
        calcscale = float(output[20:27])
        scale = vidscale
        scale_list.append(scale)
        calc_scale_list.append(calcscale)
        
        
        # load trajectories
        
        lattice = pd.read_csv(adatomfolder + 'lattice_tracks.csv')
        lattice_list.append(lattice)
    
        mo_traj = pd.read_csv(adatomfolder + 'trajectories_mo.csv')
        s_traj = pd.read_csv(adatomfolder + 'trajectories_s.csv')
        vac_traj = pd.read_csv(adatomfolder + 'trajectories_vac.csv')
        mo_traj_list.append(mo_traj)
        s_traj_list.append(s_traj)
        vac_traj_list.append(vac_traj)
        
        atoms = pd.read_csv(adatomfolder + 'tracks_cat.csv')
    
        doseparams = pd.Series()
        doseparams['vidnum'] = settings.name
        doseparams['filename'] = settings.filename
        doseparams['scale'] = vidscale
            
        # calculate dose scale
        timescale = float(origvid.original_metadata.ObjectInfo.AcquireInfo.FrameTime)
        timescale_list.append(timescale)
        doseparams['frame_time'] = timescale
        
        ss = int(float(settings.Spot_size))
        doseparams['spot_size'] = ss
        pxtime = float(settings.DwellTimePath)
        doseparams['dwell_time'] = pxtime
        px_area = (vidscale) ** 2 # nm^2
        doseparams['px_area'] = px_area
        beam_current = 0        # in pA
        if filename in ['15.26.Scanning Preview.hspy', 'old_dry_vid.hspy', 'old_wet_vid.hspy']:
            beam_current = 167.0e-12
        elif ss == 5:
            beam_current = 167.0e-12
        elif ss == 6:
            beam_current = 81.0e-12
        elif ss == 7:
            beam_current = 42.0e-12
        else:
            print('BEAM CURRENT NOT SET')
        beam_current_list.append(beam_current)
        doseparams['beam_current'] = beam_current
        electrons_per_sec = beam_current *  6.242e18
        doseparams['electrons_per_sec'] = electrons_per_sec
        electrons_per_px = electrons_per_sec * pxtime
        doseparams['electrons_per_px'] = electrons_per_px
        dose_per_nm = electrons_per_px / px_area
        doseparams['dose_per_nm'] = dose_per_nm
        dosescale = dose_per_nm / 1e6 # area dose per frame in electrons (x 10^6) per nm^2
        doseparams['dose_per_nm_million'] = dosescale
        dosescale_list.append(dosescale)
        
        doseparams_list.append(doseparams)
        
    
        # trim and normalise trajectories
        atoms = lattice_trim(atoms)
        atoms_list.append(atoms)
    
        atoms_norm = atoms_normalise(atoms, scale)
        atoms_norm_list.append(atoms_norm)
        
        lattice_norm = lattice_normalise(lattice, scale)
        lattice_norm_list.append(lattice_norm)
        
        mo_traj_norm = traj_normalise(mo_traj, scale)
        s_traj_norm = traj_normalise(s_traj, scale)
        vac_traj_norm = traj_normalise(vac_traj, scale)
        mo_traj_norm_list.append(mo_traj_norm)
        s_traj_norm_list.append(s_traj_norm)
        vac_traj_norm_list.append(vac_traj_norm)
    
    
    
        # load averaged lattice vectors
        mo_points = np.loadtxt(adatomfolder + 'mo_points.txt') * scale
        s_points = np.loadtxt(adatomfolder + 's_points.txt') * scale
        vac_points = np.loadtxt(adatomfolder + 'vac_points.txt') * scale
        mo_points_list.append(mo_points)
        s_points_list.append(s_points)
        vac_points_list.append(vac_points)
        


# choose colorlist

        
wetdry_dict = {'15.26.Scanning Preview': 'ignore',
           'Scanning Preview_15,51': 'dry',
           'Scanning Preview_15.14': 'dry',
           'Scanning Preview_15.18': 'dry', #unsure
           'Scanning Preview_15.56': 'dry',
           'Scanning Preview12': 'wet',
           'Scanning Preview13': 'wet',
           'Scanning Preview13b': 'wet',
           'Scanning Preview14': 'wet',
           'Scanning Preview15': 'wet',
           'Scanning Preview17': 'wet',
           'Scanning Preview18': 'wet',
           'Scanning Preview19': 'wet',
           'Scanning Preview20_sparse data': 'ignore',
           'Scanning Preview22': 'wet',
           'Scanning Preview23_40pA_12us': 'ignore',
           'old_dry_vid': 'dry',
           'old_wet_vid': 'wet',
           'Scanning Preview_12.36': 'dry',
           'Scanning Preview_12.43':'dry'}
           
           

wetdry_list = [wetdry_dict[filename[:-5]] for filename in filename_list]

doseparams_df = pd.concat(doseparams_list, axis=1).T.set_index('vidnum')
doseparams_df['orig_index'] = range(len(doseparams_df))
doseparams_df['beam_current_pA'] = doseparams_df['beam_current'] * 1e12
doseparams_df['wetdry'] = [wetdry_dict[filename[:-5]] for filename in doseparams_df['filename']]
doseparams_df = doseparams_df.sort_values(['wetdry', 'dose_per_nm_million'])
doseparams_df['filename_reduced'] = [filename[:-5] for filename in doseparams_df['filename']]
doseparams_df.to_csv(savefolder + 'doseparams.csv')




list_reorder = doseparams_df.orig_index.values



#reorder by wetdry, beam_current
settings_list = [settings_list[i] for i in list_reorder]
filename_list = [filename_list[i] for i in list_reorder]
wetdry_list = [wetdry_list[i] for i in list_reorder]

vid_list = [vid_list[i] for i in list_reorder]
latvid_list = [latvid_list[i] for i in list_reorder]
origvid_list = [origvid_list[i] for i in list_reorder]

vid_cut_vals_list = [vid_cut_vals_list[i] for i in list_reorder]
latvid_cut_vals_list = [latvid_cut_vals_list[i] for i in list_reorder]
origvid_cut_vals_list = [origvid_cut_vals_list[i] for i in list_reorder]

scale_list = [scale_list[i] for i in list_reorder]
calc_scale_list = [calc_scale_list[i] for i in list_reorder]
timescale_list = [timescale_list[i] for i in list_reorder]

lattice_list = [lattice_list[i] for i in list_reorder]
mo_traj_list = [mo_traj_list[i] for i in list_reorder]
s_traj_list = [s_traj_list[i] for i in list_reorder]
vac_traj_list = [vac_traj_list[i] for i in list_reorder]
atoms_list = [atoms_list[i] for i in list_reorder]

beam_current_list = [beam_current_list[i] for i in list_reorder]
dosescale_list = [dosescale_list[i] for i in list_reorder]

lattice_norm_list = [lattice_norm_list[i] for i in list_reorder]
mo_traj_norm_list = [mo_traj_norm_list[i] for i in list_reorder]
s_traj_norm_list = [s_traj_norm_list[i] for i in list_reorder]
vac_traj_norm_list = [vac_traj_norm_list[i] for i in list_reorder]
atoms_norm_list = [atoms_norm_list[i] for i in list_reorder]

mo_points_list = [mo_points_list[i] for i in list_reorder]
s_points_list = [s_points_list[i] for i in list_reorder]
vac_points_list = [vac_points_list[i] for i in list_reorder]






colors_bc = divide_colors_bc(beam_current_list, cmap_names = ('Blues', 'Greens', 'Reds'))

colors_dosescale = get_colors_dosescale(dosescale_list, minmax=2)   






numwet = np.sum([wetdry == 'wet' for wetdry in wetdry_list])
numdry = np.sum([wetdry == 'dry' for wetdry in wetdry_list])
numinclude = numwet + numdry



sl_a = np.array(scale_list)
meanscale = sl_a.mean()
stdscale = sl_a.std()


numvid = len(vid_list)


plot_indexes = []

drycount = numwet + 1
wetcount = 0
for i in range(numvid):
    wetdry = wetdry_list[i]
    if wetdry == 'wet':
        plot_indexes.append(wetcount)
        wetcount += 1
    elif wetdry == 'dry':
        plot_indexes.append(drycount)
        drycount += 1
        
        

# find averaged lattice positions

setdist = 0.315
initial_vector = np.array([setdist, 0])
mo_points_fixed = []
for i in range(6):
    angle_deg = -150 + (i*60)
    angle = np.deg2rad(angle_deg)
    out_pos_x = initial_vector[0]*np.cos(angle) - initial_vector[1]*np.sin(angle)
    out_pos_y = initial_vector[0]*np.sin(angle) + initial_vector[1]*np.cos(angle)
    mo_points_fixed.append([out_pos_x, out_pos_y])
mo_points_fixed.append([0, 0])
mo_points_fixed = np.array(mo_points_fixed)
s_points_fixed = np.array([np.mean(mo_points_fixed[[0, 1, 6]], axis=0),
                           np.mean(mo_points_fixed[[2, 3, 6]], axis=0),
                           np.mean(mo_points_fixed[[4, 5, 6]], axis=0)])
vac_points_fixed = np.array([np.mean(mo_points_fixed[[5, 0, 6]], axis=0),
                           np.mean(mo_points_fixed[[1, 2, 6]], axis=0),
                           np.mean(mo_points_fixed[[3, 4, 6]], axis=0)])
all_points_fixed = np.concatenate([mo_points_fixed, s_points_fixed])
points_fixed = [mo_points_fixed, s_points_fixed, vac_points_fixed]
    

fig, ax = plt.subplots(1,3)   
Tmat_list = []
for i in range(numvid):   
    clist=list(range(3))
    mo_points, s_points, vac_points = mo_points_list[i], s_points_list[i], vac_points_list[i]
    ax[0].scatter(s_points[:, 0], s_points[:, 1], c=list(range(3)), s=100, alpha=0.1, cmap='viridis')
    ax[0].scatter(mo_points[:, 0], mo_points[:, 1], c=list(range(7)), s=100, alpha=0.1, cmap='jet')
    s_angles = np.rad2deg(np.arctan2(s_points[:, 1], s_points[:, 0]))
    angle = s_angles[s_angles > 0].min()
    rotmat = [[np.cos(np.deg2rad(angle)), -np.sin(np.deg2rad(angle)), 0],
              [np.sin(np.deg2rad(angle)), np.cos(np.deg2rad(angle)), 0],
              [0, 0, 1]]
    mo_points_no0 = mo_points[~(mo_points == 0).all(axis=1)]
    mo_points_twist = Tunpad(np.dot(Tpad(mo_points_no0), rotmat))
    mo_points_twist = mo_points_twist[np.rad2deg(np.arctan2(mo_points_twist[:, 1], mo_points_twist[:, 0])).argsort()]
    s_points_twist = twist(s_points, -angle)
    mo_points_twist = np.concatenate([mo_points_twist, np.array([[0, 0]])], axis=0)
    ax[1].scatter(s_points_twist[:, 0], s_points_twist[:, 1], c=list(range(3)), s=100, alpha=0.1, cmap='viridis')
    ax[1].scatter(mo_points_twist[:, 0], mo_points_twist[:, 1], c=list(range(7)), s=100, alpha=0.1, cmap='jet')
    all_points = np.concatenate([mo_points_twist, s_points_twist])
    Tmat, res, rank, s = np.linalg.lstsq(Tpad(all_points), Tpad(all_points_fixed))
    Fullmat = np.dot(rotmat, Tmat)
    mo_points_transformed = Tunpad(np.dot(Tpad(mo_points), Fullmat))
    s_points_transformed = Tunpad(np.dot(Tpad(s_points), Fullmat))
    ax[2].scatter(mo_points_transformed[:, 0], mo_points_transformed[:, 1], c=list(range(7)), s=100, alpha=0.1, cmap='jet')
    ax[2].scatter(s_points_transformed[:, 0], s_points_transformed[:, 1], c=list(range(3)), s=100, alpha=0.1, cmap='viridis')
    Tmat_list.append(Fullmat)
    old_points_list = atoms_norm_list[i][['lat_x', 'lat_y']].values
    points_list_transformed = Tunpad(np.dot(Tpad(old_points_list), Fullmat))
    atoms_norm_list[i][['transformed_lat_x', 'transformed_lat_y']] = points_list_transformed
for a in ax:
    a.set_xlim(-0.4, 0.4)
    a.set_ylim(-0.4, 0.4)
    a.set_aspect('equal')
ax[0].set_title('Initial')
ax[1].set_title('Rotated')
ax[2].set_title('Affine')
fig.savefig(savefolder + 'points_align_check.png')
plt.close('all')

    


#create combined dataframes (wet and dry)
wet_atoms_norm_list = []
dry_atoms_norm_list = []
wet_lattice_norm_list = []
dry_lattice_norm_list = []           
for i in range(numvid):
    settings = settings_list[i]
    atoms_norm = atoms_norm_list[i].copy()
    atoms_norm['video'] = settings.filename[:-5]
    atoms_norm['beam_current'] = beam_current_list[i]
    atoms_norm['dosescale'] = dosescale_list[i]
    lattice_norm = lattice_norm_list[i].copy()
    lattice_norm['video'] = settings.filename[:-5]
    lattice_norm['beam_current'] = beam_current_list[i]
    lattice_norm['dosescale'] = dosescale_list[i]
    if wetdry_list[i] == 'wet':
        wet_atoms_norm_list.append(atoms_norm)
        wet_lattice_norm_list.append(lattice_norm)
    elif wetdry_list[i] == 'dry':
        dry_atoms_norm_list.append(atoms_norm)
        dry_lattice_norm_list.append(lattice_norm)
        
particle_num = 0
for i in range(len(wet_atoms_norm_list)):
    wet_atoms_norm_list[i].particle += particle_num
    particle_num += wet_atoms_norm_list[i].particle.max()
    
particle_num = 0
for i in range(len(dry_atoms_norm_list)):
    dry_atoms_norm_list[i].particle += particle_num
    particle_num += dry_atoms_norm_list[i].particle.max()
    
wet_atoms_norm_all = pd.concat(wet_atoms_norm_list)
dry_atoms_norm_all = pd.concat(dry_atoms_norm_list)

        
particle_num = 0
for i in range(len(wet_lattice_norm_list)):
    wet_lattice_norm_list[i].particle += particle_num
    particle_num += wet_lattice_norm_list[i].particle.max()
    
particle_num = 0
for i in range(len(dry_lattice_norm_list)):
    dry_lattice_norm_list[i].particle += particle_num
    particle_num += dry_lattice_norm_list[i].particle.max()
    
wet_atoms_norm_all = pd.concat(wet_atoms_norm_list)
dry_atoms_norm_all = pd.concat(dry_atoms_norm_list)
wet_lattice_norm_all = pd.concat(wet_lattice_norm_list)
dry_lattice_norm_all = pd.concat(dry_lattice_norm_list)







        
        
fig, ax = plt.subplots(2, 2, sharex='col', sharey='row', figsize=(12,8))
wet_df = doseparams_df[doseparams_df.wetdry == 'wet']
dry_df = doseparams_df[doseparams_df.wetdry == 'dry']
wet_df.plot.bar(x='filename_reduced', y='beam_current_pA', ax=ax[0, 0], legend=False, color='blue')
dry_df.plot.bar(x='filename_reduced', y='beam_current_pA', ax=ax[0, 1], legend=False, color='red')
wet_df.plot.bar(x='filename_reduced', y='dose_per_nm_million', ax=ax[1, 0], legend=False, color='blue')
dry_df.plot.bar(x='filename_reduced', y='dose_per_nm_million', ax=ax[1, 1], legend=False, color='red')
for a in ax[1]:
    a.tick_params(axis='x', labelrotation=80)
    a.set_xlabel('')
for ax_arr in ax:
    for a in ax_arr:
        a.grid(axis='y')
ax[0, 0].set_ylabel('Beam Current (pA)')
ax[1, 0].set_ylabel(r'10$^6$ e per nm$^2$ per frame')
fig.subplots_adjust(top=0.955,
                    bottom=0.25,
                    left=0.06,
                    right=0.99,
                    hspace=0.05,
                    wspace=0.025)
fig.suptitle('Dose Parameters')
fig.savefig(savefolder + 'dose_params.png')



framechoice = 5
offmax = 100
halfsize = 64
inset_params = [0.7, 0.05, 0.25, 0.25]
startoffset = np.array([37, -37])
proflen = 10
savefolder_recon = savefolder + 'reconsructed_img_compare' + os.sep
if not os.path.exists(savefolder_recon):
    os.makedirs(savefolder_recon)
    for i in tqdm(range(numvid), desc='outputting PCA recon figure'):
        
        settings = settings_list[i]
        filename = settings.filename
        
        scale = scale_list[i]
        
        vidfolder = folder_base + "5_driftcorr" + os.sep + filename + os.sep
        latfolder = folder_base + "6_latticepoints" + os.sep + filename + os.sep
        
        vid = hs.load(vidfolder + 'original_video.hspy')
        latvid = hs.load(vidfolder + 'lattice_video.hspy')
        
        latpoints_all = pd.read_csv(latfolder + 'trajectories.csv')
        latpoints = latpoints_all[latpoints_all.frame == framechoice]
        latpairs = latpoints[['x', 'y']].to_numpy()
        
        vidframe = vid.inav[framechoice].data.copy()
        gaussframe = gaussian_filter(vidframe, 0.6)
        latframe = latvid.inav[framechoice].data
    
        orig_centre = (np.array(vidframe.shape) / 2).astype(int)
        offset = np.array([random.randint(-offmax, offmax), random.randint(-offmax, offmax)])
        centre = orig_centre + offset
        
        tree = KDTree(latpairs)        
        startpoint_search = (centre + startoffset)[::-1]
        dist , ind = tree.query(startpoint_search)
        startpoint = latpairs[ind]
        
        bluepoint_base = centre + np.array([37, 32])
        redpoint_base = centre + np.array([-37, -22])
        
        searchpoints = (proflen * mo_points_list[i] / scale) + startpoint
        bluepoint_search = searchpoints[distance.cdist(searchpoints, [bluepoint_base[::-1]]).argmin()]
        redpoint_search = searchpoints[distance.cdist(searchpoints, [redpoint_base[::-1]]).argmin()]
        dist , ind = tree.query(bluepoint_search)
        bluepoint = latpairs[ind]
        dist , ind = tree.query(redpoint_search)
        redpoint = latpairs[ind]    
        
        fig = plt.figure(figsize=(12,8))
        ax_orig = plt.subplot2grid((4,6), (0,0), rowspan=2, colspan=2, fig=fig)
        ax_gauss = plt.subplot2grid((4,6), (0,2), rowspan=2, colspan=2, fig=fig)
        ax_recon = plt.subplot2grid((4,6), (0,4), rowspan=2, colspan=2, fig=fig)
        ax_profiles = plt.subplot2grid((4,6), (2,0), rowspan=2, colspan=6, fig=fig)
        ax_profiles.set_frame_on(False)
        ax_profiles.set_xticks([])
        ax_profiles.set_yticks([])
    
        ax_red = ax_profiles.inset_axes([0, 0.51, 1, 0.49])
        ax_blue = ax_profiles.inset_axes([0, 0, 1, 0.49])
        
        ax_orig.imshow(vidframe, cmap='inferno')
        ax_gauss.imshow(gaussframe, cmap='inferno')
        ax_recon.imshow(latframe, cmap='inferno')
        
        for ax in [ax_orig, ax_gauss, ax_recon]:
            ax.set_ylim(centre[0] - halfsize, centre[0] + halfsize - 1)
            ax.set_xlim(centre[1] - halfsize, centre[1] + halfsize - 1)
            ax.set_xticks([])
            ax.set_yticks([])
            sb = ScaleBar(scale, 'nm', box_alpha=0, color='white', fixed_value=1)  
            ax.add_artist(sb)
            ax.annotate("", xy=bluepoint, xytext=startpoint, arrowprops=dict(edgecolor='blue', facecolor='blue', width=2))
            ax.annotate("", xy=redpoint, xytext=startpoint, arrowprops=dict(edgecolor='red', facecolor='red', width=2))
            # ax.scatter(bluepoint_base[1], bluepoint_base[0], color='blue')
            # ax.scatter(redpoint_base[1], redpoint_base[0], color='red')
        
        vidfft = np.log(np.abs(np.fft.fftshift(np.fft.fft2(vidframe[centre[0]-halfsize: centre[0]+halfsize, centre[1]-halfsize: centre[1]+halfsize]))))
        gaussfft = np.log(np.abs(np.fft.fftshift(np.fft.fft2(gaussframe[centre[0]-halfsize: centre[0]+halfsize, centre[1]-halfsize: centre[1]+halfsize]))))
        latfft = np.log(np.abs(np.fft.fftshift(np.fft.fft2(latframe[centre[0]-halfsize: centre[0]+halfsize, centre[1]-halfsize: centre[1]+halfsize]))))
            
        ax_orig_ins = ax_orig.inset_axes(inset_params)
        ax_gauss_ins = ax_gauss.inset_axes(inset_params)
        ax_recon_ins = ax_recon.inset_axes(inset_params)
        
        ax_orig_ins.imshow(vidfft, cmap='inferno')
        ax_gauss_ins.imshow(gaussfft, cmap='inferno')
        ax_recon_ins.imshow(latfft, cmap='inferno')
        
        for ax in [ax_orig_ins, ax_gauss_ins, ax_recon_ins]:
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlim(32, 97)
            ax.set_ylim(32, 97)
        
        indframe = np.mgrid[0:vidframe.shape[0], 0:vidframe.shape[1]:]
        distframe = ((indframe[0] - startpoint[1])**2 + (indframe[1] - startpoint[0])**2)**0.5
        
        red_profile_dist = profile_line(distframe.T, startpoint, redpoint, linewidth=1, order=1, mode='nearest')
        red_profile_vid = profile_line(vidframe.T, startpoint, redpoint, linewidth=1, order=1, mode='nearest')
        red_profile_gauss = profile_line(gaussframe.T, startpoint, redpoint, linewidth=1, order=1, mode='nearest')
        red_profile_lat = profile_line(latframe.T, startpoint, redpoint, linewidth=1, order=1, mode='nearest')
        
        blue_profile_dist = profile_line(distframe.T, startpoint, bluepoint, linewidth=1, order=1, mode='nearest')
        blue_profile_vid = profile_line(vidframe.T, startpoint, bluepoint, linewidth=1, order=1, mode='nearest')
        blue_profile_gauss = profile_line(gaussframe.T, startpoint, bluepoint, linewidth=1, order=1, mode='nearest')
        blue_profile_lat = profile_line(latframe.T, startpoint, bluepoint, linewidth=1, order=1, mode='nearest')
        
        reconfactor = (red_profile_vid.mean() / red_profile_lat.mean())/1.5
        
        ax_red.plot(red_profile_dist, red_profile_vid, label='Original Video')
        ax_red.plot(red_profile_dist, red_profile_gauss, label='Gaussian Filtered')
        ax_red.plot(red_profile_dist, red_profile_lat * reconfactor, label='Reconstructed')
        
        ax_blue.plot(blue_profile_dist, blue_profile_vid, label='Original Video')
        ax_blue.plot(blue_profile_dist, blue_profile_gauss, label='Gaussian Filtered')
        ax_blue.plot(blue_profile_dist, blue_profile_lat * reconfactor, label='Reconstructed')
    
        ax_red.spines['bottom'].set_color('red')
        ax_red.spines['top'].set_color('red')
        ax_red.spines['right'].set_color('red')
        ax_red.spines['left'].set_color('red')
        ax_red.tick_params(axis='both', color='red')
        
        ax_blue.spines['bottom'].set_color('blue')
        ax_blue.spines['top'].set_color('blue')
        ax_blue.spines['right'].set_color('blue')
        ax_blue.spines['left'].set_color('blue')
        ax_blue.tick_params(axis='both', color='blue')
        
        amax = max([red_profile_dist[-1], blue_profile_dist[-1]])
        ax_red.set_xlim(0, amax)
        ax_blue.set_xlim(0, amax)
        
        ax_blue.legend(loc='upper right')
        ax_orig.add_artist(AnchoredText(r"$\bf{a}$ Original Image", loc='upper left', frameon=False, prop=dict(color='white')))  
        ax_gauss.add_artist(AnchoredText(r"$\bf{b}$ Gaussian Filtered", loc='upper left', frameon=False, prop=dict(color='white')))  
        ax_recon.add_artist(AnchoredText(r"$\bf{c}$ Reconstructed", loc='upper left', frameon=False, prop=dict(color='white')))
        ax_red.add_artist(AnchoredText(r"$\bf{d}$ Line Profile", loc='upper left', frameon=False, prop=dict(color='black')))
        ax_blue.add_artist(AnchoredText(r"$\bf{e}$ Line Profile", loc='upper left', frameon=False, prop=dict(color='black')))
        
        ax_profiles.set_ylabel('Normalised Intensity')
        ax_red.xaxis.set_ticks_position('top')
        ax_blue.set_xlabel('Distance (nm)')
        ax_red.set_yticks([])
        ax_blue.set_yticks([])
        ax_red.grid(True)
        ax_blue.grid(True)
        ax_red.set_xticklabels([])
        
        fig.savefig(savefolder_recon + settings.filename[:-5] + '.png')
        fig.savefig(savefolder_recon + settings.filename[:-5] + '.pdf')
        fig.savefig(savefolder_recon + settings.filename[:-5] + '.svg')
        
        plt.close('all')
        gc.collect()













# plot lattice number vs frame (beam damage) UNFINISHED
fig, ax = plt.subplots(2)
for i in range(numvid):
    lattice = lattice_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    vid = vid_list[i]
    wetdry = wetdry_list[i]
    framecounts = np.unique(lattice.frame, return_counts=True)
    if wetdry == 'wet':
        ax[0].plot(framecounts[0], framecounts[1] / framecounts[1].max(), color=color, label=settings.filename[:-5])
    elif wetdry == 'dry':
        ax[1].plot(framecounts[0], framecounts[1] / framecounts[1].max(), color=color, label=settings.filename[:-5])
ax[0].set_xlabel('Elapsed frames')
ax[0].set_xlim(left=1)
ax[0].set_ylim(0.45, 1.05)
ax[0].set_ylabel('Fraction of Remaining Lattice Mo')
ax[0].legend()

fig.savefig(savefolder + 'Mo Lattice Remaining.png')
fig.savefig(savefolder + 'Mo Lattice Remaining.pdf')
















# plot nearest neighbour pt distributions

fig, ax = plt.subplots(1, figsize=(16,10))
for i in range(numvid):
    atoms = atoms_norm_list[i]
    settings = settings_list[i]
    sns.distplot(atoms.pt_dist, color=colors_bc[i], ax=ax, kde_kws={"bw": 0.1, "alpha": 0.5}, hist_kws={"alpha": 0.5}, label=settings.filename[:-5])
ax.legend()
ax.set_xlim(0, 5)
ax.set_xlabel('Pt Nearest Neighbour (nm)')
fig.suptitle("Pt NN distance histogram")
fig.tight_layout()
fig.savefig(savefolder + 'Pt distance Histograms.png')






isolatedfolder = savefolder + 'isolated' + os.sep
os.makedirs(isolatedfolder, exist_ok = True)

splitvals = (0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8)

framecutval = 40

# for splitval in splitvals:
    
#     fig, ax = plt.subplots(1, 1, figsize=(16, 8))
#     ax_i = ax.inset_axes([0.05, 0.6, 0.35, 0.35])
    
#     for i in range(numvid):
#         atoms = atoms_norm_list[i]
#         settings = settings_list[i]
        
#         atoms_split = neartraj_split(atoms, splitval)

#         emsd = momentcalc(atoms, 1, 1/timescale_list[i], 2)
#         emsdfit, emsd_cov = curve_fit(modelfunc, emsd.lagt.values[0: framecutval], emsd.mmd.values[0: framecutval], p0=[0.8, 0.005], sigma=emsd.err.values[0: framecutval], maxfev=5000)

#         emsd_split = momentcalc(atoms_split, 1, 1/timescale_list[i], 2)
#         emsd_split_fit, emsd_split_cov = curve_fit(modelfunc, emsd_split.lagt.values[0: framecutval], emsd_split.mmd.values[0: framecutval], p0=[0.8, 0.005], sigma=emsd_split.err.values[0: framecutval], maxfev=5000)
        
#         ax.errorbar(emsd.lagt.values, emsd.mmd.values, yerr=emsd.err.values/2, fmt='.', label = settings.filename[:-5] + ' Original Data', color=colors_bc[i])
#         ax.plot(emsd.lagt.values, modelfunc(emsd.lagt.values, emsdfit[0], emsdfit[1]), label = settings.filename[:-5] + ' Original Fit', color=colors_bc[i])
#         ax.errorbar(emsd_split.lagt.values, emsd_split.mmd.values, yerr=emsd_split.err.values/2, fmt='.', label = settings.filename[:-5] + ' Isolated Data', color=colors_bc[i], alpha=0.5)
#         ax.plot(emsd_split.lagt.values, modelfunc(emsd_split.lagt.values, emsd_split_fit[0], emsd_split_fit[1]), label = settings.filename[:-5] + ' Isolated Fit', color=colors_bc[i], alpha=0.5)
        
#         ax_i.errorbar(emsd.lagt.values, emsd.mmd.values, yerr=emsd.err.values/2, fmt='.', label = settings.filename[:-5] + ' Original Data', color=colors_bc[i])
#         ax_i.plot(emsd.lagt.values, modelfunc(emsd.lagt.values, emsdfit[0], emsdfit[1]), label = settings.filename[:-5] + ' Original Fit', color=colors_bc[i])
#         ax_i.errorbar(emsd_split.lagt.values, emsd_split.mmd.values, yerr=emsd_split.err.values/2, fmt='.', label = settings.filename[:-5] + ' Isolated Data', color=colors_bc[i], alpha=0.5)
#         ax_i.plot(emsd_split.lagt.values, modelfunc(emsd_split.lagt.values, emsd_split_fit[0], emsd_split_fit[1]), label = settings.filename[:-5] + ' Isolated Fit', color=colors_bc[i], alpha=0.5)

#         # ax[0].annotate(r'D$_\alpha$=%.3f $\alpha$=%.3f' % (wetfit[1], wetfit[0]), xy=(40,0.5), color='blue')
#         # ax[0].annotate(r'D$_\alpha$=%.3f $\alpha$=%.3f' % (wetsplitfit[1], wetsplitfit[0]), xy=(40,0.4), color='blue', alpha=0.5)
        
#         # ax[0].annotate(r'D$_\alpha$=%.3f $\alpha$=%.3f' % (dryfit[1], dryfit[0]), xy=(40,0.3), color='red')
#         # ax[0].annotate(r'D$_\alpha$=%.3f $\alpha$=%.3f' % (drysplitfit[1], drysplitfit[0]), xy=(40,0.2), color='red', alpha=0.5)
#         # ax[0].legend(loc='lower right')

#     ax.set_xlim(0,80)
#     ax.set_ylim(-0.05,1)

#     ax_i.set_xscale('log')
#     ax_i.set_yscale('log')
    
#     ax_i.set_xlim(2,120)
#     ax_i.set_ylim(0.02,1)
    

#     ax.set_ylabel(r'MSD (nm$^2$)')
#     ax_i.set_ylabel(r'MSD (nm$^2$)')
#     ax.set_xlabel(r'Lag Time $\Delta$t (s)')
#     ax_i.set_xlabel(r'Lag Time $\Delta$t (s)')
    
#     ax.grid(which='major',axis='both', alpha = 0.1)
#     ax_i.grid(which='major',axis='both', alpha = 0.3)
#     ax_i.grid(which='minor',axis='both', alpha = 0.1)
    
#     fig.savefig(isolatedfolder + 'msdcompare_splitval_%.1f_fitframe_%d.png' %(splitval, framecutval))

#     plt.close('all')
    
    
    
    
    
    
    
    
    
    












# adatom number density plot

fig, ax1 = plt.subplots(1, 2, figsize=(20,16))
ax2_0 = ax1[0].twinx()
ax2_1 = ax1[1].twinx()
ax2 = [ax2_0, ax2_1]

adatom_counts_list = []
adatom_density_list = []

for i in range(numvid):
    
    atoms = atoms_norm_list[i]
    settings = settings_list[i]
    area = (scale_list[i] * 512) ** 2
    wetdry = wetdry_list[i]
    
    adatom_counts = []
    for frame in atoms.frame.unique():
        frame_list = atoms[atoms.frame == frame]
        adatom_counts.append(len(frame_list))
    adatom_density = np.array(adatom_counts) / area
    
    adatom_counts_list.append(adatom_counts)
    adatom_density_list.append(adatom_density)
    
    if wetdry == 'wet':
        ax1[0].plot(adatom_counts, color=colors_dosescale[i], marker='o', linestyle='', label=settings.filename[:-5])
        ax2[0].plot(adatom_density, color=colors_dosescale[i], label='_nolabel_')
        
    if wetdry == 'dry':
        ax1[1].plot(adatom_counts, color=colors_dosescale[i], marker='o', linestyle='', label=settings.filename[:-5])
        ax2[1].plot(adatom_density, color=colors_dosescale[i], label='_nolabel_')

ax1[0].set_ylabel('Counts (dots)')
ax1[0].set_xlabel('Frames')
ax2[0].set_ylabel('Adatom Density (per nm^-2) - (lines)')
ax1[0].set_title('Liquid Cell')

ax1[1].set_ylabel('Counts (dots)')
ax1[1].set_xlabel('Frames')
ax2[1].set_ylabel('Adatom Density (per nm^-2) - (lines)')
ax1[1].set_title('Vacuum')

ax1[0].set_ylim(bottom=0)
ax2[0].set_ylim(bottom=0)

ax1[1].set_ylim(ax1[0].get_ylim())
ax2[1].set_ylim(ax2[0].get_ylim())

ax1[0].legend()
ax1[1].legend()

fig.savefig(savefolder + 'areal_density.png')
fig.savefig(savefolder + 'areal_density.pdf')







# adatom number density plot (only area always in frame)

# adatom number density plot

fig, ax1 = plt.subplots(1, 2, figsize=(20,16))
ax2_0 = ax1[0].twinx()
ax2_1 = ax1[1].twinx()
ax2 = [ax2_0, ax2_1]

adatom_counts_norm_list = []
adatom_density_norm_list = []

for i in range(numvid):
    
    atoms = atoms_list[i]
    settings = settings_list[i]

    vid = vid_list[i]
    wetdry = wetdry_list[i]
    const_im = np.min(vid.data, axis=0) > 0
    const = np.where(const_im)
    const_area = const_im.sum() * (scale_list[i]**2)
    x_bool = (atoms.x >= const[1].min()) & (atoms.x <= const[1].max())
    y_bool = (atoms.y >= const[0].min()) & (atoms.y <= const[0].max())
    atoms_const = atoms[x_bool & y_bool]
    
    adatom_counts = []
    for frame in atoms_const.frame.unique():
        frame_list = atoms_const[atoms_const.frame == frame]
        adatom_counts.append(len(frame_list))
    adatom_density = np.array(adatom_counts) / const_area
    
    adatom_counts_norm_list.append(adatom_counts)
    adatom_density_norm_list.append(adatom_density)

    
    if wetdry == 'wet':
        ax1[0].plot(adatom_counts, color=colors_bc[i], marker='o', linestyle='', label=settings.filename[:-5] + '_%.1fpA' %(beam_current_list[i]*1e12))
        ax2[0].plot(adatom_density, color=colors_bc[i], label='_nolabel_')
        
    if wetdry == 'dry':
        ax1[1].plot(adatom_counts, color=colors_bc[i], marker='o', linestyle='', label=settings.filename[:-5] + '_%.1fpA' %(beam_current_list[i]*1e12))
        ax2[1].plot(adatom_density, color=colors_bc[i], label='_nolabel_')

ax1[0].set_ylabel('Counts (dots)')
ax1[0].set_xlabel('Frames')
ax2[0].set_ylabel('Adatom Density (per nm^-2) - (lines)')
ax1[0].set_title('Liquid')

ax1[1].set_ylabel('Counts (dots)')
ax1[1].set_xlabel('Frames')
ax2[1].set_ylabel('Adatom Density (per nm^-2) - (lines)')
ax1[1].set_title('Liquid')

ax1[0].set_ylim(bottom=0)
ax2[0].set_ylim(bottom=0)

ax1[1].set_ylim(ax1[0].get_ylim())
ax2[1].set_ylim(ax2[0].get_ylim())

ax1[0].legend()
ax1[1].legend()

fig.savefig(savefolder + 'areal_density_normed.png')
fig.savefig(savefolder + 'areal_density_normed.pdf')











framecutval = 20
splitval = 0.5

for i in range(numvid):

    atoms = atoms_norm_list[i]
    settings = settings_list[i]

    density = adatom_density_norm_list[i]
    bc = beam_current_list[i]
    wetdry = wetdry_list[i]
    
    savename = isolatedfolder + '%d_%s_' %(i, wetdry_list[i]) + settings.filename[:-5] + '_msdcompare_splitval_%.1f_fitframe_%d.png' %(splitval, framecutval)
    
    if not os.path.exists(savename):
        print(savename)
        atoms_split = neartraj_split(atoms, splitval)
        emsd = momentcalc(atoms, 1, 1/timescale_list[i], 2)
        emsdfit, emsd_cov = curve_fit(modelfunc, emsd.lagt.values[0: framecutval], emsd.mmd.values[0: framecutval], sigma=emsd.err.values[0: framecutval])
        
        emsd_split = momentcalc(atoms_split, 1, 1/timescale_list[i], 2)
        emsd_split_fit, emsd_split_cov = curve_fit(modelfunc, emsd_split.lagt.values[0: framecutval], emsd_split.mmd.values[0: framecutval], sigma=emsd_split.err.values[0: framecutval])
        
        fig, ax = plt.subplots(1, 1, figsize=(16, 8))
        ax_i = ax.inset_axes([0.05, 0.6, 0.35, 0.35])
        
        bplot = ~np.isnan(emsd.err)
        bplot_s = ~np.isnan(emsd_split.err)
        ax.errorbar(emsd.lagt.values[bplot], emsd.mmd.values[bplot], yerr=emsd.err.values[bplot], fmt='.', label = settings.filename[:-5] + ' Original Data', color=colors_bc[i])
        ax.plot(emsd.lagt.values, modelfunc(emsd.lagt.values, emsdfit[0], emsdfit[1]), label = settings.filename[:-5] + ' Original Fit', color=colors_bc[i])
        ax.errorbar(emsd_split.lagt.values[bplot_s], emsd_split.mmd.values[bplot_s], yerr=emsd_split.err.values[bplot_s], fmt='.', label = settings.filename[:-5] + ' Isolated Data', color=colors_bc[i], alpha=0.5)
        ax.plot(emsd_split.lagt.values, modelfunc(emsd_split.lagt.values, emsd_split_fit[0], emsd_split_fit[1]), label = settings.filename[:-5] + ' Isolated Fit', color=colors_bc[i], alpha=0.5)
        
        ax_i.errorbar(emsd.lagt.values[bplot], emsd.mmd.values[bplot], yerr=emsd.err.values[bplot], fmt='.', label = settings.filename[:-5] + ' Original Data', color=colors_bc[i])
        ax_i.plot(emsd.lagt.values, modelfunc(emsd.lagt.values, emsdfit[0], emsdfit[1]), label = settings.filename[:-5] + ' Original Fit', color=colors_bc[i])
        ax_i.errorbar(emsd_split.lagt.values[bplot_s], emsd_split.mmd.values[bplot_s], yerr=emsd_split.err.values[bplot_s], fmt='.', label = settings.filename[:-5] + ' Isolated Data', color=colors_bc[i], alpha=0.5)
        ax_i.plot(emsd_split.lagt.values, modelfunc(emsd_split.lagt.values, emsd_split_fit[0], emsd_split_fit[1]), label = settings.filename[:-5] + ' Isolated Fit', color=colors_bc[i], alpha=0.5)
        
        
        ax.set_xlim(0, 30 * timescale_list[i])
        ax.set_ylim(bottom=0, top=1)
        
        ax_i.set_xscale('log')
        ax_i.set_yscale('log')
        
        ax_i.set_xlim(timescale_list[i], 60 * timescale_list[i])
        ax_i.set_ylim(0.02,1)
        
        ax.legend(loc='lower right')
        
        ax.set_ylabel(r'MSD (nm$^2$)')
        ax_i.set_ylabel(r'MSD (nm$^2$)')
        ax.set_xlabel(r'Lag Time $\Delta$t (s)')
        ax_i.set_xlabel(r'Lag Time $\Delta$t (s)')
        
        ax.grid(which='major',axis='both', alpha = 0.3)
        ax_i.grid(which='major',axis='both', alpha = 0.6)
        ax_i.grid(which='minor',axis='both', alpha = 0.3)
        
        ax.set_title("%d %s %dpA %.2f density %s" %(i, wetdry, bc * 1e12, density.mean(), settings.filename[:-5]))
        
        fig.savefig(savename)
        
        plt.close('all')
        gc.collect()











plt.close('all')

framecutval = 20
splitval = 0.5

selection = [2, 4, 9, 16]

grid = (2,3)
fig = plt.figure(figsize=(7.06, 5))

fig.subplots_adjust(top=0.95,
                    bottom=0.095,
                    left=0.095,
                    right=0.92,
                    hspace=0.25,
                    wspace=0.25)

statcolors = ['red', 'darkorange', 'blue', 'aqua']

ax_dens = plt.subplot2grid(grid, (0, 0), fig=fig)
for i in range(numvid):
    wetdry = wetdry_list[i]
    density = adatom_density_norm_list[i]
    timescale = timescale_list[i]
    if wetdry == 'wet':
        ax_dens.plot(np.array(range(0, len(density))) * timescale, density, color='blue', alpha=0.5)
    elif wetdry == 'dry':
        ax_dens.plot(np.array(range(0, len(density))) * timescale, density, color='red', alpha=0.5)
ax_dens.set_ylim(bottom=0)
ax_dens.set_xlim(0, 200)
ax_dens.set_xlabel('Time (s)')
ax_dens.set_ylabel(r'Pt Adatom Density (nm$^{-2}$)')

ax_dist = plt.subplot2grid(grid, (1, 0), fig=fig)
bins = np.linspace(0,8,256)
for j in range(4):
    i = selection[j]
    atoms = atoms_norm_list[i]
    color = statcolors[j]
    ax_dist.hist(atoms.pt_dist, bins=bins, color=color, density=True, histtype='stepfilled', alpha=0.2)
    ax_dist.hist(atoms.pt_dist, bins=bins, color=color, density=True, histtype='step')
ax_dist.axvline(1)
ax_dist.set_xlim(0, 4)
ax_dist.set_xlabel('Adatom Nearest Neighbour distance (nm)')
ax_dist.set_ylabel('Probability Density')

ax_msd_all = plt.subplot2grid(grid, (0, 1), rowspan=2, colspan=2, fig=fig)
axis_off(ax_msd_all)
ax_msd = ax_msd_all.inset_axes([0, 0, 0.46, 1])
ax_msd_split = ax_msd_all.inset_axes([0.54, 0, 0.46, 1])

ax_msd_i = ax_msd.inset_axes([0.02, 0.54, 0.45, 0.45])
ax_msd_split_i = ax_msd_split.inset_axes([0.02, 0.54, 0.45, 0.45])

for j in range(4):
    i = selection[j]
    atoms = atoms_norm_list[i]
    atoms_split = neartraj_split(atoms, splitval)
    color = statcolors[j]
    
    emsd = momentcalc(atoms, 1, 1/timescale_list[i], 2)
    emsdfit, emsd_cov = curve_fit(modelfunc, emsd.lagt.values[0: framecutval], emsd.mmd.values[0: framecutval], sigma=emsd.err.values[0: framecutval])

    emsd_split = momentcalc(atoms_split, 1, 1/timescale_list[i], 2)
    emsd_split_fit, emsd_split_cov = curve_fit(modelfunc, emsd_split.lagt.values[0: framecutval], emsd_split.mmd.values[0: framecutval], sigma=emsd_split.err.values[0: framecutval])

    bplot = ~np.isnan(emsd.err)
    bplot_s = ~np.isnan(emsd_split.err)
    ax_msd.errorbar(emsd.lagt.values[bplot], emsd.mmd.values[bplot], yerr=emsd.err.values[bplot], fmt='.', label = '_nolabel_', color=color, alpha=0.5)
    ax_msd.plot(emsd.lagt.values, modelfunc(emsd.lagt.values, emsdfit[0], emsdfit[1]), label = r'D$_\alpha$ = %.3f, $\alpha$ = %.2f' % (emsdfit[1], emsdfit[0]), color=color, alpha=0.5)
    ax_msd_split.errorbar(emsd_split.lagt.values[bplot_s], emsd_split.mmd.values[bplot_s], yerr=emsd_split.err.values[bplot_s], fmt='.', label = '_nolabel_', color=color, alpha=0.5)
    ax_msd_split.plot(emsd_split.lagt.values, modelfunc(emsd_split.lagt.values, emsd_split_fit[0], emsd_split_fit[1]), label = r'D$_\alpha$ = %.3f, $\alpha$ = %.2f' % (emsd_split_fit[1], emsd_split_fit[0]), color=color, alpha=0.5)
    
    ax_msd_i.errorbar(emsd.lagt.values[bplot], emsd.mmd.values[bplot], yerr=emsd.err.values[bplot], fmt='.', label = settings.filename[:-5] + ' Original Data', color=color, alpha=0.5)
    ax_msd_i.plot(emsd.lagt.values, modelfunc(emsd.lagt.values, emsdfit[0], emsdfit[1]), label = settings.filename[:-5] + ' Original Fit', color=color, alpha=0.5)
    ax_msd_split_i.errorbar(emsd_split.lagt.values[bplot_s], emsd_split.mmd.values[bplot_s], yerr=emsd_split.err.values[bplot_s], fmt='.', label = settings.filename[:-5] + ' Isolated Data', color=color, alpha=0.5)
    ax_msd_split_i.plot(emsd_split.lagt.values, modelfunc(emsd_split.lagt.values, emsd_split_fit[0], emsd_split_fit[1]), label = settings.filename[:-5] + ' Isolated Fit', color=color, alpha=0.5)
    
for ax in [ax_msd, ax_msd_split]:
    ax.set_xlim(0, 30 * timescale_list[i])
    ax.set_ylim(bottom=0, top=1)
    ax.grid()
    ax.set_ylabel(r'TAMSD (nm$^2$')
    ax.set_xlabel('Time (s)')

for ax_i in [ax_msd_i, ax_msd_split_i]:
    ax_i.set_xscale('log')
    ax_i.set_yscale('log')
    ax_i.grid(which='both')
    ax_i.set_xlim(timescale_list[i], 60 * timescale_list[i])
    ax_i.set_ylim(0.02,1)
    ax_i.yaxis.tick_right()
    ax_i.yaxis.set_label_position('right')

ax_msd_split.yaxis.tick_right()
ax_msd_split.yaxis.set_label_position('right')

ax_dens.set_title(r'$\bf{a}$ Adatom Density', loc='left')
ax_dist.set_title(r'$\bf{b}$ Pt-Pt Nearest Neighbour', loc='left')
ax_msd.set_title(r'$\bf{c}$ All adatom MSD', loc='left')
ax_msd_split.set_title(r'$\bf{d}$ Isolated adatom MSD', loc='left')

legend_elements = [Line2D([0], [0], color='blue', lw=2, label='Liquid Cell'),
                   Line2D([0], [0], color='red', lw=2, label='Vacuum')]

ax_dens.legend(handles=legend_elements, loc='upper right')

legend_elements = []
for j in range(4):
    i = selection[j]
    density = adatom_density_norm_list[i].mean()
    bc = beam_current_list[i] * 1e12
    if j > 1:
        typestr = r'Liquid %dpA, density=%.2f nm$^{-2}$' %(bc, density)
    elif j < 2:
        typestr = r'Vacuum %dpA, density=%.2f nm$^{-2}$' % (bc, density)
    line = Line2D([0], [0], color=statcolors[j], lw=2, label=typestr)
    legend_elements.append(line)

ax_dist.legend(handles=legend_elements, loc='upper right')

ax_msd.legend(loc='lower right')
ax_msd_split.legend(loc='lower right')

fig.savefig(savefolder_figs + 'isolated_atoms.png', dpi=300)
fig.savefig(savefolder_figs + 'isolated_atoms.pdf', dpi=300)
fig.savefig(savefolder_figs + 'isolated_atoms.svg', dpi=300)




# plot atomic locations for each video

savefolder_loc = savefolder + 'atomic_locations' + os.sep

if not os.path.exists(savefolder_loc):
    os.makedirs(savefolder_loc)
    for i in tqdm(range(numvid), desc="plotting atomic locations"):
        
        atoms = atoms_list[i]
        settings = settings_list[i]
        vid = vid_list[i]
    
        locationplot_s(atoms, vid, (0, vid.data.shape[0]), s=30, alpha=0.1)
        plt.gcf().savefig(savefolder_loc + settings.filename[:-5] +'_s.png')
        locationplot_all(atoms, vid, (0, vid.data.shape[0]), s=30)
        plt.gcf().savefig(savefolder_loc + settings.filename[:-5] +'_all.png')
        
        plt.close('all')
        gc.collect()
    









# Pt adatom pair distributions


fig, ax = plt.subplots(figsize=(12,10))

for i in range(numvid):
    
    atoms = atoms_list[i]
    settings = settings_list[i]
    
    corrlist = []
    for fnum in atoms.frame.unique():
        atomframe = atoms[atoms.frame == fnum]
        atomnum = len(atomframe)
        if atomnum > 5:
            bins, paircorr = tp.static.pair_correlation_2d(atomframe, 60, dr=0.4, fraction=1, max_rel_ndensity=1000)
            corrlist.append(paircorr * atomnum)
            
    if len(corrlist) > 5:    
        totcorr = np.sum(corrlist, axis=0) 
            
        bincentres = (bins[1:] + bins[:-1])/2
        
        real_bincentres = scale_list[i] * bincentres
        paircorr_prob = real_bincentres * totcorr
    
        ax.plot(real_bincentres, totcorr/totcorr.sum(), color=colors_bc[i], label=settings.filename[:-5])
    
ax.axvline(0.3179)
ax.axvline(0.38)
ax.set_xlim(0, 2)    
ax.legend()
ax.set_xlabel('Distance (nm)')
fig.suptitle('Pt adatom pair correlation')
fig.savefig(savefolder + 'basic_pair_correlation.png')
fig.savefig(savefolder + 'basic_pair_correlation.pdf')







bc_vals = np.sort(wet_atoms_norm_all.beam_current.unique())
cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
dry_cmap = plt.get_cmap('Reds')
wet_cmap = plt.get_cmap('Blues')

savefolder_pairs = savefolder + 'radial_distribution_function' + os.sep
os.makedirs(savefolder_pairs, exist_ok=True)

try:
    normdists = np.load(savefolder_pairs + 'normalised_rdf.npy')
except:
    randnum = int(2e4)
    randpoints = np.random.rand(randnum, 2)
    tree = cKDTree(randpoints)
    dists, ints = tree.query(randpoints, k=randnum)
    flatdists = dists.flatten()
    normdists = flatdists[flatdists != 0]
    np.save(savefolder_pairs + 'normalised_rdf.npy', normdists)
    
bins = np.linspace(0, 20, 512)
bin_centres = (bins[1:] + bins[:-1])/2
standardhist, bin_edges = np.histogram(normdists * 512 * scale, bins=bins, density=True)

binning = 10
binspacing = 0.3

mainfig, mainax = plt.subplots(2, figsize=(12, 10))
pairsexcludelist = ['Scanning Preview22',]
wet_distlists = []
dry_distlists = []
wet_bcs = []
dry_bcs = []
for i in tqdm(range(numvid), desc = 'calculating pair dists'):
    
    atoms = atoms_norm_list[i]
    settings = settings_list[i]
    wetdry = wetdry_list[i]
    bc_val = beam_current_list[i]
    cm_val = cm_vals[np.where(bc_val == bc_vals)[0][0]]
    
    scale = scale_list[i]
    
    vid = origvid_list[i]
    if vid.axes_manager[-1].size == 512 and scale_list[i] == scale:
        normhist = standardhist
    else:
        normhist, bin_edges = np.histogram(normdists * vid.axes_manager[-1].size * vid.axes_manager[-1].scale, bins=bins, density=True)
    
    distlist = []
    for fnum in atoms.frame.unique():
        atomframe = atoms[atoms.frame == fnum]
        atomnum = len(atomframe)
        if atomnum > 1:
            points = atomframe[['x', 'y']].values
            tree = cKDTree(points)
            dist_arr, int_arr = tree.query(points, k=len(points))
            flatdist = dist_arr.flatten()
            framedists = flatdist[flatdist != 0]
            distlist.append(framedists)
  
   
    if len(distlist) > 5:    
        flatdist = np.concatenate(distlist)
        if wetdry == 'wet':
            wet_distlists.append(distlist)
            wet_bcs.append(bc_val)
        if wetdry == 'dry':
            dry_distlists.append(distlist)
            dry_bcs.append(bc_val)
        hist, bin_edges = np.histogram(flatdist, bins=bins, density=True)
        
        fig, ax = plt.subplots(1)
        ax.plot(bin_centres, hist, label = 'pair distribution function')
        ax.plot(bin_centres, normhist, label = 'random distribution function')
        ax.legend()
        ax.set_title(settings.filename[:-5])
        fig.savefig(savefolder_pairs + 'prenorm_' + settings.filename[:-5] + '.png')
        
        plt.close(fig)
        gc.collect() 
        
        fig, ax = plt.subplots(1)
        ax.plot(bin_centres, hist/normhist, label = 'norm pair distribution function')
        ax.legend()
        ax.set_title(settings.filename[:-5])
        ax.axhline(1)
        fig.savefig(savefolder_pairs + 'norm_' + settings.filename[:-5] + '.png')
        
        plt.close(fig)
        gc.collect()
        
        if not settings.filename[:-5] in pairsexcludelist:
            if wetdry == 'wet':
                mainax[0].plot(bin_centres, hist/normhist, label=str(bc_val*1e12) + ' pA', drawstyle='steps', color=wet_cmap(cm_val), alpha=0.4)
            elif wetdry == 'dry':
                mainax[1].plot(bin_centres, hist/normhist, label=str(bc_val*1e12) + ' pA', drawstyle='steps', color=dry_cmap(cm_val), alpha=0.4)
                
        fig, ax = plt.subplots(1)
        tickpos, ticklabs = [], []
        for j in range(0, len(distlist), binning):
            flatdist = np.concatenate(distlist[j: j+binning])
            hist, bin_edges = np.histogram(flatdist, bins=bins, density=True)
            plothist = (hist/normhist) - (j*binspacing)
            ax.plot(bin_centres, plothist, color='black')
            ax.fill_between(bin_centres, plothist, - (j*binspacing), color='white')
            tickpos.append(-(j*binspacing))
            ticklabs.append('Frame %d - %d' %(j, j+binning-1))
        ax.set_xlim(0, 5)
        ax.set_xlabel('Pair Separation (nm)')
        ax.set_ylabel('Normalised Probability (vs. Random)')
        ax.set_yticks(tickpos)
        ax.set_yticklabels(ticklabs)
        ax.yaxis.set_ticks_position('right')
        fig.suptitle(settings.filename[:-5])
        fig.subplots_adjust(top=0.88,
                            bottom=0.11,
                            left=0.05,
                            right=0.81,
                            hspace=0.2,
                            wspace=0.2)
        fig.savefig(savefolder_pairs + 'time_evolution_' + settings.filename[:-5] + '.png')
        
        plt.close(fig)
        gc.collect()
        
wethist, bin_edges = np.histogram(np.concatenate(np.concatenate(wet_distlists)), bins=bins, density=True)
dryhist, bin_edges = np.histogram(np.concatenate(np.concatenate(dry_distlists)), bins=bins, density=True)

mainax[0].plot(bin_centres, wethist/normhist, label='All wet', drawstyle='steps', color='black', zorder=10)
mainax[1].plot(bin_centres, dryhist/normhist, label='All dry', drawstyle='steps', color='black', zorder=10)     
         
mainax[0].axhline(1)
mainax[0].legend()

mainax[1].axhline(1)
mainax[1].legend()

mainax[0].set_ylabel('Normalised Probability (vs. random spacing)')
mainax[0].set_xlabel('Distance (nm)')
mainax[1].set_ylabel('Normalised Probability (vs. random spacing)')
mainax[1].set_xlabel('Distance (nm)')

mainfig.suptitle('Pair Distribution functions')

mainax[0].axvline(0.3179)
mainax[0].axvline(0.241)           
mainax[1].axvline(0.3179)
mainax[1].axvline(0.241)

mainax[0].set_xlim(0, 10)
mainax[1].set_xlim(0, 10)
mainax[0].set_ylim(0, 10)
mainax[1].set_ylim(0, 10)

mainax[0].add_artist(AnchoredText(r"$\bf{a}$ Liquid Cell", loc='upper left', frameon=False))
mainax[1].add_artist(AnchoredText(r"$\bf{b}$ Vacuum", loc='upper left', frameon=False))

mainfig.savefig(savefolder_pairs + 'normed_compare.png')
mainfig.savefig(savefolder_pairs + 'normed_compare.pdf')
mainfig.savefig(savefolder_pairs + 'normed_compare.svg')

mainax[0].axvline(0.3179)
mainax[0].axvline(0.241)           
mainax[1].axvline(0.3179)
mainax[1].axvline(0.241)

mainax[0].set_xlim(0, 1)
mainax[1].set_xlim(0, 1)
mainax[0].set_ylim(0, 10)
mainax[1].set_ylim(0, 10)

mainfig.savefig(savefolder_pairs + 'normed_compare_zoom.png')
mainfig.savefig(savefolder_pairs + 'normed_compare_zoom.pdf')
mainfig.savefig(savefolder_pairs + 'normed_compare_zoom.svg')

plt.close('all')
gc.collect()


bins = np.linspace(0, 20, 256)
bin_centres = (bins[1:] + bins[:-1])/2
standardhist, bin_edges = np.histogram(normdists * 512 * scale, bins=bins, density=True)
cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
dry_cmap = plt.get_cmap('Reds')
wet_cmap = plt.get_cmap('Blues')

fig, ax = plt.subplots(2)

for i in range(3):
    bc = bc_vals[i]
    cm_val = cm_vals[i]
    wet_dists = np.concatenate(np.concatenate(list(compress(wet_distlists,bc == wet_bcs))))
    dry_dists = np.concatenate(np.concatenate(list(compress(dry_distlists,bc == dry_bcs))))
    wet_hist, bin_edges = np.histogram(wet_dists, bins=bins, density=True)
    dry_hist, bin_edges = np.histogram(dry_dists, bins=bins, density=True)
    ax[0].plot(bin_centres, wet_hist/standardhist, label='Liquid %d pA'%(bc*1e12), color=wet_cmap(cm_val), zorder=5, drawstyle='steps-mid')
    ax[1].plot(bin_centres, dry_hist/standardhist, label='Vacuum %d pA'%(bc*1e12), color=dry_cmap(cm_val), zorder=5, drawstyle='steps-mid')

for a in ax:
    a.set_xlim(0, 10)
    a.set_ylim(0, 6)
    a.axhline(1, color='black', zorder=1)
    a.legend()
    a.set_ylabel('Relative Probability')    

ax[0].set_xticklabels([])
ax[1].set_xlabel('Pt separation (nm)')
fig.suptitle('Pair Distribution Function')

fig.subplots_adjust(top=0.93,
                    bottom=0.1,
                    left=0.065,
                    right=0.975,
                    hspace=0.2,
                    wspace=0.2)

fig.savefig(savefolder_pairs + 'final.png')
fig.savefig(savefolder_pairs + 'final.pdf')
fig.savefig(savefolder_pairs + 'final.svg')
    
    

    
    





















# pairlagtimes = [1, 2, 5, 10, 20, 40]
# for lagt in tqdm(pairlagtimes):
    
#     savefolder_pairdiffs = savefolder + 'pairdiffs/' + str(lagt) + '/'
#     os.makedirs(savefolder_pairdiffs, exist_ok=True)   
        
#     for i in range(numvid):
        
#         atoms = atoms_norm_list[i]
#         settings = settings_list[i]
#         startdiffs, enddiffs = get_pairs(atoms, lagt)

        
#         np.save(savefolder_pairdiffs + settings.filename[:-5] + '_startdiffs.npy', startdiffs)
#         np.save(savefolder_pairdiffs + settings.filename[:-5] + '_enddiffs.npy', enddiffs)

#         plt.figure()
#         plt.scatter(startdiffs, (enddiffs), s=1, alpha=0.05)
#         plt.savefig(savefolder_pairdiffs + settings.filename[:-5] + '_scatter.png')

#         plt.close('all')
#     gc.collect()
    
    
    









# drift calculation

atom_drifts = []
lattice_drifts = []

savefolder_drift = savefolder + 'drift' + os.sep
os.makedirs(savefolder_drift, exist_ok=True)

atom_max = 0.5
lat_max = 0.04

for i in range(numvid):
    
    atoms_norm = atoms_norm_list[i]
    lattice_norm = lattice_norm_list[i]
    settings = settings_list[i]

    drift = tp.compute_drift(atoms_norm)
    firstpoint = pd.DataFrame({"x": 0.0, "y": 0.0}, index=[0])
    drift = drift.append(firstpoint, sort=False)
    drift = drift.sort_index()
    
    drift_lat = tp.compute_drift(lattice_norm)
    drift_lat = drift_lat.append(firstpoint, sort=False)
    drift_lat = drift_lat.sort_index()
    
    atom_drifts.append(drift)
    lattice_drifts.append(drift_lat)
    
    fig, ax = plt.subplots(1,2, figsize=(12,8))
    ax[0].set_aspect('equal')
    ax[1].set_aspect('equal')
    colorlineplot(drift.x, drift.y, ax[0], 'jet')
    colorlineplot(drift_lat.x, drift_lat.y, ax[1], 'jet')
    ax[0].set_xlim(-atom_max, atom_max)
    ax[0].set_ylim(-atom_max, atom_max)
    ax[1].set_xlim(-lat_max, lat_max)
    ax[1].set_ylim(-lat_max, lat_max)
    ax[0].set_title('Atom Average Drift (nm)')
    ax[1].set_title('Lattice Average Drift (nm)')
    fig.suptitle(settings.filename[:-5] + ' Drift')
    fig.tight_layout()
    fig.savefig(savefolder_drift + settings.filename[:-5] + '_drift.png')
    
    plt.close('all')
    gc.collect()



fig, ax = plt.subplots(2,2, figsize=(16,16))
for a_arr in ax:
    for a in a_arr:
        a.set_aspect('equal')
for i in range(numvid):
    
    drift = atom_drifts[i]
    settings = settings_list[i]
    wetdry = wetdry_list[i]
    
    if wetdry == 'wet':
        ax[0, 0].plot(drift.x, drift.y, color=colors_bc[i], alpha=0.4)
        ax[0, 0].scatter(drift.x.values[-1], drift.y.values[-1], color=colors_bc[i], s=20)
        
        drift_lat = lattice_drifts[i]
        ax[0, 1].plot(drift_lat.x, drift_lat.y, color=colors_bc[i], alpha=0.4, label=settings.filename[:-5])
        ax[0, 1].scatter(drift_lat.x.values[-1], drift_lat.y.values[-1], color=colors_bc[i], s=20, label='_nolabel_')

    elif wetdry == 'dry':
        ax[1, 0].plot(drift.x, drift.y, color=colors_bc[i], alpha=0.4)
        ax[1, 0].scatter(drift.x.values[-1], drift.y.values[-1], color=colors_bc[i], s=20)
        
        drift_lat = lattice_drifts[i]
        ax[1, 1].plot(drift_lat.x, drift_lat.y, color=colors_bc[i], alpha=0.4, label=settings.filename[:-5])
        ax[1, 1].scatter(drift_lat.x.values[-1], drift_lat.y.values[-1], color=colors_bc[i], s=20, label='_nolabel_')
        
ax[0, 0].set_xlim(-atom_max, atom_max)
ax[0, 0].set_ylim(-atom_max, atom_max)
ax[0, 1].set_xlim(-lat_max, lat_max)
ax[0, 1].set_ylim(-lat_max, lat_max)
ax[0, 0].set_title('Atom Average Drift in Liquid (nm)')
ax[0, 1].set_title('Lattice Average Drift in Liquid (nm)')
ax[0, 1].legend()

ax[1, 0].set_xlim(-atom_max, atom_max)
ax[1, 0].set_ylim(-atom_max, atom_max)
ax[1, 1].set_xlim(-lat_max, lat_max)
ax[1, 1].set_ylim(-lat_max, lat_max)
ax[1, 0].set_title('Atom Average Drift in Vacuum (nm)')
ax[1, 1].set_title('Lattice Average Drift in Vacuum (nm)')
ax[1, 1].legend()

fig.tight_layout()
fig.savefig(savefolder_drift + 'combined_drift.png')

plt.close('all')
gc.collect()











savefolder_intensity = savefolder + 'adf_intensity' + os.sep
os.makedirs(savefolder_intensity, exist_ok=True)

# comparing intensity histograms plot

center_lists = []

for i in tqdm(range(numvid), desc = 'fitting intensity profiles'):
    
    origvid = origvid_list[i]
    settings = settings_list[i]
    
    savename = savefolder_intensity + settings.filename[:-5] + '_intensity_comp'


    std = np.std(origvid.data)
    medval = np.median(origvid.data)
    
    bins = np.linspace(medval - 3 * std, medval + 5 * std, 128)
    bin_centres = (bins[1:] + bins[:-1]) / 2
    
    fig = plt.figure(figsize=(14, 8))
    ax1 = plt.subplot2grid((4, 12), (0, 0), colspan=4, rowspan=4)
    ax1.get_yaxis().set_visible(False)
    ax2 = plt.subplot2grid((4, 12), (0, 4), colspan=4, rowspan=4)
    ax2.get_yaxis().set_visible(False)
    ax3 = plt.subplot2grid((4, 12), (0, 8), colspan=4, rowspan=1)
    ax3.get_xaxis().set_visible(False)
    ax4 = plt.subplot2grid((4, 12), (1, 8), colspan=4, rowspan=1, sharex=ax3)
    ax4.get_xaxis().set_visible(False)
    ax5 = plt.subplot2grid((4, 12), (2, 8), colspan=4, rowspan=1, sharex=ax3)
    ax5.get_xaxis().set_visible(False)
    ax6 = plt.subplot2grid((4, 12), (3, 8), colspan=4, rowspan=1, sharex=ax3)
    
    hist, bin_edges = np.histogram(origvid.data[:], bins=bins, density=True)
    center_guess = bin_centres[hist.argmax()]
    near_halfmax = -(hist - hist.max() / 2) ** 2
    halfmax_peaks, props = find_peaks(near_halfmax, prominence=0)
    selected_peaks = bin_centres[halfmax_peaks[np.argsort(props['prominences'])][-2:]]
    sigma_guess = np.abs(selected_peaks[0] - selected_peaks[1])

 
        
    centers = []
    amplitudes = []
    sigmas = []
    gammas = []
    correct_vid = origvid.deepcopy()
    colormap = plt.get_cmap('jet')
    
    for fnum in range(0, origvid.data.shape[0]):
        framedata = origvid.inav[fnum].data.flatten()
        hist, bin_edges = np.histogram(framedata[framedata > medval - 2 * std], bins=bins, density=True)
        ax1.plot(bin_centres, hist, linewidth=0.5, color = colormap(fnum/origvid.data.shape[0]))
        model = SkewedGaussianModel()
        params = model.make_params(amplitude=1, center=center_guess, sigma=sigma_guess, gamma=0)
        result = model.fit(hist, params, x=bin_centres)
        centers.append(result.best_values['center'])
        amplitudes.append(result.best_values['amplitude'])
        sigmas.append(result.best_values['sigma'])
        gammas.append(result.best_values['gamma'])
        shift = center_guess - result.best_values['center']
        correct_vid.inav[fnum].data[:] = origvid.inav[fnum].data[:] + shift
        hist, bin_edges = np.histogram(framedata[framedata > medval - 2 * std] + shift, bins=bins, density=True)
        ax2.plot(bin_centres, hist, linewidth=0.5, color = colormap(fnum/origvid.data.shape[0]))

    center_lists.append(centers)

    
    ax1.set_title(r'$\bf(a)$ Original Intensity histogram')
    ax2.set_title(r'$\bf(b)$ Aligned Intensity histogram')

    ax3.set_title(r'$\bf(e)$ Fit A (Amplitude)')
    ax4.set_title(r'$\bf(f)$ Fit $x_0$ (Centre)')
    ax5.set_title(r'$\bf(g)$ Fit $\sigma$ (width)')
    ax6.set_title(r'$\bf(h)$ Fit $\gamma$ (skew)')
    
    ax1.set_xlabel('HAADF Intensity')
    ax2.set_xlabel('HAADF Intensity')
    
    
    ax3.plot(amplitudes, 'b')
    ax4.plot(centers, 'b')
    ax5.plot(sigmas, 'b')   
    ax6.plot(gammas, 'b')

    ax6.set_xlabel('video frame')
    ax6.set_xlim(0,origvid.data.shape[0]+1)
    
    fig.subplots_adjust(top=0.965,
                        bottom=0.06,
                        left=0.015,
                        right=0.995,
                        hspace=0.2,
                        wspace=1.0)

    fig.savefig(savename + '.png')
    fig.savefig(savename + '.pdf')

    
    plt.close('all')
    gc.collect()
    
    


fig1, ax1 = plt.subplots(1, 2, figsize=(14, 10))
fig2, ax2 = plt.subplots(1, 2, figsize=(14, 10))

for i in range(numvid):
    centers = center_lists[i]
    settings = settings_list[i]
    color = colors_bc[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    
    if wetdry == 'wet':
        ax1[0].plot(centers, color=color, label=legendstring)
        ax2[0].plot(centers/max(centers), color=color, label=legendstring)
    elif wetdry == 'dry':
        ax1[1].plot(centers, color=color, label=legendstring)
        ax2[1].plot(centers/max(centers), color=color, label=legendstring)
    
ax1[0].set_xlabel('Video Frame')
ax2[0].set_xlabel('Video Frame')
ax1[0].set_ylabel('Absolute ADF centroid intensity')
ax2[0].set_ylabel('Relative ADF centroid intensity')
    
ax1[0].legend()
ax2[0].legend()

ax1[1].set_xlabel('Video Frame')
ax2[1].set_xlabel('Video Frame')
ax1[1].set_ylabel('Absolute ADF centroid intensity')
ax2[1].set_ylabel('Relative ADF centroid intensity')
    
ax1[1].legend()
ax2[1].legend()

ax1[0].set_title('Liquid')
ax1[1].set_title('Vacuum')

ax2[0].set_title('Liquid')
ax2[1].set_title('Vacuum')


fig1.savefig(savefolder_intensity + 'abs_intensity_compare.png')
fig2.savefig(savefolder_intensity + 'rel_intensity_compare.png')
fig1.savefig(savefolder_intensity + 'abs_intensity_compare.pdf')
fig2.savefig(savefolder_intensity + 'rel_intensity_compare.pdf')

plt.close('all')
gc.collect()






# calculate jumps basic

savefolder_jumps = savefolder + 'jumps' + os.sep
os.makedirs(savefolder_jumps, exist_ok=True)

atoms_jump_list = []
lattice_jump_list = []

for i in tqdm(range(numvid), desc = 'tabulating jump distances'):
    
    atoms_norm = atoms_norm_list[i]
    lattice_norm = lattice_norm_list[i]
    settings = settings_list[i]
    cmap = plt.get_cmap('jet')
    color = colors_bc[i]
     
    jumps = []
    for fnum in range(0, atoms_norm.frame.max() + 1):
        track = atoms_norm[atoms_norm['frame'] == fnum]
        frame_jumps = track['d_dist'].values
        jumps.append(frame_jumps)
             
    int_jumps = np.concatenate(jumps)
    int_jumps = int_jumps[~np.isnan(int_jumps)].copy()
    atoms_jump_list.append(int_jumps)
    
    lattice_jumps = []
    for fnum in range(0, atoms_norm.frame.max() + 1):
        track = lattice_norm[lattice_norm['frame'] == fnum]
        frame_jumps = track['d_dist'].values
        lattice_jumps.append(frame_jumps)

    int_lattice_jumps = np.concatenate(lattice_jumps)
    int_lattice_jumps = int_lattice_jumps[~np.isnan(int_lattice_jumps)].copy()
    lattice_jump_list.append(int_lattice_jumps)
    
    fig, ax = plt.subplots(1, 2, figsize=(18,12), dpi=80)
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]  
    sns.histplot(int_jumps, color=color, ax=ax[0], label=legendstring, element='step', alpha=0.1, stat='density')
    sns.histplot(int_lattice_jumps, color=color, ax=ax[1], label=legendstring, element='step', alpha=0.1, stat='density')
    ax[0].legend()
    ax[1].legend()
    ax[0].set_xlabel('Single Frame Jump Distance (nm)')
    ax[1].set_xlabel('Single Frame Jump Distance (nm)')
    ax[0].set_yticks([])
    ax[1].set_yticks([])
    ax[0].set_xlim(0, 1.2)
    ax[1].set_xlim(0, 0.1)
    ax[0].set_title('Pt Adatoms')
    ax[1].set_title('Mo Lattice')
    plt.tight_layout()  
    plt.savefig(savefolder_jumps + settings.filename[:-5] + 'jumps_compare.png')
    plt.savefig(savefolder_jumps + settings.filename[:-5] + 'jumps_compare.pdf') 
    
    plt.close('all')
    gc.collect()



jumps_exclude_list = ['Scanning Preview17',]

# compare single frame jumps (scaled by time, frame, dose) for wet and dry

fig, ax = plt.subplots(2, 2, figsize=(18,12), dpi=80, sharex='col', sharey='col')
fig_alt, ax_alt = plt.subplots(1, 2, figsize=(18,12), dpi=80, sharex='col', sharey='col')

for i in range(numvid):
    jumps = atoms_jump_list[i]
    lattice_jumps = lattice_jump_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    wetdry = wetdry_list[i]
    
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    
    if settings.filename[:-5] not in jumps_exclude_list:
    
        if wetdry == 'wet':
            sns.histplot(jumps, color=color, ax=ax[0, 0], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            sns.histplot(lattice_jumps, color=color, ax=ax[0, 1], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            sns.histplot(jumps, color='blue', ax=ax_alt[0], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            sns.histplot(lattice_jumps, color='blue', ax=ax_alt[1], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
        elif wetdry == 'dry':
            sns.histplot(jumps, color=color, ax=ax[1, 0], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            sns.histplot(lattice_jumps, color=color, ax=ax[1, 1], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            sns.histplot(jumps, color='red', ax=ax_alt[0], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            sns.histplot(lattice_jumps, color='red', ax=ax_alt[1], label=legendstring, element='step', alpha=0.1, stat='density', fill=False)
            
        
ax[0, 0].legend()
ax[0, 1].legend()
ax[0, 0].set_xlabel('Single Frame Jump Distance (nm)')
ax[0, 1].set_xlabel('Single Frame Jump Distance (nm)')
ax[0, 0].set_xlim(0, 1.2)
ax[0, 1].set_xlim(0, 0.1)
ax[0, 0].set_title('Pt Adatoms Liquid')
ax[0, 1].set_title('Mo Lattice Liquid')

ax[1, 0].legend()
ax[1, 1].legend()
ax[1, 0].set_xlabel('Single Frame Jump Distance (nm)')
ax[1, 1].set_xlabel('Single Frame Jump Distance (nm)')
ax[1, 0].set_xlim(0, 1.2)
ax[1, 1].set_xlim(0, 0.1)
ax[1, 0].set_title('Pt Adatoms Vacuum')
ax[1, 1].set_title('Mo Lattice Vacuum')


ax_alt[0].set_xlabel('Single Frame Jump Distance (nm)')
ax_alt[1].set_xlabel('Single Frame Jump Distance (nm)')
ax_alt[0].set_xlim(0, 1.2)
ax_alt[1].set_xlim(0, 0.1)
ax_alt[0].set_title('Pt Adatoms')
ax_alt[1].set_title('Mo Lattice')

for ax_arr in ax:
    for a in ax_arr:
        a.grid(which='major')
        
for a in ax_alt:
    a.grid(which='major')
    
for f in [fig, fig_alt]:
    f.subplots_adjust(top=0.95,
                      bottom=0.05,
                      left=0.05,
                      right=0.95,
                      hspace=0.1,
                      wspace=0.1)

fig.savefig(savefolder_jumps + 'jumps_compare.png')
fig.savefig(savefolder_jumps + 'jumps_compare.pdf')

fig_alt.savefig(savefolder_jumps + 'jumps_compare_alt.png')
fig_alt.savefig(savefolder_jumps + 'jumps_compare_alt.pdf')

plt.close('all')
gc.collect()






fig, ax = plt.subplots(1)
sns.histplot(wet_atoms_norm_all.d_dist.dropna(), ax=ax, color='blue', label='Vacuum', stat='density', element='step', fill=False)
sns.histplot(dry_atoms_norm_all.d_dist.dropna(), ax=ax, color='red', label='Liquid Cell', stat='density', element='step', fill=False)
ax_ins = ax.inset_axes([0.6, 0.6, 0.35, 0.35])
sns.histplot(wet_lattice_norm_all.d_dist.dropna(), ax=ax_ins, color='blue', label='Vacuum', stat='density', element='step', fill=False)
sns.histplot(dry_lattice_norm_all.d_dist.dropna(), ax=ax_ins, color='red', label='Liquid Cell', stat='density', element='step', fill=False)
ax.set_xlim(0, 1)
ax_ins.set_xlim(0, 0.1)
ax.set_xlabel('Pt Jump Distance (nm)')
ax_ins.set_xlabel('Mo Jump Distance (nm)')
fig.suptitle('Single Frame Jumps (nm)')
fig.savefig(savefolder_jumps + 'jumps_averaged_compare.png')
fig.savefig(savefolder_jumps + 'jumps_averaged_compare.pdf')






jumps_exclude_list = ['Scanning Preview18','Scanning Preview22']

bc_vals = np.sort(wet_atoms_norm_all.beam_current.unique())
cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
dry_cmap = plt.get_cmap('Reds')
wet_cmap = plt.get_cmap('Blues')

fig, ax = plt.subplots(1)
ax_ins = ax.inset_axes([0.6, 0.6, 0.35, 0.35])
bins_pt = np.linspace(0, 1, 101)
bins_lat = np.linspace(0, 0.1, 101)

for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    wet_color = wet_cmap(cm_vals[i])
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(jumps_exclude_list))]
    wet_lat_df = wet_lattice_norm_all[(wet_lattice_norm_all['beam_current'] == bc_val) & ~(wet_lattice_norm_all['video'].isin(jumps_exclude_list))]
    sns.histplot(wet_df.d_dist.dropna(), ax=ax, color=wet_color, label = 'Liquid %d pA' %(bc_val *1e12), stat='density', element='step', fill=False, bins=bins_pt, common_norm=True)
    sns.histplot(wet_lat_df.d_dist.dropna(), ax=ax_ins, color=wet_color, label = 'Liquid %d pA' %(bc_val *1e12), stat='density', element='step', fill=False, bins=bins_lat, common_norm=True)

for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    dry_color = dry_cmap(cm_vals[i])
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(jumps_exclude_list))]
    dry_lat_df = dry_lattice_norm_all[(dry_lattice_norm_all['beam_current'] == bc_val) & ~(dry_lattice_norm_all['video'].isin(jumps_exclude_list))]
    sns.histplot(dry_df.d_dist.dropna(), ax=ax, color=dry_color, label = 'Vacuum %d pA' %(bc_val *1e12), stat='density', element='step', fill=False, bins=bins_pt, common_norm=True)
    sns.histplot(dry_lat_df.d_dist.dropna(), ax=ax_ins, color=dry_color, label = 'Vacuum %d pA' %(bc_val *1e12), stat='density', element='step', fill=False, bins=bins_lat, common_norm=True)

ax.set_xlim(0, 0.8)
ax_ins.set_xlim(0, 0.05)
for a in [ax, ax_ins]:
    a.set_yticks([])
    a.set_xlabel('Jump Distance (nm)')
    a.set_ylabel('Probability')
ax.legend(loc='lower right')

fig.savefig(savefolder_figs + 'jumps.png')
fig.savefig(savefolder_figs + 'jumps.pdf')





















# calculate jumps per cat

savefolder_jumps_cat = savefolder_jumps + 'categorised' + os.sep
os.makedirs(savefolder_jumps_cat, exist_ok=True)

multifig, multiax = plt.subplots(1,3, sharey=True, figsize=(20, 10))

for i in tqdm(range(numvid), desc = 'tabulating jump distances'):
    
    atoms_norm = atoms_norm_list[i]
    lattice_norm = lattice_norm_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]

    jumps_bycat = listjumps_bycat(atoms_norm)
    
    fig = plt.figure()
    
    ax1 = fig.add_subplot(111)
    sns.distplot(jumps_bycat[0], ax=ax1, color='purple', kde_kws={"bw":0.05}, hist_kws={"histtype":"step", "linewidth":2}, label='From Mo')
    sns.distplot(jumps_bycat[1], ax=ax1, color='yellow', kde_kws={"bw":0.05}, hist_kws={"histtype":"step", "linewidth":2}, label='From S')
    sns.distplot(jumps_bycat[2], ax=ax1, color='red', kde_kws={"bw":0.05}, hist_kws={"histtype":"step", "linewidth":2}, label='From HC')
    
    ax1.set_xlim(0, 0.8)
    ax1.set_title('per Frame (nm) ', fontname='Arial', loc='right', pad=-12)
    ax1.set_facecolor('grey')
    ax1.set_xlabel('Jump Distance from site (nm)')
    ax1.legend()
    fig.suptitle(settings.filename[:-5])
    
    fig.savefig(savefolder_jumps_cat + settings.filename[:-5] + '_categorised_jumplengths.png')
    fig.savefig(savefolder_jumps_cat + settings.filename[:-5] + '_categorised_jumplengths.pdf')

    plt.close(fig)
    gc.collect()
    
    sns.distplot(jumps_bycat[0], ax=multiax[0], color=color, kde_kws={"bw":0.05}, hist_kws={"histtype":"step", "linewidth":2}, label=legendstring)
    sns.distplot(jumps_bycat[1], ax=multiax[1], color=color, kde_kws={"bw":0.05}, hist_kws={"histtype":"step", "linewidth":2}, label=legendstring)
    sns.distplot(jumps_bycat[2], ax=multiax[2], color=color, kde_kws={"bw":0.05}, hist_kws={"histtype":"step", "linewidth":2}, label=legendstring)
    
multiax[0].legend()

multiax[0].set_title('Mo Site')
multiax[1].set_title('S Site')
multiax[2].set_title('Vac Site')

multiax[1].set_xlabel('Jump Distance from site (nm)')

for ax in multiax:
    ax.set_xlim(0, 1.2)
    
multifig.subplots_adjust(top=0.965,
                          bottom=0.06,
                          left=0.035,
                          right=0.99,
                          hspace=0.2,
                          wspace=0.07)

multifig.savefig(savefolder_jumps_cat + 'compare_all_vids.png')
multifig.savefig(savefolder_jumps_cat + 'compare_all_vids.pdf')





# category counts plots

savefolder_count = savefolder + 'category_count' + os.sep
os.makedirs(savefolder_count, exist_ok=True)

maxrad_dist = 0.9


figbar, axbar = plt.subplots(1, numinclude + 1, figsize = (24, 12), sharey=True)
figbar2, axbar2 = plt.subplots(1, numinclude + 1, figsize = (24, 12), sharey=True)
j=0
for i in tqdm(range(numvid), desc = 'counting categories'):
    
    atoms_norm = atoms_norm_list[i]
    
    settings = settings_list[i]
    color = colors_dosescale[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    
    os.makedirs(savefolder_count + settings.filename[:-5] + os.sep, exist_ok=True)
    
    cat_count = catcount(atoms_norm[atoms_norm.nn_dist < 5], 'cat')
    cat_count_plot = np.array(cat_count) / np.sum(cat_count)
    cat_count_plot2 = np.array(cat_count)[:3] / np.sum(cat_count[:3])
    

    colors = ['purple', 'yellow', 'red', 'black']
    labels = ['Mo', 'S', 'Vac', 'None']

    fig, ax = plt.subplots(1, figsize=(6,6), dpi=200, sharex=True)
    ax.bar(range(4), cat_count, color=colors)
    ax.set_ylabel('Site Counts')
    ax.set_xticks(range(4))
    ax.set_xticklabels(labels)
    ax.set_title('Identified Adatom Site Counts')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_barcharts.png')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_barcharts.pdf')
    
    plt.close(fig)
    
    if wetdry == 'wet' or wetdry == 'dry':
        axbar[plot_indexes[j]].bar(range(4), cat_count_plot, color=colors)
        axbar[plot_indexes[j]].set_title(settings.filename[:-5], rotation=70)
        axbar[plot_indexes[j]].set_xticks(range(4))
        axbar[plot_indexes[j]].set_xticklabels(labels)
        axbar[plot_indexes[j]].set_xlabel(str(beam_current_list[i] * 1e12) + ' pA')
        axbar2[plot_indexes[j]].bar(range(3), cat_count_plot2, color=colors[:3])
        axbar2[plot_indexes[j]].set_title(settings.filename[:-5], rotation=70)
        axbar2[plot_indexes[j]].set_xticks(range(3))
        axbar2[plot_indexes[j]].set_xticklabels(labels[:3])
        axbar2[plot_indexes[j]].set_xlabel(str(beam_current_list[i] * 1e12) + ' pA')
        j+=1

    
for fig in [figbar, figbar2]:
    fig.suptitle('Category Probabilities')   
    fig.subplots_adjust(top=0.8,
                           bottom=0.05,
                           left=0.045,
                           right=0.975,
                           hspace=0.2,
                           wspace=0.15)
for a in axbar:
    a.grid(axis='y')
for a in axbar2:
    a.grid(axis='y')    
    
axbar[wetcount].axis('off')
axbar2[wetcount].axis('off')
figbar.savefig(savefolder_count + 'all_barcharts.png')
figbar.savefig(savefolder_count + 'all_barcharts.pdf')
figbar2.savefig(savefolder_count + 'all_barcharts_no_nans.png')
figbar2.savefig(savefolder_count + 'all_barcharts_no_nans.pdf')

plt.close('all')
gc.collect()




fig1, ax1 = plt.subplots(1, 2, figsize = (24, 12), sharey=True)
fig2, ax2 = plt.subplots(1, 2, figsize = (24, 12), sharey=True)
wetcount = 0
drycount = 0
axlabels_wet, axlabels_dry = np.empty(numwet * 4, dtype=object), np.empty(numdry * 4, dtype=object)
for i in tqdm(range(numvid), desc = 'counting categories'):
    
    atoms_norm = atoms_norm_list[i]
    
    settings = settings_list[i]
    color = colors_dosescale[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    
    cat_count = catcount(atoms_norm[atoms_norm.nn_dist < 5], 'cat')
    cat_count_plot = np.array(cat_count) / np.sum(cat_count)
    cat_count_plot2 = np.array(cat_count)[:3] / np.sum(cat_count[:3])
    
    colors = ['purple', 'yellow', 'red', 'black']
    labels = [str(beam_current_list[i] * 1e12) + ' pA'] * 4
    
    if wetdry == 'wet':
        positions = (np.linspace(0, 3, 4) * numwet + wetcount).astype(int)
        ax1[0].bar(positions, cat_count_plot, color=colors)
        ax2[0].bar(positions[:-1], cat_count_plot2, color=colors[:-1])
        axlabels_wet[positions] = labels
        wetcount += 1
    elif wetdry == 'dry':
        positions = (np.linspace(0, 3, 4) * numdry + drycount).astype(int)
        ax1[1].bar(positions, cat_count_plot, color=colors)
        ax2[1].bar(positions[:-1], cat_count_plot2, color=colors[:-1])
        axlabels_dry[positions] = labels
        drycount += 1
        
ax1[0].set_xticks(range(numwet*4))
ax1[1].set_xticks(range(numdry*4))
ax2[0].set_xticks(range(numwet*3))
ax2[1].set_xticks(range(numdry*3))
ax1[0].set_xticklabels(axlabels_wet, rotation=70)
ax1[1].set_xticklabels(axlabels_dry, rotation=70)
ax2[0].set_xticklabels(axlabels_wet[:-numwet], rotation=70)
ax2[1].set_xticklabels(axlabels_dry[:-numdry], rotation=70)
for ax in [ax1[0], ax2[0]]:
    ax.set_title('Liquid Cell')
for ax in [ax1[1], ax2[1]]:
    ax.set_title('Vacuum')
for fig in [fig1, fig2]:
    fig.suptitle('Category Counts')
fig1.savefig(savefolder_count + 'all_barcharts_rearranged.png')
fig1.savefig(savefolder_count + 'all_barcharts_rearranged.pdf')
fig2.savefig(savefolder_count + 'all_barcharts_no_nans_rearranged.png')
fig2.savefig(savefolder_count + 'all_barcharts_no_nans_rearranged.pdf')




fig1, ax1 = plt.subplots(1, 2, sharey=True)
fig2, ax2 = plt.subplots(1, 2, sharey=True)
wet_cat_count = catcount(wet_atoms_norm_all[wet_atoms_norm_all.nn_dist < 20], 'cat')
wet_cat_count_plot = np.array(wet_cat_count) / np.sum(wet_cat_count)
wet_cat_count_plot2 = np.array(wet_cat_count)[:3] / np.sum(wet_cat_count[:3])
dry_cat_count = catcount(dry_atoms_norm_all[dry_atoms_norm_all.nn_dist < 20], 'cat')
dry_cat_count_plot = np.array(dry_cat_count) / np.sum(dry_cat_count)
dry_cat_count_plot2 = np.array(dry_cat_count)[:3] / np.sum(dry_cat_count[:3])
colors = ['purple', 'yellow', 'red', 'black']
labels = ['Mo', 'S', 'Vac', 'None']
ax1[0].bar(range(4), wet_cat_count_plot, color=colors)
ax1[1].bar(range(4), dry_cat_count_plot, color=colors)
ax2[0].bar(range(3), wet_cat_count_plot2, color=colors[:-1])
ax2[1].bar(range(3), dry_cat_count_plot2, color=colors[:-1])
for ax in ax1:
    ax.set_xticks(range(4))
    ax.set_xticklabels(labels)
for ax in ax2:
    ax.set_xticks(range(3))
    ax.set_xticklabels(labels[:-1])
for ax in [ax1[0], ax2[0]]:
    ax.set_title('Liquid Cell')
for ax in [ax1[1], ax2[1]]:
    ax.set_title('Vacuum')
for fig in [fig1, fig2]:
    fig.suptitle('Category Counts')
fig1.savefig(savefolder_count + 'averaged_barcharts.png')
fig1.savefig(savefolder_count + 'averaged_barcharts.pdf')
fig2.savefig(savefolder_count + 'averaged_barcharts_no_nans.png')
fig2.savefig(savefolder_count + 'averaged_barcharts_no_nans.pdf')

















    

    
figbarstuck, axbarstuck = plt.subplots(1, numinclude + 1, figsize = (24, 12), sharey=True)
figbarstuck1, axbarstuck1 = plt.subplots(1, 2, figsize = (24, 12), sharey=True)
wetcount = 0
drycount = 0
axlabels_wet, axlabels_dry = np.empty(numwet * 4, dtype=object), np.empty(numdry * 4, dtype=object)
j=0
for i in tqdm(range(numvid), desc = 'counting categories'):
    
    atoms_norm = atoms_norm_list[i]
    
    settings = settings_list[i]
    color = colors_dosescale[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]

    os.makedirs(savefolder_count + settings.filename[:-5] + os.sep, exist_ok=True)
       
    stuck_prob, stuck_prob_mo, stuck_prob_s, stuck_prob_vac = stuck_counter(atoms_norm)

    fig, ax = plt.subplots(1, 1)
    ax.bar(1, stuck_prob, color=['black'])
    ax.bar(2, stuck_prob_mo, color=['purple'])
    ax.bar(3, stuck_prob_s, color=['yellow'])
    ax.bar(4, stuck_prob_vac, color=['red'])
    
    ax.set_xticks(range(1,5))
    ax.set_xticklabels(['Overall', 'Mo', 'S', 'HC'])
    ax.set_title(settings.filename[:-5] +' Stuck Probability')
    
    ax.grid(axis='y', alpha=0.3)
    
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_stuck_prob_singleframe.pdf')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_stuck_prob_singleframe.png')
    plt.close(fig)
    
    if wetdry == 'wet' or wetdry == 'dry':
        axbarstuck[plot_indexes[j]].bar(1, stuck_prob, color=['black'])
        axbarstuck[plot_indexes[j]].bar(2, stuck_prob_mo, color=['purple'])
        axbarstuck[plot_indexes[j]].bar(3, stuck_prob_s, color=['yellow'])
        axbarstuck[plot_indexes[j]].bar(4, stuck_prob_vac, color=['red'])
        axbarstuck[plot_indexes[j]].set_title(settings.filename[:-5], rotation=70)
        axbarstuck[plot_indexes[j]].set_xticks(range(1,5))
        axbarstuck[plot_indexes[j]].set_xticklabels(['Overall', 'Mo', 'S', 'HC'])
        axbarstuck[plot_indexes[j]].set_xlabel(str(beam_current_list[i] * 1e12) + ' pA')
        j += 1
        
    labels = [str(beam_current_list[i] * 1e12) + ' pA'] * 4

    if wetdry == 'wet':
        positions = (np.linspace(0, 3, 4) * numwet + wetcount).astype(int)
        axbarstuck1[0].bar(positions[0], stuck_prob, color=['black'])
        axbarstuck1[0].bar(positions[1], stuck_prob_mo, color=['purple'])
        axbarstuck1[0].bar(positions[2], stuck_prob_s, color=['yellow'])
        axbarstuck1[0].bar(positions[3], stuck_prob_vac, color=['red'])
        axlabels_wet[positions] = labels
        wetcount += 1
    if wetdry == 'dry':
        positions = (np.linspace(0, 3, 4) * numdry + drycount).astype(int)
        axbarstuck1[1].bar(positions[0], stuck_prob, color=['black'])
        axbarstuck1[1].bar(positions[1], stuck_prob_mo, color=['purple'])
        axbarstuck1[1].bar(positions[2], stuck_prob_s, color=['yellow'])
        axbarstuck1[1].bar(positions[3], stuck_prob_vac, color=['red'])
        axlabels_dry[positions] = labels
        drycount += 1
for ax in axbarstuck:
    ax.grid(axis='y')
axbarstuck[wetcount].axis('off')
figbarstuck.suptitle('Probabilities of Pt adatom in same site')   
figbarstuck.subplots_adjust(top=0.8,
                           bottom=0.05,
                           left=0.045,
                           right=0.975,
                           hspace=0.2,
                           wspace=0.15)
figbarstuck.savefig(savefolder_count + 'all_stuck_barcharts.png')
figbarstuck.savefig(savefolder_count + 'all_stuck_barcharts.pdf')
for ax in axbarstuck1:
    ax.grid(axis='y')
figbarstuck1.suptitle('Probabilities of Pt adatom in same site')
axbarstuck1[0].set_xticks(range(numwet*4))
axbarstuck1[0].set_xticklabels(axlabels_wet, rotation=70)
axbarstuck1[1].set_xticks(range(numdry*4))
axbarstuck1[1].set_xticklabels(axlabels_dry, rotation=70)
axbarstuck1[0].set_title('Liquid Cell')
axbarstuck1[1].set_title('Vacuum')

figbarstuck1.suptitle('Probabilities of Pt adatom in same site')   
figbarstuck1.subplots_adjust(top=0.945,
                             bottom=0.08,
                             left=0.045,
                             right=0.975,
                             hspace=0.2,
                             wspace=0.08)

figbarstuck1.savefig(savefolder_count + 'all_stuck_barcharts_rearranged.png')
figbarstuck1.savefig(savefolder_count + 'all_stuck_barcharts_rearranged.pdf')

fig, ax = plt.subplots(1, 2, sharey=True)
wet_probs = stuck_counter(wet_atoms_norm_all)
dry_probs = stuck_counter(dry_atoms_norm_all)
ax[0].bar(range(4), wet_probs, color=['black', 'purple', 'yellow', 'red'])
ax[1].bar(range(4), dry_probs, color=['black', 'purple', 'yellow', 'red'])
ax[0].set_xticks(range(4))
ax[0].set_xticklabels(['All', 'Mo', 'S', 'Vac'])
ax[1].set_xticks(range(4))
ax[1].set_xticklabels(['All', 'Mo', 'S', 'Vac'])
ax[0].set_title('Liquid Cell')
ax[1].set_title('Vacuum')
ax[0].grid(axis='y')
ax[1].grid(axis='y')
fig.suptitle('Probabilities of Pt adatom in same site')
fig.savefig(savefolder_count + 'all_stuck_barcharts_averaged.png')
fig.savefig(savefolder_count + 'all_stuck_barcharts_averaged.pdf')

plt.close('all')
gc.collect()


















figplots, axplots = plt.subplots(5, 2, figsize=(25, 15), sharex='col', sharey='row')
for i in tqdm(range(numvid), desc = 'counting categories'):
    
    atoms_norm = atoms_norm_list[i]
    
    settings = settings_list[i]
    color = colors_bc[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    os.makedirs(savefolder_count + settings.filename[:-5] + os.sep, exist_ok=True)
    
    catframes_count = catcount_frames(atoms_norm[atoms_norm.nn_dist < 5], 'cat')
    
    fig, ax = plt.subplots(1)
    ax.plot(catframes_count[0], color='purple')
    ax.plot(catframes_count[1], color='yellow')
    ax.plot(catframes_count[2], color='red')
    ax.plot(catframes_count[3], color='black')
    ax.set_facecolor('grey')
    ax.set_xlabel('Video Frames')
    ax.set_ylabel('Atom Site Counts')
    ax.set_title('Adatom site evolution')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_plots.png')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_plots.pdf')
    
    plt.close(fig)
    
    totlist = np.unique(atoms_norm[atoms_norm.nn_dist < 5].frame, return_counts=True)[1]
    
    if wetdry == 'wet':
        axplots[0, 0].plot(totlist, color=color)
        axplots[1, 0].plot(catframes_count[0] / totlist, color=color)
        axplots[2, 0].plot(catframes_count[1] / totlist, color=color)
        axplots[3, 0].plot(catframes_count[2] / totlist, color=color)
        axplots[4, 0].plot(catframes_count[3] / totlist, color=color, label=legendstring)
    elif wetdry == 'dry':
        axplots[0, 1].plot(totlist, color=color)
        axplots[1, 1].plot(catframes_count[0] / totlist, color=color)
        axplots[2, 1].plot(catframes_count[1] / totlist, color=color)
        axplots[3, 1].plot(catframes_count[2] / totlist, color=color)
        axplots[4, 1].plot(catframes_count[3] / totlist, color=color, label=legendstring)
    

for a in axplots:
    for ax in a:
        ax.yaxis.grid()
axplots[4, 0].legend()
axplots[4, 1].legend()
axplots[0, 0].set_ylabel('Number of Pt Atoms')
axplots[1, 0].set_ylabel('Fraction of Pt atoms on Mo site')
axplots[2, 0].set_ylabel('Fraction of Pt atoms on S site')
axplots[3, 0].set_ylabel('Fraction of Pt atoms on HC site')
axplots[4, 0].set_ylabel('Fraction of uncategorised Pt atoms')
axplots[0, 0].set_title('Liquid Environment')
axplots[0, 1].set_title('Vacuum Environment')
figplots.suptitle('Category Evolution with Frames')
figplots.subplots_adjust(top=0.955,
                         bottom=0.045,
                         left=0.03,
                         right=0.99,
                         hspace=0.2,
                         wspace=0.105)
figplots.savefig(savefolder_count + 'all_plots.png')
figplots.savefig(savefolder_count + 'all_plots.pdf')   



    

figplots_stuck, axplots_stuck = plt.subplots(4, 2, figsize=(25, 15), sharex='col', sharey='row')
for i in tqdm(range(numvid), desc = 'counting categories'):
    
    atoms_norm = atoms_norm_list[i]
    
    settings = settings_list[i]
    color = colors_bc[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    os.makedirs(savefolder_count + settings.filename[:-5] + os.sep, exist_ok=True)
      
    stuck_prob_f, stuck_prob_mo_f, stuck_prob_s_f, stuck_prob_vac_f = stuck_counter_byframes(atoms_norm)
    fig, ax = plt.subplots(1, figsize=(20, 16))
    ax.plot(stuck_prob_f, color='black', label='Overall')
    ax.plot(stuck_prob_mo_f, color='purple', label='Mo')
    ax.plot(stuck_prob_s_f, color='yellow', label='S') 
    ax.plot(stuck_prob_vac_f, color='red', label='HC')
    ax.set_xlabel('Frames')
    ax.set_ylabel('Stuck Probability')
    ax.set_title(settings.filename[:-5] + ' beam current: ' + str(beam_current_list[i] * 1e12) + ' pA' + ' Probability of adatom in same site in consecutive frames')
    ax.set_facecolor('grey')
    ax.legend()
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_stuck_prob_evolution.png')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_stuck_prob_evolution.pdf')
    plt.close(fig)
    
    if wetdry == 'wet':
        axplots_stuck[0, 0].plot(stuck_prob_f, color=color)
        axplots_stuck[1, 0].plot(stuck_prob_mo_f, color=color)
        axplots_stuck[2, 0].plot(stuck_prob_s_f, color=color)
        axplots_stuck[3, 0].plot(stuck_prob_vac_f, color=color, label=legendstring)
    elif wetdry == 'dry':
        axplots_stuck[0, 1].plot(stuck_prob_f, color=color)
        axplots_stuck[1, 1].plot(stuck_prob_mo_f, color=color)
        axplots_stuck[2, 1].plot(stuck_prob_s_f, color=color)
        axplots_stuck[3, 1].plot(stuck_prob_vac_f, color=color, label=legendstring)    
 
axplots_stuck[0, 0].set_ylabel('Overall')
axplots_stuck[1, 0].set_ylabel('Mo Site')
axplots_stuck[2, 0].set_ylabel('S Site')
axplots_stuck[3, 0].set_ylabel('HC Site')
axplots_stuck[0, 0].set_title('Liquid Environment')
axplots_stuck[0, 1].set_title('Vacuum Environment')
for a in axplots_stuck:
    for ax in a:
        ax.yaxis.grid()

axplots_stuck[3, 0].legend()
axplots_stuck[3, 1].legend()
figplots_stuck.suptitle('Stuck Probability Evolution with Frames')
figplots_stuck.subplots_adjust(top=0.955,
                         bottom=0.045,
                         left=0.03,
                         right=0.99,
                         hspace=0.2,
                         wspace=0.105)
figplots_stuck.savefig(savefolder_count + 'all_plots_stuck.png')
figplots_stuck.savefig(savefolder_count + 'all_plots_stuck.pdf')
    






figplots_cons, axplots_cons = plt.subplots(2, 4, figsize=(25, 15), sharex=True)      
for i in tqdm(range(numvid), desc = 'counting categories'):
    
    atoms_norm = atoms_norm_list[i].copy()
    
    atoms_norm.cons_stuck = atoms_norm.cons_stuck + atoms_norm.stuck
    
    settings = settings_list[i]
    color = colors_bc[i]
    wetdry = wetdry_list[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    
    os.makedirs(savefolder_count + settings.filename[:-5] + os.sep, exist_ok=True)
    
    stuck_prob_cons, stuck_prob_mo_cons, stuck_prob_s_cons, stuck_prob_vac_cons = stuck_counter_cons(atoms_norm)
    
    fig, ax = plt.subplots(1)
    ax.plot(stuck_prob_cons[0] + 1, stuck_prob_cons[1], color='black', label='Overall')
    ax.plot(stuck_prob_mo_cons[0] + 1, stuck_prob_mo_cons[1], color='purple', label='Mo')
    ax.plot(stuck_prob_s_cons[0] + 1, stuck_prob_s_cons[1], color='yellow', label='S')
    ax.plot(stuck_prob_vac_cons[0] + 1, stuck_prob_vac_cons[1], color='red', label='HC')
    ax.set_yscale('log')
    ax.set_xlim(1,11)
    ax.set_xticks(range(1,12))
    ax.set_ylim(0.001, 1)
    ax.set_ylabel('Normalised Probability')
    ax.set_xlabel('consecutive frames at same site')   
    ax.set_title(settings.filename[:-5] + ' ' + str(beam_current_list[i] * 1e12) + ' pA')
    ax.grid(which='major', axis='y', color='red')
    ax.grid(which='minor', axis='y', linestyle='--', alpha=0.2)
    ax.grid(which='major', axis='x', alpha=0.5)
    ax.legend()
    axlim = ax.get_xlim()
    axtop = ax.twiny()
    axtop.set_xlim(axlim[0] * float(settings.FrameTime), axlim[1] * float(settings.FrameTime))
    axtop.set_xlabel('time (s)')
    fig.subplots_adjust(top=0.855,
                        bottom=0.095,
                        left=0.105,
                        right=0.99,
                        hspace=0.2,
                        wspace=0.2)
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_stuck_prob_consecutive.png')
    fig.savefig(savefolder_count + settings.filename[:-5] + os.sep + '_stuck_prob_consecutive.pdf')
    
    plt.close(fig)
    
    if wetdry == 'wet':
        axplots_cons[0, 0].plot(stuck_prob_cons[0] + 1, stuck_prob_cons[1], color=color, label=legendstring)
        axplots_cons[0, 1].plot(stuck_prob_mo_cons[0] + 1, stuck_prob_mo_cons[1], color=color)
        axplots_cons[0, 2].plot(stuck_prob_s_cons[0] + 1, stuck_prob_s_cons[1], color=color)
        axplots_cons[0, 3].plot(stuck_prob_vac_cons[0] + 1, stuck_prob_vac_cons[1], color=color)
    elif wetdry == 'dry':
        axplots_cons[1, 0].plot(stuck_prob_cons[0] + 1, stuck_prob_cons[1], color=color, label=legendstring)
        axplots_cons[1, 1].plot(stuck_prob_mo_cons[0] + 1, stuck_prob_mo_cons[1], color=color)
        axplots_cons[1, 2].plot(stuck_prob_s_cons[0] + 1, stuck_prob_s_cons[1], color=color)
        axplots_cons[1, 3].plot(stuck_prob_vac_cons[0] + 1, stuck_prob_vac_cons[1], color=color)    

for a in axplots_cons:
    for ax in a:
        ax.set_yscale('log')
        ax.set_xlim(1,11)
        ax.set_xticks(range(1,12))
        ax.set_ylim(0.001, 1)
        ax.grid(which='major', axis='y', color='red')
        ax.grid(which='minor', axis='y', linestyle='--', alpha=0.2)
        ax.grid(which='major', axis='x', alpha=0.5)
        axlim = ax.get_xlim()
        axtop = ax.twiny()
        axtop.set_xlim(axlim[0] * float(settings.FrameTime), axlim[1] * float(settings.FrameTime))
        axtop.set_xlabel('time (s)')
    
axplots_cons[0, 0].legend()    
axplots_cons[1, 0].legend()
axplots_cons[0, 0].set_ylabel('Liquid Environment')
axplots_cons[1, 0].set_ylabel('Vacuum Environment')
axplots_cons[0, 0].set_xlabel('consecutive frames at same site')
axplots_cons[0, 1].set_xlabel('consecutive frames at same (Mo) site')
axplots_cons[0, 2].set_xlabel('consecutive frames at same (S) site')
axplots_cons[0, 3].set_xlabel('consecutive frames at same (HC) site')
     
figplots_cons.subplots_adjust(bottom=0.04,
                            left=0.045,
                            right=0.99,
                            hspace=0.2,
                            wspace=0.2)
figplots_cons.savefig(savefolder_count + 'all_plots_cons.png')
figplots_cons.savefig(savefolder_count + 'all_plots_cons.pdf')







excludelist = ['Scanning Preview19',]

wet_atoms_norm = wet_atoms_norm_all.copy()
wet_atoms_norm = wet_atoms_norm[~wet_atoms_norm.video.isin(excludelist)]
wet_atoms_norm.cons_stuck = wet_atoms_norm.cons_stuck + wet_atoms_norm.stuck
dry_atoms_norm = dry_atoms_norm_all.copy()
dry_atoms_norm = dry_atoms_norm[~dry_atoms_norm.video.isin(excludelist)]
dry_atoms_norm.cons_stuck = dry_atoms_norm.cons_stuck + dry_atoms_norm.stuck
wet_stuck_prob_cons, wet_stuck_prob_mo_cons, wet_stuck_prob_s_cons, wet_stuck_prob_vac_cons = stuck_counter_cons(wet_atoms_norm)
dry_stuck_prob_cons, dry_stuck_prob_mo_cons, dry_stuck_prob_s_cons, dry_stuck_prob_vac_cons = stuck_counter_cons(dry_atoms_norm)

fig, ax = plt.subplots(1, 2)
ax[0].plot(wet_stuck_prob_cons[0] + 1, wet_stuck_prob_cons[1], color='black', label='Overall')
ax[0].plot(wet_stuck_prob_mo_cons[0] + 1, wet_stuck_prob_mo_cons[1], color='purple', label='Mo')
ax[0].plot(wet_stuck_prob_s_cons[0] + 1, wet_stuck_prob_s_cons[1], color='yellow', label='S')
ax[0].plot(wet_stuck_prob_vac_cons[0] + 1, wet_stuck_prob_vac_cons[1], color='red', label='HC')
ax[1].plot(dry_stuck_prob_cons[0] + 1, dry_stuck_prob_cons[1], color='black', label='Overall')
ax[1].plot(dry_stuck_prob_mo_cons[0] + 1, dry_stuck_prob_mo_cons[1], color='purple', label='Mo')
ax[1].plot(dry_stuck_prob_s_cons[0] + 1, dry_stuck_prob_s_cons[1], color='yellow', label='S')
ax[1].plot(dry_stuck_prob_vac_cons[0] + 1, dry_stuck_prob_vac_cons[1], color='red', label='HC')
tvals, tcounts = np.unique(timescale_list, return_counts=True)
usetime = tvals[tcounts.argmax()]
for a in ax:
    a.set_yscale('log')
    a.set_xlim(1,8)
    a.set_xticks(range(1,9))
    a.set_ylim(0.001, 1)
    a.grid(which='major', axis='y', color='red')
    a.grid(which='minor', axis='y', linestyle='--', alpha=0.2)
    a.grid(which='major', axis='x', alpha=0.5)
    axlim = a.get_xlim()
    axtop = a.twiny()
    axtop.set_xlim(axlim[0] * float(usetime), axlim[1] * float(settings.FrameTime))
    axtop.set_xlabel('time (s)')
    a.set_xlabel('frames')
    a.set_ylabel('Stuck Probability')
ax[0].set_title('Liquid Cell')
ax[1].set_title('Vacuum')
fig.subplots_adjust(top=0.855,
                    bottom=0.11,
                    left=0.11,
                    right=0.985,
                    hspace=0.2,
                    wspace=0.2)
fig.savefig(savefolder_count + 'all_plots_av.png')
fig.savefig(savefolder_count + 'all_plots_av.pdf')


fig, ax = plt.subplots(1)

ax.plot(wet_stuck_prob_cons[0] + 1, wet_stuck_prob_cons[1], color='black', label='Overall Liquid')
ax.plot(wet_stuck_prob_mo_cons[0] + 1, wet_stuck_prob_mo_cons[1], color='purple', label='Mo Liquid')
ax.plot(wet_stuck_prob_s_cons[0] + 1, wet_stuck_prob_s_cons[1], color='yellow', label='S Liquid')
ax.plot(wet_stuck_prob_vac_cons[0] + 1, wet_stuck_prob_vac_cons[1], color='red', label='HC Liquid')
ax.plot(dry_stuck_prob_cons[0] + 1, dry_stuck_prob_cons[1], color='black', label='Overall Vacuum', ls='dotted')
ax.plot(dry_stuck_prob_mo_cons[0] + 1, dry_stuck_prob_mo_cons[1], color='purple', label='Mo Vacuum', ls='dotted')
ax.plot(dry_stuck_prob_s_cons[0] + 1, dry_stuck_prob_s_cons[1], color='yellow', label='S Vacuum', ls='dotted')
ax.plot(dry_stuck_prob_vac_cons[0] + 1, dry_stuck_prob_vac_cons[1], color='red', label='HC Vacuum', ls='dotted')
tvals, tcounts = np.unique(timescale_list, return_counts=True)
usetime = tvals[tcounts.argmax()]
for a in [ax,]:
    a.set_yscale('log')
    a.set_xlim(1,8)
    a.set_xticks(range(1,9))
    a.set_ylim(0.001, 1)
    a.grid(which='major', axis='y', color='red')
    a.grid(which='minor', axis='y', linestyle='--', alpha=0.2)
    a.grid(which='major', axis='x', alpha=0.5)
    axlim = a.get_xlim()
    axtop = a.twiny()
    axtop.set_xlim(axlim[0] * float(usetime), axlim[1] * float(settings.FrameTime))
    axtop.set_xlabel('time (s)')
    a.set_xlabel('frames')
    a.set_ylabel('Stuck Probability')
    a.set_facecolor('grey')
fig.subplots_adjust(top=0.855,
                    bottom=0.11,
                    left=0.11,
                    right=0.985,
                    hspace=0.2,
                    wspace=0.2)
fig.savefig(savefolder_count + 'all_plots_av2.png')
fig.savefig(savefolder_count + 'all_plots_av2.pdf')

plt.close('all')
gc.collect()
























savefolder_hopping_base = savefolder + 'hopping_analysis' + os.sep
os.makedirs(savefolder_hopping_base, exist_ok=True)

savefolder_hopping_pickle = savefolder_hopping_base + 'pickled' + os.sep
os.makedirs(savefolder_hopping_pickle, exist_ok=True)

try:
    
    atoms_class_list = pickle.load(open(savefolder_hopping_pickle + 'atoms_class_list.p', 'rb'))
    
    mo_shifts_list = pickle.load(open(savefolder_hopping_pickle + 'mo_shifts_list.p', 'rb'))
    mo_shifts_2_list = pickle.load(open(savefolder_hopping_pickle + 'mo_shifts_2_list.p', 'rb'))
    mo_shifts_bar_list = pickle.load(open(savefolder_hopping_pickle + 'mo_shifts_bar_list.p', 'rb'))
    mo_shifts_bar_2_list = pickle.load(open(savefolder_hopping_pickle + 'mo_shifts_bar_2_list.p', 'rb'))
    mo_shifts_bar_comb_list = pickle.load(open(savefolder_hopping_pickle + 'mo_shifts_bar_comb_list.p', 'rb'))
    mo_shifts_bar_comb_2_list = pickle.load(open(savefolder_hopping_pickle + 'mo_shifts_bar_comb_2_list.p', 'rb'))
    
    s_shifts_list = pickle.load(open(savefolder_hopping_pickle + 's_shifts_list.p', 'rb'))
    s_shifts_2_list = pickle.load(open(savefolder_hopping_pickle + 's_shifts_2_list.p', 'rb'))
    s_shifts_bar_list = pickle.load(open(savefolder_hopping_pickle + 's_shifts_bar_list.p', 'rb'))
    s_shifts_bar_2_list = pickle.load(open(savefolder_hopping_pickle + 's_shifts_bar_2_list.p', 'rb'))
    s_shifts_bar_comb_list = pickle.load(open(savefolder_hopping_pickle + 's_shifts_bar_comb_list.p', 'rb'))
    s_shifts_bar_comb_2_list = pickle.load(open(savefolder_hopping_pickle + 's_shifts_bar_comb_2_list.p', 'rb'))
    
    vac_shifts_list = pickle.load(open(savefolder_hopping_pickle + 'vac_shifts_list.p', 'rb'))
    vac_shifts_2_list = pickle.load(open(savefolder_hopping_pickle + 'vac_shifts_2_list.p', 'rb'))
    vac_shifts_bar_list = pickle.load(open(savefolder_hopping_pickle + 'vac_shifts_bar_list.p', 'rb'))
    vac_shifts_bar_2_list = pickle.load(open(savefolder_hopping_pickle + 'vac_shifts_bar_2_list.p', 'rb'))
    vac_shifts_bar_comb_list = pickle.load(open(savefolder_hopping_pickle + 'vac_shifts_bar_comb_list.p', 'rb'))
    vac_shifts_bar_comb_2_list = pickle.load(open(savefolder_hopping_pickle + 'vac_shifts_bar_comb_2_list.p', 'rb'))


except:
    
    atoms_class_list = []
    
    mo_shifts_list = []
    mo_shifts_2_list = []
    s_shifts_list = []
    s_shifts_2_list = []
    vac_shifts_list = []
    vac_shifts_2_list = []
    
    mo_shifts_bar_list = []
    mo_shifts_bar_2_list = []
    mo_shifts_bar_comb_list = []
    mo_shifts_bar_comb_2_list = []
    
    s_shifts_bar_list = []
    s_shifts_bar_2_list = []
    s_shifts_bar_comb_list = []
    s_shifts_bar_comb_2_list = []
    
    vac_shifts_bar_list = []
    vac_shifts_bar_2_list = []
    vac_shifts_bar_comb_list = []
    vac_shifts_bar_comb_2_list = []
    
    for i in tqdm(range(numvid), desc = 'hopping analysis'):
        
    
        
        atoms_norm = atoms_norm_list[i]
        atoms_norm.cons_stuck = atoms_norm.cons_stuck + atoms_norm.stuck
        atoms_class = classify_atoms(atoms_norm)
        atoms_class_Tmat = class_Tmat(atoms_class, Tmat_list[i])
        atoms_class_list.append(atoms_class_Tmat)
        
        settings = settings_list[i]
        color = colors_bc[i]
        legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
        
        savefolder_hopping = savefolder_hopping_base + settings.filename[:-5] + os.sep
        os.makedirs(savefolder_hopping, exist_ok=True)    
    
        points_set = (mo_points_fixed, s_points_fixed, vac_points_fixed)
    
        mo_shifts = get_mo_shifts(atoms_class_Tmat, points_set)
        s_shifts = get_s_shifts(atoms_class_Tmat, points_set)
        vac_shifts = get_vac_shifts(atoms_class_Tmat, points_set)
        
        mo_shifts_2 = get_mo_shifts_double(atoms_class_Tmat, points_set)
        s_shifts_2 = get_s_shifts_double(atoms_class_Tmat, points_set)
        vac_shifts_2 = get_vac_shifts_double(atoms_class_Tmat, points_set)
        
        mo_shifts_list.append(mo_shifts)
        mo_shifts_2_list.append(mo_shifts_2)
        s_shifts_list.append(s_shifts)
        s_shifts_2_list.append(s_shifts_2)
        vac_shifts_list.append(vac_shifts)
        vac_shifts_2_list.append(vac_shifts_2)
    
    
        fig = plt.figure(figsize=(25,16))
        ax_mo = fig.add_subplot(231, aspect='equal')
        plotshifts(ax_mo, 'Mo', mo_shifts, points_set)
        ax_mo.set_ylabel('Single Frame Jump Probability')
        ax_mo.set_title('From Mo Site')
        ax_s = fig.add_subplot(232, aspect='equal')
        plotshifts(ax_s, 'S', s_shifts, points_set)
        ax_s.set_title('From S Site')
        ax_vac = fig.add_subplot(233, aspect='equal')
        plotshifts(ax_vac, 'Vac', vac_shifts, points_set)
        ax_vac.set_title('From Hollow Site')
        ax_mo2 = fig.add_subplot(234, aspect='equal')
        plotshifts(ax_mo2, 'Mo', mo_shifts_2, points_set)
        ax_mo2.set_ylabel('2 frames Jump Probability')
        ax_s2 = fig.add_subplot(235, aspect='equal')
        plotshifts(ax_s2, 'S', s_shifts_2, points_set)
        ax_vac2 = fig.add_subplot(236, aspect='equal')
        plotshifts(ax_vac2, 'Vac', vac_shifts_2, points_set)
        for ax in [ax_mo, ax_mo2, ax_s, ax_s2, ax_vac, ax_vac2]:
            ax.set_facecolor('grey')
        fig.suptitle(legendstring)
        fig.subplots_adjust(top=0.985,
                            bottom=0.01,
                            left=0.025,
                            right=0.98,
                            hspace=0.05,
                            wspace=0.05)
        fig.savefig(savefolder_hopping + settings.filename[:-5] + '_hopping_2D.png')
        fig.savefig(savefolder_hopping + settings.filename[:-5] + '_hopping_2D.pdf')
        
        plt.close(fig)
        gc.collect()
    
        if settings.filename != 'Scanning Preview20_sparse data.hspy':
    
            mo_shifts_bar = barify_shifts(mo_shifts, 0.1)
            s_shifts_bar = barify_shifts(s_shifts, 0.1)
            vac_shifts_bar = barify_shifts(vac_shifts, 0.1)
            
            mo_shifts_bar_2 = barify_shifts(mo_shifts_2, 0.1)
            s_shifts_bar_2 = barify_shifts(s_shifts_2, 0.1)
            vac_shifts_bar_2 = barify_shifts(vac_shifts_2, 0.1)
        
        
            mo_shifts_bar_comb = combineshifts_bar(mo_shifts_bar)
            s_shifts_bar_comb = combineshifts_bar(s_shifts_bar)
            vac_shifts_bar_comb = combineshifts_bar(vac_shifts_bar)
            
            mo_shifts_bar_comb_2 = combineshifts_bar(mo_shifts_bar_2)
            s_shifts_bar_comb_2 = combineshifts_bar(s_shifts_bar_2)
            vac_shifts_bar_comb_2 = combineshifts_bar(vac_shifts_bar_2)
            
            mo_shifts_bar_list.append(mo_shifts_bar)
            mo_shifts_bar_2_list.append(mo_shifts_bar_2)
            mo_shifts_bar_comb_list.append(mo_shifts_bar_comb)
            mo_shifts_bar_comb_2_list.append(mo_shifts_bar_comb_2)
            
            s_shifts_bar_list.append(s_shifts_bar)
            s_shifts_bar_2_list.append(s_shifts_bar_2)
            s_shifts_bar_comb_list.append(s_shifts_bar_comb)
            s_shifts_bar_comb_2_list.append(s_shifts_bar_comb_2)
            
            vac_shifts_bar_list.append(vac_shifts_bar)
            vac_shifts_bar_2_list.append(vac_shifts_bar_2)
            vac_shifts_bar_comb_list.append(vac_shifts_bar_comb)
            vac_shifts_bar_comb_2_list.append(vac_shifts_bar_comb_2)
        
        
            fig, ax = plt.subplots(2,3, figsize=(22,12))
            
            barplot(ax[0, 0], mo_shifts_bar_comb, lims=((0,0.64),(0,60)))
            ax[0,0].set_yticklabels([])
            barplot(ax[0, 1], s_shifts_bar_comb, lims=((0,0.64),(0,60)))
            ax[0,1].set_yticklabels([])
            barplot(ax[0, 2], vac_shifts_bar_comb, lims=((0,0.64),(0,60)))
            
            barplot(ax[1, 0], mo_shifts_bar_comb_2, lims=((0,0.64),(0,50)))
            ax[1,0].set_yticklabels([])
            barplot(ax[1, 1], s_shifts_bar_comb_2, lims=((0,0.64),(0,50)))
            ax[1,1].set_yticklabels([])
            barplot(ax[1, 2], vac_shifts_bar_comb_2, lims=((0,0.64),(0,50)))
            
            ax[0, 0].set_ylabel('Probability after one frame')
            ax[1, 0].set_ylabel('Probability after two frames')
            
            ax[1, 0].set_xlabel('From Mo Site')
            ax[1, 1].set_xlabel('From S Site')
            ax[1, 2].set_xlabel('From HC Site')
            
            ax[0, 2].yaxis.set_label_position('right')
            ax[0, 2].set_ylabel('Relative Probability (%)')
            ax[1, 2].yaxis.set_label_position('right')
            ax[1, 2].set_ylabel('Relative Probability (%)')
            
            for axis in ax[0]:
                axis.set_xlabel('distance (nm)')
                
            fig.suptitle(legendstring)
            fig.subplots_adjust(top=0.915,
                                bottom=0.03,
                                left=0.02,
                                right=0.96,
                                hspace=0.12,
                                wspace=0.1)
            
            fig.savefig(savefolder_hopping + settings.filename[:-5] + '_jump_probablities_bar_chart.png')
            fig.savefig(savefolder_hopping + settings.filename[:-5] + '_jump_probablities_bar_chart.pdf')
        
            plt.close(fig)
            gc.collect()
        
            fig, ax = plt.subplots(2, figsize=(12,8))
        
            ax[0].axhline(1, color='red')
            ax[0].bar(1, mo_shifts_bar_comb.iloc[1].prob / mo_shifts_bar_comb.iloc[2].prob)
            ax[0].bar(2, s_shifts_bar_comb.iloc[1].prob / s_shifts_bar_comb.iloc[2].prob)
            ax[0].bar(3, vac_shifts_bar_comb.iloc[1].prob / vac_shifts_bar_comb.iloc[2].prob)
            
            ax[1].bar(1, 2*mo_shifts_bar_comb.iloc[3].prob / (mo_shifts_bar_comb.iloc[4].prob + mo_shifts_bar_comb.iloc[5].prob))
            ax[1].bar(2, 2*s_shifts_bar_comb.iloc[3].prob / (s_shifts_bar_comb.iloc[4].prob + s_shifts_bar_comb.iloc[5].prob))
            ax[1].bar(3, 2*vac_shifts_bar_comb.iloc[3].prob / (vac_shifts_bar_comb.iloc[4].prob + vac_shifts_bar_comb.iloc[5].prob))
            ax[1].bar(4, mo_shifts_bar_comb.iloc[4].prob / mo_shifts_bar_comb.iloc[5].prob)
            ax[1].bar(5, s_shifts_bar_comb.iloc[4].prob / s_shifts_bar_comb.iloc[5].prob)
            ax[1].bar(6, vac_shifts_bar_comb.iloc[4].prob / vac_shifts_bar_comb.iloc[5].prob)
            
            for axis in ax:
                axis.grid(axis='y')
                axis.axhline(1, color='red') 
                
            ax[0].set_xticks(range(1,4))    
            ax[1].set_xticks(range(1,7))
            
            ticklist1 = [r'$\frac{Mo\rightarrow S}{Mo\rightarrow HC}$',
                        r'$\frac{S\rightarrow Mo}{S\rightarrow HC}$',
                        r'$\frac{HC\rightarrow Mo}{HC\rightarrow S}$']
            ticklist2 = [r'$\frac{Mo\rightarrow Mo}{Mo\rightarrow S/HC}$',
                        r'$\frac{S\rightarrow S}{S\rightarrow Mo/HC}$',
                        r'$\frac{HC\rightarrow HC}{HC\rightarrow Mo/S}$',
                        r'$\frac{Mo\rightarrow S}{Mo\rightarrow HC}$',
                        r'$\frac{S\rightarrow Mo}{S\rightarrow HC}$',
                        r'$\frac{Hc\rightarrow Mo}{HC\rightarrow S}$']
            ax[0].set_xticklabels(ticklist1)
            ax[1].set_xticklabels(ticklist2)
            ax[0].set_ylabel(r'Nearest Neighbour (1.9$\mathring{A}$)') 
            ax[1].set_ylabel(r'2nd Nearest Neighbour (3.3 - 3.8$\mathring{A}$)')
        
            fig.suptitle(legendstring + ' Relative Transition Probabilities')
            
            fig.savefig(savefolder_hopping + settings.filename[:-5] + '_singleframe_transition_probs.png')
            fig.savefig(savefolder_hopping + settings.filename[:-5] + '_singleframe_transition_probs.pdf')
            
            plt.close(fig)
            gc.collect()  
            
            fig, ax = plt.subplots(2, figsize=(12,8))
        
            ax[0].axhline(1, color='red')
            ax[0].bar(1, mo_shifts_bar_comb_2.iloc[1].prob / mo_shifts_bar_comb_2.iloc[2].prob)
            ax[0].bar(2, s_shifts_bar_comb_2.iloc[1].prob / s_shifts_bar_comb_2.iloc[2].prob)
            ax[0].bar(3, vac_shifts_bar_comb_2.iloc[1].prob / vac_shifts_bar_comb_2.iloc[2].prob)
            
            ax[1].bar(1, 2*mo_shifts_bar_comb_2.iloc[3].prob / (mo_shifts_bar_comb_2.iloc[4].prob + mo_shifts_bar_comb.iloc[5].prob))
            ax[1].bar(2, 2*s_shifts_bar_comb_2.iloc[3].prob / (s_shifts_bar_comb_2.iloc[4].prob + s_shifts_bar_comb.iloc[5].prob))
            ax[1].bar(3, 2*vac_shifts_bar_comb_2.iloc[3].prob / (vac_shifts_bar_comb_2.iloc[4].prob + vac_shifts_bar_comb.iloc[5].prob))
            ax[1].bar(4, mo_shifts_bar_comb_2.iloc[4].prob / mo_shifts_bar_comb_2.iloc[5].prob)
            ax[1].bar(5, s_shifts_bar_comb_2.iloc[4].prob / s_shifts_bar_comb_2.iloc[5].prob)
            ax[1].bar(6, vac_shifts_bar_comb_2.iloc[4].prob / vac_shifts_bar_comb_2.iloc[5].prob)
            
            for axis in ax:
                axis.grid(axis='y')
                axis.axhline(1, color='red') 
                
            ax[0].set_xticks(range(1,4))    
            ax[1].set_xticks(range(1,7))
            
            ticklist1 = [r'$\frac{Mo\rightarrow S}{Mo\rightarrow HC}$',
                        r'$\frac{S\rightarrow Mo}{S\rightarrow HC}$',
                        r'$\frac{HC\rightarrow Mo}{HC\rightarrow S}$']
            ticklist2 = [r'$\frac{Mo\rightarrow Mo}{Mo\rightarrow S/HC}$',
                        r'$\frac{S\rightarrow S}{S\rightarrow Mo/HC}$',
                        r'$\frac{HC\rightarrow HC}{HC\rightarrow Mo/S}$',
                        r'$\frac{Mo\rightarrow S}{Mo\rightarrow HC}$',
                        r'$\frac{S\rightarrow Mo}{S\rightarrow HC}$',
                        r'$\frac{Hc\rightarrow Mo}{HC\rightarrow S}$']
            ax[0].set_xticklabels(ticklist1)
            ax[1].set_xticklabels(ticklist2)
            ax[0].set_ylabel(r'Nearest Neighbour (1.9$\mathring{A}$)') 
            ax[1].set_ylabel(r'2nd Nearest Neighbour (3.3 - 3.8$\mathring{A}$)')
        
            fig.suptitle(legendstring + ' Relative Transition Probabilities')
            
            fig.savefig(savefolder_hopping + settings.filename[:-5] + '_doubleframe_transition_probs.png')
            fig.savefig(savefolder_hopping + settings.filename[:-5] + '_doubleframe_transition_probs.pdf')
        
        else:
            mo_shifts_bar_list.append([])
            mo_shifts_bar_2_list.append([])
            mo_shifts_bar_comb_list.append([])
            mo_shifts_bar_comb_2_list.append([])
            
            s_shifts_bar_list.append([])
            s_shifts_bar_2_list.append([])
            s_shifts_bar_comb_list.append([])
            s_shifts_bar_comb_2_list.append([])
            
            vac_shifts_bar_list.append([])
            vac_shifts_bar_2_list.append([])
            vac_shifts_bar_comb_list.append([])
            vac_shifts_bar_comb_2_list.append([])
    
    pickle.dump(atoms_class_list, open(savefolder_hopping_pickle + 'atoms_class_list.p', 'wb'))
    
    pickle.dump(mo_shifts_list, open(savefolder_hopping_pickle + 'mo_shifts_list.p', 'wb'))
    pickle.dump(mo_shifts_2_list, open(savefolder_hopping_pickle + 'mo_shifts_2_list.p', 'wb'))
    pickle.dump(mo_shifts_bar_list, open(savefolder_hopping_pickle + 'mo_shifts_bar_list.p', 'wb'))
    pickle.dump(mo_shifts_bar_2_list, open(savefolder_hopping_pickle + 'mo_shifts_bar_2_list.p', 'wb'))
    pickle.dump(mo_shifts_bar_comb_list, open(savefolder_hopping_pickle + 'mo_shifts_bar_comb_list.p', 'wb'))
    pickle.dump(mo_shifts_bar_comb_2_list, open(savefolder_hopping_pickle + 'mo_shifts_bar_comb_2_list.p', 'wb'))
    
    pickle.dump(s_shifts_list, open(savefolder_hopping_pickle + 's_shifts_list.p', 'wb'))
    pickle.dump(s_shifts_2_list, open(savefolder_hopping_pickle + 's_shifts_2_list.p', 'wb'))
    pickle.dump(s_shifts_bar_list, open(savefolder_hopping_pickle + 's_shifts_bar_list.p', 'wb'))
    pickle.dump(s_shifts_bar_2_list, open(savefolder_hopping_pickle + 's_shifts_bar_2_list.p', 'wb'))
    pickle.dump(s_shifts_bar_comb_list, open(savefolder_hopping_pickle + 's_shifts_bar_comb_list.p', 'wb'))
    pickle.dump(s_shifts_bar_comb_2_list, open(savefolder_hopping_pickle + 's_shifts_bar_comb_2_list.p', 'wb'))
    
    pickle.dump(vac_shifts_list, open(savefolder_hopping_pickle + 'vac_shifts_list.p', 'wb'))
    pickle.dump(vac_shifts_2_list, open(savefolder_hopping_pickle + 'vac_shifts_2_list.p', 'wb'))
    pickle.dump(vac_shifts_bar_list, open(savefolder_hopping_pickle + 'vac_shifts_bar_list.p', 'wb'))
    pickle.dump(vac_shifts_bar_2_list, open(savefolder_hopping_pickle + 'vac_shifts_bar_2_list.p', 'wb'))
    pickle.dump(vac_shifts_bar_comb_list, open(savefolder_hopping_pickle + 'vac_shifts_bar_comb_list.p', 'wb'))
    pickle.dump(vac_shifts_bar_comb_2_list, open(savefolder_hopping_pickle + 'vac_shifts_bar_comb_2_list.p', 'wb'))
    










figsize= 8, 46
fig_single_mo = plt.figure(figsize=figsize)
fig_single_s = plt.figure(figsize=figsize)
fig_single_vac = plt.figure(figsize=figsize)
fig_double_mo = plt.figure(figsize=figsize)
fig_double_s = plt.figure(figsize=figsize)
fig_double_vac = plt.figure(figsize=figsize)

wetcount, drycount = 1, 1
for i in tqdm(range(numvid), desc = 'hopping analysis'):
    
    settings = settings_list[i]    
    wetdry = wetdry_list[i]
    
    if wetdry != 'ignore':
    
        atoms_norm = atoms_norm_list[i]
        atoms_norm.cons_stuck = atoms_norm.cons_stuck + atoms_norm.stuck
        atoms_class = atoms_class_list[i]        
    
        color = colors_dosescale[i]
        legendstring = settings.filename[:-5]
    
        points_set = (mo_points_fixed, s_points_fixed, vac_points_fixed)
    
        mo_shifts = mo_shifts_list[i]
        s_shifts = s_shifts_list[i]
        vac_shifts = vac_shifts_list[i]
        
        mo_shifts_2 = mo_shifts_2_list[i]
        s_shifts_2 = s_shifts_2_list[i]
        vac_shifts_2 = vac_shifts_2_list[i]
        
        numcols = max([numwet, numdry])   
        
        if wetdry == 'wet':
            ax_mo = fig_single_mo.add_subplot(numcols, 2, (2 * wetcount) - 1, aspect='equal')
            ax_s = fig_single_s.add_subplot(numcols, 2, (2 * wetcount) - 1, aspect='equal')
            ax_vac = fig_single_vac.add_subplot(numcols, 2, (2 * wetcount) - 1, aspect='equal')
            ax_mo2 = fig_double_mo.add_subplot(numcols, 2, (2 * wetcount) - 1, aspect='equal')
            ax_s2 = fig_double_s.add_subplot(numcols, 2, (2 * wetcount) - 1, aspect='equal')
            ax_vac2 = fig_double_vac.add_subplot(numcols, 2, (2 * wetcount) - 1, aspect='equal')
            wetcount += 1
        elif wetdry == 'dry':
            ax_mo = fig_single_mo.add_subplot(numcols, 2, (2 * drycount), aspect='equal')
            ax_s = fig_single_s.add_subplot(numcols, 2, (2 * drycount), aspect='equal')
            ax_vac = fig_single_vac.add_subplot(numcols, 2, (2 * drycount), aspect='equal')
            ax_mo2 = fig_double_mo.add_subplot(numcols, 2, (2 * drycount), aspect='equal')
            ax_s2 = fig_double_s.add_subplot(numcols, 2, (2 * drycount), aspect='equal')
            ax_vac2 = fig_double_vac.add_subplot(numcols, 2, (2 * drycount), aspect='equal')
            drycount +=1
            
        plotshifts(ax_mo, 'Mo', mo_shifts, points_set)
        plotshifts(ax_s, 'S', s_shifts, points_set)
        plotshifts(ax_vac, 'Vac', vac_shifts, points_set)
        plotshifts(ax_mo2, 'Mo', mo_shifts_2, points_set)
        plotshifts(ax_s2, 'S', s_shifts_2, points_set)
        plotshifts(ax_vac2, 'Vac', vac_shifts_2, points_set)
        for ax in [ax_mo, ax_mo2, ax_s, ax_s2, ax_vac, ax_vac2]:
            ax.set_facecolor('grey')
            ax.set_title(settings.filename[:-5], color=color)
                
fig_single_mo.suptitle('Hopping probability from Mo Site', color='purple')
fig_single_s.suptitle('Hopping probability from S Site', color='yellow')
fig_single_vac.suptitle('Hopping probability from HC Site', color='red')
fig_double_mo.suptitle('2 frames hopping probability from Mo Site', color='purple')
fig_double_s.suptitle('2 frames hopping probability from S Site', color='yellow')
fig_double_vac.suptitle('2 frames hopping probability from HC Site', color='red')
for fig in [fig_single_mo, fig_single_s, fig_single_vac, fig_double_mo, fig_double_s, fig_double_vac]:
    fig.subplots_adjust(top=0.96,
                        bottom=0.01,
                        left=0.025,
                        right=0.98,
                        hspace=0.2,
                        wspace=0.05)
fig_single_mo.savefig(savefolder_hopping_base + 'single_hopping_2D_mo.png')
fig_single_s.savefig(savefolder_hopping_base + 'single_hopping_2D_s.png')
fig_single_vac.savefig(savefolder_hopping_base + 'single_hopping_2D_hc.png')
fig_double_mo.savefig(savefolder_hopping_base + 'double_hopping_2D_mo.png')
fig_double_s.savefig(savefolder_hopping_base + 'double_hopping_2D_s.png')
fig_double_vac.savefig(savefolder_hopping_base + 'double_hopping_2D_hc.png')

plt.close('all')
gc.collect()
    
    
    
    


figsize= 12, 46
fig_single_mo = plt.figure(figsize=figsize)
fig_single_s = plt.figure(figsize=figsize)
fig_single_vac = plt.figure(figsize=figsize)
fig_double_mo = plt.figure(figsize=figsize)
fig_double_s = plt.figure(figsize=figsize)
fig_double_vac = plt.figure(figsize=figsize)

wetcount, drycount = 1, 1
for i in tqdm(range(numvid), desc = 'hopping analysis'):
    
    settings = settings_list[i]    
    wetdry = wetdry_list[i]
    
    if wetdry != 'ignore':
    
        atoms_norm = atoms_norm_list[i]
        atoms_class = atoms_class_list[i]   
    
        color = colors_dosescale[i]
        legendstring = settings.filename[:-5]
    
        points_set = (mo_points_fixed, s_points_fixed, vac_points_fixed)

        mo_shifts = mo_shifts_list[i]
        s_shifts = s_shifts_list[i]
        vac_shifts = vac_shifts_list[i]
        
        mo_shifts_2 = mo_shifts_2_list[i]
        s_shifts_2 = s_shifts_2_list[i]
        vac_shifts_2 = vac_shifts_2_list[i]

        mo_shifts_bar_comb = mo_shifts_bar_comb_list[i]
        s_shifts_bar_comb = s_shifts_bar_comb_list[i]
        vac_shifts_bar_comb = vac_shifts_bar_comb_list[i]
        
        mo_shifts_bar_comb_2 = mo_shifts_bar_comb_2_list[i]
        s_shifts_bar_comb_2 = s_shifts_bar_comb_2_list[i]
        vac_shifts_bar_comb_2 = vac_shifts_bar_comb_2_list[i]

        numcols = max([numwet, numdry])
        
        if wetdry == 'wet':
            ax_mo = fig_single_mo.add_subplot(numcols, 2, (2 * wetcount) - 1)
            ax_s = fig_single_s.add_subplot(numcols, 2, (2 * wetcount) - 1)
            ax_vac = fig_single_vac.add_subplot(numcols, 2, (2 * wetcount) - 1)
            ax_mo2 = fig_double_mo.add_subplot(numcols, 2, (2 * wetcount) - 1)
            ax_s2 = fig_double_s.add_subplot(numcols, 2, (2 * wetcount) - 1)
            ax_vac2 = fig_double_vac.add_subplot(numcols, 2, (2 * wetcount) - 1)
            wetcount += 1
        elif wetdry == 'dry':
            ax_mo = fig_single_mo.add_subplot(numcols, 2, (2 * drycount))
            ax_s = fig_single_s.add_subplot(numcols, 2, (2 * drycount))
            ax_vac = fig_single_vac.add_subplot(numcols, 2, (2 * drycount))
            ax_mo2 = fig_double_mo.add_subplot(numcols, 2, (2 * drycount))
            ax_s2 = fig_double_s.add_subplot(numcols, 2, (2 * drycount))
            ax_vac2 = fig_double_vac.add_subplot(numcols, 2, (2 * drycount))
            drycount +=1
        
        barplot(ax_mo, mo_shifts_bar_comb, lims=((0,0.64),(0,60)))
        barplot(ax_s, s_shifts_bar_comb, lims=((0,0.64),(0,60)))
        barplot(ax_vac, vac_shifts_bar_comb, lims=((0,0.64),(0,60)))     
        barplot(ax_mo2, mo_shifts_bar_comb_2, lims=((0,0.64),(0,50)))
        barplot(ax_s2, s_shifts_bar_comb_2, lims=((0,0.64),(0,50)))
        barplot(ax_vac2, vac_shifts_bar_comb_2, lims=((0,0.64),(0,50)))
        
        for ax in [ax_mo, ax_mo2, ax_s, ax_s2, ax_vac, ax_vac2]:
            ax.set_facecolor('grey')
            ax.set_title(settings.filename[:-5], color=color)
            ax.set_xlabel('distance (nm)')
            ax.yaxis.set_label_position('right')
            ax.set_ylabel('Relative Probability (%)')
            
for fig in [fig_single_mo, fig_single_s, fig_single_vac, fig_double_mo, fig_double_s, fig_double_vac]:
    fig.subplots_adjust(top=0.96,
                        bottom=0.03,
                        left=0.01,
                        right=0.97,
                        hspace=0.265,
                        wspace=0.09)    
        
fig_single_mo.suptitle('Hopping probability from Mo Site', color='purple')
fig_single_s.suptitle('Hopping probability from S Site', color='yellow')
fig_single_vac.suptitle('Hopping probability from HC Site', color='red')
fig_double_mo.suptitle('2 frames hopping probability from Mo Site', color='purple')
fig_double_s.suptitle('2 frames hopping probability from S Site', color='yellow')
fig_double_vac.suptitle('2 frames hopping probability from HC Site', color='red')
        
fig_single_mo.savefig(savefolder_hopping_base + 'single_hopping_bar_2D_mo.png')
fig_single_s.savefig(savefolder_hopping_base + 'single_hopping_bar_2D_s.png')
fig_single_vac.savefig(savefolder_hopping_base + 'single_hopping_bar_2D_hc.png')
fig_double_mo.savefig(savefolder_hopping_base + 'double_hopping_bar_2D_mo.png')
fig_double_s.savefig(savefolder_hopping_base + 'double_hopping_bar_2D_s.png')
fig_double_vac.savefig(savefolder_hopping_base + 'double_hopping_bar_2D_hc.png')

plt.close('all')
gc.collect()










ticklist1 = [r'$\frac{Mo\rightarrow S}{Mo\rightarrow HC}$',
            r'$\frac{S\rightarrow Mo}{S\rightarrow HC}$',
            r'$\frac{HC\rightarrow Mo}{HC\rightarrow S}$']
ticklist2 = [r'$\frac{Mo\rightarrow Mo}{Mo\rightarrow S/HC}$',
            r'$\frac{S\rightarrow S}{S\rightarrow Mo/HC}$',
            r'$\frac{HC\rightarrow HC}{HC\rightarrow Mo/S}$',
            r'$\frac{Mo\rightarrow S}{Mo\rightarrow HC}$',
            r'$\frac{S\rightarrow Mo}{S\rightarrow HC}$',
            r'$\frac{Hc\rightarrow Mo}{HC\rightarrow S}$']

figsize= 12, 46
fig_hoppingratios = plt.figure(figsize=figsize)

wetcount, drycount = 1, 1
for i in tqdm(range(numvid), desc = 'hopping analysis'):
    
    settings = settings_list[i]    
    wetdry = wetdry_list[i]
    
    if wetdry != 'ignore':
    
        atoms_norm = atoms_norm_list[i]
        atoms_class = atoms_class_list[i]   
    
        color = colors_dosescale[i]
        legendstring = settings.filename[:-5]
    
        points_set = (mo_points_fixed, vac_points_fixed, s_points_fixed)

        mo_shifts_bar_comb = mo_shifts_bar_comb_list[i]
        s_shifts_bar_comb = s_shifts_bar_comb_list[i]
        vac_shifts_bar_comb = vac_shifts_bar_comb_list[i]
        
        mo_shifts_bar_comb_2 = mo_shifts_bar_comb_2_list[i]
        s_shifts_bar_comb_2 = s_shifts_bar_comb_2_list[i]
        vac_shifts_bar_comb_2 = vac_shifts_bar_comb_2_list[i]

        numcols = max([numwet, numdry])
        
        if wetdry == 'wet':
            ax_single = fig_hoppingratios.add_subplot(numcols, 4, (4 * wetcount) - 3)
            ax_double = fig_hoppingratios.add_subplot(numcols, 4, (4 * wetcount) - 2)
            wetcount += 1
        elif wetdry == 'dry':
            ax_single = fig_hoppingratios.add_subplot(numcols, 4, (4 * drycount) - 1)
            ax_double = fig_hoppingratios.add_subplot(numcols, 4, (4 * drycount) - 0)
            drycount +=1

        ax_single.bar(1, mo_shifts_bar_comb_2.iloc[1].prob / mo_shifts_bar_comb_2.iloc[2].prob)
        ax_single.bar(2, s_shifts_bar_comb_2.iloc[1].prob / s_shifts_bar_comb_2.iloc[2].prob)
        ax_single.bar(3, vac_shifts_bar_comb_2.iloc[1].prob / vac_shifts_bar_comb_2.iloc[2].prob)
        
        ax_double.bar(1, 2*mo_shifts_bar_comb_2.iloc[3].prob / (mo_shifts_bar_comb_2.iloc[4].prob + mo_shifts_bar_comb.iloc[5].prob))
        ax_double.bar(2, 2*s_shifts_bar_comb_2.iloc[3].prob / (s_shifts_bar_comb_2.iloc[4].prob + s_shifts_bar_comb.iloc[5].prob))
        ax_double.bar(3, 2*vac_shifts_bar_comb_2.iloc[3].prob / (vac_shifts_bar_comb_2.iloc[4].prob + vac_shifts_bar_comb.iloc[5].prob))
        ax_double.bar(4, mo_shifts_bar_comb_2.iloc[4].prob / mo_shifts_bar_comb_2.iloc[5].prob)
        ax_double.bar(5, s_shifts_bar_comb_2.iloc[4].prob / s_shifts_bar_comb_2.iloc[5].prob)
        ax_double.bar(6, vac_shifts_bar_comb_2.iloc[4].prob / vac_shifts_bar_comb_2.iloc[5].prob)
        
        for axis in [ax_single, ax_double]:
            axis.grid(axis='y')
            axis.axhline(1, color='red')
            axis.set_ylim(0, 3)
            
        ax_single.set_xticks(range(1,4))    
        ax_double.set_xticks(range(1,7))
        ax_single.set_xticklabels(ticklist1)
        ax_double.set_xticklabels(ticklist2)
        
        ax_single.set_xlabel(settings.filename[:-5], color=color)
        ax_double.set_xlabel(settings.filename[:-5], color=color)
        
        if (wetdry == 'wet' and wetcount==2):
            ax_single.set_title(r'Liquid NN hopping probabilities (1.9$\mathring{A}$)') 
            ax_double.set_title(r'Liquid 2nd NN (3.3 - 3.8$\mathring{A}$)')        
        elif (wetdry=='dry' and drycount==2):
            ax_single.set_title(r'Vacuum NN hopping probabilities (1.9$\mathring{A}$)') 
            ax_double.set_title(r'Vacuum 2nd NN (3.3 - 3.8$\mathring{A}$)')  

fig_hoppingratios.subplots_adjust(  top=0.99,
                                    bottom=0.03,
                                    left=0.035,
                                    right=0.995,
                                    hspace=0.265,
                                    wspace=0.29)
        
fig_hoppingratios.savefig(savefolder_hopping_base + 'single_hopping_bar_2D_mo.png')

plt.close('all')
gc.collect()




wet_shifts_list = []
dry_shifts_list = []
wet_shifts2_list = []
dry_shifts2_list = []
excludelist = ['Scanning Preview19',]
for i in tqdm(range(numvid), desc = 'hopping analysis'):
    
    settings = settings_list[i]    
    wetdry = wetdry_list[i]
    
    if wetdry != 'ignore':
        
        if settings.filename[:-5] not in excludelist:
    
            atoms_norm = atoms_norm_list[i]
            atoms_class = atoms_class_list[i]   
    
            points_set = (mo_points_fixed, s_points_fixed, vac_points_fixed)
        
            mo_shifts = mo_shifts_list[i]
            s_shifts = s_shifts_list[i]
            vac_shifts = vac_shifts_list[i]
            
            mo_shifts_2 = mo_shifts_2_list[i]
            s_shifts_2 = s_shifts_2_list[i]
            vac_shifts_2 = vac_shifts_2_list[i]
            
            if wetdry == 'wet':
                wet_shifts_list.append([mo_shifts, s_shifts, vac_shifts])
                wet_shifts2_list.append([mo_shifts_2, s_shifts_2, vac_shifts_2]) 
            elif wetdry == 'dry':
                dry_shifts_list.append([mo_shifts, s_shifts, vac_shifts])
                dry_shifts2_list.append([mo_shifts_2, s_shifts_2, vac_shifts_2]) 

wet_shifts_sum = combine_shifts_fixed(wet_shifts_list)
dry_shifts_sum = combine_shifts_fixed(dry_shifts_list)
wet_shifts2_sum = combine_shifts_fixed(wet_shifts2_list)
dry_shifts2_sum = combine_shifts_fixed(dry_shifts2_list)

def plotshifts2(ax, typestr, shifts, points_set, lims = (-0.4, 0.4), offset = (0.01, 0.01), dotscale=50):
    mo_points, s_points, vac_points = points_set
    ax.set_xlim(lims)
    ax.set_ylim(lims)
    if typestr == 'Mo':
        molplot_mo(ax, mo_points, s_points, alpha=0.6)
    elif typestr == 'S':
        molplot_s(ax, mo_points, s_points, alpha=0.6)
    elif typestr == 'Vac':
        molplot_vac(ax, mo_points, vac_points, alpha=0.6)
    to_mo_counts, to_s_counts, to_vac_counts = shifts
    if len(to_mo_counts[0]) > 0:
        ax.scatter(to_mo_counts[1][:, 0], to_mo_counts[1][:, 1], color='purple', s=to_mo_counts[3] * dotscale, zorder=5)
        textplot2(to_mo_counts, ax, offset, 'black')
    if len(to_s_counts[0]) > 0:
        ax.scatter(to_s_counts[1][:, 0], to_s_counts[1][:, 1], color='yellow', s=to_s_counts[3] * dotscale, zorder=5)
        textplot2(to_s_counts, ax, offset, 'black')
    if len(to_vac_counts[0]) > 0:
        ax.scatter(to_vac_counts[1][:, 0], to_vac_counts[1][:, 1], color='red', s=to_vac_counts[3] * dotscale, zorder=5)
        textplot2(to_vac_counts, ax, offset, 'black')
    ax.set_xticks([])
    ax.set_yticks([])
    scalebar=ScaleBar(1, 'nm', label=None, box_alpha=0, color='black', location='lower left')
    ax.add_artist(scalebar)
    
def textplot2(cat_shifts, ax, offset, color):
    xlims=ax.get_xlim()
    ylims=ax.get_ylim()
    for i in range(len(cat_shifts[0])):
        if xlims[0] < cat_shifts[1][i, 0] + offset[0] < xlims[1] and ylims[0] < cat_shifts[1][i, 1] + offset[1] < ylims[1]:
            ax.text(cat_shifts[1][i, 0] + offset[0], cat_shifts[1][i, 1] + offset[1], str(str(np.round(cat_shifts[3][i], 1)) + '%'), zorder=10)


fig, ax = plt.subplots(2, 3, figsize=(12, 10))
plotshifts2(ax[0, 0], 'Mo', wet_shifts_sum[0], points_fixed)
plotshifts2(ax[0, 1], 'S', wet_shifts_sum[1], points_fixed)
plotshifts2(ax[0, 2], 'Vac', wet_shifts_sum[2], points_fixed)
plotshifts2(ax[1, 0], 'Mo', dry_shifts_sum[0], points_fixed)
plotshifts2(ax[1, 1], 'S', dry_shifts_sum[1], points_fixed)
plotshifts2(ax[1, 2], 'Vac', dry_shifts_sum[2], points_fixed)
ax[1, 0].set_xlabel('From Mo Site', fontweight='bold')
ax[1, 1].set_xlabel('From S Site', fontweight='bold')
ax[1, 2].set_xlabel('From HC Site', fontweight='bold')
ax[0, 0].set_ylabel('Liquid Cell', fontweight='bold')
ax[1, 0].set_ylabel('Vacuum', fontweight='bold')
for axarr in ax:
    for a in axarr:
        a.set_aspect('equal')
fig.savefig(savefolder_hopping_base + '2dplots.png')


fig, ax = plt.subplots(2, 3, figsize=(12, 10))
plotshifts2(ax[0, 0], 'Mo', wet_shifts2_sum[0], points_fixed)
plotshifts2(ax[0, 1], 'S', wet_shifts2_sum[1], points_fixed)
plotshifts2(ax[0, 2], 'Vac', wet_shifts2_sum[2], points_fixed)
plotshifts2(ax[1, 0], 'Mo', dry_shifts2_sum[0], points_fixed)
plotshifts2(ax[1, 1], 'S', dry_shifts2_sum[1], points_fixed)
plotshifts2(ax[1, 2], 'Vac', dry_shifts2_sum[2], points_fixed)
ax[1, 0].set_xlabel('From Mo Site', fontweight='bold')
ax[1, 1].set_xlabel('From S Site', fontweight='bold')
ax[1, 2].set_xlabel('From HC Site', fontweight='bold')
ax[0, 0].set_ylabel('Liquid Cell', fontweight='bold')
ax[1, 0].set_ylabel('Vacuum', fontweight='bold')
for axarr in ax:
    for a in axarr:
        a.set_aspect('equal')
fig.savefig(savefolder_hopping_base + '2dplots_double.png')

























fig = plt.figure(figsize=(16, 18))
grid = (14, 12)
fig.subplots_adjust(top=0.975,
                    bottom=0.03,
                    left=0.03,
                    right=0.96,
                    hspace=0.1,
                    wspace=0.2)

excludelist = ['Scanning Preview19',]

wet_atoms_norm = wet_atoms_norm_all.copy()
wet_atoms_norm = wet_atoms_norm[~wet_atoms_norm.video.isin(excludelist)]
dry_atoms_norm = dry_atoms_norm_all.copy()
dry_atoms_norm = dry_atoms_norm[~dry_atoms_norm.video.isin(excludelist)]
exwetnum = len(wet_atoms_norm.video.unique())
exdrynum = len(dry_atoms_norm.video.unique())

axstuck = plt.subplot2grid(grid, (0,0), rowspan=5, colspan=6, fig=fig, frameon=False)
axstuck.set_xticks([])
axstuck.set_yticks([])
axstuckwet = axstuck.inset_axes([0, 0, 0.5, 1])
axstuckdry = axstuck.inset_axes([0.5, 0, 0.5, 1])
wetcount = 0
drycount = 0
axlabels_wet, axlabels_dry = np.empty(exwetnum * 4, dtype=object), np.empty(exdrynum * 4, dtype=object)
for i in tqdm(range(numvid), desc = 'counting categories'):
    atoms_norm = atoms_norm_list[i]
    settings = settings_list[i]
    color = colors_dosescale[i]
    wetdry = wetdry_list[i]
    if (wetdry != 'ignore') and (~np.isin(settings.filename[:-5],excludelist)):
        legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
        stuck_count = stuck_counter(atoms_norm[atoms_norm.nn_dist < 20])
        colors = ['grey', 'purple', 'yellow', 'red']
        labels = [str(beam_current_list[i] * 1e12)] * 4
        if wetdry == 'wet':
            positions = (np.linspace(0, 3, 4) * exwetnum + wetcount).astype(int)
            axstuckwet.bar(positions, stuck_count, color=colors, alpha=0.5, edgecolor='black', zorder=10)
            axlabels_wet[positions] = labels
            wetcount += 1
        elif wetdry == 'dry':
            positions = (np.linspace(0, 3, 4) * exdrynum + drycount).astype(int)
            axstuckdry.bar(positions, stuck_count, color=colors, alpha=0.5, edgecolor='black', zorder=10)
            axlabels_dry[positions] = labels
            drycount += 1
axstuckwet.set_xticks(range(exwetnum*4))
axstuckdry.set_xticks(range(exdrynum*4))
axstuckwet.set_xticklabels(axlabels_wet, rotation=70)
axstuckdry.set_xticklabels(axlabels_dry, rotation=70)

wet_stuck_count = stuck_counter(wet_atoms_norm[wet_atoms_norm.nn_dist < 20])
dry_stuck_count = stuck_counter(dry_atoms_norm[dry_atoms_norm.nn_dist < 20])
wet_avticks = [split.mean() for split in np.split(axstuckwet.get_xticks(), 4)]
dry_avticks = [split.mean() for split in np.split(axstuckdry.get_xticks(), 4)]
wet_avtickwidth = [len(split) for split in np.split(axstuckwet.get_xticks(), 4)]
dry_avtickwidth = [len(split) for split in np.split(axstuckdry.get_xticks(), 4)]
axstuckwet.bar(wet_avticks, wet_stuck_count, wet_avtickwidth, color=colors, alpha=0.5, edgecolor='black', zorder=5)
axstuckdry.bar(dry_avticks, dry_stuck_count, dry_avtickwidth, color=colors, alpha=0.5, edgecolor='black', zorder=5)

for ax in [axstuckwet, axstuckdry]:
    ax.set_facecolor('lightgrey')
    ax.set_ylim(0, 0.6)
    ax.yaxis.grid()
axstuckdry.set_yticklabels([])
axstuckdry.yaxis.tick_right()
    
axstuckwet.add_artist(AnchoredText('Liquid Cell', loc='upper right', frameon=False))
axstuckdry.add_artist(AnchoredText('Vacuum', loc='upper right', frameon=False))

axstuck.set_title(r' $\bf{a}$ Single Frame Stuck Probability', loc='left')
axstuckwet.set_ylabel('Probability')

legend_elements = [Patch(facecolor='grey', edgecolor='black', alpha=0.5, label='Overall'),
                   Patch(facecolor='purple', edgecolor='black', alpha=0.5, label='Mo Site'),
                   Patch(facecolor='yellow', edgecolor='black', alpha=0.5, label='S Site'),
                   Patch(facecolor='red', edgecolor='black', alpha=0.5, label='HC Site')]

axstuckwet.legend(handles=legend_elements, loc='upper left')
axstuck.set_xlabel('Beam Current (pA)', labelpad=40)



axstuck_cons = plt.subplot2grid(grid, (0,6), rowspan=5, colspan=6, fig=fig, frameon=False)
axstuck_cons.set_xticks([])
axstuck_cons.set_yticks([])
axstuckwet_cons = axstuck_cons.inset_axes([0, 0, 0.48, 0.9])
axstuckdry_cons = axstuck_cons.inset_axes([0.52, 0, 0.48, 0.9])

excludelist = ['Scanning Preview19',]
wet_atoms_norm = wet_atoms_norm_all.copy()
wet_atoms_norm = wet_atoms_norm[~wet_atoms_norm.video.isin(excludelist)]
wet_atoms_norm.cons_stuck = wet_atoms_norm.cons_stuck + wet_atoms_norm.stuck
dry_atoms_norm = dry_atoms_norm_all.copy()
dry_atoms_norm = dry_atoms_norm[~dry_atoms_norm.video.isin(excludelist)]
dry_atoms_norm.cons_stuck = dry_atoms_norm.cons_stuck + dry_atoms_norm.stuck
wet_stuck_prob_cons, wet_stuck_prob_mo_cons, wet_stuck_prob_s_cons, wet_stuck_prob_vac_cons = stuck_counter_cons(wet_atoms_norm)
dry_stuck_prob_cons, dry_stuck_prob_mo_cons, dry_stuck_prob_s_cons, dry_stuck_prob_vac_cons = stuck_counter_cons(dry_atoms_norm)

ax = [axstuckwet_cons, axstuckdry_cons]
ax[0].plot(wet_stuck_prob_cons[0] + 1, wet_stuck_prob_cons[1], color='black', label='Overall')
ax[0].plot(wet_stuck_prob_mo_cons[0] + 1, wet_stuck_prob_mo_cons[1], color='purple', label='Mo')
ax[0].plot(wet_stuck_prob_s_cons[0] + 1, wet_stuck_prob_s_cons[1], color='yellow', label='S')
ax[0].plot(wet_stuck_prob_vac_cons[0] + 1, wet_stuck_prob_vac_cons[1], color='red', label='HC')
ax[1].plot(dry_stuck_prob_cons[0] + 1, dry_stuck_prob_cons[1], color='black', label='Overall')
ax[1].plot(dry_stuck_prob_mo_cons[0] + 1, dry_stuck_prob_mo_cons[1], color='purple', label='Mo')
ax[1].plot(dry_stuck_prob_s_cons[0] + 1, dry_stuck_prob_s_cons[1], color='yellow', label='S')
ax[1].plot(dry_stuck_prob_vac_cons[0] + 1, dry_stuck_prob_vac_cons[1], color='red', label='HC')
tvals, tcounts = np.unique(timescale_list, return_counts=True)
usetime = tvals[tcounts.argmax()]
for a in ax:
    a.set_yscale('log')
    a.set_xlim(1,8)
    a.set_xticks(range(1,9))
    a.set_ylim(0.001, 1)
    a.grid(which='major', axis='y', color='red')
    a.grid(which='minor', axis='y', linestyle='--', alpha=0.2)
    a.grid(which='major', axis='x', alpha=0.5)
    axlim = a.get_xlim()
    axtop = a.twiny()
    axtop.set_xlim(axlim[0] * float(usetime), axlim[1] * float(settings.FrameTime))
    axtop.set_xlabel('time (s)')
    a.set_xlabel('frames')
    a.set_facecolor('lightgrey')
    
ax[1].set_ylabel('Stuck Probability')
ax[1].yaxis.tick_right()
ax[1].yaxis.set_label_position("right")
ax[0].set_yticklabels([])

ax[0].add_artist(AnchoredText('Liquid Cell', loc='upper right', frameon=False))
ax[1].add_artist(AnchoredText('Vacuum', loc='upper right', frameon=False))
legend_elements = [Line2D([0], [0], color='black', lw=2, label='Overall'),
                   Line2D([0], [0], color='purple', lw=2, label='Mo Site'),
                   Line2D([0], [0], color='yellow', lw=2, label='S Site'), 
                   Line2D([0], [0], color='red', lw=2, label='HC Site')]

ax[1].legend(handles=legend_elements, loc='lower left')
axstuck_cons.set_title(r' $\bf{b}$ Time at same site', loc='left')

ax00 = plt.subplot2grid(grid, (6, 0), rowspan=4, colspan=4)
ax01 = plt.subplot2grid(grid, (6, 4), rowspan=4, colspan=4)
ax02 = plt.subplot2grid(grid, (6, 8), rowspan=4, colspan=4)
ax10 = plt.subplot2grid(grid, (10, 0), rowspan=4, colspan=4)
ax11 = plt.subplot2grid(grid, (10, 4), rowspan=4, colspan=4)
ax12 = plt.subplot2grid(grid, (10, 8), rowspan=4, colspan=4)

ax = np.array([[ax00, ax01, ax02],
      [ax10, ax11, ax12]])

plotshifts2(ax[0, 0], 'Mo', wet_shifts_sum[0], points_fixed)
plotshifts2(ax[0, 1], 'S', wet_shifts_sum[1], points_fixed)
plotshifts2(ax[0, 2], 'Vac', wet_shifts_sum[2], points_fixed)
plotshifts2(ax[1, 0], 'Mo', dry_shifts_sum[0], points_fixed)
plotshifts2(ax[1, 1], 'S', dry_shifts_sum[1], points_fixed)
plotshifts2(ax[1, 2], 'Vac', dry_shifts_sum[2], points_fixed)
ax[1, 0].set_xlabel('From Mo Site', fontweight='bold')
ax[1, 1].set_xlabel('From S Site', fontweight='bold')
ax[1, 2].set_xlabel('From HC Site', fontweight='bold')
ax[0, 0].set_ylabel('Liquid Cell', fontweight='bold')
ax[1, 0].set_ylabel('Vacuum', fontweight='bold')
for axarr in ax:
    for a in axarr:
        a.set_aspect('equal')
        a.set_facecolor('lightgrey')
        
ax00.set_title(r' $\bf{c}$ Single Frame Hopping Probability', loc='left')

fig.savefig(savefolder_figs + 'hopping_probabilities.png', dpi=300)
fig.savefig(savefolder_figs + 'hopping_probabilities.pdf', dpi=300)


























# # trackpy emsd analysis

# fitcut_f = 42.0
# fitcut_d = 217

# savefolder_tp = savefolder_msd + 'tp_crockergrier' + os.sep
# os.makedirs(savefolder_tp, exist_ok=True)

# emsd_frame_list = []
# fit_frame_list = []

# for i in tqdm(range(numvid), desc = 'msd analysis'):
    
#     atoms_norm = atoms_norm_list[i]
#     settings = settings_list[i]
#     color = colors_bc[i]
#     legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
#     timescale = timescale_list[i]
#     wetdry = wetdry_list[i]
    
#     if wetdry != 'ignore':

#         fake_atoms_norm = atoms_norm.copy()
#         fake_atoms_norm['x'] = fake_atoms_norm['fake_x']
#         fake_atoms_norm['y'] = fake_atoms_norm['fake_y']   
        
#         xvals = np.arange(1, atoms_norm.frame.max() + 2)
#         fitcut_s = atoms_norm.frame.max()/2
        
#         emsd_time = tp.emsd(atoms_norm, 1, 1/timescale)
#         fit_time = tp.utils.fit_powerlaw(emsd_time[emsd_time.index < fitcut_s*timescale], plot=False)
        
#         fake_emsd_time = tp.emsd(fake_atoms_norm, 1, 1/timescale)
#         fake_fit_time = tp.utils.fit_powerlaw(fake_emsd_time[fake_emsd_time.index < fitcut_s*timescale], plot=False)
        
#         emsd_frame = tp.emsd(atoms_norm, 1, 1)
#         fit_frame = tp.utils.fit_powerlaw(emsd_frame[emsd_frame.index < fitcut_s*timescale], plot=False)
        
#         fake_emsd_frame = tp.emsd(fake_atoms_norm, 1, 1)
#         fake_fit_frame = tp.utils.fit_powerlaw(fake_emsd_frame[fake_emsd_frame.index < fitcut_s*timescale], plot=False)
        
#         fig, ax = plt.subplots(1, 2)
#         for axis in ax:
#             axis.plot(emsd_frame, 'o', color='blue', alpha=0.3, label='Identified Pt')
#             axis.plot(fake_emsd_frame, 'o', color='red', alpha=0.3, label='Closest Site')
#             axis.plot(xvals, fit_frame.A[0]*pow(xvals, fit_frame.n[0]), color='blue', label='_nolabel_')
#             axis.plot(xvals, fake_fit_frame.A[0]*pow(xvals, fake_fit_frame.n[0]), color='red', label='_nolabel_')
#             axis.axvline(fitcut_s)
#             axis.set_xlabel('Frames')
#         ax[1].set_yscale('log')
#         ax[1].set_xscale('log')
#         ax[1].annotate(r'At$^n$: ' + 'A=%.3f, n=%.2f' % (fit_frame.A[0], fit_frame.n[0]), (1.4, 1), color='blue')
#         ax[1].annotate(r'At$^n$: ' + 'A=%.3f, n=%.2f' % (fake_fit_frame.A[0], fake_fit_frame.n[0]), (1.4, 0.8), color='red')
#         ax[0].legend()
#         ax[0].set_ylabel('EMSD')
#         fig.suptitle(legendstring + ' (trackpy / no weights)')
#         if wetdry == 'wet':
#             fig.savefig(savefolder_tp + settings.filename[:-5] + '_wet_tp_emsd.png')
#             fig.savefig(savefolder_tp + settings.filename[:-5] + '_wet_tp_emsd.pdf')
#         elif wetdry == 'dry':
#             fig.savefig(savefolder_tp + settings.filename[:-5] + '_dry_tp_emsd.png')
#             fig.savefig(savefolder_tp + settings.filename[:-5] + '_dry_tp_emsd.pdf')
#         plt.close(fig)
        
#         emsd_frame_list.append(emsd_frame)
#         fit_frame_list.append(fit_frame)
        
#     else:
#         emsd_frame_list.append([])
#         fit_frame_list.append([])
        
        
        
        
        
        
        
        
        
        
        
# # tpmsd comparison

# inset_params = [0.05, 0.6, 0.35, 0.35]
# fig_tpmsd, [ax_tpmsd_wet, ax_tpmsd_dry] = plt.subplots(1, 2, figsize=(24,16))
# ax_tpmsd_wet_log = ax_tpmsd_wet.inset_axes(inset_params)
# ax_tpmsd_dry_log = ax_tpmsd_dry.inset_axes(inset_params)        

# for i in tqdm(range(numvid), desc = 'msd analysis'):
    
#     atoms_norm = atoms_norm_list[i]
#     settings = settings_list[i]
#     color = colors_bc[i]
#     legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
#     timescale = timescale_list[i]
#     wetdry = wetdry_list[i]
    
#     if wetdry != 'ignore': 
#         xvals = np.arange(1, atoms_norm.frame.max() + 2)
#         emsd_frame = emsd_frame_list[i]
#         fit_frame = fit_frame_list[i]
        
#         if wetdry == 'wet':
#             ax_tpmsd_wet.plot(emsd_frame, 'o', color=color, alpha=0.3, label=settings.filename[:-5] + ' %d pA' %(beam_current_list[i] * 1e12) + r' At$^n$: ' + 'A=%.3f, n=%.2f' % (fit_frame.A[0], fit_frame.n[0]))
#             ax_tpmsd_wet.plot(xvals, fit_frame.A[0]*pow(xvals, fit_frame.n[0]), color=color, label='_nolabel_')
#             ax_tpmsd_wet_log.plot(emsd_frame, 'o', color=color, alpha=0.3, label='_nolabel_')
#             ax_tpmsd_wet_log.plot(xvals, fit_frame.A[0]*pow(xvals, fit_frame.n[0]), color=color)

#         if wetdry == 'dry':
#             ax_tpmsd_dry.plot(emsd_frame, 'o', color=color, alpha=0.3, label=settings.filename[:-5] + ' %d pA' %(beam_current_list[i] * 1e12) + r' At$^n$: ' + 'A=%.3f, n=%.2f' % (fit_frame.A[0], fit_frame.n[0]))
#             ax_tpmsd_dry.plot(xvals, fit_frame.A[0]*pow(xvals, fit_frame.n[0]), color=color, label='_nolabel_')
#             ax_tpmsd_dry_log.plot(emsd_frame, 'o', color=color, alpha=0.3, label='_nolabel_')
#             ax_tpmsd_dry_log.plot(xvals, fit_frame.A[0]*pow(xvals, fit_frame.n[0]), color=color)

# for ax in [ax_tpmsd_wet_log, ax_tpmsd_dry_log]:
#     ax.set_yscale('log')
#     ax.set_xscale('log')
#     ax.grid(which='both')
#     ax.set_xlim(1, 40)
#     ax.set_ylim(0.02, 2)
# for ax in [ax_tpmsd_wet, ax_tpmsd_dry]:
#     ax.legend(loc='lower right')
#     ax.set_xlim(0, 40)
#     ax.set_ylim(0, 1.5)
#     ax.grid(which='both')
#     ax.set_ylabel('EMSD')
#     ax.set_xlabel('Frames')
# ax_tpmsd_wet.set_title('Liquid Cell')
# ax_tpmsd_dry.set_title('Vacuum')

# fig_tpmsd.subplots_adjust(top=0.955,
#                           bottom=0.045,
#                           left=0.04,
#                           right=0.975,
#                           hspace=0.2,
#                           wspace=0.12)

# fig_tpmsd.savefig(savefolder_msd + 'all_tp_emsd.png')
# fig_tpmsd.savefig(savefolder_msd + 'all_tp_emsd.pdf')

# plt.close('all')
# gc.collect()














# calculate moments of stationarity

savefolder_stationarity = savefolder + 'stationarity' + os.sep
os.makedirs(savefolder_stationarity, exist_ok=True)

savefolder_stationarity_calc = savefolder_stationarity + 'calculated_moments' + os.sep
os.makedirs(savefolder_stationarity_calc, exist_ok=True)

lagtimes = range(1, 51)

for i in tqdm(range(numvid), desc = 'tamsd stationarity calculation'):
    
    atoms_norm = atoms_norm_list[i]
    settings = settings_list[i]

    if not os.path.exists(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt'):
        profiles = np.full((len(lagtimes), atoms_norm.frame.max()), np.nan, dtype='float64')
        for lagt in tqdm(lagtimes):
            profile = moment_stationarity_calc(atoms_norm, 2, lagt)
            profiles[lagt-1][0:len(profile)] = profile

        np.savetxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt', profiles)
        np.savetxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt', lagtimes)


savefolder_stationarity_individual = savefolder_stationarity + 'individual' + os.sep
os.makedirs(savefolder_stationarity_individual, exist_ok=True)

# analysis of moments of stationarity

for i in tqdm(range(numvid), desc = 'tamsd stationarity analysis'):
    
    atoms_norm = atoms_norm_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]

    profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
    lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)

    profiles_norm = profiles.copy()

    for j in range(0, len(lagtimes)):
        lagt = lagtimes[j]
        profiles_norm[j] = profiles[j] / lagt

    lagt_mat = profiles_norm.copy()
    
    t_real_mat = lagt_mat.copy()

    for j in range(0, len(lagtimes)):
        lagt = lagtimes[j]
        for t in range(0, len(lagt_mat[j])):
            if not np.isnan(lagt_mat[j, t]):
                lagt_mat[j, t] = lagt * timescale
                t_real_mat[j, t] = ((t + t + lagt) / 2) * timescale
    
    #plot msd(t,lagT) in 2D
    
    fig = plt.figure(figsize=(12,18))
    
    cmap = get_cmap('jet')
    cmap2 = get_cmap('inferno')
    
    ax00 = fig.add_subplot(211)
    ax00.set_xlabel('Lag time (s)')
    ax00.set_ylabel('<MSD(t, lag time)>')
    for j in range(0, profiles.data.shape[1]):
        ax00.plot(lagtimes[0:12] * timescale, profiles[0:12,j], color=cmap(j / profiles.data.shape[1]), alpha=0.3)
    sm = ScalarMappable(cmap=cmap)
    sm.set_array(np.linspace(0, profiles.data.shape[1]-1, profiles.data.shape[1]) * timescale)
    cbar = fig.colorbar(sm)
    cbar.set_label('Real Time (s)')
    ax00.set_ylim(bottom=0)
    ax00.set_xlim(0, lagtimes[11] * timescale)
    
    ax01 = fig.add_subplot(212)
    ax01.set_xlabel('Real Time (s)')
    ax01.set_ylabel('<MSD(t, lag time)>')
    for j in range(0, profiles.data.shape[0]):
        ax01.plot(np.arange(1, profiles.shape[1] + 1) * timescale ,profiles[j,:], color=cmap2(j / profiles.data.shape[0]))
    sm = ScalarMappable(cmap=cmap2)
    sm.set_array(np.linspace(0, profiles.data.shape[0]-1, profiles.data.shape[0]) * timescale)
    cbar = fig.colorbar(sm)
    cbar.set_label('Lag Time (s)')
    ax01.set_ylim(bottom=0)
    ax01.set_xlim(0, (profiles.shape[1] + 1) * timescale)
    
    fig.suptitle(legendstring)
    fig.subplots_adjust(top=0.96,
                        bottom=0.03,
                        left=0.06,
                        right=1.0,
                        hspace=0.1,
                        wspace=0.175)
    
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_1d_plots.png')
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_1d_plots.pdf')
    
    
    #plot msd(t,lagT) normed in 2D
    
    fig = plt.figure(figsize=(12,18))
    
    cmap = get_cmap('jet')
    cmap2 = get_cmap('inferno')
    
    ax00 = fig.add_subplot(211)
    ax00.set_xlabel('Lag time (s)')
    ax00.set_ylabel('<MSD(t, lag time)>/lag time')
    for j in range(0, profiles.data.shape[1]):
        ax00.plot(lagtimes[0:12] * timescale, profiles_norm[0:12,j], color=cmap(j / profiles.data.shape[1]), alpha=0.3)
    sm = ScalarMappable(cmap=cmap)
    sm.set_array(np.linspace(0, profiles.data.shape[1]-1, profiles.data.shape[1]) * timescale)
    cbar = fig.colorbar(sm)
    cbar.set_label('Real Time (s)')
    ax00.set_ylim(bottom=0)
    ax00.set_xlim(0, lagtimes[11] * timescale)
    
    ax01 = fig.add_subplot(212)
    ax01.set_xlabel('Real Time (s)')
    ax01.set_ylabel('<MSD(t, lag time)>')
    for j in range(0, profiles.data.shape[0]):
        ax01.plot(np.arange(1, profiles.shape[1] + 1) * timescale ,profiles[j,:], color=cmap2(j / profiles.data.shape[0]))
    sm = ScalarMappable(cmap=cmap2)
    sm.set_array(np.linspace(0, profiles.data.shape[0]-1, profiles.data.shape[0]) * timescale)
    cbar = fig.colorbar(sm)
    cbar.set_label('Lag Time (s)')
    ax01.set_ylim(bottom=0)
    ax01.set_xlim(0, (profiles.shape[1] + 1) * timescale)
    
    fig.suptitle(legendstring)
    fig.subplots_adjust(top=0.96,
                        bottom=0.03,
                        left=0.06,
                        right=1.0,
                        hspace=0.1,
                        wspace=0.175)
    
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_1d_plots_normed.png')
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_1d_plots_normed.pdf')
    


    savefolder_lagtcut = savefolder_stationarity + 'fit_msd_function_time' + os.sep + settings.filename[:-5] + os.sep
    os.makedirs(savefolder_lagtcut, exist_ok=True)
    
    lagtrange = list(range(4, 18, 2)) + list(range(16, 41, 4))
    
    for lagtcut in tqdm(lagtrange):
        
        savename = savefolder_lagtcut + 'lagtcut_%d' % lagtcut
        
        if not os.path.exists(savename + '.png'):
    
            fig = plt.figure(figsize=(14, 18))
        
            axplot = fig.add_subplot(311)
            exponents = []
            coefficients = []
            tnum = []
            for j in range(0, profiles.data.shape[1] - 10):
                timemsd = profiles[:,j][np.isfinite(profiles[:,j])]
                if len(timemsd) > 2:
                    axplot.plot(lagtimes[0: len(timemsd)] * timescale, timemsd, '.', markersize=0.1, color = get_cmap('jet')(j / profiles.data.shape[1]))
                    popt, pcov = curve_fit(modelfunc, (lagtimes[0: len(timemsd)] * timescale)[:lagtcut], timemsd[:lagtcut], p0=[0.8, 0.02], bounds=(0, [2, 0.8]))
                    xvals = np.linspace(0, 40 * timescale, 800)
                    yvals = modelfunc(xvals, popt[0], popt[1])
                    axplot.plot(xvals, yvals, color = get_cmap('jet')(j / profiles.data.shape[1]), alpha=0.2)
                    exponents.append(popt[0])
                    coefficients.append(popt[1])
                    tnum.append(j)
            tnum = np.array(tnum)
            axplot.set_title(r' $\bf{(a)}$ TAMSD', loc='left', pad=-12)
            axplot.set_xlabel('lag time (s)')
            axplot.set_ylabel('MSD') 
            axplot.axvline(lagtcut * timescale)
            axplot.set_xlim(timescale, 40 * timescale)
            axplot.set_ylim(0,1)
            sm = ScalarMappable(cmap='jet')
            sm.set_array(np.linspace(0, profiles.data.shape[1]-1, profiles.data.shape[1]) * timescale)
            cbar = fig.colorbar(sm)
            cbar.set_label('time (s)')
        
            axfit = fig.add_subplot(312)
            axfit.plot((tnum + 1) * timescale, exponents, color='red')
            axfit.set_ylabel(r'Fit exponent $\alpha$', color='red')
            axfit.set_xlabel('Time (s)')
            axfit.tick_params(axis='y', color='red', labelcolor='red')
            axfit.set_xlim(timescale, (profiles.data.shape[1] - 10) * timescale)
            axfit.set_ylim(0.5, 1.2)
            axtwin = axfit.twinx()
            axtwin.plot((tnum + 1) * timescale, coefficients, color='blue')
            axtwin.set_ylabel(r'Fit Coefficient D$_\alpha$', color='blue')
            axtwin.tick_params(axis='y', color='blue', labelcolor='blue')
            axtwin.set_ylim(0.00, 0.1)
            axfit.set_title(r' $\bf{(b)}$ Fit Exponents', loc='left', pad=-12)
            
            axdist = fig.add_subplot(313)
            axdist.set_title(r' $\bf{(c)}$ Wet', loc='left', pad=-12)
            axdist.set_xlabel('Time (s)')
            timestep = lagtcut * timescale
            timestep_i = int(np.round(timestep))       
            axdist.set_ylabel('MSD (t, lagt = %ds)' % timestep_i)
            for j in range(0, profiles.data.shape[1] - 10):
                if j in tnum:
                    jval = np.where(tnum == j)[0][0]
                    xval = (j+1) * timescale
                    yval = modelfunc(timestep, exponents[jval], coefficients[jval])
                    color = get_cmap('jet')(j / profiles.data.shape[1])
                    axdist.scatter(xval, yval, color=color)
                
        
        
            fig.subplots_adjust(top=0.995,
                                bottom=0.075,
                                left=0.085,
                                right=0.925,
                                hspace=0.24,
                                wspace=0.53)
        
            fig.savefig(savename + '.png')
            fig.savefig(savename + '.pdf')
    
            plt.close('all')
            collect()
            
    plt.close('all')
    collect()


    #plot msd(t,lagT) in 2D
    
    stationarity_trisurf = np.stack((t_real_mat, lagt_mat, profiles), axis=-1)

    x = stationarity_trisurf[:,:,0].flatten()
    x = x[~np.isnan(x)]
    y = stationarity_trisurf[:,:,1].flatten()
    y = y[~np.isnan(y)]
    z = stationarity_trisurf[:,:,2].flatten()
    z = z[~np.isnan(z)]
    
    fig = plt.figure()
    
    ax = fig.add_subplot(111, projection='3d', proj_type ='ortho')
    ax.view_init(-90,270)
    plot = ax.plot_trisurf(x, y, z, cmap='inferno', shade=False)
    plot.set_facecolors((0, 0, 0, 0))
    ax.set_xlabel(r'Real time $t$ (frames)')
    ax.set_ylabel(r'Lag Time $\Delta t$ (frames)')
    ax.set_xlim(0, 171)
    ax.set_ylim(1, 80)
    ax.set_zticks([])
    m = ScalarMappable(cmap='inferno')
    m.set_array(z)
    cb = plt.colorbar(m)
    cb.set_label(r'MSD($\Delta t$/$t$)')
    
    fig.subplots_adjust(top=0.99,
                        bottom=0.01,
                        left=0.0,
                        right=1.0,
                        hspace=0.065,
                        wspace=0.2)
    
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_2dplots.png')
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_2dplots.pdf')
    
    
    #plot msd(t,lagT) normed in 2D
    
    stationarity_trisurf = np.stack((t_real_mat, lagt_mat,profiles_norm), axis=-1)

    x = stationarity_trisurf[:,:,0].flatten()
    x = x[~np.isnan(x)]
    y = stationarity_trisurf[:,:,1].flatten()
    y = y[~np.isnan(y)]
    z = stationarity_trisurf[:,:,2].flatten()
    z = z[~np.isnan(z)]
    
    fig = plt.figure()
    
    ax = fig.add_subplot(111, projection='3d', proj_type ='ortho')
    ax.view_init(-90,270)
    plot = ax.plot_trisurf(x, y, z, cmap='inferno', shade=False)
    plot.set_facecolors((0, 0, 0, 0))
    ax.set_xlabel(r'Real time $t$ (frames)')
    ax.set_ylabel(r'Lag Time $\Delta t$ (frames)')
    ax.set_xlim(0, 171)
    ax.set_ylim(1, 80)
    ax.set_zticks([])
    m = ScalarMappable(cmap='inferno')
    m.set_array(z)
    cb = plt.colorbar(m)
    cb.set_label(r'MSD($\Delta t$/$t$)')
    
    fig.subplots_adjust(top=0.99,
                        bottom=0.01,
                        left=0.0,
                        right=1.0,
                        hspace=0.065,
                        wspace=0.2)
    
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_normed_2dplots.png')
    fig.savefig(savefolder_stationarity_individual + settings.filename[:-5] + '_stationarity_normed_2dplots.pdf')
    





# stationarity ensemble plot
profiles_list = []
lagt_list = []

for i in tqdm(range(numvid), desc = 'loading stationarity profiles'):
    
    settings = settings_list[i]

    profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
    lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)
    
    profiles_list.append(profiles)
    lagt_list.append(lagtimes)
    
from scipy.signal import savgol_filter

savefolder_stationarity_ensemble = savefolder_stationarity + 'ensemble' + os.sep
os.makedirs(savefolder_stationarity_ensemble, exist_ok=True)

if (lagt_list == lagt_list[0]).all():
    lagtimes_plot = lagt_list[0]

for lagt_plot in tqdm(lagtimes_plot[:-1:3], desc = 'ensemble stationarity analysis'):

    fig, ax = plt.subplots(1, 2, sharey=True, figsize=(16,10))
    
    for i in range(numvid):
        settings = settings_list[i]
        color = colors_bc[i]
        legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
        timescale = timescale_list[i]
        wetdry = wetdry_list[i]
    
        profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
        lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)
        profnorm = profiles / lagtimes[:,None]
        plotprof = profnorm[lagt_plot,:].copy()
        nanprof = plotprof[~np.isnan(plotprof)]
        try:
            smoothprof = savgol_filter(nanprof, 21, 3)
            
            if wetdry == 'wet':
                ax[0].plot(np.arange(1, len(smoothprof) + 1) * timescale, smoothprof, color=color, label=legendstring)
            elif wetdry == 'dry':
                ax[1].plot(np.arange(1, len(smoothprof) + 1) * timescale, smoothprof, color=color, label=legendstring)   
        except:
            continue
    for axis in ax:
        axis.legend()
        axis.set_ylim(bottom=0, top=0.1)
        axis.set_xlim(left=0)
        axis.set_xlabel('Real Time (s)')
        axis.set_ylabel('MSD(lagt = %.2f)' % (lagt_plot * timescale))
        axis.grid(axis='y')
    ax[0].set_title('Liquid Cell')
    ax[1].set_title('Vacuum')
    
    fig.suptitle('LagT = %d frames, %.2f s' %(lagt_plot, lagt_plot * timescale))
    fig.subplots_adjust(top=0.95,
                        bottom=0.055,
                        left=0.045,
                        right=0.99,
                        hspace=0.095,
                        wspace=0.055)
                            
    fig.savefig(savefolder_stationarity_ensemble + '_lagt_%d_plots.png' %lagt_plot)
    fig.savefig(savefolder_stationarity_ensemble + '_lagt_%d_plots.pdf' %lagt_plot)
    
    plt.close('all')
    gc.collect()




# fig, ax = plt.subplots(2, numwet)
# plotwet = 0
# plotdry = 0
# for i in range(numvid):
#     settings = settings_list[i]
#     wetdry = wetdry_list[i]
#     profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
#     lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)
#     profnorm = profiles / lagtimes[:,None]
    
#     if wetdry != 'ignore':
#         if wetdry == 'wet':
#             axis = ax[0, plotwet]
#             plotwet += 1
#         elif wetdry == 'dry':
#             axis = ax[1, plotdry]
#             plotdry += 1
#         axis.imshow(profnorm)
#         axis.set_title(settings.filename[:-5])















savefolder_msd_compare = savefolder_msd + 'EAMSD_vs_TAMSD' + os.sep
os.makedirs(savefolder_msd_compare, exist_ok=True)

savefolder_msd_pickle = savefolder_msd + 'pickle' + os.sep
os.makedirs(savefolder_msd_pickle, exist_ok=True)

try:
    
    enfirst_msd_list = pickle.load(open(savefolder_msd_pickle + 'enfirst_msd_list.p', 'rb'))
    enfirst_msd_err_list = pickle.load(open(savefolder_msd_pickle + 'enfirst_msd_err_list.p', 'rb'))
    timefirst_msd_list = pickle.load(open(savefolder_msd_pickle + 'timefirst_msd_list.p', 'rb'))
    lat_timefirst_msd_list = pickle.load(open(savefolder_msd_pickle + 'lat_timefirst_msd_list.p', 'rb'))
    fit_ens_list = pickle.load(open(savefolder_msd_pickle + 'fit_ens_list.p', 'rb'))
    fit_time_list = pickle.load(open(savefolder_msd_pickle + 'fit_time_list.p', 'rb'))
    fit_time_lat_list = pickle.load(open(savefolder_msd_pickle + 'fit_time_lat_list.p', 'rb'))
    trajlengths_list = pickle.load(open(savefolder_msd_pickle + 'trajlengths_list.p', 'rb'))
    
except:

    enfirst_msd_list = []
    enfirst_msd_err_list = []
    timefirst_msd_list = []
    lat_timefirst_msd_list = []
    fit_ens_list = []
    fit_time_list = []
    fit_time_lat_list = []
    
    for i in tqdm(range(numvid), desc = 'EAMSD / TAMSD Compare'):
        
        atoms_norm = atoms_norm_list[i]
        lattice_norm = lattice_norm_list[i]
        settings = settings_list[i]
        color = colors_bc[i]
        legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
        timescale = timescale_list[i]
    
        profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
        lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)
    
        # compare timefirst averaged msds with ensemble first averaged msds
        
        enfirst_msd = np.nanmean(profiles, axis=1)
        enfirst_msd_err = np.nanstd(profiles, axis=1) / ((~np.isnan(profiles)).sum(axis=1))**0.5
        timefirst_msd = momentcalc(atoms_norm, 1, 1/timescale, 2)
        
        enfirst_msd_reduced = enfirst_msd[enfirst_msd_err > 0]
        enfirst_msd_err_reduced = enfirst_msd_err[enfirst_msd_err > 0]
        lagtimes_reduced = lagtimes[enfirst_msd_err > 0]
    
        popt_ens, pcov_ens = curve_fit(modelfunc, lagtimes_reduced * timescale, enfirst_msd_reduced, sigma=enfirst_msd_err_reduced)
        popt_time, pcov_time = curve_fit(modelfunc, timefirst_msd.lagt.values[:40], timefirst_msd.mmd.values[:40], sigma=timefirst_msd.err.values[:40])
    
        lat_timefirst_msd = momentcalc(lattice_norm, 1, 1/timescale, 2)
        popt_time_lat, pcov_time_lat = curve_fit(modelfunc, lat_timefirst_msd.lagt.values[0:40], lat_timefirst_msd.mmd.values[0:40], sigma=lat_timefirst_msd.err.values[:40])
        
        enfirst_msd_list.append(enfirst_msd)
        enfirst_msd_err_list.append(enfirst_msd_err)
        timefirst_msd_list.append(timefirst_msd)
        lat_timefirst_msd_list.append(lat_timefirst_msd)
        fit_ens_list.append([popt_ens, pcov_ens])
        fit_time_list.append([popt_time, pcov_time])
        fit_time_lat_list.append([popt_time_lat, pcov_time_lat])
    
        fig, ax1 = plt.subplots(1, 1, figsize=(8, 8))
        ax2 = ax1.inset_axes([0.12, 0.63, 0.3, 0.35])
        ax = [ax1, ax2]
        
        time_label = (r'D$_\alpha$=%.3f $\alpha$=%.3f' % (popt_time[1], popt_time[0]))
        ens_label = (r'D$_\alpha$=%.3f $\alpha$=%.3f' % (popt_ens[1], popt_ens[0]))
        lat_label = (r'D$_\alpha$=%.3f $\alpha$=%.3f' % (popt_time_lat[1], popt_time_lat[0]))
        ax[0].errorbar(timefirst_msd.lagt.values, timefirst_msd.mmd.values, yerr=timefirst_msd.err.values, fmt='.', label = 'TAMSD', color='blue', alpha=0.5)
        ax[0].plot(timefirst_msd.lagt.values, modelfunc(timefirst_msd.lagt.values, popt_time[0], popt_time[1]), label=time_label, color='blue', alpha=0.5)
        ax[0].errorbar(lagtimes * timescale, enfirst_msd, yerr=enfirst_msd_err, fmt='.', label = 'EAMSD', color='red', alpha=0.5)
        ax[0].plot(lagtimes * timescale, modelfunc(lagtimes * timescale, popt_ens[0], popt_ens[1]), label=ens_label, color='red', alpha=0.5) 
        ax[0].errorbar(lat_timefirst_msd.lagt.values, lat_timefirst_msd.mmd.values, yerr=lat_timefirst_msd.err.values, fmt='.', label = 'TAMSD Lattice', color='black', alpha=0.5)
        ax[0].plot(lat_timefirst_msd.lagt.values, modelfunc(lat_timefirst_msd.lagt.values, popt_time_lat[0], popt_time_lat[1]), label=lat_label, color='black', alpha=0.5)
        
        ax[0].set_xlim(0,80)
        ax[0].set_ylim(-0.05,1)
        
        ax[1].errorbar(timefirst_msd.lagt.values, timefirst_msd.mmd.values, yerr=timefirst_msd.err.values, fmt='.', label = 'TAMSD', color='blue', alpha=0.5)
        ax[1].plot(timefirst_msd.lagt.values, modelfunc(timefirst_msd.lagt.values, popt_time[0], popt_time[1]), label=time_label, color='blue', alpha=0.5)
        ax[1].errorbar(lagtimes * timescale, enfirst_msd, yerr=enfirst_msd_err, fmt='.', label = 'EAMSD', color='red', alpha=0.5)
        ax[1].plot(lagtimes * timescale, modelfunc(lagtimes * timescale, popt_ens[0], popt_ens[1]), label=ens_label, color='red', alpha=0.5) 
        ax[1].errorbar(lat_timefirst_msd.lagt.values, lat_timefirst_msd.mmd.values, yerr=lat_timefirst_msd.err.values, fmt='.', label = 'TAMSD Lattice', color='black', alpha=0.5)
        ax[1].plot(lat_timefirst_msd.lagt.values, modelfunc(lat_timefirst_msd.lagt.values, popt_time_lat[0], popt_time_lat[1]), label=lat_label, color='black', alpha=0.5)
        
        ax[1].set_xscale('log')
        ax[1].set_yscale('log')
    
        ax[1].set_xlim(2,120)
        ax[1].set_ylim(0.0002,1)
    
        ax[0].legend(loc='center right')
    
        ax[0].set_ylabel(r'MSD (nm$^2$)')
        ax[1].set_ylabel(r'MSD (nm$^2$)')
        ax[0].set_xlabel(r'Lag Time $\Delta$t (s)')
        ax[1].set_xlabel(r'Lag Time $\Delta$t (s)')
    
        ax[0].grid(which='major',axis='both', alpha = 0.1)
        ax[1].grid(which='major',axis='both', alpha = 0.3)
        ax[1].grid(which='minor',axis='both', alpha = 0.1)
        
        fig.suptitle(settings.filename[:-5], color=color)
        fig.subplots_adjust(top=0.95,
                            bottom=0.065,
                            left=0.08,
                            right=0.98,
                            hspace=0.2,
                            wspace=0.2)
    
        fig.savefig(savefolder_msd_compare + settings.filename[:-5] + '_msd_figplot.png')
        fig.savefig(savefolder_msd_compare + settings.filename[:-5] + '_msd_figplot.pdf')
        
        plt.close('all')
        gc.collect()
        
    pickle.dump(enfirst_msd_list, open(savefolder_msd_pickle + 'enfirst_msd_list.p', 'wb'))
    pickle.dump(enfirst_msd_err_list, open(savefolder_msd_pickle + 'enfirst_msd_err_list.p', 'wb'))
    pickle.dump(timefirst_msd_list, open(savefolder_msd_pickle + 'timefirst_msd_list.p', 'wb'))
    pickle.dump(lat_timefirst_msd_list, open(savefolder_msd_pickle + 'lat_timefirst_msd_list.p', 'wb'))
    pickle.dump(fit_ens_list, open(savefolder_msd_pickle + 'fit_ens_list.p', 'wb'))
    pickle.dump(fit_time_list, open(savefolder_msd_pickle + 'fit_time_list.p', 'wb'))
    pickle.dump(fit_time_lat_list, open(savefolder_msd_pickle + 'fit_time_lat_list.p', 'wb'))
        
    
    
    
    
    
savefolder_msd_trajstats = savefolder_msd + 'trajectory_statistics' + os.sep
os.makedirs(savefolder_msd_trajstats, exist_ok=True)

try:
    trajlengths_list = pickle.load(open(savefolder_msd_pickle + 'trajlengths_list.p', 'rb'))

except:
    trajlengths_list = []
    
    for i in tqdm(range(numvid), desc = 'Trajectory Statistics'):
        
        atoms_norm = atoms_norm_list[i]
        lattice_norm = lattice_norm_list[i]
        settings = settings_list[i]
        color = colors_bc[i]
        legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
        timescale = timescale_list[i]
    
        profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
        lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)
    
        # compare timefirst averaged msds with ensemble first averaged msds
        
        enfirst_msd = enfirst_msd_list[i]
        enfirst_msd_err = enfirst_msd_err_list[i]
        timefirst_msd = timefirst_msd_list[i]
    
        popt_ens, pcov_ens = fit_ens_list[i]
        popt_time, pcov_time = fit_time_list[i]
    
        lat_timefirst_msd = lat_timefirst_msd_list[i]
        popt_time_lat, pcov_time_lat = fit_time_lat_list[i]
        
        # timefirst number per lagtime and trajectory length histograms
        
        fig, ax = plt.subplots(2)
        ax[0].plot(timefirst_msd.lagt, timefirst_msd.real_N, label='Number of Trajectories')
        ax[0].plot(timefirst_msd.lagt, timefirst_msd.N, label='Number of Statisically independent Trajectories')
        ax[0].legend()
        ax[0].set_ylabel('Number of Trajectories')
        ax[0].set_xlabel(r'Lag Time $\Delta t$')
    
        trajlengths = []
        for pnum in atoms_norm.particle:
            particle = atoms_norm[atoms_norm.particle == pnum]
            trajlengths.append(len(particle))
        sns.distplot(trajlengths, label='wet', ax=ax[1])
        ax[1].set_ylabel('Number of Trajectories')
        fig.savefig(savefolder + 'trajlength_histograms.png')
        fig.savefig(savefolder + 'trajlength_histograms.pdf')
    
        fig.suptitle(legendstring)
        fig.savefig(savefolder_msd_trajstats + settings.filename[:-5] + '_trajectory_stats.png')
    
        plt.close(fig)
        
        trajlengths_list.append(trajlengths)
        
    
    pickle.dump(trajlengths_list, open(savefolder_msd_pickle + 'trajlengths_list.p', 'wb'))








































figmain_ens, axmain_ens = plt.subplots(1, 2, figsize=(24,14), sharey=True)
ax2_ens = axmain_ens[0].inset_axes([0.05, 0.6, 0.35, 0.35])
ax3_ens = axmain_ens[1].inset_axes([0.05, 0.6, 0.35, 0.35])
axmain_ens = [axmain_ens[0], axmain_ens[1], ax2_ens, ax3_ens]

figmain_time, axmain_time = plt.subplots(1, 2, figsize=(24,14), sharey=True)
ax2_time = axmain_time[0].inset_axes([0.05, 0.6, 0.35, 0.35])
ax3_time = axmain_time[1].inset_axes([0.05, 0.6, 0.35, 0.35])
axmain_time = [axmain_time[0], axmain_time[1], ax2_time, ax3_time]


for i in tqdm(range(numvid), desc = 'Trajectory Statistics'):
    
    atoms_norm = atoms_norm_list[i]
    lattice_norm = lattice_norm_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]
    wetdry = wetdry_list[i]
    plot_times = np.linspace(0, 50, 100) * timescale
    
    if (wetdry != 'ignore') and (settings.filename[:-5] not in ['Scanning Preview17', 'Scanning Preview_12.36', 'Scanning Preview_12.43']):

        profiles = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity.txt')
        lagtimes = np.loadtxt(savefolder_stationarity_calc + settings.filename[:-5] + '_stationarity_lagt.txt').astype(int)
    
        # compare timefirst averaged msds with ensemble first averaged msds
        
        enfirst_msd = enfirst_msd_list[i]
        enfirst_msd_err = enfirst_msd_err_list[i]
        timefirst_msd = timefirst_msd_list[i]
    
        popt_ens, pcov_ens = fit_ens_list[i]
        popt_time, pcov_time = fit_time_list[i]
    
        lat_timefirst_msd = lat_timefirst_msd_list[i]
        popt_time_lat, pcov_time_lat = fit_time_lat_list[i]
        
        time_label = (r'D$_\alpha$=%.3f $\alpha$=%.3f' % (popt_time[1], popt_time[0]))
        ens_label = (r'D$_\alpha$=%.3f $\alpha$=%.3f' % (popt_ens[1], popt_ens[0]))
        
        if wetdry == 'wet':
            axmain_time[0].errorbar(timefirst_msd.lagt.values[:40], timefirst_msd.mmd.values[:40], yerr=timefirst_msd.err.values[:40], fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_time[0].plot(plot_times, modelfunc(plot_times, popt_time[0], popt_time[1]), label=time_label, color=color, alpha=0.5)
            
            axmain_time[2].errorbar(timefirst_msd.lagt.values[:40], timefirst_msd.mmd.values[:40], yerr=timefirst_msd.err.values[:40], fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_time[2].plot(plot_times, modelfunc(plot_times, popt_time[0], popt_time[1]), label=time_label, color=color, alpha=0.5)   
            
            axmain_ens[0].errorbar(lagtimes * timescale, enfirst_msd, yerr=enfirst_msd_err/2, fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_ens[0].plot(plot_times, modelfunc(plot_times, popt_ens[0], popt_ens[1]), label=ens_label, color=color, alpha=0.5) 
    
            axmain_ens[2].errorbar(lagtimes * timescale, enfirst_msd, yerr=enfirst_msd_err/2, fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_ens[2].plot(plot_times, modelfunc(plot_times, popt_ens[0], popt_ens[1]), label=ens_label, color=color, alpha=0.5)
            
        elif wetdry == 'dry':
            axmain_time[1].errorbar(timefirst_msd.lagt.values[:40], timefirst_msd.mmd.values[:40], yerr=timefirst_msd.err.values[:40], fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_time[1].plot(plot_times, modelfunc(plot_times, popt_time[0], popt_time[1]), label=time_label, color=color, alpha=0.5)
            
            axmain_time[3].errorbar(timefirst_msd.lagt.values[:40], timefirst_msd.mmd.values[:40], yerr=timefirst_msd.err.values[:40], fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_time[3].plot(plot_times, modelfunc(plot_times, popt_time[0], popt_time[1]), label=time_label, color=color, alpha=0.5)   
            
            axmain_ens[1].errorbar(lagtimes * timescale, enfirst_msd, yerr=enfirst_msd_err/2, fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_ens[1].plot(plot_times, modelfunc(plot_times, popt_ens[0], popt_ens[1]), label=ens_label, color=color, alpha=0.5) 
    
            axmain_ens[3].errorbar(lagtimes * timescale, enfirst_msd, yerr=enfirst_msd_err/2, fmt='.', label = legendstring, color=color, alpha=0.5)
            axmain_ens[3].plot(plot_times, modelfunc(plot_times, popt_ens[0], popt_ens[1]), label=ens_label, color=color, alpha=0.5)

for ax in axmain_ens[2:] + axmain_time[2:]:
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_xlim(1,80)
    ax.set_ylim(0.01,2)
    ax.grid(which='both', color='grey', alpha=0.5)

for ax in axmain_ens[:2] + axmain_time[:2]:
    ax.set_xlim(0, 80)
    ax.set_ylim(0, 1.6)
    ax.legend(loc='lower right')
    ax.set_ylabel(r'MSD (nm$^2$)')
    ax.set_xlabel(r'Lag Time $\Delta$t (s)')
    ax.grid(which='major', color='grey', alpha=0.5)
    
for ax in [axmain_ens[0]] + [axmain_time[0]]:
    ax.set_title('Liquid Cell')
    
for ax in [axmain_ens[1]] + [axmain_time[1]]:
    ax.set_title('Vacuum')
    
figmain_ens.suptitle('Ensemble FIRST averaged mean square displacement (EAMSD)')
figmain_time.suptitle('Time FIRST averaged mean square displacement (TAMSD)')

for figmain in [figmain_ens] + [figmain_time]:
    figmain.subplots_adjust(top=0.955,
                            bottom=0.05,
                            left=0.03,
                            right=0.99,
                            hspace=0.2,
                            wspace=0.075)

figmain_ens.savefig(savefolder_msd + 'EAMSD_compare.png')
figmain_ens.savefig(savefolder_msd + 'EAMSD_compare.pdf')

figmain_time.savefig(savefolder_msd + 'TAMSD_compare.png')
figmain_time.savefig(savefolder_msd + 'TAMSD_compare.pdf')

plt.close('all')
gc.collect()








fig1_time, ax1_time = plt.subplots(1,2)
#fig2_time, ax2_time = plt.subplots(1)
fig1_ens, ax1_ens = plt.subplots(1,2)
#fig2_ens, ax2_ens = plt.subplots(1)
plt.get_cmap('inferno')([0, 0.5, 1])
wet_params_time = []
dry_params_time = []
wet_params_ens = []
dry_params_ens = []
wet_color_list = []
dry_color_list = []
bc_max, bc_min = np.max(beam_current_list), np.min(beam_current_list)
dummyvar = []
for i in range(numvid):
    wetdry = wetdry_list[i]
    settings = settings_list[i]
    if (wetdry != 'ignore') and (settings.filename[:-5] not in ['Scanning Preview17', 'Scanning Preview_12.36', 'Scanning Preview_12.43']):
        popt_time, pcov_time = fit_time_list[i]
        err_time = np.sqrt(np.diag(pcov_time))
        popt_ens, pcov_ens = fit_ens_list[i]
        err_ens = np.sqrt(np.diag(pcov_ens))
        if wetdry == 'wet':
            wet_params_time.append([popt_time, err_time])
            wet_params_ens.append([popt_ens, err_ens])
            wet_color_list.append([beam_current_list[i], dummyvar])
            print(popt_time, settings.filename[:-5], i)
        elif wetdry == 'dry':
            dry_params_time.append([popt_time, err_time])   
            dry_params_ens.append([popt_ens, err_ens])
            dry_color_list.append([beam_current_list[i], dummyvar])
wet_params_time_arr = np.array(wet_params_time)
wet_params_ens_arr = np.array(wet_params_ens)
dry_params_time_arr = np.array(dry_params_time)
dry_params_ens_arr = np.array(dry_params_ens)
dry_color_list_arr = np.array(dry_color_list)
wet_color_list_arr = np.array(wet_color_list)
for bc in np.unique(beam_current_list):
    wet_params_time = wet_params_time_arr[wet_color_list_arr[:,0] == bc]
    dry_params_time = dry_params_time_arr[dry_color_list_arr[:,0] == bc]
    wet_params_ens = wet_params_ens_arr[wet_color_list_arr[:,0] == bc]
    dry_params_ens = dry_params_ens_arr[dry_color_list_arr[:,0] == bc]
    bc_norm = (bc - bc_min) / (bc_max - bc_min)
    color_all = plt.get_cmap('cool')(bc_norm)
    ax1_time[0].errorbar(wet_params_time[:, 0, 0], wet_params_time[:, 0, 1], yerr=wet_params_time[:, 1, 1], xerr=wet_params_time[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
    ax1_time[1].errorbar(dry_params_time[:, 0, 0], dry_params_time[:, 0, 1], yerr=dry_params_time[:, 1, 1], xerr=dry_params_time[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
    ax1_ens[0].errorbar(wet_params_ens[:, 0, 0], wet_params_ens[:, 0, 1], yerr=wet_params_ens[:, 1, 1], xerr=wet_params_ens[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
    ax1_ens[1].errorbar(dry_params_ens[:, 0, 0], dry_params_ens[:, 0, 1], yerr=dry_params_ens[:, 1, 1], xerr=dry_params_ens[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
for ax in [ax1_time[0], ax1_time[1], ax1_ens[0], ax1_ens[1]]:
    ax.set_xlim(0.5, 1)
    ax.set_ylim(0, 0.1)
    ax.legend()
    ax.grid()
    ax.set_ylabel(r'Anomalous Diffusion Coefficient D$_/alpha$')
    ax.set_xlabel(r'Anomalous Diffusion Exponent $/alpha$')
for ax in ax1_time:
    ax.set_xlim(0.7, 1.0)
    ax.set_ylim(0, 0.06)
ax1_time[0].set_title('Liquid Cell')
ax1_time[1].set_title('Vacuum')
ax1_ens[0].set_title('Liquid Cell')
ax1_ens[1].set_title('Vacuum')
fig1_time.suptitle('TAMSD Fit Parameters')
fig1_ens.suptitle('EAMSD Fit Parameters')
for fig in [fig1_ens, fig1_time]:
    fig.subplots_adjust(top=0.89,
                        bottom=0.11,
                        left=0.11,
                        right=0.975,
                        hspace=0.2,
                        wspace=0.33)
fig1_ens.savefig(savefolder_msd + 'EAMSD_fitparams_compare.png')
fig1_time.savefig(savefolder_msd + 'TAMSD_fitparams_compare.png')
fig1_ens.savefig(savefolder_msd + 'EAMSD_fitparams_compare.pdf')
fig1_time.savefig(savefolder_msd + 'TAMSD_fitparams_compare.pdf')








savefolder_msd_distcomp = savefolder_msd + 'distance_compare' + os.sep
os.makedirs(savefolder_msd_distcomp, exist_ok=True)

timevals = np.linspace(10, 50, 5)
for timeval in timevals:
    fig_ens, ax_ens = plt.subplots(1, 2)
    fig_time, ax_time = plt.subplots(1, 2)
    for i in range(numvid):
        wetdry = wetdry_list[i]
        settings = settings_list[i]
        if (wetdry != 'ignore') and (settings.filename[:-5] not in ['Scanning Preview17', 'Scanning Preview_12.36', 'Scanning Preview_12.43']):
            popt_time, pcov_time = fit_time_list[i]
            err_time = np.sqrt(np.diag(pcov_time))
            popt_ens, pcov_ens = fit_ens_list[i]
            err_ens = np.sqrt(np.diag(pcov_ens))
            val_ens = popt_ens[1] * (timeval ** popt_ens[0])
            val_time = popt_time[1] * (timeval ** popt_time[0])
            if wetdry=='wet':
                ax_ens[0].scatter(beam_current_list[i] * 1e12, val_ens, color='blue')
                ax_ens[1].scatter(dosescale_list[i], val_ens, color='blue')
                ax_time[0].scatter(beam_current_list[i] * 1e12, val_time, color='blue')
                ax_time[1].scatter(dosescale_list[i], val_time, color='blue')
            elif wetdry=='dry':
                ax_ens[0].scatter(beam_current_list[i] * 1e12, val_ens, color='red')
                ax_ens[1].scatter(dosescale_list[i], val_ens, color='red')
                ax_time[0].scatter(beam_current_list[i] * 1e12, val_time, color='red')
                ax_time[1].scatter(dosescale_list[i], val_time, color='red')
    ax_ens[0].set_ylabel('MSD (%d s)'%timeval)
    ax_time[0].set_ylabel('MSD (%d s)'%timeval)
    ax_ens[0].set_xlabel('Beam Current (pA)')
    ax_ens[1].set_xlabel('Fluence (10^6 e per nm^2 per frame')
    ax_time[0].set_xlabel('Beam Current (pA)')
    ax_time[1].set_xlabel('Fluence (10^6 e per nm^2 per frame')
    fig_ens.suptitle('EAMSD %d s'%timeval)
    fig_time.suptitle('TAMSD %d s'%timeval)
    fig_ens.savefig(savefolder_msd_distcomp + 'EAMSD_%d_s.png'%timeval)
    fig_time.savefig(savefolder_msd_distcomp + 'TAMSD_%d_s.png'%timeval)














# MSD figureplots

lagtimes = range(1, 51)
msd_exclude_list = ['Scanning Preview17',]
bc_vals = np.sort(wet_atoms_norm_all.beam_current.unique())

for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(msd_exclude_list))]
    wet_lat_df = wet_lattice_norm_all[(wet_lattice_norm_all['beam_current'] == bc_val) & ~(wet_lattice_norm_all['video'].isin(msd_exclude_list))]
    wet_profilename = 'wet_bc_%d'%(bc_val*1e12)
    if not os.path.exists(savefolder_stationarity_calc + wet_profilename + '_stationarity.txt'):
        profiles = np.full((len(lagtimes), wet_df.frame.max()), np.nan, dtype='float64')
        for lagt in tqdm(lagtimes):
            profile = moment_stationarity_calc(wet_df, 2, lagt)
            profiles[lagt-1][0:len(profile)] = profile
        np.savetxt((savefolder_stationarity_calc + wet_profilename + '_stationarity.txt'), profiles)
        np.savetxt((savefolder_stationarity_calc + wet_profilename + '_stationarity_lagt.txt'), lagtimes)
 
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(msd_exclude_list))]
    dry_lat_df = dry_lattice_norm_all[(dry_lattice_norm_all['beam_current'] == bc_val) & ~(dry_lattice_norm_all['video'].isin(msd_exclude_list))]
    dry_profilename = 'dry_bc_%d'%(bc_val*1e12)
    if not os.path.exists(savefolder_stationarity_calc + dry_profilename + '_stationarity.txt'):
        profiles = np.full((len(lagtimes), dry_df.frame.max()), np.nan, dtype='float64')
        for lagt in tqdm(lagtimes):
            profile = moment_stationarity_calc(dry_df, 2, lagt)
            profiles[lagt-1][0:len(profile)] = profile
        np.savetxt((savefolder_stationarity_calc + dry_profilename + '_stationarity.txt'), profiles)
        np.savetxt((savefolder_stationarity_calc + dry_profilename + '_stationarity_lagt.txt'), lagtimes)

    # if not os.path.exists(savefolder_stationarity_calc + wet_profilename + '_lattice_stationarity.txt'):
    #     profiles = np.full((len(lagtimes), wet_lat_df.frame.max()), np.nan, dtype='float64')
    #     for lagt in tqdm(lagtimes):
    #         profile = moment_stationarity_calc(wet_lat_df, 2, lagt)
    #         profiles[lagt-1][0:len(profile)] = profile
    #     np.savetxt((savefolder_stationarity_calc + wet_profilename + '_lattice_stationarity.txt'), profiles)
    #     np.savetxt((savefolder_stationarity_calc + wet_profilename + '_lattice_stationarity_lagt.txt'), lagtimes)

    # if not os.path.exists(savefolder_stationarity_calc + dry_profilename + '_lattice_stationarity.txt'):
    #     profiles = np.full((len(lagtimes), dry_lat_df.frame.max()), np.nan, dtype='float64')
    #     for lagt in tqdm(lagtimes):
    #         profile = moment_stationarity_calc(dry_lat_df, 2, lagt)
    #         profiles[lagt-1][0:len(profile)] = profile
    #     np.savetxt((savefolder_stationarity_calc + dry_profilename + '_lattice_stationarity.txt'), profiles)
    #     np.savetxt((savefolder_stationarity_calc + dry_profilename + '_lattice_stationarity_lagt.txt'), lagtimes)










cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
dry_cmap = plt.get_cmap('Reds')
wet_cmap = plt.get_cmap('Blues')

fig_ens, ax_ens = plt.subplots(1)
ax_ens_ins = ax_ens.inset_axes([0.05, 0.6, 0.35, 0.35])
fig_time, ax_time = plt.subplots(1)
ax_time_ins = ax_time.inset_axes([0.05, 0.6, 0.35, 0.35])

tvals, tcounts = np.unique(timescale_list, return_counts=True)
usetime = tvals[tcounts.argmax()]

for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    wet_color = wet_cmap(cm_vals[i])
    dry_color = dry_cmap(cm_vals[i])
    
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(msd_exclude_list))]
    wet_lat_df = wet_lattice_norm_all[(wet_lattice_norm_all['beam_current'] == bc_val) & ~(wet_lattice_norm_all['video'].isin(msd_exclude_list))]
    wet_profilename = 'wet_bc_%d'%(bc_val*1e12)
    
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(msd_exclude_list))]
    dry_lat_df = dry_lattice_norm_all[(dry_lattice_norm_all['beam_current'] == bc_val) & ~(dry_lattice_norm_all['video'].isin(msd_exclude_list))]
    dry_profilename = 'dry_bc_%d'%(bc_val*1e12)
    
    wet_profiles = np.loadtxt(savefolder_stationarity_calc + wet_profilename + '_stationarity.txt')
    enfirst_msd_wet = np.nanmean(wet_profiles, axis=1)
    enfirst_msd_wet_err = np.nanstd(wet_profiles, axis=1) / ((~np.isnan(wet_profiles)).sum(axis=1))**0.5

    timefirst_msd_wet = momentcalc(wet_df, 1, 1/usetime, 2)

    popt_ens_wet, pcov_ens_wet = curve_fit(modelfunc, lagtimes * usetime, enfirst_msd_wet, sigma=enfirst_msd_wet_err)
    popt_time_wet, pcov_time_wet = curve_fit(modelfunc, timefirst_msd_wet.lagt.values[:40], timefirst_msd_wet.mmd.values[:40], sigma=timefirst_msd_wet.err.values[:40])

    dry_profiles = np.loadtxt(savefolder_stationarity_calc + dry_profilename + '_stationarity.txt')
    enfirst_msd_dry = np.nanmean(dry_profiles, axis=1)
    enfirst_msd_dry_err = np.nanstd(dry_profiles, axis=1) / ((~np.isnan(dry_profiles)).sum(axis=1))**0.5

    timefirst_msd_dry = momentcalc(dry_df, 1, 1/usetime, 2)

    popt_ens_dry, pcov_ens_dry = curve_fit(modelfunc, lagtimes * usetime, enfirst_msd_dry, sigma=enfirst_msd_dry_err)
    popt_time_dry, pcov_time_dry = curve_fit(modelfunc, timefirst_msd_dry.lagt.values[:40], timefirst_msd_dry.mmd.values[:40], sigma=timefirst_msd_dry.err.values[:40])

    ax_time.errorbar(timefirst_msd_wet.lagt, timefirst_msd_wet.mmd.values, yerr=timefirst_msd_wet.err, fmt='.', color=wet_color, label=r'Liquid %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_time_wet[1], popt_time_wet[0]))
    ax_time.plot(timefirst_msd_wet.lagt, modelfunc(timefirst_msd_wet.lagt, popt_time_wet[0], popt_time_wet[1]), color=wet_color, label='_nolegend_', alpha=0.5)
    ax_ens.errorbar(lagtimes * usetime, enfirst_msd_wet, yerr=enfirst_msd_wet_err, fmt='.', color=wet_color, label=r'Liquid %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_ens_wet[1], popt_ens_wet[0]))
    ax_ens.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_wet[0], popt_ens_wet[1]), color=wet_color, label='_nolegend_', alpha=0.5)

    ax_time_ins.errorbar(timefirst_msd_wet.lagt, timefirst_msd_wet.mmd.values, yerr=timefirst_msd_wet.err, fmt='.', color=wet_color)
    ax_time_ins.plot(timefirst_msd_wet.lagt, modelfunc(timefirst_msd_wet.lagt, popt_time_wet[0], popt_time_wet[1]), color=wet_color, alpha=0.5)
    ax_ens_ins.errorbar(lagtimes * usetime, enfirst_msd_wet, yerr=enfirst_msd_wet_err, fmt='.', color=wet_color)
    ax_ens_ins.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_wet[0], popt_ens_wet[1]), color=wet_color, alpha=0.5)

    if bc_val != bc_vals[0]:
        ax_time.errorbar(timefirst_msd_dry.lagt, timefirst_msd_dry.mmd.values, yerr=timefirst_msd_dry.err, fmt='.', color=dry_color, label=r'Vacuum %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_time_dry[1], popt_time_dry[0]))
        ax_time.plot(timefirst_msd_dry.lagt, modelfunc(timefirst_msd_dry.lagt, popt_time_dry[0], popt_time_dry[1]), color=dry_color, label='_nolegend_', alpha=0.5)
        ax_ens.errorbar(lagtimes * usetime, enfirst_msd_dry, yerr=enfirst_msd_dry_err, fmt='.', color=dry_color, label=r'Vacuum %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_ens_dry[1], popt_ens_dry[0]))
        ax_ens.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_dry[0], popt_ens_dry[1]), color=dry_color, label='_nolegend_', alpha=0.5)
    
        ax_time_ins.errorbar(timefirst_msd_dry.lagt, timefirst_msd_dry.mmd.values, yerr=timefirst_msd_dry.err, fmt='.', color=dry_color)
        ax_time_ins.plot(timefirst_msd_dry.lagt, modelfunc(timefirst_msd_dry.lagt, popt_time_dry[0], popt_time_dry[1]), color=dry_color, alpha=0.5)
        ax_ens_ins.errorbar(lagtimes * usetime, enfirst_msd_dry, yerr=enfirst_msd_dry_err, fmt='.', color=dry_color)
        ax_ens_ins.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_dry[0], popt_ens_dry[1]), color=dry_color, alpha=0.5)

for a in [ax_ens_ins, ax_time_ins]:
    a.set_xscale('log')
    a.set_yscale('log')
    a.set_xlim(2,100)
    a.grid(which='both')
    a.yaxis.tick_right()
for a in [ax_ens, ax_time]:
    a.set_xlim(0, 80)
    a.set_ylim(0,1.4)
    a.legend(loc='lower right')
fig_ens.savefig(savefolder_figs + 'eamsd.png')
fig_ens.savefig(savefolder_figs + 'eamsd.pdf')
fig_ens.savefig(savefolder_figs + 'eamsd.svg')
fig_time.savefig(savefolder_figs + 'tamsd.png')
fig_time.savefig(savefolder_figs + 'tamsd.pdf')
fig_time.savefig(savefolder_figs + 'tamsd.svg')


















cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
dry_cmap = plt.get_cmap('Reds')
wet_cmap = plt.get_cmap('Blues')

fig, (ax_ens, ax_time) = plt.subplots(1, 2, figsize=(7.08661,5))
ax_ens_ins = ax_ens.inset_axes([0.05, 0.6, 0.35, 0.35])
ax_time_ins = ax_time.inset_axes([0.05, 0.6, 0.35, 0.35])
fig.subplots_adjust(top=0.95,
                    bottom=0.11,
                    left=0.09,
                    right=0.91,
                    hspace=0.2,
                    wspace=0.075)

tvals, tcounts = np.unique(timescale_list, return_counts=True)
usetime = tvals[tcounts.argmax()]

for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    wet_color = wet_cmap(cm_vals[i])
    dry_color = dry_cmap(cm_vals[i])
    
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(msd_exclude_list))]
    wet_lat_df = wet_lattice_norm_all[(wet_lattice_norm_all['beam_current'] == bc_val) & ~(wet_lattice_norm_all['video'].isin(msd_exclude_list))]
    wet_profilename = 'wet_bc_%d'%(bc_val*1e12)
    
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(msd_exclude_list))]
    dry_lat_df = dry_lattice_norm_all[(dry_lattice_norm_all['beam_current'] == bc_val) & ~(dry_lattice_norm_all['video'].isin(msd_exclude_list))]
    dry_profilename = 'dry_bc_%d'%(bc_val*1e12)
    
    wet_profiles = np.loadtxt(savefolder_stationarity_calc + wet_profilename + '_stationarity.txt')
    enfirst_msd_wet = np.nanmean(wet_profiles, axis=1)
    enfirst_msd_wet_err = np.nanstd(wet_profiles, axis=1) / ((~np.isnan(wet_profiles)).sum(axis=1))**0.5

    timefirst_msd_wet = momentcalc(wet_df, 1, 1/usetime, 2)

    popt_ens_wet, pcov_ens_wet = curve_fit(modelfunc, lagtimes * usetime, enfirst_msd_wet, sigma=enfirst_msd_wet_err)
    popt_time_wet, pcov_time_wet = curve_fit(modelfunc, timefirst_msd_wet.lagt.values[:40], timefirst_msd_wet.mmd.values[:40], sigma=timefirst_msd_wet.err.values[:40])

    dry_profiles = np.loadtxt(savefolder_stationarity_calc + dry_profilename + '_stationarity.txt')
    enfirst_msd_dry = np.nanmean(dry_profiles, axis=1)
    enfirst_msd_dry_err = np.nanstd(dry_profiles, axis=1) / ((~np.isnan(dry_profiles)).sum(axis=1))**0.5

    timefirst_msd_dry = momentcalc(dry_df, 1, 1/usetime, 2)

    popt_ens_dry, pcov_ens_dry = curve_fit(modelfunc, lagtimes * usetime, enfirst_msd_dry, sigma=enfirst_msd_dry_err)
    popt_time_dry, pcov_time_dry = curve_fit(modelfunc, timefirst_msd_dry.lagt.values[:40], timefirst_msd_dry.mmd.values[:40], sigma=timefirst_msd_dry.err.values[:40])

    ax_time.errorbar(timefirst_msd_wet.lagt, timefirst_msd_wet.mmd.values, yerr=timefirst_msd_wet.err, fmt='.', color=wet_color, label=r'Liquid %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_time_wet[1], popt_time_wet[0]))
    ax_time.plot(timefirst_msd_wet.lagt, modelfunc(timefirst_msd_wet.lagt, popt_time_wet[0], popt_time_wet[1]), color=wet_color, label='_nolegend_', alpha=0.5)
    ax_ens.errorbar(lagtimes * usetime, enfirst_msd_wet, yerr=enfirst_msd_wet_err, fmt='.', color=wet_color, label=r'Liquid %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_ens_wet[1], popt_ens_wet[0]))
    ax_ens.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_wet[0], popt_ens_wet[1]), color=wet_color, label='_nolegend_', alpha=0.5)

    ax_time_ins.errorbar(timefirst_msd_wet.lagt, timefirst_msd_wet.mmd.values, yerr=timefirst_msd_wet.err, fmt='.', color=wet_color)
    ax_time_ins.plot(timefirst_msd_wet.lagt, modelfunc(timefirst_msd_wet.lagt, popt_time_wet[0], popt_time_wet[1]), color=wet_color, alpha=0.5)
    ax_ens_ins.errorbar(lagtimes * usetime, enfirst_msd_wet, yerr=enfirst_msd_wet_err, fmt='.', color=wet_color)
    ax_ens_ins.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_wet[0], popt_ens_wet[1]), color=wet_color, alpha=0.5)

    if bc_val != bc_vals[0]:
        ax_time.errorbar(timefirst_msd_dry.lagt, timefirst_msd_dry.mmd.values, yerr=timefirst_msd_dry.err, fmt='.', color=dry_color, label=r'Vacuum %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_time_dry[1], popt_time_dry[0]))
        ax_time.plot(timefirst_msd_dry.lagt, modelfunc(timefirst_msd_dry.lagt, popt_time_dry[0], popt_time_dry[1]), color=dry_color, label='_nolegend_', alpha=0.5)
        ax_ens.errorbar(lagtimes * usetime, enfirst_msd_dry, yerr=enfirst_msd_dry_err, fmt='.', color=dry_color, label=r'Vacuum %d pA D$_\alpha$ = %.3f, $\alpha$ = %.2f' %(bc_val * 1e12, popt_ens_dry[1], popt_ens_dry[0]))
        ax_ens.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_dry[0], popt_ens_dry[1]), color=dry_color, label='_nolegend_', alpha=0.5)
    
        ax_time_ins.errorbar(timefirst_msd_dry.lagt, timefirst_msd_dry.mmd.values, yerr=timefirst_msd_dry.err, fmt='.', color=dry_color)
        ax_time_ins.plot(timefirst_msd_dry.lagt, modelfunc(timefirst_msd_dry.lagt, popt_time_dry[0], popt_time_dry[1]), color=dry_color, alpha=0.5)
        ax_ens_ins.errorbar(lagtimes * usetime, enfirst_msd_dry, yerr=enfirst_msd_dry_err, fmt='.', color=dry_color)
        ax_ens_ins.plot(lagtimes * usetime, modelfunc(lagtimes * usetime, popt_ens_dry[0], popt_ens_dry[1]), color=dry_color, alpha=0.5)

for a in [ax_ens_ins, ax_time_ins]:
    a.set_xscale('log')
    a.set_yscale('log')
    a.set_xlim(2,100)
    a.set_ylim(0.02, 3)
    a.grid(which='both')
    a.yaxis.tick_right()
for a in [ax_ens, ax_time]:
    a.set_xlim(0, 80)
    a.set_ylim(0,1.4)
    a.legend(loc='lower right')
    a.set_xlabel(r'Lag Time $\Delta$T (s)')
    a.set_ylabel(r'MSD (nm$^2$)')
    

ax_ens.set_title(r'$\bf{a}$ EAMSD', loc='left')
ax_time.set_title(r'$\bf{b}$ TAMSD', loc='left')
ax_time.yaxis.tick_right()
ax_time.yaxis.set_label_position('right')
    
fig.savefig(savefolder_figs + 'eamsd_tamsd_compare.png')
fig.savefig(savefolder_figs + 'eamsd_tamsd_compare.pdf')
fig.savefig(savefolder_figs + 'eamsd_tamsd_compare.svg')

plt.close('all')
gc.collect()



fig1_time = plt.figure(figsize=(10,10))
ax1_time1 = plt.subplot2grid((3, 4), (0,0), rowspan=2, colspan=2)
ax1_time2 = plt.subplot2grid((3, 4), (0,2), rowspan=2, colspan=2)
ax1_msd = plt.subplot2grid((3,4), (2,0), rowspan=1, colspan=4)
ax1_time = [ax1_time1, ax1_time2]
figtime = 10
inset_params = [0.6, 0.6, 0.38, 0.38]
plt.get_cmap('inferno')([0, 0.5, 1])
wet_params_time = []
dry_params_time = []
wet_params_ens = []
dry_params_ens = []
wet_color_list = []
dry_color_list = []
bc_max, bc_min = np.max(beam_current_list), np.min(beam_current_list)
dummyvar = []
msdval_list = []
for i in range(numvid):
    wetdry = wetdry_list[i]
    settings = settings_list[i]
    if (wetdry != 'ignore') and (settings.filename[:-5] not in ['Scanning Preview17', 'Scanning Preview_12.36', 'Scanning Preview_12.43']):
        popt_time, pcov_time = fit_time_list[i]
        err_time = np.sqrt(np.diag(pcov_time))
        popt_ens, pcov_ens = fit_ens_list[i]
        err_ens = np.sqrt(np.diag(pcov_ens))
        
        exponent = figtime**popt_time[0]
        exponent_err = (exponent**2)*((np.log(figtime)*err_time[0])**2)
        exponent_ferr2 = (exponent_err / exponent)**2
        coeff_ferr2 = (err_time[1] / popt_time[1])
        msdval = exponent * popt_time[1]
        msdval_ferr = (exponent_ferr2 + coeff_ferr2)**0.5
        msdval_err = msdval_ferr * msdval
        
        bc = beam_current_list[i]
        msdval_series = pd.DataFrame()
        msdval_series['msdval'] = [msdval]
        msdval_series['err'] = [msdval_err]
        msdval_series['bc'] = [bc]
        msdval_series['wetdry'] = [wetdry]
        msdval_list.append(msdval_series)
            
        if wetdry == 'wet':
            wet_params_time.append([popt_time, err_time])
            wet_params_ens.append([popt_ens, err_ens])
            wet_color_list.append([beam_current_list[i], dummyvar])
            print(popt_time, settings.filename[:-5], i)
        elif wetdry == 'dry':
            dry_params_time.append([popt_time, err_time])   
            dry_params_ens.append([popt_ens, err_ens])
            dry_color_list.append([beam_current_list[i], dummyvar])
            
msdval_df = pd.concat(msdval_list)
wet_params_time_arr = np.array(wet_params_time)
wet_params_ens_arr = np.array(wet_params_ens)
dry_params_time_arr = np.array(dry_params_time)
dry_params_ens_arr = np.array(dry_params_ens)
dry_color_list_arr = np.array(dry_color_list)
wet_color_list_arr = np.array(wet_color_list)
for bc in np.unique(beam_current_list):
    wet_params_time = wet_params_time_arr[wet_color_list_arr[:,0] == bc]
    dry_params_time = dry_params_time_arr[dry_color_list_arr[:,0] == bc]
    wet_params_ens = wet_params_ens_arr[wet_color_list_arr[:,0] == bc]
    dry_params_ens = dry_params_ens_arr[dry_color_list_arr[:,0] == bc]
    bc_norm = (bc - bc_min) / (bc_max - bc_min)
    color_all = plt.get_cmap('cool')(bc_norm)
    ax1_time[0].errorbar(wet_params_time[:, 0, 0], wet_params_time[:, 0, 1], yerr=wet_params_time[:, 1, 1], xerr=wet_params_time[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
    ax1_time[1].errorbar(dry_params_time[:, 0, 0], dry_params_time[:, 0, 1], yerr=dry_params_time[:, 1, 1], xerr=dry_params_time[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
#     ax1_ens[0].errorbar(wet_params_ens[:, 0, 0], wet_params_ens[:, 0, 1], yerr=wet_params_ens[:, 1, 1], xerr=wet_params_ens[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))
#     ax1_ens[1].errorbar(dry_params_ens[:, 0, 0], dry_params_ens[:, 0, 1], yerr=dry_params_ens[:, 1, 1], xerr=dry_params_ens[:, 1, 0], fmt='.', color=color_all, label='%d pA' %(bc * 1e12))

spacing = 2
coldict = {'wet':'blue', 'dry':'red'}
for bc in msdval_df.bc.unique():
    msdval_bc = msdval_df[msdval_df.bc == bc].sort_values('wetdry', ascending=False)
    num = len(msdval_bc)
    bc_pA = bc * 1e12
    xvals = np.linspace(bc_pA, bc_pA + num*spacing, num) - num*spacing/2
    colorlist = [coldict[wetdry] for wetdry in msdval_bc.wetdry]
    ax1_msd.errorbar(xvals, msdval_bc.msdval, yerr = msdval_bc.err, fmt='none', ecolor=colorlist)
    ax1_msd.scatter(xvals, msdval_bc.msdval, color=colorlist)
    
legend_elements = [Line2D([0], [0], color='b', label='Liquid Environment'),
                   Line2D([0], [0], color='r', label='Vacuum Environment')]
ax1_msd.legend(handles=legend_elements, loc='upper right')

for ax in ax1_time:
    ax.set_xlim(0.7, 1)
    ax.set_ylim(0, 0.06)
    ax.legend()
    ax.grid()
    ax.set_ylabel(r'Anomalous Diffusion Coefficient D$_\alpha$')
    ax.set_xlabel(r'Anomalous Diffusion Exponent $\alpha$')

ax1_msd.set_ylim(0, 0.4)
ax1_msd.set_xticks(msdval_df.bc.unique() * 1e12)
ax1_msd.grid(axis='y')
ax1_msd.set_xlabel('Beam Current (pA)')
ax1_msd.set_ylabel(r'TAMSD($\delta$T = %ds) (nm$^2$)'%figtime)
    
ax1_time[0].set_title(r'$\bf{a}$ Liquid Cell', loc='left')
ax1_time[1].set_title(r'$\bf{b}$ Vacuum', loc='left')
ax1_time[1].yaxis.tick_right()
ax1_time[1].yaxis.set_label_position('right')
ax1_msd.set_title(r'$\bf{c}$ Displacement at %ds'%figtime, loc='left')

fig1_time.subplots_adjust(top=0.975,
                        bottom=0.055,
                        left=0.07,
                        right=0.935,
                        hspace=0.37,
                        wspace=0.24)
fig1_time.savefig(savefolder_figs + 'TAMSD_fitparams_compare.png')
fig1_time.savefig(savefolder_figs + 'TAMSD_fitparams_compare.pdf')
fig1_time.savefig(savefolder_figs + 'TAMSD_fitparams_compare.svg')




cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
dry_cmap = plt.get_cmap('Reds')
wet_cmap = plt.get_cmap('Blues')

fig, ax = plt.subplots(5, 2, figsize=(8,14))

tvals, tcounts = np.unique(timescale_list, return_counts=True)
usetime = tvals[tcounts.argmax()]

bins = np.linspace(-1, 1, 33)

for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    wet_color = wet_cmap(cm_vals[i])
    dry_color = dry_cmap(cm_vals[i])
    
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(msd_exclude_list))]
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(msd_exclude_list))]
    
    wet_vh_pos_y = wet_df.set_index(['frame', 'particle'])['y'].unstack()
    dry_vh_pos_y = dry_df.set_index(['frame', 'particle'])['y'].unstack()
    
    wet_vh_pos_x = wet_df.set_index(['frame', 'particle'])['x'].unstack()
    dry_vh_pos_x = dry_df.set_index(['frame', 'particle'])['x'].unstack()
    
    for i in range(5):
        lagt = [1, 2, 3, 4, 5][i]
        
        wet_vh = tp.vanhove(wet_vh_pos_x, lagt, bins=bins)
        wet_vh_sum = wet_vh.sum(axis=1)
        wet_vh_norm = wet_vh_sum / wet_vh_sum.sum()
        wet_vh_norm.plot(ax=ax[i][0], color=wet_color, alpha=0.5)
        
        dry_vh = tp.vanhove(dry_vh_pos_x, lagt, bins=bins)
        dry_vh_sum = dry_vh.sum(axis=1)
        dry_vh_norm = dry_vh_sum / dry_vh_sum.sum()
        dry_vh_norm.plot(ax=ax[i][0], color=dry_color, alpha=0.5)

        wet_vh = tp.vanhove(wet_vh_pos_y, lagt, bins=bins)
        wet_vh_sum = wet_vh.sum(axis=1)
        wet_vh_norm = wet_vh_sum / wet_vh_sum.sum()
        wet_vh_norm.plot(ax=ax[i][1], color=wet_color, alpha=0.5)
        
        dry_vh = tp.vanhove(dry_vh_pos_y, lagt, bins=bins)
        dry_vh_sum = dry_vh.sum(axis=1)
        dry_vh_norm = dry_vh_sum / dry_vh_sum.sum()
        dry_vh_norm.plot(ax=ax[i][1], color=dry_color, alpha=0.5)
        
    for ax_arr in ax:
        for axis in ax_arr:
            axis.set_xlim(-1, 1)
            axis.set_ylim(0, 0.5)
            axis.axvline(0, color='black')
            
fig.savefig(savefolder_msd + 'vhs.png')





























# imsd analysis

savefolder_imsd = savefolder_msd + 'imsd' + os.sep
os.makedirs(savefolder_imsd, exist_ok=True)

savefolder_imsd_individual = savefolder_imsd + 'individual' + os.sep
os.makedirs(savefolder_imsd_individual, exist_ok=True)

savefolder_imsd_outliers = savefolder_imsd + 'outlier_maps' + os.sep
os.makedirs(savefolder_imsd_outliers, exist_ok=True)

try:
    
    imsd_list = pickle.load(open(savefolder_msd_pickle + 'imsd_list.p', 'rb'))
    fitparams_df_list = pickle.load(open(savefolder_msd_pickle + 'imsd_fitparams_df_list.p', 'rb'))
    
except:

    imsd_list = []
    fitparams_df_list = []
    
    for i in tqdm(range(numvid), desc = 'IMSD Analysis'):
        
        vid=vid_list[i]
        
        atoms_norm = atoms_norm_list[i]
        atoms = atoms_list[i]
        lattice_norm = lattice_norm_list[i]
        settings = settings_list[i]
        color = colors_bc[i]
        legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
        timescale = timescale_list[i]
    
    
        #distribution of indivudual TAMSD power law fit params
        
        imsd = tp.imsd(atoms_norm, 1, 1/timescale, max_lagtime=160)
        imsd_list.append(imsd)
        timefirst_msd = timefirst_msd_list[i]
        popt_time, pcov_time = fit_time_list[i]
        
        fig = plt.figure(figsize=(16,14), dpi=100)
        ax_imsd = plt.subplot2grid((3,4), (0,0), rowspan=2, colspan=4, fig=fig)
        
        fitparams_list = []
        for pnum in imsd.columns:
            track = imsd[pnum]
            track = track[~np.isnan(track)]
            ax_imsd.plot(track.index.values, track.values, linewidth=0.5, alpha=0.2)
            if len(track) >= 5:
                track = track[:10*timescale + 0.1]
                popt, pcov = curve_fit(modelfunc, track.index.values[:40], track.values[:40], p0=popt_time ,bounds=([0., 0.], [5., 1.]))
                fitparams = tp.utils.fit_powerlaw(track, plot=False)
                fitparams.n = popt[0]
                fitparams.A = popt[1]
                fitparams['len'] = len(imsd[pnum])
                if not fitparams.isnull().values.any():
                    fitparams_list.append(fitparams)
        fitparams_df = pd.concat(fitparams_list)
        fitparams_df_list.append(fitparams_df)
        
        ax_imsd.plot(timefirst_msd.index.values * timescale, timefirst_msd.mmd.values, linewidth=2, alpha=1, color='red')
        ax_imsd.set_xscale('log')
        ax_imsd.set_yscale('log')
        ax_imsd.set_ylim(bottom = 0.001)
        ax_imsd.set_ylabel('Trajectory MSD')
        ax_imsd.set_xlabel(r'Lag Time $\Delta$T (s)')
        ax_imsd.set_xlim(left=timescale)
        ax_imsd.set_title(r' $\bf{(a)}$ Trajectory MSDs', loc='left', pad=-15)
        ax_imsd.grid(which='major', linewidth=0.5)
        ax_imsd.grid(which='minor', linewidth=0.1)  
        
        ax_ins = ax_imsd.inset_axes([0.65, 0.1, 0.30, 0.30])
        ax_ins.scatter(fitparams_df.n, fitparams_df.A, color='blue', s=10, alpha=0.2)
        
        ax_ins.scatter(popt_time[0], popt_time[1], color='red')
        ax_ins.set_xlabel(r'$\alpha$')
        ax_ins.set_ylabel(r'D$_\alpha$')
        ax_ins.set_title(r' $\bf{(b)}$ Fit Parameter Distribution', loc='left', pad=-15)
        ax_ins.set_xlim(0, 2)
        ax_ins.set_ylim(0, 0.1)
    
        ax_d = plt.subplot2grid((3,4), (2,0), rowspan=1, colspan=2, fig=fig)
        sns.distplot(fitparams_df.A.values, rug=True, ax=ax_d, norm_hist=True, hist_kws={"histtype":"step", "linewidth":2}, kde_kws={"bw":0.1})
        ax_d.set_title(r' $\bf{(c)}$ D$_\alpha$ Distribution', loc='left', pad=-15)
        ax_d.set_xlim(left=0)
        ax_d.set_xlabel(r'D$_\alpha$')
        
        ax_alpha = plt.subplot2grid((3,4), (2,2), rowspan=1, colspan=2, fig=fig)
        sns.distplot(fitparams_df.n.values, rug=True, ax=ax_alpha, norm_hist=True, hist_kws={"histtype":"step", "linewidth":2}, kde_kws={"bw":0.1})
        ax_alpha.set_title(r' $\bf{(d)}$ $\alpha$ Distribution', loc='left', pad=-15)
        ax_alpha.set_xlim(left=0)
        ax_alpha.set_xlabel(r'$\alpha$')
    
        fig.suptitle(legendstring)
        fig.subplots_adjust(top=0.955,
                            bottom=0.035,
                            left=0.045,
                            right=0.995,
                            hspace=0.26,
                            wspace=0.26)
        
        fig.savefig(savefolder_imsd_individual + settings.filename[:-5] + '_imsd_fitparams.png')
        fig.savefig(savefolder_imsd_individual + settings.filename[:-5] + '_imsd_fitparams.pdf')
        
        plt.close(fig)
    
    
        sorted_n = fitparams_df.n.sort_values().values
        sorted_A = fitparams_df.A.sort_values().values
        cut = 32
        if len(sorted_n) > cut * 1.5:
    
            n_vals = sorted_n[[cut, -cut]]
            A_vals = sorted_A[[cut, -cut]]
        
            fast_list = fitparams_df[fitparams_df.A > A_vals[1]].index.values
            slow_list = fitparams_df[fitparams_df.A < A_vals[0]].index.values
            sub_list = fitparams_df[fitparams_df.n < n_vals[0]].index.values
            super_list = fitparams_df[fitparams_df.n > n_vals[1]].index.values
        
            fig, ax = plt.subplots(1, 2, figsize=(16,8), dpi=200)
            
            axis_off(ax[0])
            ax[0].imshow(vid.inav[30].data, cmap='inferno')
            for particle in fast_list:
                ptrack_df = atoms[atoms.particle == particle]
                ax[0].plot(ptrack_df.x, ptrack_df.y, color='green', linewidth=0.75)
            for particle in slow_list:
                ptrack_df = atoms[atoms.particle == particle]
                ax[0].plot(ptrack_df.x, ptrack_df.y, color='red', linewidth=0.75)
            ax[0].set_title(r'$\bf{(a)}$ D$_\alpha$ Outliers')
            sb = ScaleBar(scale_list[i], units='nm', box_alpha=0, color='white')
            ax[0].add_artist(sb)
            
            axis_off(ax[1])
            ax[1].imshow(vid.inav[30].data, cmap='inferno')
            for particle in super_list:
                ptrack_df = atoms[atoms.particle == particle]
                ax[1].plot(ptrack_df.x, ptrack_df.y, color='green', linewidth=0.75)
            for particle in sub_list:
                ptrack_df = atoms[atoms.particle == particle]
                ax[1].plot(ptrack_df.x, ptrack_df.y, color='red', linewidth=0.75)
            ax[1].set_title(r'$\bf{(c)}$ $\alpha$ Outliers')
            sb = ScaleBar(scale_list[i], units='nm', box_alpha=0, color='white')
            ax[1].add_artist(sb)
        
            fig.subplots_adjust(top=0.965,
                                bottom=0.0,
                                left=0.0,
                                right=1.0,
                                hspace=0.1,
                                wspace=0.1)
            
            fig.savefig(savefolder_imsd_outliers + settings.filename[:-5] + '_imsd_outliers.png')
            fig.savefig(savefolder_imsd_outliers + settings.filename[:-5] + '_imsd_outliers.pdf')
            
            plt.close(fig)
            
    pickle.dump(imsd_list, open(savefolder_msd_pickle + 'imsd_list.p', 'wb'))
    pickle.dump(fitparams_df_list, open(savefolder_msd_pickle + 'imsd_fitparams_df_list.p', 'wb'))
    










fig_imsd_params = plt.figure(figsize=(20,14))
ax_imsd_params_wet = plt.subplot2grid((3,8), (0,0), rowspan=2, colspan=4, fig=fig_imsd_params)
ax_d_wet = plt.subplot2grid((3,8), (2,0), rowspan=1, colspan=2, fig=fig_imsd_params)
ax_alpha_wet = plt.subplot2grid((3,8), (2,2), rowspan=1, colspan=2, fig=fig_imsd_params)
ax_imsd_params_dry = plt.subplot2grid((3,8), (0,4), rowspan=2, colspan=4, fig=fig_imsd_params)
ax_d_dry = plt.subplot2grid((3,8), (2,4), rowspan=1, colspan=2, fig=fig_imsd_params)
ax_alpha_dry = plt.subplot2grid((3,8), (2,6), rowspan=1, colspan=2, fig=fig_imsd_params)

fitparams_cat_list = []
    
for i in tqdm(range(numvid), desc = 'IMSD Ensemble'):

    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]    
    wetdry = wetdry_list[i]
    fitparams_df = fitparams_df_list[i].copy()
    fitparams_long = fitparams_df[fitparams_df.len > 5]
    fitparams_df['wetdry'] = wetdry
    fitparams_cat_list.append(fitparams_df)
    
    popt_time, pcov_time = fit_time_list[i]
    
    if wetdry == 'wet':
        ax_imsd_params_wet.scatter(fitparams_long.n, fitparams_long.A, color=color, s=60, alpha=0.2, label=legendstring)
        ax_imsd_params_wet.scatter(popt_time[0], popt_time[1], color=color, s=600, alpha=1, label='_nolabel_', marker='X')
        sns.distplot(fitparams_long.A.values, rug=True, ax=ax_d_wet, norm_hist=True, hist_kws={"histtype":"step", "linewidth":2}, kde_kws={"bw":0.2}, color=color)
        sns.distplot(fitparams_long.n.values, rug=True, ax=ax_alpha_wet, norm_hist=True, hist_kws={"histtype":"step", "linewidth":2}, kde_kws={"bw":0.2}, color=color)
    elif wetdry == 'dry':
        ax_imsd_params_dry.scatter(fitparams_long.n, fitparams_long.A, color=color, s=60, alpha=0.2, label=legendstring)
        ax_imsd_params_dry.scatter(popt_time[0], popt_time[1], color=color, s=600, alpha=1, label='_nolabel_', marker='X')
        sns.distplot(fitparams_long.A.values, rug=True, ax=ax_d_dry, norm_hist=True, hist_kws={"histtype":"step", "linewidth":2}, kde_kws={"bw":0.2}, color=color)
        sns.distplot(fitparams_long.n.values, rug=True, ax=ax_alpha_dry, norm_hist=True, hist_kws={"histtype":"step", "linewidth":2}, kde_kws={"bw":0.2}, color=color)

for ax in [ax_imsd_params_wet, ax_imsd_params_dry]:
    ax.set_xlabel(r'$\alpha$')
    ax.set_ylabel(r'D$_\alpha$')
    ax.set_xlim(0, 2)
    ax.set_ylim(0, 0.2)
    ax.legend()
    
for ax in [ax_alpha_wet, ax_alpha_dry]:
    ax.set_xlim(0, 2)
    ax.set_xlabel(r'$\alpha$')

for ax in [ax_d_wet, ax_d_dry]:
    ax.set_xlim(0, 0.2)
    ax.set_xlabel(r'D$_\alpha$')
    
ax_imsd_params_wet.set_title('Liquid Cell Trajectory Paramater Distribution', loc='left', pad=-15)
ax_imsd_params_dry.set_title('Vacuum', loc='left', pad=-15)
ax_alpha_wet.set_title(r'Liquid Cell $\alpha$ Distribution', loc='left', pad=-15)
ax_alpha_dry.set_title(r'Vacuum $\alpha$ Distribution', loc='left', pad=-15)
ax_d_wet.set_title(r'Liquid Cell D$_\alpha$ Distribution', loc='left', pad=-15)
ax_d_dry.set_title(r'Vacuum D$_\alpha$ Distribution', loc='left', pad=-15)

fig_imsd_params.subplots_adjust(top=0.955,
                                bottom=0.035,
                                left=0.045,
                                right=0.995,
                                hspace=0.26,
                                wspace=0.26)
    
fig_imsd_params.savefig(savefolder_imsd + 'all_imsd_fitparams.png')
fig_imsd_params.savefig(savefolder_imsd + 'all_imsd_fitparams.pdf')

plt.close('all')
gc.collect()



fitparams_total = pd.concat(fitparams_cat_list)
fitparams_total = fitparams_total[fitparams_total.wetdry != 'ignore']
fitparams_total = fitparams_total[fitparams_total.len > 18]
fitparams_total = fitparams_total.reset_index(drop=True)

fp_wet = fitparams_total[fitparams_total.wetdry == 'wet']
fp_dry = fitparams_total[fitparams_total.wetdry == 'dry']



fig_joint = sns.JointGrid(height=12)
fig_joint.ax_joint.scatter(fp_wet.n, fp_wet.A, color='blue', label='Liquid Cell', alpha=0.2)
fig_joint.ax_joint.scatter(fp_dry.n, fp_dry.A, color='red', label='Vacuum', alpha=0.2)

fig_joint.ax_joint.set_xlim(0, 2)
fig_joint.ax_joint.set_ylim(0, 0.2)

sns.kdeplot(fp_wet.n, ax=fig_joint.ax_marg_x, color='blue', bw=0.2, common_norm=True)
sns.kdeplot(fp_dry.n, ax=fig_joint.ax_marg_x, color='red', bw=0.2, common_norm=True)

sns.kdeplot(y=fp_wet.A, ax=fig_joint.ax_marg_y, color='blue', bw=0.02, common_norm=True)
sns.kdeplot(y=fp_dry.A, ax=fig_joint.ax_marg_y, color='red', bw=0.02, common_norm=True)

fig_joint.ax_joint.set_xlabel(r'$\alpha$')
fig_joint.ax_joint.set_ylabel(r'D$_\alpha$')
fig_joint.ax_joint.legend()

fig_joint.fig.suptitle('IMSD Fit Parameters')
fig_joint.fig.subplots_adjust(top=0.973,
                    bottom=0.052,
                    left=0.077,
                    right=0.987,
                    hspace=0.2,
                    wspace=0.2)
fig_joint.fig.savefig(savefolder_imsd + 'all_imsd_fitparams_jointplot.png')
fig_joint.fig.savefig(savefolder_imsd + 'all_imsd_fitparams_jointplot.pdf')



# # IMSD fig

# fp_details_list = []
# for i in range(numvid):
#     fitparams = fitparams_cat_list[i].copy()
#     fitparams['bc_val'] = beam_current_list[i]
#     fitparams['name'] = settings_list[i].filename[:-5]
#     fp_details_list.append(fitparams)
# fp_details_df = pd.concat(fp_details_list)

# fp_details_df.groupby('name').mean()
    
# cm_vals = np.linspace(0, 1, len(bc_vals)+2)[1:-1]
# dry_cmap = plt.get_cmap('Reds')
# wet_cmap = plt.get_cmap('Blues')
# tvals, tcounts = np.unique(timescale_list, return_counts=True)
# usetime = tvals[tcounts.argmax()]

# imsd_exclude_list = ['Scanning Preview20_sparse data', 'Scanning Preview_12.43',]

# fig = plt.figure()
# ax = fig.add_subplot()
# for i in range(len(bc_vals)):
#     bc_val = bc_vals[i]
#     wet_color = wet_cmap(cm_vals[i])
#     dry_color = dry_cmap(cm_vals[i])
    
#     excludebool = ~fp_details_df.name.isin(imsd_exclude_list)
#     bc_bool = fp_details_df.bc_val == bc_val
#     wet_bool = fp_details_df.wetdry == 'wet'
#     dry_bool = fp_details_df.wetdry == 'dry'
#     flat_bool = fp_details_df.n > 1e-9
#     super_bool = fp_details_df.A > 1e-5
    
#     wet_df = fp_details_df[excludebool & bc_bool & wet_bool & flat_bool & super_bool]
#     dry_df = fp_details_df[excludebool & bc_bool & dry_bool & flat_bool & super_bool]
    
#     if len(wet_df) > 10:
#         ax.scatter(wet_df.n, wet_df.A, color=wet_color, alpha=0.5)
#     if len(dry_df) > 10:
#         ax.scatter(dry_df.n, dry_df.A, color=dry_color, alpha=0.5)
        
#     ax.set_xlim(0, 2)
#     ax.set_ylim(0, 0.5)
    










































# IMSD violin plots

savefolder_violins = savefolder_imsd + 'violins' + os.sep
os.makedirs(savefolder_violins, exist_ok=True)

imsd_melted_list = []

for i in tqdm(range(numvid), desc = 'IMSD Analysis'):
    
    vid=vid_list[i]
    
    atoms_norm = atoms_norm_list[i]
    atoms = atoms_list[i]
    lattice_norm = lattice_norm_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]
    wetdry = wetdry_list[i]
    
    imsd = imsd_list[i]
    imsd_melt = imsd.transpose().melt().dropna()
    imsd_melt['lag frames'] = (imsd_melt['lag time [s]'].values / timescale).astype(int)
    imsd_melt['wetdry'] = wetdry
    imsd_melted_list.append(imsd_melt)
    imsd_plot = imsd_melt[imsd_melt['lag frames'] < 31]

    fig, ax = plt.subplots(1, figsize=(30, 12))
    sns.violinplot(data=imsd_plot, x='lag frames', y='value', inner=None, ax=ax, bw=.1, cut=0, scale='width')
    ax.set_ylim(0, 5)
    ax.set_ylabel('MSD (nm)')
    ax.set_title(settings.filename[:-5], color=color)
    fig.subplots_adjust(top=0.965,
                        bottom=0.06,
                        left=0.04,
                        right=0.99,
                        hspace=0.2,
                        wspace=0.2)
    fig.savefig(savefolder_violins + settings.filename[:-5] + '_imsd_violins.png')
    
    plt.close(fig)




imsd_melted_all = pd.concat(imsd_melted_list).reset_index(drop=True)
imsd_melted_plot = imsd_melted_all[imsd_melted_all['lag frames'] < 31]
imsd_melted_plot = imsd_melted_plot[imsd_melted_plot['wetdry'] != 'ignore']
fig, ax = plt.subplots(1, figsize=(30, 12))
sns.violinplot(data=imsd_melted_plot, x='lag frames', y='value', split=True, inner=None, hue='wetdry', ax=ax, bw=.1, cut=0, scale='width')
ax.set_ylim(0, 2)
ax.set_ylabel('MSD (nm)')
fig.subplots_adjust(top=0.965,
                    bottom=0.06,
                    left=0.04,
                    right=0.99,
                    hspace=0.2,
                    wspace=0.2)
fig.savefig(savefolder_violins + '_imsd_violins.png')








# output data table
df = pd.DataFrame(index=filename_list)
df['Status'] = wetdry_list
df['beam_current_pA'] = [bc * 1e12 for bc in beam_current_list]
df['dosescale'] = dosescale_list
df['mean adatom count'] = [np.mean(lis) for lis in adatom_counts_list]
df['mean adatom density'] = [np.mean(lis) for lis in adatom_density_list]
df['D (TAMSD)'] = [fit_time[0][1] for fit_time in fit_time_list]
df['alpha (TAMSD)'] = [fit_time[0][0] for fit_time in fit_time_list]
df['displacement_30s (TAMSD)'] = [modelfunc(30, fit_time[0][0], fit_time[0][1]) for fit_time in fit_time_list]
df['D (EAMSD)'] = [fit_ens[0][1] for fit_ens in fit_ens_list]
df['alpha (EAMSD)'] = [fit_ens[0][0] for fit_ens in fit_ens_list]
df['displacement_30s (EAMSD)'] = [modelfunc(30, fit_ens[0][0], fit_ens[0][1]) for fit_ens in fit_ens_list]

df = df[df.Status != 'ignore']

df.to_csv(savefolder + 'data_table.csv')


def nullfunc(a,b):
    return ''

# output first and last video frames

figwet, ax_arrwet = plt.subplots(numwet, 2, figsize=(6, 3 * numwet))
figdry, ax_arrdry = plt.subplots(numdry, 2, figsize=(7, 3 * numdry))
filtwidth = 0.8
wetcount = 0
drycount = 0
for i in tqdm(range(numvid), desc = 'Frame Outputs'):
    
    if wetdry_list[i] != 'ignore':

        vid=vid_list[i]
        vidlen = vid.data.shape[0]
        atoms = atoms_list[i]
        settings = settings_list[i]
        timescale = timescale_list[i]
        
        temp_df = pd.DataFrame(columns=['maxdist', 'len'])
        for pnum in atoms.particle.unique():
            track = atoms[atoms.particle == pnum]
            maxdist = track.sum_dist.max()
            flen = len(track)
            temp_df.loc[pnum] = maxdist, flen
            
        vid_flat = vid.data.flatten()
        vid_dat = vid_flat[vid_flat != 0]
        per_indices = np.array([0.1, 99.5])
        indices = (len(vid_dat) * (per_indices / 100)).astype(int)
        vmin = np.partition(vid_dat, indices[0])[indices[0]]
        vmax = np.partition(vid_dat, indices[1])[indices[1]]
        
        # x_int = np.where(vid.data.min(axis=(0)).sum(axis=0))[0]
        # y_int = np.where(vid.data.min(axis=(0)).sum(axis=1))[0]
        # xlim = x_int[0], x_int[-1]
        # ylim = y_int[0], y_int[-1]
        
        long_list = temp_df.sort_values('len').index.values
        dist_list = temp_df.sort_values('maxdist').index.values
        
        plotnumber = 60
        long_list = temp_df.sort_values('len').index.values[-plotnumber:]
        dist_list = temp_df.sort_values('maxdist').index.values[-plotnumber:]
        
        common_list = np.intersect1d(long_list, dist_list)
        if len(common_list < plotnumber):
            remaindernum = (plotnumber - len(common_list))/2
            long_remainders = long_list[np.isin(long_list, common_list)]
            dist_remainders = dist_list[np.isin(dist_list, common_list)]
            common_list = np.append(common_list, long_remainders[:int(remaindernum)])
            common_list = np.append(common_list, dist_remainders[:int(remaindernum)])
            
        if wetdry_list[i] == 'wet':
            ax_duo = ax_arrwet[wetcount]
            wetcount += 1
        elif wetdry_list[i] == 'dry':
            ax_duo = ax_arrdry[drycount]
            drycount += 1
        for j, fnum in zip([0,1], [0, vidlen-1]):
            ax = ax_duo[j]
            ax.axis('off')
            frame = vid.inav[fnum]
            frame_df = atoms[atoms.frame <= fnum]
            frame_df = frame_df[np.isin(frame_df.particle, common_list)]
            framedata = gaussian_filter(frame.data, filtwidth)
            ax.imshow(framedata, cmap='inferno', vmin=vmin, vmax=vmax)
            ax.add_artist(AnchoredText('t = %.2f s' % (fnum * timescale), loc='lower left', frameon=False, prop=dict(color='white')))
            ax.add_artist(AnchoredText('Sequence %d' %(wetcount + drycount), loc='upper left', frameon=False, prop=dict(color='white')))
            if j == 1:
                # for particle in common_list:
                #     ptrack_df = frame_df[frame_df.particle == particle]
                #     for k in range(len(ptrack_df)-1):
                #         ax.plot(ptrack_df.x[k:k+2], ptrack_df.y[k:k+2], color=plt.get_cmap('rainbow')(ptrack_df.frame.values[k]/vidlen), linewidth=1)
                ax.yaxis.set_label_position('right')
                ax.add_artist(AnchoredText('%d pA' %(beam_current_list[i] * 1e12), loc='lower right', frameon=False, prop=dict(color='white')))
            else:
                ax.add_artist(AnchoredText(settings.filename[:-5], loc='lower right', frameon=False, prop=dict(color='white')))
            ax.yaxis.label.set_fontsize(6)
            ax.set_xticks([])
            ax.set_yticks([])
            curframe_df = frame_df[frame_df.frame == fnum]
            # ax.scatter(curframe_df.x, curframe_df.y, s=30, color=plt.get_cmap('rainbow')(fnum/framechoice[-1]))
            # ax.set_xlim(xlim)
            # ax.set_ylim(ylim)
            sb = ScaleBar(scale_list[i], 'nm', box_alpha=0, color='white', fixed_value=2, fixed_units='nm', label_formatter=nullfunc)
            ax.add_artist(sb)
for fig in [figwet, figdry]:
    fig.subplots_adjust(top=1.0,
                        bottom=0.0,
                        left=0.04,
                        right=0.965,
                        hspace=0.05,
                        wspace=0.05)
figwet.savefig(savefolder + 'overview_wet.png')
figwet.savefig(savefolder + 'overview_wet.pdf')
figdry.savefig(savefolder + 'overview_dry.png')
figdry.savefig(savefolder + 'overview_dry.pdf')  
    
    
    








# output video frames with marked trajectories

savefolder_frames = savefolder + 'frames_output' + os.sep
os.makedirs(savefolder_frames, exist_ok=True)

filtwidth = 0.6

for i in tqdm(range(numvid), desc = 'Frame Outputs'):
    
    vid=vid_list[i]
    vidlen = vid.data.shape[0]
    atoms = atoms_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]
    
    savename = savefolder_frames + settings.filename[:-5] + '_framelist_'
    
    if not os.path.exists(savename + '6.png'):
    
        temp_df = pd.DataFrame(columns=['maxdist', 'len'])
        for pnum in atoms.particle.unique():
            track = atoms[atoms.particle == pnum]
            maxdist = track.sum_dist.max()
            flen = len(track)
            temp_df.loc[pnum] = maxdist, flen
            
        vid_flat = vid.data.flatten()
        vid_dat = vid_flat[vid_flat != 0]
        vmin, vmax = vid_dat.min(), vid_dat.max()
        
        x_int = np.where(vid.data.min(axis=(0)).sum(axis=0))[0]
        y_int = np.where(vid.data.min(axis=(0)).sum(axis=1))[0]
        xlim = x_int[0], x_int[-1]
        ylim = y_int[0], y_int[-1]
            
        plotnumber = 60
        
        long_list = temp_df.sort_values('len').index.values[-plotnumber:]
        dist_list = temp_df.sort_values('maxdist').index.values[-plotnumber:]
        
        common_list = np.intersect1d(long_list, dist_list)
        if len(common_list < plotnumber):
            remaindernum = (plotnumber - len(common_list))/2
            long_remainders = long_list[np.isin(long_list, common_list)]
            dist_remainders = dist_list[np.isin(dist_list, common_list)]
            common_list = np.append(common_list, long_remainders[:int(remaindernum)])
            common_list = np.append(common_list, dist_remainders[:int(remaindernum)])
            
        
        framestep = int((vidlen-1) / 5)
        framechoice = np.arange(0, vidlen, framestep)
        
        fig, ax_arr = plt.subplots(2, 3, figsize=(19,12))
        for j in range(len(framechoice)):
            ax = ax_arr.flatten()[j]
            fnum = framechoice[j]
            axis_off(ax)
            frame = vid.inav[fnum]
            frame_df = atoms[atoms.frame <= fnum]
            frame_df = frame_df[np.isin(frame_df.particle, common_list)]
            framedata = gaussian_filter(frame.data, filtwidth)
            ax.imshow(framedata, cmap='inferno', vmin=vmin, vmax=vmax)
            ax.text(0.05, 0.05, 't = %.2f s' % (fnum * timescale), transform=ax.transAxes, color='white')
            for particle in common_list:
                ptrack_df = frame_df[frame_df.particle == particle]
                for k in range(len(ptrack_df)-1):
                    ax.plot(ptrack_df.x[k:k+2], ptrack_df.y[k:k+2], color=plt.get_cmap('rainbow')(ptrack_df.frame.values[k]/framechoice[-1]), linewidth=1)
            curframe_df = frame_df[frame_df.frame == fnum]
            ax.scatter(curframe_df.x, curframe_df.y, s=30, color=plt.get_cmap('rainbow')(fnum/framechoice[-1]))
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            sb = ScaleBar(scale_list[i], 'nm', box_alpha=0, color='white')
            ax.add_artist(sb)
        fig.subplots_adjust(top=1.0,
                            bottom=0.0,
                            left=0.0,
                            right=1.0,
                            hspace=0.025,
                            wspace=0.025)
    
        fig.savefig(savename + '6.png')
        fig.savefig(savename + '6.pdf')
        
        plt.close(fig)
        
        framestep = int((vidlen-1)/ 3)
        framechoice = np.arange(0, vidlen, framestep)
        
        fig, ax_arr = plt.subplots(2, 2, figsize=(16,16))
        for j in range(len(framechoice)):
            ax = ax_arr.flatten()[j]
            fnum = framechoice[j]
            axis_off(ax)
            frame = vid.inav[fnum]
            frame_df = atoms[atoms.frame <= fnum]
            frame_df = frame_df[np.isin(frame_df.particle, common_list)]
            framedata = gaussian_filter(frame.data, filtwidth)
            ax.imshow(framedata, cmap='inferno', vmin=vmin, vmax=vmax)
            ax.text(0.05, 0.05, 't = %.2f s' % (fnum * timescale), transform=ax.transAxes, color='white')
            for particle in common_list:
                ptrack_df = frame_df[frame_df.particle == particle]
                for k in range(len(ptrack_df)-1):
                    ax.plot(ptrack_df.x[k:k+2], ptrack_df.y[k:k+2], color=plt.get_cmap('rainbow')(ptrack_df.frame.values[k]/framechoice[-1]), linewidth=1)
            curframe_df = frame_df[frame_df.frame == fnum]
            ax.scatter(curframe_df.x, curframe_df.y, s=30, color=plt.get_cmap('rainbow')(fnum/framechoice[-1]))
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            sb = ScaleBar(scale_list[i], 'nm', box_alpha=0, color='white')
            ax.add_artist(sb)
        fig.subplots_adjust(top=1.0,
                            bottom=0.0,
                            left=0.0,
                            right=1.0,
                            hspace=0.025,
                            wspace=0.025)
    
        fig.savefig(savename + '4.png')
        fig.savefig(savename + '4.pdf')
    
        plt.close(fig)
    
    
    
    

plt.figure()
sm = plt.cm.ScalarMappable(cmap='rainbow', norm=plt.Normalize(vmin=0, vmax=1))
plt.colorbar(sm, fraction = 0.05, pad = -0.2)
plt.savefig(savefolder_frames + 'framelist_colorbar.pdf')

plt.close('all')
gc.collect()



#################################################################################






savefolder_mo_dists = savefolder + 'mo_dists' + os.sep
os.makedirs(savefolder_mo_dists, exist_ok=True)

distlists_wet, distlists_dry, distlists_all = [], [], []

for i in tqdm(range(numvid), desc = 'position analysis'):
    
    atoms_norm = atoms_norm_list[i]
    lattice_norm = lattice_norm_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]
    wetdry = wetdry_list[i]
    distlist = []
    for fnum in lattice_norm.frame.unique():
        atoms_frame = atoms_norm[atoms_norm.frame == fnum]
        lattice_frame = lattice_norm[lattice_norm.frame == fnum]
        latpoints = lattice_frame[['x', 'y']]
        atompoints = atoms_frame[['x', 'y']]
        tree = KDTree(latpoints)
        dists, indexes = tree.query(atompoints)
        distlist.append(dists)
    alldists = np.concatenate(distlist)
    
    fig, ax = plt.subplots(1)
    sns.distplot(alldists, kde_kws={'bw':0.02})    
    ax.set_xlim(0, 0.2)    
    ax.set_title(settings.filename[:-5], color=color)
    fig.savefig(savefolder_mo_dists + settings.filename[:-5] + '.png')
    
    plt.close('all')
    
    distlists_all.append(alldists)
    if wetdry == 'wet':
        distlists_wet.append(alldists)
    elif wetdry == 'dry':
        distlists_dry.append(alldists)
        
alldists_wet = np.concatenate(distlists_wet)
alldists_dry = np.concatenate(distlists_dry)
    
fig, ax = plt.subplots(1)
sns.distplot(alldists_wet, label='Liquid Cell', color='blue', ax=ax, kde_kws={'bw':0.0001}, bins=600)
sns.distplot(alldists_dry, label='Vacuum', color='red', ax=ax, kde_kws={'bw':0.0001}, bins=600)
ax.set_xlim(0, 0.2) 
fig.savefig(savefolder_mo_dists + 'all_wetdry.png')

    
    
    
    
    
    
    
    
    
    
    
    
    
    
##KDE plotting calculations   


savefolder_kde = savefolder + 'KDE_plot' + os.sep
os.makedirs(savefolder_kde, exist_ok=True)

# bandwidths = [0.03, 0.035, 0.04, 0.045, 0.05, 0.055, 0.06, 0.07, 0.08, 0.09,  0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.2, 0.25]
# bandwidths = [0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.11]
bandwidths = [0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1]
kde_plotlim = 0.4




#dry ensemble

mo_points, s_points, vac_points = mo_points_fixed, s_points_fixed, vac_points_fixed
atoms_norm = dry_atoms_norm_all

savefolder_kdefile = savefolder_kde + 'raw' + os.sep + 'all_dry' + os.sep
os.makedirs(savefolder_kdefile, exist_ok=True)

mo_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in mo_points], 8), axis=0)
s_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in s_points], 8), axis=0)
vac_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in vac_points], 8), axis=0)

lat_x_initial, lat_y_initial = atoms_norm['transformed_lat_x'], atoms_norm['transformed_lat_y']
nancols = (np.isnan(lat_x_initial)) | (np.isnan(lat_y_initial))
lat_x, lat_y = lat_x_initial[~nancols], lat_y_initial[~nancols]

repeat_x, repeat_y = zip(*[(x + point[0], y + point[1]) for point in mo_points_ext for x, y in zip(lat_x, lat_y)])
        
# reshape
values = np.vstack([repeat_x, repeat_y])


# choose KDE grid for plotting (same for both)
plotlim = kde_plotlim
xx, yy = np.mgrid[-plotlim:plotlim:256j, -plotlim:plotlim:256j]
gridnum = len(xx[0])
positions = np.vstack([xx.ravel(), yy.ravel()])

# calculate mask for single repeat

index_0 = np.where((mo_points_fixed == [0, 0]).all(1))[0][0]
mask = np.rot90(closest_index(xx, yy, mo_points_fixed) == index_0)
   
plot_as_scatter(values, mask, plotlim, 0.05, 10, savefolder_kdefile + 'scatter.png')
plot_as_hexbin(values, mask, plotlim, 120, savefolder_kdefile + 'hexbin120.png')
plot_as_hexbin(values, mask, plotlim, 60, savefolder_kdefile + 'hexbin60.png')

# choose KDE bandwidths, calculate and output images

for bw in bandwidths:
    KDE_output_test(positions, values, bw, savefolder_kdefile, mask)

plt.close('all')
gc.collect()




# wet ensemble
mo_points, s_points, vac_points = mo_points_fixed, s_points_fixed, vac_points_fixed
atoms_norm = wet_atoms_norm_all

savefolder_kdefile = savefolder_kde + 'raw' + os.sep + 'all_wet' + os.sep
os.makedirs(savefolder_kdefile, exist_ok=True)

mo_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in mo_points], 8), axis=0)
s_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in s_points], 8), axis=0)
vac_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in vac_points], 8), axis=0)

lat_x_initial, lat_y_initial = atoms_norm['transformed_lat_x'], atoms_norm['transformed_lat_y']
nancols = (np.isnan(lat_x_initial)) | (np.isnan(lat_y_initial))
lat_x, lat_y = lat_x_initial[~nancols], lat_y_initial[~nancols]

repeat_x, repeat_y = zip(*[(x + point[0], y + point[1]) for point in mo_points_ext for x, y in zip(lat_x, lat_y)])
        
# reshape
values = np.vstack([repeat_x, repeat_y])


# choose KDE grid for plotting (same for both)
plotlim = kde_plotlim
xx, yy = np.mgrid[-plotlim:plotlim:256j, -plotlim:plotlim:256j]
gridnum = len(xx[0])
positions = np.vstack([xx.ravel(), yy.ravel()])

# calculate mask for single repeat

index_0 = np.where((mo_points_fixed == [0, 0]).all(1))[0][0]
mask = np.rot90(closest_index(xx, yy, mo_points_fixed) == index_0)
   
plot_as_scatter(values, mask, plotlim, 0.05, 10, savefolder_kdefile + 'scatter.png')
plot_as_hexbin(values, mask, plotlim, 120, savefolder_kdefile + 'hexbin120.png')
plot_as_hexbin(values, mask, plotlim, 60, savefolder_kdefile + 'hexbin60.png')

# choose KDE bandwidths, calculate and output images

for bw in bandwidths:
    KDE_output_test(positions, values, bw, savefolder_kdefile, mask)

plt.close('all')
gc.collect()







mo_points, s_points, vac_points = mo_points_fixed, s_points_fixed, vac_points_fixed
mo_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in mo_points], 8), axis=0)
s_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in s_points], 8), axis=0)
vac_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in vac_points], 8), axis=0)
    

for i in tqdm(range(numvid), desc = 'KDE Plotting'):
    
    vid=vid_list[i]
    vidlen = vid.data.shape[0]
    atoms_norm = atoms_norm_list[i]
    settings = settings_list[i]
    color = colors_bc[i]
    legendstring = settings.filename[:-5] + ' Dose: %.2f x 10E6 e per nm^2' %dosescale_list[i]
    timescale = timescale_list[i]
    scale = scale_list[i]


    savefolder_kdefile = savefolder_kde + 'raw' + os.sep + settings.filename[:-5] + os.sep
    os.makedirs(savefolder_kdefile, exist_ok=True)

    mo_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in mo_points], 8), axis=0)
    s_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in s_points], 8), axis=0)
    vac_points_ext = np.unique(np.round([point1 + point2 for point1 in mo_points for point2 in vac_points], 8), axis=0)
    
    lat_x_initial, lat_y_initial = atoms_norm['transformed_lat_x'], atoms_norm['transformed_lat_y']
    nancols = (np.isnan(lat_x_initial)) | (np.isnan(lat_y_initial))
    lat_x, lat_y = lat_x_initial[~nancols], lat_y_initial[~nancols]
    
    repeat_x, repeat_y = zip(*[(x + point[0], y + point[1]) for point in mo_points_ext for x, y in zip(lat_x, lat_y)])
            
    # reshape and twist
    
    values = np.vstack([repeat_x, repeat_y])

    # choose KDE grid for plotting (same for both)
    plotlim = kde_plotlim
    xx, yy = np.mgrid[-plotlim:plotlim:256j, -plotlim:plotlim:256j]
    gridnum = len(xx[0])
    positions = np.vstack([xx.ravel(), yy.ravel()])
    
    # calculate mask for single repeat
    
    index_0 = np.where((mo_points_fixed == [0, 0]).all(1))[0][0]
    mask = np.rot90(closest_index(xx, yy, mo_points_fixed) == index_0)
       
    plot_as_scatter(values, mask, plotlim, 0.05, 10, savefolder_kdefile + 'scatter.png')
    plot_as_hexbin(values, mask, plotlim, 120, savefolder_kdefile + 'hexbin120.png')
    plot_as_hexbin(values, mask, plotlim, 60, savefolder_kdefile + 'hexbin60.png')

    # choose KDE bandwidths, calculate and output images
    
    # bandwidths = [0.03, 0.035, 0.04, 0.045, 0.05, 0.055, 0.06, 0.07, 0.08, 0.09,  0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.2, 0.25]
    # bandwidths = [0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.11]
    bandwidths = [0.06, 0.07, 0.08, 0.09, 0.1]
    
    for bw in bandwidths:
        KDE_output_test(positions, values, bw, savefolder_kdefile, mask)

    plt.close('all')
    gc.collect()












bw_dict = {'15.26.Scanning Preview': 0.08,
           'Scanning Preview_15,51': 0.08,
           'Scanning Preview_15.14': 0.09,
           'Scanning Preview_15.18': 0.1,
           'Scanning Preview_15.56': 0.09,
           'Scanning Preview12': 0.09,
           'Scanning Preview13': 0.08,
           'Scanning Preview13b': 0.08,
           'Scanning Preview14': 0.06,
           'Scanning Preview15': 0.06,
           'Scanning Preview17': 0.1,
           'Scanning Preview18': 0.07,
           'Scanning Preview19': 0.1,
           'Scanning Preview20_sparse data': 0.1,
           'Scanning Preview22': 0.08,
           'Scanning Preview23_40pA_12us': 0.07,
           'old_dry_vid': 0.1,
           'old_wet_vid': 0.08,
           'Scanning Preview_12.36': 0.08,
           'Scanning Preview_12.43': 0.08,
           'all_wet': 0.05, 
           'all_dry': 0.05
           }


bw_vals = []
kdemaps = []
fig, ax = plt.subplots(2, figsize=(16, 12))

for i in tqdm(range(numvid), desc = 'KDE Load All'):
    
    color = colors_bc[i]
    settings = settings_list[i]
    savefolder_kdefile = savefolder_kde + 'raw' + os.sep + settings.filename[:-5] + os.sep
    wetdry = wetdry_list[i]
    
    bw_val = bw_dict[settings.filename[:-5]]
    kdenorm = np.loadtxt(savefolder_kdefile + 'norm_prob_bw%.5f.txt' % bw_val, dtype='float')
    
    bw_vals.append(bw_val)
    kdemaps.append(kdenorm)
    
    if wetdry == 'wet':
        sns.distplot(kdenorm.flatten(), color=color, ax=ax[0], label=settings.filename[:-5], kde_kws={'bw':bw_val}, norm_hist=True)
    elif wetdry == 'dry':
        sns.distplot(kdenorm.flatten(), color=color, ax=ax[1], label=settings.filename[:-5], kde_kws={'bw':bw_val}, norm_hist=True)


savefolder_kdefile = savefolder_kde + 'raw' + os.sep + 'all_wet' + os.sep
bw_val_wet = bw_dict['all_wet']
kde_all_wet = np.loadtxt(savefolder_kdefile + 'norm_prob_bw%.5f.txt' % bw_val_wet, dtype='float')
sns.distplot(kdenorm.flatten(), color='blue', ax=ax[0], label='all wet', kde_kws={'bw':bw_val_wet}, norm_hist=True)

savefolder_kdefile = savefolder_kde + 'raw' + os.sep + 'all_dry' + os.sep
bw_val_dry = bw_dict['all_dry']
kde_all_dry = np.loadtxt(savefolder_kdefile + 'norm_prob_bw%.5f.txt' % bw_val_dry, dtype='float')
sns.distplot(kdenorm.flatten(), color='red', ax=ax[1], label='all dry', kde_kws={'bw':bw_val_dry}, norm_hist=True)

ax[0].legend()
ax[1].legend()

ax[0].set_title('Liquid Cell')
ax[1].set_title('Vacuum')

vmin = np.floor(np.min(kdemaps) * 10) / 10
vmax = np.ceil(np.max(kdemaps) * 10) / 10

for a in ax:
    a.axvline(vmin)
    a.axvline(vmax)
    a.set_xlabel('Adatom location probability histogram')

fig.savefig(savefolder_kde + 'probability_intensity_histogram.png')







vmin, vmax = 0.5, 2.5

molplotsettings = dict(bond_alpha=1,
                       bond_linewidth=1,
                       mo_size=600,
                       mo_color='purple',
                       s_size=400,
                       s_color='yellow',
                       marker_linewidth=4)

savefolder_kde_individual = savefolder_kde + 'individual' + os.sep
os.makedirs(savefolder_kde_individual, exist_ok=True)

s_points_fixed_red = np.concatenate([s_points_fixed[[0, 2]] , [mo_points_fixed[0] + s_points_fixed[2],]])
mo_points_fixed_red = mo_points_fixed[[0, 5, 6]]
vac_points_fixed_red = vac_points_fixed[[0,]]

s_profile = s_points_fixed_red[0][1]
mo_profile = mo_points_fixed_red[0][1]
profloc_real = np.mean([s_profile, mo_profile])
profloc = int(np.rint((kde_plotlim - profloc_real) / (kde_plotlim*2) * kdenorm.shape[1]))
mo_points_prof = mo_points_ext[np.abs(mo_points_ext[:, 1] - profloc_real) < 0.08][:, 0]
s_points_prof = s_points_ext[np.abs(s_points_ext[:, 1] - profloc_real) < 0.08][:, 0]
vac_points_prof = vac_points_ext[np.abs(vac_points_ext[:, 1] - profloc_real) < 0.08][:, 0]
    
for i in tqdm(range(numvid + 2)):
    
    if i < numvid:
        settings = settings_list[i]
        savefolder_kdefile = savefolder_kde + settings.filename[:-5] + os.sep
        bw_val = bw_vals[i]
        kdenorm = kdemaps[i].copy()
    elif i == numvid:
        settings = pd.Series()
        settings['filename'] = 'all_wet.hspy'
        savefolder_kdefile = savefolder_kde + settings.filename[:-5] + os.sep
        bw_val = bw_val_wet
        kdenorm = kde_all_wet
    elif i == numvid + 1:
        settings = pd.Series()
        settings.filename = 'all_dry.hspy'
        savefolder_kdefile = savefolder_kde + settings.filename[:-5] + os.sep
        bw_val = bw_val_dry
        kdenorm = kde_all_dry
    
    profile = kdenorm[profloc, :].copy()
    # kdenorm[profloc, :] = 2
    x_prof = np.linspace(-kde_plotlim, kde_plotlim, kdenorm.data.shape[0])
    

    
    fig = plt.figure(figsize=(8, 12))
    axkde = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
    axprof = plt.subplot2grid((3, 1), (2, 0))
    implot = axkde.imshow(kdenorm, extent=[-kde_plotlim, kde_plotlim, -kde_plotlim, kde_plotlim], cmap='Blues_r', aspect='auto')
    axis_off(axkde)
    molplot(axkde, mo_points_fixed_red, s_points_fixed_red, **molplotsettings)
    axkde.scatter(vac_points_fixed_red[:, 0], vac_points_fixed_red[:, 1], s=320, color='red', marker='.')
    axkde.set_xlim(-kde_plotlim, kde_plotlim)
    axkde.set_ylim(-kde_plotlim, kde_plotlim)
    arrow = FancyArrowPatch((-kde_plotlim, profloc_real), (kde_plotlim, profloc_real), arrowstyle='-|>', mutation_scale=40, ls='--', lw=2)
    axkde.add_patch(arrow)
    axprof.plot(x_prof, profile)
    for val in mo_points_prof:
        axprof.axvline(val, color='purple')
    for val in s_points_prof:
        axprof.axvline(val, color='yellow')
    for val in vac_points_prof:
        axprof.axvline(val, color='red')
    axprof.set_xlim(-kde_plotlim, kde_plotlim)
    axprof.axhline(1, color='black', alpha=0.5)
    divider = make_axes_locatable(axkde)
    ax_cb = divider.append_axes('top', size='5%', pad=0.1)
    plt.colorbar(implot, cax=ax_cb, orientation='horizontal')
    ax_cb.xaxis.tick_top()
    axprof.set_xlabel('Position along arrow (nm)')
    axprof.set_ylabel('Pt Occupancy Probability')
    sb = ScaleBar(1, units='nm', color='white', box_alpha=0)
    axkde.add_artist(sb)
    fig.suptitle(settings.filename[:-5])
    fig.subplots_adjust(top=0.94,
                        bottom=0.04,
                        left=0.085,
                        right=0.98,
                        hspace=0.025,
                        wspace=0.2)
    fig.savefig(savefolder_kde_individual + settings.filename[:-5] + '_.png')
    fig.savefig(savefolder_kde_individual + settings.filename[:-5] + '_.pdf')
    
    plt.close(fig)
    gc.collect()












wetcount, drycount = 1, 1
numcols = max([numwet, numdry])
figsize= 36, 12
fig_kdeoutputs = plt.figure(figsize=figsize)
shape = (2, numcols-1)

for i in tqdm(range(numvid), desc = 'kde_output'):
    
    settings = settings_list[i]    
    wetdry = wetdry_list[i]
    
    if wetdry != 'ignore':
    
        color = colors_bc[i]
        settings = settings_list[i]
        
        if settings.filename[:-5] not in ['Scanning Preview19', 'Scanning Preview_12.36', 'Scanning Preview_12.43']:
            savefolder_kdefile = savefolder_kde + settings.filename[:-5] + os.sep
            
            bw_val = bw_vals[i]
            kdenorm = kdemaps[i].copy()
            
            profile = kdenorm[profloc, :].copy()
            # kdenorm[profloc, :] = 2
            x_prof = np.linspace(-kde_plotlim, kde_plotlim, kdenorm.data.shape[0])
            
            if wetdry == 'wet':
                ax = plt.subplot2grid(shape , (0, (wetcount - 1)), aspect=3/2, fig=fig_kdeoutputs)
                wetcount += 1
            elif wetdry == 'dry':
                ax = plt.subplot2grid(shape , (1, (drycount - 1)), aspect=3/2, fig=fig_kdeoutputs)
                drycount += 1
                
            axis_off(ax)
            axkde = ax.inset_axes([0, 0.34, 1, 0.66])
            axprof = ax.inset_axes([0, 0, 1, 0.34])
            implot = axkde.imshow(kdenorm, extent=[-kde_plotlim, kde_plotlim, -kde_plotlim, kde_plotlim], cmap='Blues_r', aspect='auto', vmin=vmin, vmax=vmax)
            axis_off(axkde)
            molplot(axkde, mo_points_fixed_red, s_points_fixed_red, **molplotsettings)
            axkde.scatter(vac_points_fixed_red[:, 0], vac_points_fixed_red[:, 1], s=320, color='red', marker='.')
            axkde.set_xlim(-kde_plotlim, kde_plotlim)
            axkde.set_ylim(-kde_plotlim, kde_plotlim)
            arrow = FancyArrowPatch((-kde_plotlim, profloc_real), (kde_plotlim, profloc_real), arrowstyle='-|>', mutation_scale=40, ls='--', lw=2)
            axkde.add_patch(arrow)
            axprof.plot(x_prof, profile)
            for val in mo_points_prof:
                axprof.axvline(val, color='purple')
            for val in s_points_prof:
                axprof.axvline(val, color='yellow')
            for val in vac_points_prof:
                axprof.axvline(val, color='red')
            axprof.set_xlim(-kde_plotlim, kde_plotlim)
            axprof.set_ylim(vmin, vmax)
            axprof.axhline(1, color='black', alpha=0.5)
            ax.set_title('%.0f_'%(beam_current_list[i] * 1e12) + ' pA ' + settings.filename[:-5], color=color)

cbax = fig_kdeoutputs.add_axes([0.05, 0.98, 0.9, 0.02])
plt.colorbar(implot, cax=cbax, orientation='horizontal')
cbax.set_xlim(vmin, vmax)
    
fig_kdeoutputs.set_facecolor('grey')
fig_kdeoutputs.subplots_adjust(top=0.92,
                               bottom=0.01,
                               left=0.01,
                               right=0.985,
                               hspace=0.115,
                               wspace=0.12)
fig_kdeoutputs.savefig(savefolder_kde + 'all_compare_plots.png')
fig_kdeoutputs.savefig(savefolder_kde + 'all_compare_plots.pdf')

plt.close('all')
gc.collect()




def kde_axplot(kdenorm, ax, molplotsettings, profcolor, ax_xticks=True, ax_yticks=True, ax_yrev=False, title=False, counts=False, scalebar=True, textcolor='black'):
    profile = kdenorm[profloc, :].copy()
    x_prof = np.linspace(-kde_plotlim, kde_plotlim, kdenorm.data.shape[0])
    axis_off(ax)
    axkde = ax.inset_axes([0, 0.25, 1, 0.75])
    axprof = ax.inset_axes([0, 0, 1, 0.25])
    implot = axkde.imshow(kdenorm, extent=[-kde_plotlim, kde_plotlim, -kde_plotlim, kde_plotlim], cmap=kde_cmap, aspect='auto', vmin=kde_im_vmin, vmax=kde_im_vmax)
    axkde.set_xticks([])
    axkde.set_yticks([])
    molplot_new(axkde, mo_points_fixed_red, s_points_fixed_red, vac_points_fixed_red, **molplotsettings)
    axkde.set_xlim(-kde_plotlim, kde_plotlim)
    axkde.set_ylim(-kde_plotlim, kde_plotlim)
    arrow = FancyArrowPatch((-kde_plotlim, profloc_real), (kde_plotlim, profloc_real), arrowstyle='-|>', mutation_scale=40, ls='--', lw=2, color=profcolor)
    axkde.add_patch(arrow)
    axprof.plot(x_prof, profile, color=profcolor)
    for val in mo_points_prof:
        axprof.axvline(val, color='purple')
    for val in s_points_prof:
        axprof.axvline(val, color='yellow')
    for val in vac_points_prof:
        axprof.axvline(val, color='red')
    axprof.set_xlim(-kde_plotlim, kde_plotlim)
    axprof.set_ylim(kde_plot_vmin, kde_plot_vmax)
    axprof.axhline(1, color='black', alpha=0.5)
    if scalebar == True:
        sb = ScaleBar(1, units='nm', box_alpha=0, location='upper right', fixed_value=100, fixed_units='pm', color=textcolor)
        axkde.add_artist(sb)
    if title != False:
        axkde.add_artist(AnchoredText(title, loc='upper left', frameon=False, prop=dict(color=textcolor)))
    if counts != False:
        axkde.add_artist(AnchoredText('%d Pt Counts'%counts, loc='lower left', frameon=False, prop=dict(color=textcolor))) 
    if ax_xticks == False:
        axprof.set_xticks([])
    else:
        axprof.set_xlabel('Distance (nm)')
    if ax_yticks == False:
        axprof.set_yticks([])
    else:
        axprof.set_ylabel('Rel. Probability')
    if ax_yrev == True:
        axprof.yaxis.tick_right()
        axprof.yaxis.set_label_position("right")
    



molplotsettings_large = dict(bond_alpha=1,
                             bond_linewidth=1,
                             mo_size=500,
                             mo_color='purple',
                             s_size=300,
                             s_color='yellow',
                             vac_size=200,
                             vac_color='red',
                             marker_linewidth=4)

molplotsettings_small = dict(bond_alpha=1,
                             bond_linewidth=0.5,
                             mo_size=250,
                             mo_color='purple',
                             s_size=150,
                             s_color='yellow',
                             vac_size=100,
                             vac_color='red',
                             marker_linewidth=2)

kde_im_vmin, kde_im_vmax = 0.5, 2
kde_plot_vmin, kde_plot_vmax = 0, 3
kde_cmap = 'Blues_r'

lat_counts = []
for i in range(numvid):
    atoms_norm = atoms_norm_list[i]
    lat_counts.append(len(atoms_norm.lat_x.notna()))



shape = (4, 4)
fig = plt.figure(figsize=(9, 12))
axwet = plt.subplot2grid(shape , (0, 0), rowspan=2, colspan=2, aspect=4/3, fig=fig)
kde_axplot(kde_all_wet, axwet, molplotsettings_large, 'blue', ax_xticks=False, title=r'$\bf{a}$ Liquid Cell - all videos', counts=len(wet_atoms_norm_all.lat_x.notna()))
axdry = plt.subplot2grid(shape , (2, 0), rowspan=2, colspan=2, aspect=4/3, fig=fig)
kde_axplot(kde_all_dry, axdry, molplotsettings_large, 'red', title=r'$\bf{f}$ Vacuum - all videos', counts=len(dry_atoms_norm_all.lat_x.notna()))
axwet00 = plt.subplot2grid(shape , (0, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[10], axwet00, molplotsettings_small, 'blue', ax_xticks=False, ax_yticks=False, title=r'$\bf{b}$ Liquid Cell', counts=lat_counts[10])
axwet01 = plt.subplot2grid(shape , (0, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[12], axwet01, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{c}$ Liquid Cell', counts=lat_counts[12])
axwet10 = plt.subplot2grid(shape , (1, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[17], axwet10, molplotsettings_small, 'blue', ax_xticks=False, ax_yticks=False, title=r'$\bf{d}$ Liquid Cell', counts=lat_counts[17])
axwet11 = plt.subplot2grid(shape , (1, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[18], axwet11, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{e}$ Liquid Cell', counts=lat_counts[18])
axdry00 = plt.subplot2grid(shape , (2, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[2], axdry00, molplotsettings_small, 'red', ax_xticks=False, ax_yticks=False, title=r'$\bf{g}$ Vacuum', counts=lat_counts[2])
axdry01 = plt.subplot2grid(shape , (2, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[4], axdry01, molplotsettings_small, 'red', ax_xticks=False, ax_yrev=True, title=r'$\bf{h}$ Vacuum', counts=lat_counts[4])
axdry10 = plt.subplot2grid(shape , (3, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[5], axdry10, molplotsettings_small, 'red', ax_yticks=False, ax_yrev=True, title=r'$\bf{i}$ Vacuum', counts=lat_counts[5])
axdry11 = plt.subplot2grid(shape , (3, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[6], axdry11, molplotsettings_small, 'red', ax_yrev=True, title=r'$\bf{j}$ Vacuum', counts=lat_counts[6])

cbax = fig.add_axes([0.05, 0.98, 0.9, 0.02])
matplotlib.colorbar.ColorbarBase(cbax, orientation='horizontal', cmap=plt.get_cmap(kde_cmap), norm=matplotlib.colors.Normalize(kde_im_vmin, kde_im_vmax), label='Pt Location Relative Probability', ticks=[kde_im_vmin, 1, kde_im_vmax])

fig.subplots_adjust(top=0.93,
                    bottom=0.035,
                    left=0.055,
                    right=0.955,
                    hspace=0.1,
                    wspace=0.1)

fig.savefig(savefolder_kde + '4plots.png')
fig.savefig(savefolder_kde + '4plots.pdf')








shape = (6, 4)
fig = plt.figure(figsize=(7, 14))
axwet = plt.subplot2grid(shape , (0, 0), rowspan=3, colspan=3, aspect=4/3, fig=fig)
kde_axplot(kde_all_wet, axwet, molplotsettings_large, 'blue', ax_xticks=False, title=r'$\bf{a}$ Liquid Cell - all videos', counts=len(wet_atoms_norm_all.lat_x.notna()))
axdry = plt.subplot2grid(shape , (3, 0), rowspan=3, colspan=3, aspect=4/3, fig=fig)
kde_axplot(kde_all_dry, axdry, molplotsettings_large, 'red', title=r'$\bf{e}$ Vacuum - all videos', counts=len(dry_atoms_norm_all.lat_x.notna()))


axwet0 = plt.subplot2grid(shape , (0, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[12], axwet0, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{b}$ Liquid Cell', counts=lat_counts[12])
axwet1 = plt.subplot2grid(shape , (1, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[17], axwet1, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{c}$ Liquid Cell', counts=lat_counts[17])
axwet2 = plt.subplot2grid(shape , (2, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[18], axwet2, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{d}$ Liquid Cell', counts=lat_counts[18])

axdry0 = plt.subplot2grid(shape , (3, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[2], axdry0, molplotsettings_small, 'red', ax_xticks=False, ax_yrev=True, title=r'$\bf{f}$ Vacuum', counts=lat_counts[2])
axdry1 = plt.subplot2grid(shape , (4, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[4], axdry1, molplotsettings_small, 'red', ax_xticks=False, ax_yrev=True, title=r'$\bf{g}$ Vacuum', counts=lat_counts[4])
axdry2 = plt.subplot2grid(shape , (5, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[5], axdry2, molplotsettings_small, 'red', ax_yrev=True, title=r'$\bf{h}$ Vacuum', counts=lat_counts[5])


cbax = fig.add_axes([0.05, 0.98, 0.9, 0.02])
matplotlib.colorbar.ColorbarBase(cbax, orientation='horizontal', cmap=plt.get_cmap(kde_cmap), norm=matplotlib.colors.Normalize(kde_im_vmin, kde_im_vmax), label='Pt Location Relative Probability', ticks=[kde_im_vmin, 1, kde_im_vmax])

fig.subplots_adjust(top=0.925,
                    bottom=0.05,
                    left=0.065,
                    right=0.925,
                    hspace=0.0,
                    wspace=0.09)

fig.savefig(savefolder_kde + '3plots.png')
fig.savefig(savefolder_kde + '3plots.pdf')





kde_im_vmin, kde_im_vmax = 0, 2
kde_plot_vmin, kde_plot_vmax = 0, 3
kde_cmap = 'coolwarm'





shape = (4, 4)
fig = plt.figure(figsize=(9, 12))
axwet = plt.subplot2grid(shape , (0, 0), rowspan=2, colspan=2, aspect=4/3, fig=fig)
kde_axplot(kde_all_wet, axwet, molplotsettings_large, 'blue', ax_xticks=False, title=r'$\bf{a}$ Liquid Cell - all videos', counts=len(wet_atoms_norm_all.lat_x.notna()))
axdry = plt.subplot2grid(shape , (2, 0), rowspan=2, colspan=2, aspect=4/3, fig=fig)
kde_axplot(kde_all_dry, axdry, molplotsettings_large, 'red', title=r'$\bf{f}$ Vacuum - all videos', counts=len(dry_atoms_norm_all.lat_x.notna()))
axwet00 = plt.subplot2grid(shape , (0, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[10], axwet00, molplotsettings_small, 'blue', ax_xticks=False, ax_yticks=False, title=r'$\bf{b}$ Liquid Cell', counts=lat_counts[10])
axwet01 = plt.subplot2grid(shape , (0, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[12], axwet01, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{c}$ Liquid Cell', counts=lat_counts[12])
axwet10 = plt.subplot2grid(shape , (1, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[17], axwet10, molplotsettings_small, 'blue', ax_xticks=False, ax_yticks=False, title=r'$\bf{d}$ Liquid Cell', counts=lat_counts[17])
axwet11 = plt.subplot2grid(shape , (1, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[18], axwet11, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{e}$ Liquid Cell', counts=lat_counts[18])
axdry00 = plt.subplot2grid(shape , (2, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[2], axdry00, molplotsettings_small, 'red', ax_xticks=False, ax_yticks=False, title=r'$\bf{g}$ Vacuum', counts=lat_counts[2])
axdry01 = plt.subplot2grid(shape , (2, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[4], axdry01, molplotsettings_small, 'red', ax_xticks=False, ax_yrev=True, title=r'$\bf{h}$ Vacuum', counts=lat_counts[4])
axdry10 = plt.subplot2grid(shape , (3, 2), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[5], axdry10, molplotsettings_small, 'red', ax_yticks=False, ax_yrev=True, title=r'$\bf{i}$ Vacuum', counts=lat_counts[5])
axdry11 = plt.subplot2grid(shape , (3, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[6], axdry11, molplotsettings_small, 'red', ax_yrev=True, title=r'$\bf{j}$ Vacuum', counts=lat_counts[6])

cbax = fig.add_axes([0.05, 0.98, 0.9, 0.02])
matplotlib.colorbar.ColorbarBase(cbax, orientation='horizontal', cmap=plt.get_cmap(kde_cmap), norm=matplotlib.colors.Normalize(kde_im_vmin, kde_im_vmax), label='Pt Location Relative Probability', ticks=[kde_im_vmin, 1, kde_im_vmax])

fig.subplots_adjust(top=0.93,
                    bottom=0.035,
                    left=0.055,
                    right=0.955,
                    hspace=0.1,
                    wspace=0.1)

fig.savefig(savefolder_kde + '4plots_alt.png')
fig.savefig(savefolder_kde + '4plots_alt.pdf')








shape = (6, 4)
fig = plt.figure(figsize=(7, 14))
axwet = plt.subplot2grid(shape , (0, 0), rowspan=3, colspan=3, aspect=4/3, fig=fig)
kde_axplot(kde_all_wet, axwet, molplotsettings_large, 'blue', ax_xticks=False, title=r'$\bf{a}$ Liquid Cell - all videos', counts=len(wet_atoms_norm_all.lat_x.notna()))
axdry = plt.subplot2grid(shape , (3, 0), rowspan=3, colspan=3, aspect=4/3, fig=fig)
kde_axplot(kde_all_dry, axdry, molplotsettings_large, 'red', title=r'$\bf{e}$ Vacuum - all videos', counts=len(dry_atoms_norm_all.lat_x.notna()))


axwet0 = plt.subplot2grid(shape , (0, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[12], axwet0, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{b}$ Liquid Cell', counts=lat_counts[12])
axwet1 = plt.subplot2grid(shape , (1, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[17], axwet1, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{c}$ Liquid Cell', counts=lat_counts[17])
axwet2 = plt.subplot2grid(shape , (2, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[18], axwet2, molplotsettings_small, 'blue', ax_xticks=False, ax_yrev=True, title=r'$\bf{d}$ Liquid Cell', counts=lat_counts[18])

axdry0 = plt.subplot2grid(shape , (3, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[2], axdry0, molplotsettings_small, 'red', ax_xticks=False, ax_yrev=True, title=r'$\bf{f}$ Vacuum', counts=lat_counts[2])
axdry1 = plt.subplot2grid(shape , (4, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[4], axdry1, molplotsettings_small, 'red', ax_xticks=False, ax_yrev=True, title=r'$\bf{g}$ Vacuum', counts=lat_counts[4])
axdry2 = plt.subplot2grid(shape , (5, 3), rowspan=1, colspan=1, aspect=4/3, fig=fig)
kde_axplot(kdemaps[5], axdry2, molplotsettings_small, 'red', ax_yrev=True, title=r'$\bf{h}$ Vacuum', counts=lat_counts[5])


cbax = fig.add_axes([0.05, 0.98, 0.9, 0.02])
matplotlib.colorbar.ColorbarBase(cbax, orientation='horizontal', cmap=plt.get_cmap(kde_cmap), norm=matplotlib.colors.Normalize(kde_im_vmin, kde_im_vmax), label='Pt Location Relative Probability', ticks=[kde_im_vmin, 1, kde_im_vmax])

fig.subplots_adjust(top=0.925,
                    bottom=0.05,
                    left=0.065,
                    right=0.925,
                    hspace=0.0,
                    wspace=0.09)

fig.savefig(savefolder_kde + '3plots_alt.png')
fig.savefig(savefolder_kde + '3plots_alt.pdf')








def kde_axplot2(kdenorm, ax, molplotsettings, profcolor, ax_xticks=True, ax_yticks=True, ax_yrev=False, title=False, counts=False, scalebar=True, textcolor='black', plotprof=True, dose=False):
    profile = kdenorm[profloc, :].copy()
    x_prof = np.linspace(-kde_plotlim, kde_plotlim, kdenorm.data.shape[0])
    axis_off(ax)
    if plotprof == True:
        axkde = ax.inset_axes([0, 0.25, 1, 0.75])
        axprof = ax.inset_axes([0, 0, 1, 0.25])
    elif plotprof == False:
        axkde = ax.inset_axes([0, 0, 1, 1])
    implot = axkde.imshow(kdenorm, extent=[-kde_plotlim, kde_plotlim, -kde_plotlim, kde_plotlim], cmap=kde_cmap, aspect='auto', vmin=kde_im_vmin, vmax=kde_im_vmax)
    axkde.set_xticks([])
    axkde.set_yticks([])
    molplot_new(axkde, mo_points_fixed_red, s_points_fixed_red, vac_points_fixed_red, **molplotsettings)
    axkde.set_xlim(-kde_plotlim, kde_plotlim)
    axkde.set_ylim(-kde_plotlim, kde_plotlim)
    arrow = FancyArrowPatch((-kde_plotlim, profloc_real), (kde_plotlim, profloc_real), arrowstyle='-|>', mutation_scale=40, ls='--', lw=2, color=profcolor)
    axkde.add_patch(arrow)
    if plotprof == True:
        axprof.plot(x_prof, profile, color=profcolor, zorder=10)
        for val in mo_points_prof:
            axprof.axvline(val, color='purple')
        for val in s_points_prof:
            axprof.axvline(val, color='yellow')
        for val in vac_points_prof:
            axprof.axvline(val, color='red')
        axprof.set_xlim(-kde_plotlim, kde_plotlim)
        axprof.set_ylim(kde_plot_vmin, kde_plot_vmax)
        axprof.axhline(1, color='black', alpha=0.5)
    if scalebar == True:
        sb = ScaleBar(1, units='nm', box_alpha=0, location='lower right', fixed_value=100, fixed_units='pm', color=textcolor)
        axkde.add_artist(sb)
    if dose != False:
        axkde.add_artist(AnchoredText('%d pA' % (dose * 1e12), loc='upper right', frameon=False, prop=dict(color=textcolor)))
    if title != False:
        axkde.add_artist(AnchoredText(title, loc='upper left', frameon=False, prop=dict(color=textcolor)))
    if counts != False:
        axkde.add_artist(AnchoredText('%d Pt Counts'%counts, loc='lower left', frameon=False, prop=dict(color=textcolor))) 
    if (ax_xticks == False) and (plotprof == True):
        axprof.set_xticks([])
    elif(ax_xticks == True) and (plotprof == True):
        axprof.set_xlabel('Distance (nm)')
    if (ax_yticks == False) and (plotprof == True):
        axprof.set_yticks([])
    elif (ax_yticks == False) and (plotprof == True):
        axprof.set_ylabel('Rel. Probability')
    if (ax_yrev == True) and (plotprof == True):
        axprof.yaxis.tick_right()
        axprof.yaxis.set_label_position("right")
    if plotprof == False:
        return x_prof, profile
    elif plotprof == True:
        return axprof
    
    



molplotsettings_large = dict(bond_alpha=1,
                             bond_linewidth=1,
                             mo_size=500,
                             mo_color='purple',
                             s_size=300,
                             s_color='yellow',
                             vac_size=200,
                             vac_color='red',
                             marker_linewidth=4)

molplotsettings_small = dict(bond_alpha=1,
                             bond_linewidth=0.5,
                             mo_size=250,
                             mo_color='purple',
                             s_size=150,
                             s_color='yellow',
                             vac_size=100,
                             vac_color='red',
                             marker_linewidth=2)



wetplots = [12, 17, 18]
dryplots = [2, 4, 5]

colors = np.concatenate([sns.color_palette("winter", 4), sns.color_palette("autumn", 4)])

shape = (6, 4)
fig = plt.figure(figsize=(4.8, 9))
axwet = plt.subplot2grid(shape , (0, 0), rowspan=3, colspan=3, aspect=4/3, fig=fig)
axprofwet = kde_axplot2(kde_all_wet, axwet, molplotsettings_large, colors[0], ax_xticks=False, title=r'$\bf{a}$ Liquid Cell - all videos', counts=len(wet_atoms_norm_all.lat_x.notna()))

axwet0 = plt.subplot2grid(shape , (0, 3), rowspan=1, colspan=1, aspect=1, fig=fig)
profwet_0 = kde_axplot2(kdemaps[wetplots[0]], axwet0, molplotsettings_small, colors[1], ax_xticks=False, ax_yrev=True, title=r'$\bf{c}$ Liquid Cell', counts=lat_counts[wetplots[0]], plotprof=False, dose=beam_current_list[wetplots[0]])
axwet1 = plt.subplot2grid(shape , (1, 3), rowspan=1, colspan=1, aspect=1, fig=fig)
profwet_1 = kde_axplot2(kdemaps[wetplots[1]], axwet1, molplotsettings_small, colors[2], ax_xticks=False, ax_yrev=True, title=r'$\bf{d}$ Liquid Cell', counts=lat_counts[wetplots[1]], plotprof=False, dose=beam_current_list[wetplots[1]])
axwet2 = plt.subplot2grid(shape , (2, 3), rowspan=1, colspan=1, aspect=1, fig=fig)
profwet_2 = kde_axplot2(kdemaps[wetplots[2]], axwet2, molplotsettings_small, colors[3], ax_xticks=False, ax_yrev=True, title=r'$\bf{e}$ Liquid Cell', counts=lat_counts[wetplots[2]], plotprof=False, dose=beam_current_list[wetplots[2]])

axprofwet.plot(profwet_0[0], profwet_0[1], color=colors[1], zorder=5, alpha=0.6)
axprofwet.plot(profwet_1[0], profwet_1[1], color=colors[2], zorder=5, alpha=0.6)
axprofwet.plot(profwet_2[0], profwet_2[1], color=colors[3], zorder=5, alpha=0.6)

axdry = plt.subplot2grid(shape , (3, 0), rowspan=3, colspan=3, aspect=4/3, fig=fig)
axprofdry = kde_axplot2(kde_all_dry, axdry, molplotsettings_large, colors[4], title=r'$\bf{f}$ Vacuum - all videos', counts=len(dry_atoms_norm_all.lat_x.notna()))

axdry0 = plt.subplot2grid(shape , (3, 3), rowspan=1, colspan=1, aspect=1, fig=fig)
profdry_0 = kde_axplot2(kdemaps[dryplots[0]], axdry0, molplotsettings_small, colors[5], ax_xticks=False, ax_yrev=True, title=r'$\bf{h}$ Vacuum', counts=lat_counts[dryplots[0]], plotprof=False, dose=beam_current_list[dryplots[0]])
axdry1 = plt.subplot2grid(shape , (4, 3), rowspan=1, colspan=1, aspect=1, fig=fig)
profdry_1 = kde_axplot2(kdemaps[dryplots[1]], axdry1, molplotsettings_small, colors[6], ax_xticks=False, ax_yrev=True, title=r'$\bf{i}$ Vacuum', counts=lat_counts[dryplots[1]], plotprof=False, dose=beam_current_list[dryplots[1]])
axdry2 = plt.subplot2grid(shape , (5, 3), rowspan=1, colspan=1, aspect=1, fig=fig)
profdry_2 = kde_axplot2(kdemaps[dryplots[2]], axdry2, molplotsettings_small, colors[7], ax_xticks=False, ax_yrev=True, title=r'$\bf{j}$ Vacuum', counts=lat_counts[dryplots[2]], plotprof=False, dose=beam_current_list[dryplots[2]])

axprofdry.plot(profdry_0[0], profdry_0[1], color=colors[5], zorder=5, alpha=0.6)
axprofdry.plot(profdry_1[0], profdry_1[1], color=colors[6], zorder=5, alpha=0.6)
axprofdry.plot(profdry_2[0], profdry_2[1], color=colors[7], zorder=5, alpha=0.6)

axprofwet.add_artist(AnchoredText(r"$\bf{b}$ Intensity Profiles", loc='upper left', frameon=False, prop=dict(color='black')))
axprofdry.add_artist(AnchoredText(r"$\bf{g}$ Intensity Profiles", loc='upper left', frameon=False, prop=dict(color='black')))

cbax = fig.add_axes([0.05, 0.98, 0.9, 0.02])
matplotlib.colorbar.ColorbarBase(cbax, orientation='horizontal', cmap=plt.get_cmap(kde_cmap), norm=matplotlib.colors.Normalize(kde_im_vmin, kde_im_vmax), label='Pt Location Relative Probability', ticks=[kde_im_vmin, 1, kde_im_vmax])

fig.subplots_adjust(top=0.925,
                    bottom=0.05,
                    left=0.0,
                    right=0.94,
                    hspace=0.1,
                    wspace=0.0)

fig.savefig(savefolder_figs + 'kdeplot.png', dpi=300)
fig.savefig(savefolder_figs + 'kdeplot.pdf', dpi=300)
fig.savefig(savefolder_figs + 'kdeplot.svg', dpi=300)











































savefolder_consecutive = savefolder + 'consecutivity_plots' + os.sep
os.makedirs(savefolder_consecutive, exist_ok=True)


def consecutive(data, stepsize=1):
    return np.split(data, np.where(np.diff(data) != stepsize)[0]+1)
from matplotlib.ticker import (MultipleLocator, AutoMinorLocator)

barwidth = 0.5
cmap = plt.get_cmap("tab10")
for i in tqdm(range(numvid), desc='continuity plots'):
    atoms = atoms_norm_list[i]
    barlist = []
    for pnum in atoms.particle.unique():
        flist = atoms[atoms.particle == pnum].frame.values
        groupends = [[group[0], group[-1]] for group in consecutive(flist)]
        barlist.append(groupends)
    
    fig, ax = plt.subplots(1, figsize=(6,9))
    for j in range(len(barlist)):
        bars = barlist[j]
        colindex = j
        while colindex > 10:
            colindex -= 10
        color = cmap(colindex)
        for bar in bars:
            ax.plot(bar, [-j,-j], color=color)
            ax.scatter(bar, [-j,-j], color=color, s=6)
            
    ax.yaxis.set_minor_locator(MultipleLocator(1))
    ax.set_yticklabels([])
    ax.xaxis.set_major_locator(MultipleLocator(10))
    ax.xaxis.set_major_formatter('{x:.0f}')
    ax.xaxis.set_minor_locator(MultipleLocator(1))
    ax.grid(True, which='both')
    
    ax.set_xlabel('Video Frames')
    ax.set_ylabel('Labelled adatoms')
    ax.set_title(settings_list[i].filename[:-5] + ' ' + wetdry_list[i])
    
    ax.set_xlim(-1, atoms.frame.max() + 1)
    ax.set_ylim(-len(barlist)-1, 1)
    
    fig.subplots_adjust(top=0.97,
                        bottom=0.05,
                        left=0.06,
                        right=0.995,
                        hspace=0.2,
                        wspace=0.2)
    
    fig.savefig(savefolder_consecutive + settings_list[i].filename[:-5] + '.png')
    fig.savefig(savefolder_consecutive + settings_list[i].filename[:-5] + '.pdf')
    
    plt.close('all')
    gc.collect()

        
        


mo_points, s_points, vac_points = np.array([[0, 0],]), s_points_fixed, vac_points_fixed

randnum = int(1e5)
randpoints_gen = (np.random.rand(randnum, 2) - 0.5)/2.5
full_mo_tree = cKDTree(mo_points_fixed)
gen_dists, gen_inds = full_mo_tree.query(randpoints_gen)
ind0 = np.where((mo_points_fixed == np.array([0,0])).all(axis=1))[0][0]
randpoints = randpoints_gen[gen_inds == ind0]

atoms_norm_dry = dry_atoms_norm_all
atoms_norm_wet = wet_atoms_norm_all

lat_x_initial_dry, lat_y_initial_dry = atoms_norm_dry['transformed_lat_x'], atoms_norm_dry['transformed_lat_y']
nancols_dry = (np.isnan(lat_x_initial_dry)) | (np.isnan(lat_y_initial_dry))
lat_x_dry, lat_y_dry = lat_x_initial_dry[~nancols_dry], lat_y_initial_dry[~nancols_dry]

lat_x_initial_wet, lat_y_initial_wet = atoms_norm_wet['transformed_lat_x'], atoms_norm_wet['transformed_lat_y']
nancols_wet = (np.isnan(lat_x_initial_wet)) | (np.isnan(lat_y_initial_wet))
lat_x_wet, lat_y_wet = lat_x_initial_wet[~nancols_wet], lat_y_initial_wet[~nancols_wet]

fig, ax = plt.subplots(1,3)
ax[0].scatter(lat_x_wet, lat_y_wet, s=2, alpha=0.01)
ax[1].scatter(lat_x_dry, lat_y_dry, s=2, alpha=0.05)
ax[2].scatter(randpoints[:,0], randpoints[:,1], s=2, alpha=0.05)
for axis in ax:
    axis.set_aspect('equal')
    axis.scatter(mo_points[:,0], mo_points[:,1], color='purple', s=100)
    axis.scatter(s_points[:,0], s_points[:,1], color='yellow', s=100)
    axis.scatter(vac_points[:,0], vac_points[:,1], color='red', s=100)

lat_wet = np.vstack([lat_x_wet, lat_y_wet]).T
lat_dry = np.vstack([lat_x_dry, lat_y_dry]).T

mo_tree, s_tree, vac_tree = cKDTree(mo_points), cKDTree(s_points), cKDTree(vac_points)

mo_dists_wet, mo_inds_wet = mo_tree.query(lat_wet)
s_dists_wet, s_inds_wet = s_tree.query(lat_wet)
vac_dists_wet, vac_inds_wet = vac_tree.query(lat_wet)
mo_dists_dry, mo_inds_dry = mo_tree.query(lat_dry)
s_dists_dry, s_inds_dry = s_tree.query(lat_dry)
vac_dists_dry, vac_inds_dry = vac_tree.query(lat_dry)
mo_dists_ran, mo_inds_ran = mo_tree.query(randpoints)
s_dists_ran, s_inds_ran = s_tree.query(randpoints)
vac_dists_ran, vac_inds_ran = vac_tree.query(randpoints)

wet_dists_min = np.sort(np.vstack([mo_dists_wet, s_dists_wet, vac_dists_wet]).min(axis=0))
dry_dists_min = np.sort(np.vstack([mo_dists_dry, s_dists_dry, vac_dists_dry]).min(axis=0))
ran_dists_min = np.sort(np.vstack([mo_dists_ran, s_dists_ran, vac_dists_ran]).min(axis=0))

Y_wet = np.linspace(0, 1, len(lat_wet))
Y_dry = np.linspace(0, 1, len(lat_dry))
Y_ran = np.linspace(0, 1, len(randpoints))

mo_dists_wet_s = np.sort(mo_dists_wet)
s_dists_wet_s = np.sort(s_dists_wet)
vac_dists_wet_s = np.sort(vac_dists_wet)
mo_dists_dry_s = np.sort(mo_dists_dry)
s_dists_dry_s = np.sort(s_dists_dry)
vac_dists_dry_s = np.sort(vac_dists_dry)
mo_dists_dry_s = np.sort(mo_dists_dry)
s_dists_dry_s = np.sort(s_dists_dry)
vac_dists_dry_s = np.sort(vac_dists_dry)

fig, ax = plt.subplots(1,3)
ax[0].plot(mo_dists_wet_s, Y_wet, color='purple', label='Mo Site')
ax[0].plot(s_dists_wet_s, Y_wet, color='yellow', label='S Site')
ax[0].plot(vac_dists_wet_s, Y_wet, color='red', label='HC Site')
ax[0].plot(wet_dists_min, Y_wet, color='black', label='All Sites')
ax[1].plot(mo_dists_dry_s, Y_dry, color='purple', label='Mo Sites')
ax[1].plot(s_dists_dry_s, Y_dry, color='yellow', label='S Sites')
ax[1].plot(vac_dists_dry_s, Y_dry, color='red', label='HS Sites')
ax[1].plot(dry_dists_min, Y_dry, color='black', label='All Sites')
ax[2].plot(mo_dists_ran, Y_ran, color='purple', label='Mo Sites')
ax[2].plot(s_dists_ran, Y_ran, color='yellow', label='S Sites')
ax[2].plot(vac_dists_ran, Y_ran, color='red', label='HS Sites')
ax[2].plot(ran_dists_min, Y_ran, color='black', label='All Sites')
for a in ax:
    a.set_xlim(0, 0.1)
    a.set_ylim(0, 1)
    a.grid()
    a.set_xlabel('Radial Distance(nm)')
    a.set_ylabel('Cumulative Distribution Function')
    a.legend()
ax[0].set_title('Liquid Cell')
ax[1].set_title('Vacuum')
ax[2].set_title('Rand. Dist.')
fig.subplots_adjust(top=0.88,
                    bottom=0.105,
                    left=0.09,
                    right=0.965,
                    hspace=0.2,
                    wspace=0.315)
fig.savefig(savefolder_figs + 'CDF_Pt_position.png')












savefolder_source = savefolder + 'source_data' + os.sep
os.makedirs(savefolder_source, exist_ok=True)

wetdry_sourcelist = []
bc_sourcelist = []
jumphist_sourcelist = []
latjumphist_sourcelist = []
tamsd_sourcelist = []
for i in range(len(bc_vals)):
    bc_val = bc_vals[i]
    
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(jumps_exclude_list))]
    wet_lat_df = wet_lattice_norm_all[(wet_lattice_norm_all['beam_current'] == bc_val) & ~(wet_lattice_norm_all['video'].isin(jumps_exclude_list))]
    hist_wet_adatom, bins_wet_adatom = np.histogram(wet_df.d_dist.dropna(), bins=bins_pt)
    hist_wet_lattice, bins_wet_lattice = np.histogram(wet_lat_df.d_dist.dropna(), bins=bins_lat)
    wetdry_sourcelist.append('Liquid Cell')
    bc_sourcelist.append(bc_val)
    jumphist_sourcelist.append(hist_wet_adatom)
    latjumphist_sourcelist.append(hist_wet_lattice)
    
    
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(jumps_exclude_list))]
    dry_lat_df = dry_lattice_norm_all[(dry_lattice_norm_all['beam_current'] == bc_val) & ~(dry_lattice_norm_all['video'].isin(jumps_exclude_list))]
    hist_dry_adatom, bins_dry_adatom = np.histogram(dry_df.d_dist.dropna(), bins=bins_pt)
    hist_dry_lattice, bins_dry_lattice = np.histogram(dry_lat_df.d_dist.dropna(), bins=bins_lat)
    wetdry_sourcelist.append('Vacuum')
    bc_sourcelist.append(bc_val)
    jumphist_sourcelist.append(hist_dry_adatom)
    latjumphist_sourcelist.append(hist_dry_lattice)
    
    wet_df = wet_atoms_norm_all[(wet_atoms_norm_all['beam_current'] == bc_val) & ~(wet_atoms_norm_all['video'].isin(msd_exclude_list))]
    timefirst_msd_wet = momentcalc(wet_df, 1, 1/usetime, 2)
    tamsd_sourcelist.append(timefirst_msd_wet)
    
    dry_df = dry_atoms_norm_all[(dry_atoms_norm_all['beam_current'] == bc_val) & ~(dry_atoms_norm_all['video'].isin(msd_exclude_list))]
    timefirst_msd_dry = momentcalc(dry_df, 1, 1/usetime, 2)
    tamsd_sourcelist.append(timefirst_msd_dry)


colnames = [str(a) + ' ' + str(int(b * 10**12)) + ' pA' for a,b in zip(wetdry_sourcelist, bc_sourcelist)]

bins_pt_sourcelist_range = ['%.2f - %.2f'%(a,b) for a,b in zip(bins_pt[:-1], bins_pt[1:])]
bins_lat_sourcelist_range = ['%.3f - %.3f'%(a,b) for a,b in zip(bins_lat[:-1], bins_lat[1:])]



sourcedat_4c_pt = pd.DataFrame()
sourcedat_4c_pt['Jump Distance (nm)'] = bins_pt_sourcelist_range
for colname, jumphist in zip(colnames, jumphist_sourcelist):
    sourcedat_4c_pt[colname + ' count'] = jumphist
    sourcedat_4c_pt[colname + ' density'] = (jumphist / jumphist.sum())/bins_pt[1]
sourcedat_4c_pt.to_csv(savefolder_source + '4c_pt.csv')



sourcedat_4c_lat = pd.DataFrame()
sourcedat_4c_lat['Jump Distance (nm)'] = bins_lat_sourcelist_range
for colname, jumphist in zip(colnames, latjumphist_sourcelist):
    sourcedat_4c_lat[colname + ' count'] = jumphist
    sourcedat_4c_lat[colname + ' density'] = (jumphist / jumphist.sum())/bins_lat[1]
sourcedat_4c_lat.to_csv(savefolder_source + '4c_lat.csv')




sourcedat_4d = pd.DataFrame()
sourcedat_4d['lagT'] = timefirst_msd_wet.lagt
for colname, msd in zip(colnames, tamsd_sourcelist):
    sourcedat_4d[colname + ' TAMSD'] = msd.mmd
    sourcedat_4d[colname + ' Err'] = msd.err
    sourcedat_4d[colname + ' Real N'] = msd.real_N
    sourcedat_4d[colname + ' Effective N'] = msd.N
sourcedat_4d.to_csv(savefolder_source + '4d.csv')








